///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_BILATERAL_FILTER_HH_
#define _WISH_BILATERAL_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {

  //! The bilateral filter is a smoothing filter that uses both spatial
  //! distances and feature space distances in order to smooth the image.
  //! The user have to specify both a spatial sigma and a range sigma.
  //! A lookup table for the exponents of range differences is used,
  //! and this is only implemented for integers so far. If you want to
  //! use the filter for float images you might need to compute 
  //! the exponent inside the inner loop instead, i.e, remove the
  //! EXP_LOOKUP and use exp(-0.5*...). 
  //! 
  //! The old version where the whole 3D mask was used is commented
  //! at the end of this file.
  //! 
  //! To really see the effects you can try the filter on an image 
  //! and use a large spatial sigma and a small range sigma to see how 
  //! edges are preserved during the process.     
  //!
  //! For reading, see
  //!
  //! C. Tomasi and R. Manduchi, 
  //! "Bilateral Filtering for Gray and Color Images",
  //! Proceedings of the 1998 IEEE International Conference on Computer Vision 
  //!
  template<class InputImage, class OutputImage>
  class BilateralFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< BilateralFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    BilateralFilter() : 
      m_sigma_domain(1.0),
      m_sigma_range(1.0),
      m_max_mask_radius(20)
    {}
    
    //! set sigma
    void setSigmaDomain(double _sigma_domain) {
      m_sigma_domain = _sigma_domain;
    }
    
    //! set sigma
    void setSigmaRange(double _sigma_range) {
      m_sigma_range = _sigma_range;
    }
    
    //! set normalized sigma range, i.e., input is in [0,1]
    void setNormalizedSigmaRange(double _sigma_range) {
      m_sigma_range = 
	_sigma_range*(double(InputImage::maxValue())-
		      double(InputImage::minValue()));
    }
    
    //! set sigma
    void setSigmas(double _sigma_domain, double _sigma_range) {
      m_sigma_domain = _sigma_domain;
      m_sigma_range = _sigma_range;
    }
    
    //! set mask radius
    void setMaxMaskRadius(int _mask_radius) {
      m_max_mask_radius = _mask_radius;
    }

  protected:
    double m_sigma_domain, m_sigma_range;
    int m_max_mask_radius;
    
    virtual void process();
    
  private:
    // purposely not implemented
    BilateralFilter(const BilateralFilter&);
    BilateralFilter& operator=(const BilateralFilter&);
  };

  template<class InputImage, class OutputImage>
  void BilateralFilter<InputImage, OutputImage>::process() {

    // references
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    out.resize(in);
    out.reset( OutputImageVoxelType(0) );
    out.setLabel( in.getLabel() + "_BIL" );

    // generate lookup for exponentials (OBS! integers only, fix!?)
    double sigmar2 = util::square(m_sigma_range);
    int maxval = in.getMax()-in.getMin();
    vector<double> EXP_LOOKUP(maxval+1); 
    for(int diff=0; diff<=maxval; diff++) {
      EXP_LOOKUP[diff] = exp((-0.5*diff)*diff/sigmar2);
    }
    
    // dimensions
    int W = in.width();
    int H = in.height();
    int D = in.depth();
    
    // set the mask size for each dimension
    Vec3f spacing = in.getSpacing();
    double aspecty = spacing.y/spacing.x;
    double aspectz = spacing.z/spacing.x;
    double sigmax = m_sigma_domain;
    double sigmay = m_sigma_domain/aspecty;
    double sigmaz = m_sigma_domain/aspectz;
    int wx,wy,wz;
    wx = util::minOf( int(ceil(3*sigmax)), m_max_mask_radius );
    wy = util::minOf( int(ceil(3*sigmay)), m_max_mask_radius );
    wz = util::minOf( int(ceil(3*sigmaz)), m_max_mask_radius );
    
    // check two dimensions
    if( in.getDim() == 2 )
      wz = 0;

    cout << "BilateralFilter: Mask size = " 
	 << 2*wx+1 << " " 
	 << 2*wy+1 << " " 
	 << 2*wz+1 << endl;
    
    // allocate spatial mask    
    Image<double> mask(2*wx+1,2*wy+1,2*wz+1);
    
    // squared sigmas
    double sigmax2 = util::square(sigmax);
    double sigmay2 = util::square(sigmay);
    double sigmaz2 = util::square(sigmaz);
    
    // compute spatial mask values
    int m,n,p;
    for(p=-wz; p<=wz; p++) {
      for(n=-wy; n<=wy; n++) {
	for(m=-wx; m<=wx; m++) {
	  mask(m+wx,n+wy,p+wz)=exp(-0.5*(m*m/sigmax2+
					 n*n/sigmay2+
					 p*p/sigmaz2));
	}
      }
    }
    
    double tempfloat1,tempfloat2,tempfloat3;
    int tempdiff, tempcenter, tempcurrent;
    
    int i,j,k;
    
    // CENTER LOOP
    for(k=wz; k<D-wz; k++) {
      this->updateLiveProgress((100.0*k)/D);
      cout << "BilateralFilter: Processing slice " << k+1
	   << " (" << D << ")" << endl;
      for(j=wy; j<H-wy; j++) {
	for(i=wx; i<W-wx; i++) {	
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // LOW INDEX LOOP Z
    for(k=0; k<wz; k++) {
      for(j=0; j<H; j++) {
	for(i=0; i<W; i++) {
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // HIGH INDEX LOOP Z
    for(k=D-wz; k<D; k++) {
      for(j=0; j<H; j++) {
	for(i=0; i<W; i++) {	
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // LOW INDEX LOOP Y
    for(k=wz; k<D-wz; k++) {
      for(j=0; j<wy; j++) {
	for(i=0; i<W; i++) {
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // HIGH INDEX LOOP Y
    for(k=wz; k<D-wz; k++) {
      for(j=H-wy; j<H; j++) {
	for(i=0; i<W; i++) {	
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // LOW INDEX LOOP X
    for(k=wz; k<D-wz; k++) {
      for(j=wy; j<H-wy; j++) {
	for(i=0; i<wx; i++) {
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
    // HIGH INDEX LOOP X
    for(k=wz; k<D-wz; k++) {
      for(j=wy; j<H-wy; j++) {
	for(i=W-wx; i<W; i++) {	
	  tempfloat1=0.0; tempfloat2=0.0;
	  tempcenter = in.getSafeMirror(i,j,k);
	  for(p=-wz; p<=wz; p++) {
	    for(n=-wy; n<=wy; n++) {
	      for(m=-wx; m<=wx; m++) {
		tempcurrent = in.getSafeMirror(i+m,j+n,k+p);
		tempdiff = abs(tempcenter-tempcurrent);
		tempfloat3 = mask(m+wx,n+wy,p+wz)*EXP_LOOKUP[tempdiff];
		tempfloat1 += tempfloat3;
		tempfloat2 += tempfloat3*tempcurrent;
	      }
	    }
	  }
	  // result
	  out(i,j,k) = OutputImageVoxelType(tempfloat2/tempfloat1+0.5);
	}
      }
    }
    
  }
  
} // end namespace wish

#endif
