///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_BINARY_DILATION_FILTER_HH_
#define _WISH_BINARY_DILATION_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  template<class InputImage, class OutputImage>
  class BinaryDilationFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< BinaryDilationFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    BinaryDilationFilter() :
      m_neighborhood(C26),
      m_output_value(OutputImageVoxelType(1))
    {}
    
    //! set neighborhood
    void setNeighborhood(NeighborhoodType _nb) {
      m_neighborhood = _nb;
    }

    //! set output value
    void setOutputValue(OutputImageVoxelType _ov) {
      m_output_value = _ov;
    }
    
  protected:
    NeighborhoodType m_neighborhood;
    OutputImageVoxelType m_output_value;
    
    virtual void process();
    
  private:
    // purposely not implemented
    BinaryDilationFilter(const BinaryDilationFilter&);
    BinaryDilationFilter& operator=(const BinaryDilationFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void BinaryDilationFilter<InputImage, OutputImage>::process() {
    
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    out.resize(in);
    out.reset( OutputImageVoxelType(0) );
    
    // neighborhood iterator
    NeighborhoodIterator<InputImage> nb(&in,this->m_neighborhood);
    nb.setBox( this->m_box );
    nb.first();
    nb.setDefaultValue(InputImageVoxelType(0));
    
    // loop
    do {
      if( nb.value() ) {
	out(nb.getPosition()) = this->m_output_value;
	do { // neighborhood loop
	  out.setSafe(this->m_output_value, nb.getNeighborPosition());
	} while( nb.nextnb() );
      }
    } while( nb.next() );
    
  }    
  
} // end namespace wish

#endif
