///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_COLOR_HH_
#define _WISH_COLOR_HH_

#include "wishutil.hh"

namespace wish {
  
  // simple color classes
  template<class T> class Color3;
  template<class T> class Color4;
  
  //! Color3
  template<class T>
  class Color3 {
  public:
    
    //! unions to be able to use different names
    union { 
      struct {
	T r; T g; T b;
      };
      struct {
	T h; T s; T v;
      };
      T data[3];
    };
    
    //! clamp value
    static double clampMin();
    static double clampMax();
    
    T* getPointer() const {
      return const_cast<T*>(&data[0]);
    }
    
    //! convenience operator, glColor??v( v )
    operator T*() {
      return getPointer();
    }
    
    T& operator[](const int i) const {
      return getPointer()[i];
    }
    
    //! default constructor
    Color3() {
      r=0; g=0; b=0;
    }
    
    //! initializing constructor
    Color3(T _r, T _g, T _b) :
      r(_r), g(_g), b(_b) {}
    
    //! copy constructor
    Color3(const Color3<T> &c);
    
    //! cast copy constructor
    template<class X>
    Color3(const Color3<X> &c);
    
    //! construct from Color4
    Color3(const Color4<T> &c);
    
    //! color addition
    Color3<T> operator+(const Color3<T> &c) const {
      return Color3<T>(r+c.r,g+c.g,b+c.b);
    }
    
    //! multiplication with double
    Color3<T> operator*(const double &d) const {
      return Color3<T>(T(r*d),T(g*d),T(b*d));
    }
    
    //! division with double
    Color3<T> operator/(const double &d) const {
      return Color3<T>(T(r/d),T(g/d),T(b/d));
    }
    
    //! multiplication with double
    friend Color3<T> operator*(const double &d, const Color3<T> &c) {
      return c*d;
    }
        
    //! hsv to rgb
    Color3<T> hsv2rgb() const;
    
    //! input stream
    friend istream& operator>>(istream &is, Color3<T> &c) {
      is >> c.r >> c.g >> c.b;
      return is;
    }
    
    //! output stream
    friend ostream& operator<<(ostream &os, const Color3<T> &c) {
      os << float(c.r) << " " << float(c.g) << " " << float(c.b);
      return os;
    }
  };

  //! Color4
  template<class T>
  class Color4 {
  public:
    
    //! unions to be able to use different names
    union {
      struct {
	T r; T g; T b; T a;
      };
      struct {
	T h; T s; T v; T aa;
      };
      T data[4];
    };
    
    //! clamp value
    static double clampMin();
    static double clampMax();
    
    T* getPointer() const {
      return const_cast<T*>(&data[0]);
    }
    
    //! convenience operator, glColor??v( v )
    operator T*() {
      return getPointer();
    }

    T& operator[](const int i) const {
      return getPointer()[i];
    }
    
    //! default constructor
    Color4();
    
    //! initializing constructor
    Color4(T _r, T _g, T _b, T _a) :
      r(_r), g(_g), b(_b), a(_a) {}
    
    //! copy constructor
    Color4(const Color4<T> &c);
    
    //! cast copy constructor
    template<class X>
    Color4(const Color4<X> &c);
    
    //! construct from Color3
    Color4(const Color3<T> &c);
    
    //! color addition
    Color4<T> operator+(const Color4<T> &c) const {
      return Color4<T>(r+c.r,g+c.g,b+c.b,a+c.a);
    }
    
    //! multiplication with double
    Color4<T> operator*(const double &d) const {
      return Color4<T>(T(r*d),T(g*d),T(b*d),T(a*d));
    }
    
    //! division with double
    Color4<T> operator/(const double &d) const {
      return Color4<T>(T(r/d),T(g/d),T(b/d),T(a/d));
    }
    
    //! multiplication with double
    friend Color4<T> operator*(const double &d, const Color4<T> &c) {
      return c*d;
    }
    
    //! input stream
    friend istream& operator>>(istream &is, Color4<T> &c) {
      is >> c.r >> c.g >> c.b >> c.a;
      return is;
    }
    
    //! output stream
    friend ostream& operator<<(ostream &os, const Color4<T> &c) {
      os << float(c.r) << " " << float(c.g) << " " << float(c.b) << " " 
	 << float(c.a);
      return os;
    }
  };
  
  template<class T> 
  double Color3<T>::clampMin() {
    return double( numeric_limits<T>::min() );
  }

  template<class T> 
  double Color3<T>::clampMax() {
    return double( numeric_limits<T>::max() );
  }

  template<class T> 
  double Color4<T>::clampMin() {
    return double( numeric_limits<T>::min() );
  }
  
  template<class T> 
  double Color4<T>::clampMax() {
    return double( numeric_limits<T>::max() );
  }
  
  template<> 
  inline double Color3<float>::clampMin() {
    return 0.0;
  }

  template<> 
  inline double Color3<float>::clampMax() {
    return 1.0;
  }

  template<> 
  inline double Color4<float>::clampMin() {
    return 0.0;
  }

  template<> 
  inline double Color4<float>::clampMax() {
    return 1.0;
  }
  
  template<class T>
  Color3<T>::Color3(const Color3<T> &c) {
    r = c.r;
    g = c.g;
    b = c.b;
  }

  template<class T>
  template<class X>
  Color3<T>::Color3(const Color3<X> &c) {
    r = T(c.r);
    g = T(c.g);
    b = T(c.b);
  }
  
  template<class T>
  Color4<T>::Color4(const Color4<T> &c) {
    r = c.r;
    g = c.g;
    b = c.b;
    a = c.a;
  }

  template<class T>
  template<class X>
  Color4<T>::Color4(const Color4<X> &c) {
    r = T(c.r);
    g = T(c.g);
    b = T(c.b);
    a = T(c.a);
  }
  
  //! cast copy constructor between Color3<uint8> and Color3<float>
  template<>
  template<>
  inline Color3<float>::Color3(const Color3<uint8> &c) {
    r = float(c.r/c.clampMax());
    g = float(c.g/c.clampMax());
    b = float(c.b/c.clampMax());
  }
  
  //! cast copy constructor between Color3<uint8> and Color3<float>
  template<>
  template<>
  inline Color3<uint8>::Color3(const Color3<float> &c) {
    r = uint8( c.r*clampMax() );
    g = uint8( c.g*clampMax() );
    b = uint8( c.b*clampMax() );
  }
  
  //! cast copy constructor between Color4<uint8> and Color4<float>
  template<>
  template<>
  inline Color4<float>::Color4(const Color4<uint8> &c) {
    r = float(c.r/c.clampMax());
    g = float(c.g/c.clampMax());
    b = float(c.b/c.clampMax());
    a = float(c.a/c.clampMax());
  }
  
  //! cast copy constructor between Color4<uint8> and Color4<float>
  template<>
  template<>
  inline Color4<uint8>::Color4(const Color4<float> &c) {
    r = uint8( c.r*clampMax() );
    g = uint8( c.g*clampMax() );
    b = uint8( c.b*clampMax() );
    a = uint8( c.a*clampMax() );
  }
  
  template<class T>
  Color3<T>::Color3(const Color4<T> &c) {
    r = c.r; g = c.g; b = c.b;
  }
  
  template<class T>
  Color4<T>::Color4(const Color3<T> &c) {
    r = c.r; g = c.g; b = c.b; a = T(clampMax());
  }
  
  template<class T>
  Color4<T>::Color4() {
    r = 0; g = 0; b = 0; a = T(clampMax());
  }
  
  //! hsv to rgb (SHOULD IMPLEMENT RGB TO HSV AS WELL)
  template<class T>
  Color3<T> Color3<T>::hsv2rgb() const {
    
    if( s == 0 )
      return Color3<T>(v,v,v);
    
    float hue = h/clampMax();
    float sat = s/clampMax();
    float val = v/clampMax();
    
    float base = float(clampMax())*(1.0f-sat)*val;
    float frac = hue*(360.0f/60.0f);
    float mod = (int(360.0f*hue)%60)/60.0f;
    
    T rr,gg,bb;
    
    switch (int(frac)) {
    case 0:
      rr = T(clampMax()*val);
      gg = T((clampMax()*val-base)*frac + base);
      bb = T(base);
      break;
    case 1:
      rr = T((clampMax()*val-base)*(1.0-mod+base));
      gg = T(clampMax()*val);
      bb = T(base);
      break;
    case 2:
      rr = T(base);
      gg = T(clampMax()*val);
      bb = T((clampMax()*val-base)* (mod + base));
      break;
    case 3:
      rr = T(base);
      gg = T(((clampMax()*val-base)*(1.0-mod + base)));
      bb = T((clampMax()*val));
      break;
    case 4:
      rr = T(((clampMax()*val-base)*(mod + base)));
      gg = T(base);
      bb = T((clampMax()*val));
      break;
    case 5:
      rr = T((clampMax()*val));
      gg = T(base);
      bb = T(((clampMax()*val-base)*(1.0-mod + base)));
      break;
    default:
      break;
    }
    
    return Color3<T>(rr,gg,bb);
  }
  
  typedef Color3<float> Color3f;
  typedef Color4<float> Color4f;
  typedef Color3f RGBf;
  typedef Color4f RGBAf;
  typedef Color3<uint8> RGB8;
  typedef Color4<uint8> RGBA8;
  
} // end namespace wish

#endif
