///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_COMPARESEGMENTATIONS_HH_
#define _WISH_COMPARESEGMENTATIONS_HH_

#include "wishImage.hh"

namespace wish {
  
  class CompareSegmentations {
    
    //! Compares segmentations according to the mesures in
    //!
    //! J.K. Udupa et al., 
    //! "A framework for evaluating segmentation algorithms",
    //! Computerized Medical Imaging and Graphics 30 (2006)
    //!
    //! It is designed to compare a segmentation against a true segmentation
    //! in terms of sensitivity and specificity, but if you need to compare 
    //! two segmentations in terms of precision you can do that as well. Just 
    //! put one of the segmenations as true delineation.
    //!
    //! The cardinalities are doubles because segmentations can be fuzzy.
    
  protected:
    ImageObject *im_Ctd; // true delinetaion
    ImageObject *im_Cd;  // delinetaion
    
    bool treat_as_fuzzy; // if fuzzy segmentations
    
    // when executed, these cardinalities are computed 
    double Ctd;  // cardinality of true delineation
    double Ctdc; // cardinality of true delineation complement
    double Cd;   // cardinality of delineation
    double Cu;   // cardinality of union
    double CTP;  // cardinality of true positives (intersection)
    double CFP;  // cardinality of false positives 
    double CTN;  // cardinality of true negatives
    double CFN;  // cardinality of false negatives
    
    // the volume of a voxel
    double voxelvolume;
    
    virtual void process();
    
  public:
    //! constructor
    CompareSegmentations() :
      im_Ctd(0),
      im_Cd(0),
      treat_as_fuzzy(true)
    {}
    
    //! set true delineation
    void setTrueDelineation(ImageObject *im) {
      im_Ctd = im;
      
      // determines the voxel volume
      if( im_Ctd ) {
	Vec3d spc = im_Ctd->getSpacing();
	voxelvolume = spc.x * spc.y * spc. z;
      }
    }
    
    //! set delineation
    void setDelineation(ImageObject *im) {
      im_Cd = im;
    }
    
    //! explicitly set the voxel volume (set after you set true delineation) 
    void setVoxelVolume(const double &v) {
      voxelvolume = v;
    }
    double getVoxelVolume() const {
      return voxelvolume;
    }
    
    //! set if segmentations should be treated as fuzzy
    //! this option can safely be set to true even if you have binary 
    //! segmentations if:
    //!   a background voxel has the value 0 and a foreground voxel
    //!   has the maximum integer value, e.g., 255 or 65535.
    virtual void setTreatAsFuzzy(const bool &b) {
      treat_as_fuzzy = b;
    }
    
    virtual bool getTreatAsFuzzy() const {
      return treat_as_fuzzy;
    }
    
    //! execute if images are set
    virtual void execute() {
      if( im_Ctd && im_Cd ) {
	if( im_Ctd->width()==im_Cd->width() &&
	    im_Ctd->height()==im_Cd->height() &&
	    im_Ctd->depth()==im_Cd->depth() )
	  process();
	else
	  cerr << __FUNCTION__ << "Dimension mismatch!!!" << endl;
      } else 
	cerr << __FUNCTION__ << "Bad input!!!" << endl;
    }
    
    //! get measures
    double getTruePositiveVolumeFraction() const {
      return CTP/Ctd;
    }
    double getTrueNegativeVolumeFraction() const {
      return CTN/Ctdc;
    }
    double getFalsePositiveVolumeFraction() const {
      return CFP/Ctdc;
    }
    double getFalseNegativeVolumeFraction() const {
      return CFN/Ctd;
    }
    
    double getTrueDelineationCardinality() const {
      return Ctd;
    }
    double getDelineationCardinality() const {
      return Cd;
    }
    double getIntersectionCardinality() const {
      return CTP;
    }
    double getUnionCardinality() const {
      return Cu;
    }
    
    double getTrueDelineationVolume() const {
      return voxelvolume*getTrueDelineationCardinality();
    }
    double getDelineationVolume() const {
      return voxelvolume*getDelineationCardinality();
    }
    double getIntersectionVolume() const {
      return voxelvolume*getIntersectionCardinality();
    }
    double getUnionVolume() const {
      return voxelvolume*getUnionCardinality();
    }
    
    double getPrecision() const {
      return CTP/Cu;
    }
    double getSensitivity() const {
      return getTruePositiveVolumeFraction();
    }
    double getSpecificity() const {
      return 1.0-getFalsePositiveVolumeFraction();
    }
    
  };
  
} // end namespace wish

#endif

