///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_DISTANCE_MAP_FILTER_HH_
#define _WISH_DISTANCE_MAP_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  template<class InputImage, class OutputImage>
  class DistanceMapFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< DistanceMapFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    DistanceMapFilter() : 
      m_weight_a(3.0), 
      m_weight_b(4.0), 
      m_weight_c(5.0), 
      m_internal(true)
    {}
    
    //! set weights
    void setWeights(double _weight_a=3.0, 
		    double _weight_b=4.0, 
		    double _weight_c=5.0) {
      m_weight_a = _weight_a;
      m_weight_b = _weight_b;
      m_weight_c = _weight_c;
    }
    
    //! set internal or not
    void setInternal(bool _internal) {
      m_internal = _internal;
    }
    
  protected:
    double m_weight_a, m_weight_b, m_weight_c;
    bool m_internal;
    
    virtual void process();
    
  private:
    // purposely not implemented
    DistanceMapFilter(const DistanceMapFilter&);
    DistanceMapFilter& operator=(const DistanceMapFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void DistanceMapFilter<InputImage, OutputImage>::process() {
    
    this->m_output->resize( *(this->m_input) );
    this->m_output->reset( OutputImageVoxelType(0) );
    
    this->m_output->setLabel( this->m_input->getLabel() + "_DM" );
    
    // check dimensions and select neighborhood
    NeighborhoodType fwd, bwd;
    
    vector<double> dtweights;
    
    if( this->m_input->getDim() == 2 ) {
      fwd = CHFWD8;
      bwd = CHBWD8;
      
      dtweights.resize(4);	 
      for(int aa=0; aa<2; aa++)
	dtweights[aa] = m_weight_a;
      for(int bb=2; bb<4; bb++)
	dtweights[bb] = m_weight_b;
      
    }
    
    else {
      fwd = CHFWD26;
      bwd = CHBWD26;
      
      dtweights.resize(13);
      for(int aa=0; aa<3; aa++)
	dtweights[aa] = m_weight_a;
      for(int bb=3; bb<9; bb++)
	dtweights[bb] = m_weight_b;
      for(int cc=9; cc<13; cc++)
	dtweights[cc] = m_weight_c;
    }
    
    // initialize image by setting object pixels to infinity and
    // background pixels to zero
    OutputImageVoxelType INF = OutputImageVoxelType( this->m_output->maxValue() );
    OutputImageVoxelType ZERO = OutputImageVoxelType(0);
    
    // defval is used at the border only
    OutputImageVoxelType defval = (m_internal) ? ZERO:INF;
    
    VoxelIterator<InputImage> pin( this->m_input );
    VoxelIterator<OutputImage> pout( this->m_output );
    
    pin.setBox( this->m_box );
    pout.setBox( this->m_box );
    
    pin.first();
    pout.first();
    
    // check if we shall compute from object or from background
    // internal => background = 0, object = INF  
    // !internal => object = 0, background = INF
    
    cout << "Initializing DT..." << endl;
    
    // initialize DT
    do {
      if( pin.value() )
	pout.value() = (m_internal) ? INF:ZERO;
      else
	pout.value() = (m_internal) ? ZERO:INF;
      
    } while( pin.next() && pout.next() );
    
    // neighborhood for forward scan
    NeighborhoodIterator<OutputImage> nb(this->m_output,fwd);
    nb.setBox( this->m_box );
    nb.first();
    nb.setDefaultValue(defval);
    
    cout << "Forward pass..." << endl;
    // forward pass
    
    OutputImageVoxelType curr;
    double temp;
    do {
      if( nb.value() ) {
	curr = INF;
	int i=0;
	do { // neighborhood loop
	  temp = double(nb.valuenbbox()) + dtweights[i];
	  if( temp < double(curr) ) 
	    curr = OutputImageVoxelType(temp); 
	  i++;
	} while( nb.nextnb() );
	nb.value() = curr;
      }
    } while( nb.next() );
    
    // neighborhood for backward scan 
    NeighborhoodIterator<OutputImage> nb2(this->m_output,bwd);
    nb2.setBox( this->m_box );
    nb2.last();
    nb2.setDefaultValue(defval);
    
    cout << "Backward pass..." << endl;
    // backward pass
    do {
      if( nb2.value() ) {
	curr = nb2.value();	
	int i=0;
	do { // neighborhood loop
	  temp = double(nb2.valuenbbox()) + dtweights[i];
	  if(temp < double(curr)) 
	    curr = OutputImageVoxelType(temp); 
	  i++;
	} while( nb2.nextnb() );
	nb2.value() = curr;
      }
    } while( nb2.prev() );
    
    cout << "Distance transform finished!" << endl;
  }    
  
} // end namespace wish

#endif
