///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishExternalForce.hh"

namespace wish {
  
  Vec3d InteractionForce::getMeanCollectedForces() const {
    Vec3d f(0,0,0);
    int N = (int)collected_forces.size();
    if( N <= 0 )
      return f;
    
    for(int i=0; i<N; i++) {
      f += collected_forces[i];
    }
    f = f/N;
    
    return f;
  }
  
  Vec3d InteractionForce::computeForce(const Vec3d &pos,
				       const Vec3d &norm,
				       const bool &selected) {
    
    // interaction only with selected vertices
    if( !selected )
      return Vec3d(0,0,0);
    
    // direction to interaction position
    Vec3d dir = interaction_position-pos;
    double l = dir.norm();
    if( l < 1e-3 ) 
      return Vec3d(0,0,0);
    
    // normalize direction
    dir.normalize();
    
    Vec3d f = this->weight*(dir*norm)*norm;
    
    // update accumulated force
    if( collect_forces )
      collected_forces.push_back( f );
    
    return f;
  }
  
  Vec3d PropagationForce::computeForce(const Vec3d &pos,
				       const Vec3d &norm,
				       const bool &selected) {
    
    if( !image )
      return Vec3d(0,0,0);
    
    // this is dependent on the coordinate systems used,
    // for now we assume that the input coords are given in world coords
    Vec3d p = Mw2i*pos; 
    
    if( this->image->getSafeNormalized(p) <= zerolevel )
      return -1.0*this->weight*norm;
    else
      return 1.0*this->weight*norm;
  }
  
  Vec3d GradientMagnitudeForce::computeForce(const Vec3d &pos,
					     const Vec3d &norm,
					     const bool &selected) {
    
    
    if( !image )
      return Vec3d(0,0,0);
        
    // this is dependent on the coordinate systems used,
    // for now we assume that the input coords are given in world coords
    Vec3d p = Mw2i*pos;
    int x = int( floor(p.x) );
    int y = int( floor(p.y) );
    int z = int( floor(p.z) );
    
    double value;
    double maxvalue = this->image->getSafeAsDouble(x,y,z);
    Vec3d maxpos = Vec3d(x,y,z);   
    
    for(int k=-search_radius; k<=search_radius; k++) {
      for(int j=-search_radius; j<=search_radius; j++) {
	for(int i=-search_radius; i<=search_radius; i++) {
	  value = this->image->getSafeAsDouble(x+i, y+j, z+k);
	  if( value > maxvalue ) {
	    maxvalue = value;
	    maxpos = Vec3d(x+i, y+j, z+k);
	  }
	}
      }
    }
    
    Vec3d tmp = Mi2w*maxpos-pos;
    return this->weight*(tmp*norm)*norm;
    
  }
  
  
  Vec3d VectorFieldForce::computeForce(const Vec3d &pos,
				       const Vec3d &norm,
				       const bool &selected) {
    
    // it must be a Vec3f image
    Image<Vec3f> *imtmp = dynamic_cast<Image<Vec3f>*>(image);
    if( !imtmp )
      return Vec3d(0,0,0);
    
    // this is dependent on the coordinate systems used,
    // for now we assume that the input coords are given in world coords
    Vec3d p = Mw2i*pos; 
    
    // normalize this (or use a high weight) to make it faster
    Vec3d tmp = imtmp->getInterp(p);
    
    // rotate about x
    tmp = Mi2w.getRotation()*tmp; 
    tmp.normalizeSafe();
    return this->weight*(tmp*norm)*norm;
    
  }
  
  Vec3d PotentialFieldForce::computeForce(const Vec3d &pos,
					  const Vec3d &norm,
					  const bool &selected) {
    
    if( !image )
      return Vec3d(0,0,0);
    
    // this is dependent on the coordinate systems used,
    // for now we assume that the input coords are given in world coords
    Vec3d p = Mw2i*pos; 
    Vec3d tmp = this->image->getGradientInterp(p);
    
    // rotate among x and use negative direction    
    tmp = -(Mi2w.getRotation()*tmp); 
    
    tmp.normalizeSafe();
    
    return this->weight*(tmp*norm)*norm;
    
  }

  
}

