///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_EXTERNAL_FORCE_HH_
#define _WISH_EXTERNAL_FORCE_HH_

#include "wishImage.hh"

namespace wish {
  
  class ExternalForce;
  typedef ExternalForce* ExternalForcePointer;
  
  // simple class for external data forces able to drive a deformable mesh
  class ExternalForce {
  protected:
    double weight;
    
  public:
    ExternalForce() : weight(1.0) {}
    
    void setWeight(double _w) {
      weight = _w;
    }
    
    virtual Vec3d computeForce(const Vec3d &pos, 
			       const Vec3d &norm, 
			       const bool &selected=true) {
      return Vec3d(0,0,0);
    }
  };
  
  class InteractionForce : public ExternalForce {
  protected:
    Vec3d interaction_position;
    vector<Vec3d> collected_forces;
    bool collect_forces;
    
  public:
    InteractionForce() : interaction_position(Vec3d(0,0,0)) 
    {
      resetCollectedForces();
      collect_forces = false;
    }
    
    void setInteractionPosition(const Vec3d &_p) {
      interaction_position = _p;
    }
    
    void setCollectForces( const bool &b ) {
      collect_forces = b;
    }
    
    void resetCollectedForces() {
      collected_forces.clear();
    }
    
    const vector<Vec3d>& getCollectedForces() const {
      return collected_forces;
    }
    
    Vec3d getMeanCollectedForces() const;

    virtual Vec3d computeForce(const Vec3d &pos, 
			       const Vec3d &norm,
			       const bool &selected=true);
  };
  
  class ImageForce : public ExternalForce {
  protected:
    ImageObject *image;
    Matrix4d Mi2w;
    Matrix4d Mw2i;
    
  public:
    
    ImageForce(ImageObject *_im=0) : image(_im) {}
    
    void setImage(ImageObject *_im) {
      image = _im;
      if( image ) {
	Mi2w = Matrix4d( image->getImageToWorldMatrix() );
	Mw2i = Matrix4d( image->getWorldToImageMatrix() );
      }
    }
    
  };
  
  // gradient magnitude 
  // This force is designed for a gradient magnitude map, 
  // not the original image
  class GradientMagnitudeForce : public ImageForce {
  protected:
    int search_radius;
    
  public:
    GradientMagnitudeForce(ImageObject *_im=0) : 
      ImageForce(_im), search_radius(1) {
    }
    
    void setSearchRadius(int _r) {
      search_radius = _r;
    }
    
    Vec3d computeForce(const Vec3d &pos, 
		       const Vec3d &norm, 
		       const bool &selected=true);
  };
  
  // propagation 
  // This force is a speed function that lets the surface propagate in 
  // its normal direction. Above the zero level we have positive 
  // speed and below we have negative speed. Can be used on a binary
  // image to refine a segmentation. 
  class PropagationForce : public ImageForce {
  protected:
    double zerolevel;
    
  public:
    PropagationForce(ImageObject *_im=0) : 
      ImageForce(_im), zerolevel(0.0) {
    }
    
    void setZeroLevel(double _zl) {
      zerolevel = _zl;
    }
    
    virtual Vec3d computeForce(const Vec3d &pos, 
			       const Vec3d &norm,
			       const bool &selected=true);
  };
  
  // vector field 
  class VectorFieldForce : public ImageForce {
    
  public:
    VectorFieldForce(ImageObject *_im=0) : ImageForce(_im) {
    }
    
    virtual Vec3d computeForce(const Vec3d &pos, 
			       const Vec3d &norm,
			       const bool &selected=true);
  };
  
  // potential field 
  class PotentialFieldForce : public ImageForce {
    
  public:
    PotentialFieldForce(ImageObject *_im=0) : ImageForce(_im) {
    }
    
    virtual Vec3d computeForce(const Vec3d &pos, 
			       const Vec3d &norm,
			       const bool &selected=true);
  };

  
}

#endif
