///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_FASTMARCHING_FILTER_HH_
#define _WISH_FASTMARCHING_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishMinHeap.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  //! Fast Marching filter
  //! - output is a time image (i.e., use float or double)
  //! - input is a COST image (1/speed) (e.g., gradient magnitude)
  //! - additional input: seed image
  //! - additional input/output: trial image (segmentation if energy is used)
  
  //! Fast Marching is the optimal way to solve 
  //! the Eikonal equation. The following text is copied from Sethians
  //! web-page http://math.berkeley.edu/~sethian/
  //
  //! The central idea of level set methods is to track a propagating 
  //! interface by embedding it as the zero level set of a higher-dimensional 
  //! function. This is an initial value partial differential equation; 
  //! the initial front gamma(t=0) provides initial data as the zero level 
  //! set of the level set function phi, and tracking the evolution of phi 
  //! is identified with the evolution of the interface under its normal 
  //! speed F. In the case where the front is moving with a speed F which 
  //! is always positive, front propagation problems can be recast as a 
  //! stationary boundary value problem. In the special case where this 
  //! speed depends only on position, the resulting Eikonal equation 
  //! is a static Hamilton-Jacobi equation with a long history; 
  //! it has applications in such areas as geometric optics, 
  //! seismology, and electromagnetics. More precisely, given a interface 
  //! initially at gamma(t=0), let T(x,y) be the time T at which the 
  //! interface crosses the point (x,y). Then a static Hamilton-Jacobi 
  //! equation of the form
  //
  //! F | nabla T | = 1 with T=0 on gamma(t=0)
  //
  //! can be given. There are several numerical techniques to solve 
  //! this equation. However, using the previous work on the theory of 
  //! curve and surface evolution, viscosity theory, upwind schemes, 
  //! and sorting/causality relationships borrowed from computer science, 
  //! a particular fast optimal technique, known as the Fast Marching Method 
  //! can be developed.
  
  template<class InputImage, class OutputImage>
  class FastMarchingFilter : 
    public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< FastMarchingFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! typedef for heap element
    typedef HeapElement<OutputImageVoxelType> fElem;
    typedef MinHeap<OutputImageVoxelType> fMinHeap;
    
    //! typedef for seed iterator
    typedef vector<Vec3i>::iterator SeedsIterator;
    
    //! constructor
    FastMarchingFilter() :
      m_seeds(vector<Vec3i>(0)),
      m_trial(0),
      m_trial_owner(true),
      m_accepted_value(1),
      m_use_spacing(false),
      m_use_constant_cost(false),
      m_constant_cost(1.0f),
      m_time_max(1.0e9),
      m_time_log(vector<float>(0)),
      m_front_log(vector<float>(0)),
      m_accepted_log(vector<float>(0))
    {}
    
    ~FastMarchingFilter() {
      if( m_trial && m_trial_owner )
	delete m_trial;
    }
    
    //! the value for accepted voxels (default 1)
    void setAcceptedValue(int _acc) {
      if( _acc <= 0 || _acc >= 255 )
	cerr << "Accepted value " << _acc << " not allowed!" << endl;
      else
	m_accepted_value = _acc;
    }
    
    //! set the seeds vector
    void setSeeds(const vector<Vec3i> &_seeds) {
      m_seeds = _seeds;
    }
    
    //! add a seed to the seeds vector
    void addSeed(const Vec3i &_seed) {
      m_seeds.push_back(_seed);
    }
    
    //! use a binary image to set the seeds vector
    void setSeedsFromImage(Image<uint8> *_seed_image);
    
    void setSeedsFromImage(const Image<uint8>::Pointer &_seed_image) {
      initializeFromImage(_seed_image.get());
    }

    //! clear all seeds
    void clearSeeds() {
      m_seeds.resize(0);
    }
    
    //! set the trial image (specify output)
    void setTrialImage(Image<uint8> *_tim) {
      if( m_trial && m_trial_owner )
	delete m_trial;
      
      m_trial = _tim;
      m_trial_owner = false;
    }
    
    void setTrialImage(const Image<uint8>::Pointer &_tim) {
      this->setTrialImage(_tim.get());
    }
    
    //! release trial ownership and return trial (use to create an auto_ptr)
    Image<uint8>* releaseTrial() {
      if( m_trial_owner ) {
	m_trial_owner = false;
	Image<uint8> *tmp = m_trial;
	m_trial = 0;
	return tmp;
      }
      else
	return 0;
    }
    
    //! gets the trial image
    Image<uint8>* getTrial() {
      return m_trial;
    }
    
    //! set if to use constant cost
    void setUseConstantCost(const bool &_cc) {
      m_constant_cost = _cc;
    }
    
    //! set constant cost
    void setConstantCost(const float &_cost) {
      m_constant_cost = _cost;
    }
    
    //! set if spacing should be used 
    void setUseSpacing(bool _spc) {
      m_use_spacing = _spc;
    }
    
    //! the time threshold for propagation ( default 1.0e9 (infinity) )
    void setTimeMax(double _tm) {
      m_time_max = _tm;
    }
    
    //! get the time log
    const vector<float>& getTimeLog() const {
      return m_time_log;
    }
    
    //! get the front log
    const vector<float>& getFrontLog() const {
      return m_front_log;
    }
    
    //! get the accepted log
    const vector<float>& getAcceptedLog() const {
      return m_accepted_log;
    }
    
  protected:
    
    //! the time image (contains heap elements)
    //! we also have m_output that contains only the times
    ImageBase<fElem> m_time;
    
    //! the seed vector
    vector<Vec3i> m_seeds;
    
    //! use constant cost or not
    float m_constant_cost;
    bool m_use_constant_cost;
    
    //! the trial image
    Image<uint8> *m_trial;
    uint8 m_accepted_value;
    bool m_trial_owner;
    
    //! log the time, front, and accepted values
    vector<float> m_time_log;
    vector<float> m_front_log;
    vector<float> m_accepted_log;
    
    //! use spacing or not
    bool m_use_spacing;
    
    //! time threshold (do not iterate more if time exceeds this value)
    double m_time_max;
    
    //! static constants
    static const uint8 FARAWAY;
    static const uint8 TRIAL;
    static const double T_MAX;
    
    virtual void process();
    virtual double computeTime(const Vec3i &pos);
    virtual double computeTimeWithSpacing(const Vec3i &pos);
    
  private:
    //! purposely not implemented
    FastMarchingFilter(const FastMarchingFilter&);
    FastMarchingFilter& operator=(const FastMarchingFilter&);
  };
  
  //! constants
  template<class InputImage, class OutputImage>
  const uint8 FastMarchingFilter<InputImage, OutputImage>::FARAWAY = uint8(0);
  
  template<class InputImage, class OutputImage>
  const uint8 FastMarchingFilter<InputImage, OutputImage>::TRIAL = uint8(255);
  
  template<class InputImage, class OutputImage>
  const double FastMarchingFilter<InputImage, OutputImage>::T_MAX = 1.0e9;
  
  //! Fill the seeds vector with voxels from the seed image.
  //! So far we consider all voxels > 0 as seed points,
  //! maybe this should be changed later so that we can have several regions
  //! growing simultaneously. Note that the seed vector is cleared! 
  template<class InputImage, class OutputImage>
  void FastMarchingFilter<InputImage, OutputImage>::
  setSeedsFromImage(Image<uint8> *_seed_image) {
    m_seeds.clear();
    for(int i=0; i<_seed_image->totalSize(); i++) {
      if( (*_seed_image)[i] > 0 ) {
	m_seeds.push_back(_seed_image->indexToPos(i));
      }
    }
  }
  
  //! compute the time at a certain position
  template<class InputImage, class OutputImage>
  double FastMarchingFilter<InputImage, OutputImage>::computeTime(const Vec3i &pos) {
    
    if( m_use_spacing )
      return computeTimeWithSpacing(pos);
    
    int dim = this->m_input->getDim();
    
    // compute the differences and find min  
    double a[3];
    a[0] = util::minOf(m_time.getSafe(pos.x-1,pos.y,pos.z).value, 
		       m_time.getSafe(pos.x+1,pos.y,pos.z).value);
    a[1] = util::minOf(m_time.getSafe(pos.x,pos.y-1,pos.z).value, 
		       m_time.getSafe(pos.x,pos.y+1,pos.z).value);
    
    if( dim == 3 ) 
      a[2] = util::minOf(m_time.getSafe(pos.x,pos.y,pos.z-1).value, 
			 m_time.getSafe(pos.x,pos.y,pos.z+1).value);
    
    // sort them
    util::bubblesort(a,dim);
    
    // risk for overflow if not casting
    // we add 1.0, because we can not have a cost of zero
    double f2;
    if( this->m_use_constant_cost )
      f2 = util::square(m_constant_cost);
    else
      f2 = util::square( double(this->m_input->at(pos))+1.0 );
    
    double m;    
    double asum, sum, ssum, sqsum, t; 
    
    if( dim == 3 ) {
      // update step
      asum = (a[2]-a[0])*(a[2]-a[0]) + (a[2]-a[1])*(a[2]-a[1]);
      if( asum <= f2 ) {
	m = 3;
	sum   = a[0] + a[1] + a[2];
	ssum  = sum*sum;
	sqsum = a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
	t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
      }
      else {
	m = 2;
	asum = (a[1]-a[0])*(a[1]-a[0]);
	if( asum <= f2 ) {
	  sum   = a[0] + a[1];
	  ssum  = sum*sum;
	  sqsum = a[0]*a[0] + a[1]*a[1];
	  t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
	}
	else {
	  t = a[0] + ( double(this->m_input->at(pos)) + 1.0 );
	}
      }
    }
    
    else { // dim = 2   
      // update step
      asum = (a[1]-a[0])*(a[1]-a[0]);
      if( asum <= f2 ) {
	m = 2;
	sum   = a[0] + a[1];
	ssum  = sum*sum;
	sqsum = a[0]*a[0] + a[1]*a[1];
	t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
      }
      else {
	t = a[0] + ( double(this->m_input->at(pos)) + 1.0 );
      }
    }
    
    return t;
  }
  
  //! compute the time at a certain position
  template<class InputImage, class OutputImage>
  double FastMarchingFilter<InputImage, OutputImage>::computeTimeWithSpacing(const Vec3i &pos) {
    
    int dim = this->m_input->getDim();
    
    if( dim < 3 ) {
      cerr << "FastMarchingFilter: Can not use spacing in 2D!" << endl;
      exit(1);
    }
    
    Vec3f spc = this->m_input->getSpacing();
    double aspecty = spc.y/spc.x;
    double aspectz = spc.z/spc.x;
    
    //cerr << "Spacing is used: " << spc << endl;

    // compute the differences and find min  
    int index[3] = {0,1,2};
    double d[3] = {1.0,aspecty,aspectz};
    double a[3];
    a[0] = util::minOf(m_time.getSafe(pos.x-1,pos.y,pos.z).value, 
		       m_time.getSafe(pos.x+1,pos.y,pos.z).value);
    a[1] = util::minOf(m_time.getSafe(pos.x,pos.y-1,pos.z).value, 
		       m_time.getSafe(pos.x,pos.y+1,pos.z).value);
    
    if( dim == 3 ) 
      a[2] = util::minOf(m_time.getSafe(pos.x,pos.y,pos.z-1).value, 
			 m_time.getSafe(pos.x,pos.y,pos.z+1).value);
    
    // sort them
    util::bubblesort(a,index,dim);
    
    // the square of d
    double d2[3];
    d2[0] = d[index[0]]*d[index[0]];
    d2[1] = d[index[1]]*d[index[1]];
    d2[2] = d[index[2]]*d[index[2]];
    
    // risk for overflow if not casting
    // we add 1.0, because we can not have a cost of zero
    double f2;
    if( this->m_use_constant_cost )
      f2 = util::square(m_constant_cost);
    else
      f2 = util::square( double(this->m_input->at(pos))+1.0 );
    
    double m;    
    double asum, sum, ssum, sqsum, t; 
    double dsum2;
    
    if( dim == 3 ) {
      
      // update step
      asum = 
	(a[index[2]]-a[index[0]])*(a[index[2]]-a[index[0]]) + 
	(a[index[2]]-a[index[1]])*(a[index[2]]-a[index[1]]);
      
      if( asum <= f2 ) {
	m = 3;
	
	dsum2 = 1.0/d2[0] + 1.0/d2[1] + 1.0/d2[2]; 
	
	sum   = 
	  a[index[0]]/d2[0] + 
	  a[index[1]]/d2[1] + 
	  a[index[2]]/d2[2];
	
	ssum  = sum*sum;
	
	sqsum = 
	  a[index[0]]*a[index[0]]/d2[0] + 
	  a[index[1]]*a[index[1]]/d2[1] + 
	  a[index[2]]*a[index[2]]/d2[2];
	
	t = (sum+sqrt(dsum2*f2+ssum-dsum2*sqsum))/dsum2;
      }
      else {
	// update step
	asum = 
	  (a[index[1]]-a[index[0]])*(a[index[1]]-a[index[0]]);
	
	if( asum <= f2 ) {
	  m = 2;
	  
	  dsum2 = 1.0/d2[0] + 1.0/d2[1];
	  
	  sum   = 
	    a[index[0]]/d2[0] + 
	    a[index[1]]/d2[1];
	  
	  ssum  = sum*sum;
	  
	  sqsum = 
	    a[index[0]]*a[index[0]]/d2[0] + 
	    a[index[1]]*a[index[1]]/d2[1]; 
	  
	  t = (sum+sqrt(dsum2*f2+ssum-dsum2*sqsum))/dsum2;
	}
	else {
	  // update step
	  asum = 0.0;
	  
	  if( asum <= f2 ) {
	    m = 1;
	    
	    dsum2 = 1.0/d2[0];
	    
	    sum   = 
	      a[index[0]]/d2[0];
	    
	    ssum  = sum*sum;
	    
	    sqsum = 
	      a[index[0]]*a[index[0]]/d2[0];
	    
	    t = (sum+sqrt(dsum2*f2+ssum-dsum2*sqsum))/dsum2;
	  }
	}
      }
    }
    
    return t;
  }
  
  //! the computation
  template<class InputImage, class OutputImage>
  void FastMarchingFilter<InputImage, OutputImage>::process() {
    
    // trial
    if( !m_trial ) {
      m_trial = new Image<uint8>;
      m_trial_owner = true;
    }
    
    m_trial->resize( *(this->m_input) );
    m_trial->reset( FARAWAY );
    
    // time
    m_time.resize( (*(this->m_input)) );
    m_time.reset( fElem((float)T_MAX) );
    
    // logs
    m_time_log.resize(0);
    m_front_log.resize(0);
    m_accepted_log.resize(0);
    
    // dimensions
    int W = this->m_input->width();
    int H = this->m_input->height();
    int D = this->m_input->depth();
    
    // energy variables
    int nrAccepted = 0;
    int nrFront = 0;
    double sumAccepted = 0.0;
    double sumFront = 0.0;
    
    // the heap (allocate for one fourth of the total number of voxels)
    fMinHeap Ht((W*H*D)/4);
    
    // select neighborhood depending on dimension
    NeighborhoodType nt = C6;
    int dim = this->m_input->getDim();
    if( dim == 2 )
      nt = C4;
    
    // iterator over the neighborhood
    NeighborhoodIterator< Image<uint8> > labelNb(m_trial, nt);
    
    // set the box
    if( this->m_use_box )
      labelNb.setBox(this->m_box);
    
    cout << "FastMarchingFilter: Initializing heap..." << endl;
    
    for(SeedsIterator it=m_seeds.begin();
	it!=m_seeds.end();
	it++) {
      
      Vec3i current = *it;
      
      // set the seeds to ACCEPTED
      m_trial->at(current) = m_accepted_value;
      
      // step accepted counter and sum
      nrAccepted++;
      sumAccepted += this->m_input->at(current);
      
      // set time to zero
      m_time.at(current) = fElem(0.0);
    }
    
    // set the seeds neighbors to TRIAL and compute times
    labelNb.first();
    do {
      
      // if accepted
      if( labelNb.value() == m_accepted_value ) {
	
	labelNb.firstnbbox();
	do {
	  
	  // find surface voxels and add them to trial
	  Vec3i pos = labelNb.getNeighborPosition();
	  if( labelNb.valuenb() == FARAWAY ) {
	    labelNb.valuenb() = TRIAL;
	    m_time.at(pos) = fElem((float)computeTime(pos));
	    Ht.insert(&(m_time.at(pos)));
	   
	    // update front counter and sum
	    nrFront++;
	    sumFront += this->m_input->at(pos);
	  }
	  
	  // if the voxel already is TRIAL, check the time and update if 
	  // necessary
	  else if( labelNb.valuenb() == TRIAL ) {
	    double temptime = computeTime(pos);
	    if( temptime < m_time.at(pos).value ) {
	      m_time.at(pos).value = (float)temptime;
	      Ht.upHeap(m_time.at(pos).index);
	    }
	  }
	} while( labelNb.nextnbbox() );
	
      }
    } while ( labelNb.next() );
    
    cout << "FastMarchingFilter: Initializing procedure finished!" << endl;
    
    double mintime = 0.0, current_time = 0.0;
    
    // main loop
    while ( !(Ht.empty()) && mintime <= m_time_max ) {
      
      // find smallest
      fElem *tmin = Ht.deleteRoot();
      mintime = tmin->value;
      Vec3i pmin = m_time.indexToPos(tmin-&m_time[0]);
      
      // set the smallest to ACCEPTED
      labelNb.moveTo(pmin);
      labelNb.value() = m_accepted_value;
      
      // update accepted counters
      nrAccepted++;
      sumAccepted += this->m_input->at(pmin);
      
      // update front counters
      nrFront--;
      sumFront -= this->m_input->at(pmin);
      
      Vec3i pos;
      
      // examine the neighbors
      labelNb.firstnbbox();
      do {
	pos = labelNb.getNeighborPosition();
	current_time = computeTime(pos);
	
	// FARAWAY neighbor, move to TRIAL set
	if( labelNb.valuenb() == FARAWAY ) {
	  labelNb.valuenb() = TRIAL;
	  
	  m_time.at(pos) = fElem((float)current_time);
	  Ht.insert(&(m_time.at(pos))); 
	  
	  // update front counters
	  nrFront++;
	  sumFront += this->m_input->at(pos);
	}
	
	// TRIAL neighbor, update time and restore heap
	else if( labelNb.valuenb() == TRIAL ) {
	  
	  fElem *t = &(m_time.at(pos));
	  
	  // update only if current_time is smaller
	  if( current_time < t->value ) {
	    t->value = (float)current_time;
	    Ht.upHeap(t->index);
	  }
	  
	}
      } while( labelNb.nextnbbox() );
      
      
      // log the front and accepted values      
      if( nrFront > 0 ) {
	m_time_log.push_back((float)mintime);
	m_front_log.push_back(float(sumFront)/nrFront);
	m_accepted_log.push_back(float(sumAccepted)/nrAccepted);
      }
    }
    
    cout << "FastMarchingFilter: Main loop finished!" << endl;
    cout << "FastMarchingFilter: Time      = " << current_time << endl;
    cout << "FastMarchingFilter: Ht.size() = " << Ht.size() << endl;
    
    // generate time image (m_output)
    this->m_output->resize( *(this->m_input) );
    this->m_output->setLabel("FastMarching");
    for(int j=0; j<m_time.totalSize(); j++) {
      if( m_trial->at(j)==m_accepted_value )
	(*(this->m_output))[j] = m_time[j].value;
      else
	(*(this->m_output))[j] = (float)mintime;
    }
    cout << "FastMarchingFilter: Finished!" << endl;  
  }
  
} // end namespace wish

#endif
