///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_FILTER_OBJECT_HH_
#define _WISH_FILTER_OBJECT_HH_

#include "wishImage.hh"
#include "wishBoundingBox.hh"
#include "wishVoxelIterator.hh"

namespace wish {
  
  //! abstract base class
  template<class InputImage, class OutputImage>
  class FilterObject {

    //! Base class for filters.
    //! There is an update function for each filter: live_progress_func
    //! which you can use to output timings. Then you need to call
    //! the updateLiveProgress in the filters process function.
    //! Note that you should, in the beginning of process(), use
    //! the function resize to allocate memory for the output image
    //! in order to provide the right dimensions AND spacing.
    //!     


  public:
    
    //! typedefs
    typedef auto_ptr< FilterObject<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    FilterObject() : 
      m_input(0),
      m_output(0),
      m_output_owner(true),
      m_use_box(false),
      m_process_time(-1.0),
      m_live_progress(0.0),
      m_live_progress_func(0)
    {
    }
    
    //! destructor
    virtual ~FilterObject() {
      if( m_output && m_output_owner ) {
	delete m_output;
      }
    }
    
    //! setInputAndOutput
    void setInputAndOutput(InputImage *_input, 
			   OutputImage *_output) {
      m_input = _input;
      
      if( m_output && m_output_owner )
	delete m_output;
      
      m_output = _output;
      m_output_owner = false;

      if( !m_use_box )
	resetBox();
    }
    
    //! setInputAndOutput
    void setInputAndOutput(const InputImagePointer &_input, 
			   const OutputImagePointer &_output) {
      setInputAndOutput(_input.get(),_output.get());
    }
    
    //! set input
    void setInput(InputImage *_input) {
      m_input = _input;
      
      if( !m_use_box )
	resetBox();
    }
    
    //! set input
    void setInput(const InputImagePointer &_input) {
      setInput(_input.get());
    }
    
    //! set input
    void setOutput(OutputImage *_output) {
      if( m_output && m_output_owner )
	delete m_output;
      
      m_output = _output;
      m_output_owner = false;
    }
    
    //! set input
    void setOutput(const OutputImagePointer &_output) {
      setOutput(_output.get());
    }
    
    //! set the bounding box
    void setBox(const BoundingBox &_box) {
      
      if( m_input ) {
	// no weird bounding boxes!
	assert( _box.start.x >= 0 && _box.start.y >= 0 && _box.start.z >= 0 );
	assert( _box.start.x < m_input->width() && 
		_box.start.y < m_input->height() && 
		_box.start.z < m_input->depth() );
	assert( _box.stop.x >= 0 && 
		_box.stop.y >= 0 && 
		_box.stop.z >= 0 );    
	assert( _box.stop.x < m_input->width()  && 
		_box.stop.y < m_input->height() && 
		_box.stop.z < m_input->depth() );
      }
      
      m_box = _box;
      m_use_box = true;
    }
    
    //! set the bounding box
    void setBox(const Vec3i &start, 
		const Vec3i &stop) {
      setBox(BoundingBox(start,stop));
    }
    
    //! reset box
    void resetBox() {
      if( m_input ) {
	setBox(Vec3i(0,0,0), 
	       Vec3i(m_input->width()-1,
		     m_input->height()-1,
		     m_input->depth()-1));
      }
      else
	setBox(BoundingBox());
      
      m_use_box = false;
    }
    
    //! execute if the input and output is set
    virtual void execute() {
      if( m_input ) {
	if( !m_output ) {
	  m_output = new OutputImage;
	  m_output_owner = true;
	}
	double t0=util::timer(); 
	updateLiveProgress(0.0);
	process();
	m_process_time = util::timer()-t0;
	updateLiveProgress(100.0);
      }
    }
    
    //! release output ownership and return output (use to create an auto_ptr)
    OutputImage* releaseOutput() {
      if( m_output_owner ) {
	m_output_owner = false;
	OutputImage *tmp = m_output;
	m_output = 0;
	return tmp;
      }
      else
	return 0;
    }
    
    //! gets the output
    OutputImage* getOutput() const {
      return m_output;
    }
    
    //! gets the input
    InputImage* getInput() const {
      return m_input;
    }
    
    //! set / get output ownership
    bool getOutputOwner() const {
      return m_output_owner;
    }
    
    //! set / get output ownership
    void setOutputOwner(bool b) {
      m_output_owner = b;
    }
    
    //! get process time
    double getProcessTime() const {
      return m_process_time;
    }
    
    //! set live progress func
    void setLiveProgressFunc( void (*func)(double t) ) {
      m_live_progress_func = func;
    }
    

    
  protected:

    //! update progress
    void updateLiveProgress(double t) {
      m_live_progress = t;
      
      if( m_live_progress_func )
	m_live_progress_func(t);
    }
    
    //! process filter    
    virtual void process() = 0;  
    
    InputImage *m_input;
    OutputImage *m_output;
    bool m_output_owner;
    BoundingBox m_box;
    bool m_use_box;
    double m_process_time;
    double m_live_progress;
    void (*m_live_progress_func)(double t);
  };
  
} // end namespace wish

#endif
