///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#if defined(WIN32) || defined(__WIN32__) // Windows specific so far

#include "wishFreeImageReader.hh"

namespace wish {
  
  /** Generic image loader
      @param lpszPathName Pointer to the full file name
      @param flag Optional load flag constant
      @return Returns the loaded dib if successful, returns NULL otherwise
  */
  FIBITMAP* FreeImageReader::GenericLoader(const char* lpszPathName, 
					   int flag) {
    FREE_IMAGE_FORMAT fif = FIF_UNKNOWN;
    
    // check the file signature and deduce its format
    // (the second argument is currently not used by FreeImage)
    fif = FreeImage_GetFileType(lpszPathName, 0);
    if(fif == FIF_UNKNOWN) {
      // no signature ?
      // try to guess the file format from the file extension
      fif = FreeImage_GetFIFFromFilename(lpszPathName);
    }
    // check that the plugin has reading capabilities ...
    if((fif != FIF_UNKNOWN) && FreeImage_FIFSupportsReading(fif)) {
      // ok, let's load the file
      FIBITMAP *dib = FreeImage_Load(fif, lpszPathName, flag);
      // unless a bad file format, we are done !
      return dib;
    }
    return 0;
  }
  
  bool FreeImageReader::readHeader() {
    
    // first we generate a filelist from the filename (can be a pattern)
    size_t slashpos = m_filename.rfind("\\");
    m_path = ".\\";
    if( slashpos != string::npos ) 
      m_path = m_filename.substr(0,slashpos+1);
    m_pattern = m_filename.substr(slashpos+1, string::npos);
        
    // issue the dir command to get a file list (this is not nice...)
    string tempfile = "FREEIMAGE_READER_TEMPORARY_FILE.txt";
    util::systemCommand("dir /B /ON " + m_path+m_pattern + " > " + tempfile); 
    
    // store the filenames in the list
    ifstream is( tempfile.c_str() );
    string tmp;
    getline(is, tmp);
    int filecount = 0;
    m_filelist.clear();
    while( tmp.size() > 0 ) {
      filecount++;
      m_filelist.push_back(tmp);
      getline(is, tmp);
    }
    
    // close and delete the temporary file
    is.close();
    util::systemCommand("del " + tempfile);
    
    // check if there were any files
    if( filecount == 0 ) {
      cerr << "FreeImageReader: No image(s) to read!" << endl;
      return false;
    }
    
    // load first image to get dimensions
    FIBITMAP *fibm = GenericLoader((m_path+m_filelist[0]).c_str(), 0);
    if( fibm == 0 ) {
      cerr << "FreeImageReader: Could not read first file " << m_filelist[0]
	   << endl;
      FreeImage_Unload(fibm);      
      return false;
    }
    
    // parameters
    int w = FreeImage_GetWidth(fibm);
    int h = FreeImage_GetHeight(fibm);
    int bpp = FreeImage_GetBPP(fibm);
    
    // image type
    FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(fibm);
    
    // unload
    FreeImage_Unload(fibm);
    
    // find datatype
    string typestring;
    switch( image_type ) {
    case FIT_BITMAP:
      if( bpp == 8 )
	typestring = "uint8";
      else if( bpp == 24 )
	typestring = "RGB8";
      else if( bpp == 32 )
	typestring = "RGBA8";
      else {
	cerr << "FreeImageReader: Unsupported image, bpp=" << bpp << endl;
	return false;
      }
      break;
    case FIT_UINT16:
      typestring = "uint16";
      break;
    default:
      cerr << "FreeImageReader: Unsupported image, bpp=" << bpp << endl;
      return false;
    }
    
    // set the info
    this->m_info.dim = Vec3i(w,h,filecount);
    this->m_info.spc = Vec3f(1.0,1.0,1.0); // default spacing
    this->m_info.datatype = typestring;
    
    return true;
  }
  
  template<class T>
  bool FreeImageReader::readScalarSequence(Image<T> *im) {
    cerr << "ScalarSequence" << endl;

    im->allocate(this->m_info.dim);
    im->setSpacing(this->m_info.spc);
    im->setInfo(this->m_info.infostr);
    
    // read the sequence
    for(int z=0; z<nFiles(); z++) {
      FIBITMAP *slc = GenericLoader((m_path+m_filelist[z]).c_str(), 0);
      if( slc==0 ) {
	cerr << "FreeImageReader: Failed to load " 
	     << (m_path+m_filelist[z]) << endl;
	return false;
      }
      
      // parameters
      int w = FreeImage_GetWidth(slc);
      int h = FreeImage_GetHeight(slc);
      
      // check dimensions
      if( w != m_info.dim.x || h != m_info.dim.y ) {
	cerr << "FreeImageReader: Dimension mismatch" << endl;
	FreeImage_Unload(slc);
	return false;
      }
      
      // copy bitmap to our buffer
      for(int y=0; y<h; y++) {
	T *bits = (T*) FreeImage_GetScanLine(slc,y);
	for(int x=0; x<w; x++) {
	  (*im)(x,h-1-y,z) = bits[x];
	}
      }
      
      // unload the bitmap
      FreeImage_Unload(slc);	
    }

    return true;
    
  }

  template<class T>
  bool FreeImageReader::readColorSequence(Image<T> *im) {
    cerr << "ColorSequence" << endl;
	
    m_info.print();

    im->allocate(this->m_info.dim);
    im->setSpacing(this->m_info.spc);
    im->setInfo(this->m_info.infostr);
    
    // read the sequence
    for(int z=0; z<nFiles(); z++) {
      FIBITMAP *slc = GenericLoader((m_path+m_filelist[z]).c_str(), 0);
      if( slc==0 ) {
	cerr << "FreeImageReader: Failed to load " 
	     << (m_path+m_filelist[z]) << endl;
	return false;
      }
      
      // parameters
      int w = FreeImage_GetWidth(slc);
      int h = FreeImage_GetHeight(slc);
      int pitch = FreeImage_GetPitch(slc);
      
      // check dimensions
      if( w != m_info.dim.x || h != m_info.dim.y ) {
	cerr << "FreeImageReader: Dimension mismatch" << endl;
	FreeImage_Unload(slc);
	return false;
      }
      
      // copy bitmap to our buffer
      if( im->getDataType() == "RGB8" ) {
	BYTE *bits = (BYTE*) FreeImage_GetBits(slc);
	for(int y=0; y<h; y++, bits+=pitch) {
	  BYTE *pixel = (BYTE*) bits;
	  for(int x=0; x<w; x++, pixel+=3) {
	    (*im)(x,h-1-y,z)[0] = pixel[FI_RGBA_RED];
	    (*im)(x,h-1-y,z)[1] = pixel[FI_RGBA_GREEN];
	    (*im)(x,h-1-y,z)[2] = pixel[FI_RGBA_BLUE];
	  }
	}
      }
      
      else if( im->getDataType() == "RGBA8" ) {
	BYTE *bits = (BYTE*) FreeImage_GetBits(slc);
	for(int y=0; y<h; y++, bits+=pitch) {
	  BYTE *pixel = (BYTE*) bits;
	  for(int x=0; x<w; x++, pixel+=4) {
	    (*im)(x,h-1-y,z)[0] = pixel[FI_RGBA_RED];
	    (*im)(x,h-1-y,z)[1] = pixel[FI_RGBA_GREEN];
	    (*im)(x,h-1-y,z)[2] = pixel[FI_RGBA_BLUE];
	    (*im)(x,h-1-y,z)[3] = pixel[FI_RGBA_ALPHA];
	  }
	}
      }
      
      // unload the bitmap
      FreeImage_Unload(slc);	
    }
    
    return true;
    
  }
  
  bool FreeImageReader::readData() {
    if( this->m_info.datatype == "uint8" ) {
      Image<uint8> *tmp = new Image<uint8>;
      if( readScalarSequence(tmp) )
	this->m_output = tmp;
      else
	return false;
    }
    else if( this->m_info.datatype == "uint16" ) {
      Image<uint16> *tmp = new Image<uint16>;
      if( readScalarSequence(tmp) )
	this->m_output = tmp;
      else
	return false;    
    }
    else if( this->m_info.datatype == "RGB8" ) {
      Image<RGB8> *tmp = new Image<RGB8>;
      if( readColorSequence(tmp) )
	this->m_output = tmp;
      else
	return false;    
    }
    else if( this->m_info.datatype == "RGBA8" ) {
      Image<RGBA8> *tmp = new Image<RGBA8>;
      if( readColorSequence(tmp) )
	this->m_output = tmp;
      else
	return false;    
    }
    else { // should not happen
      this->printSelf();
      cerr << ": Error reading file " << m_filename << ", bad datatype: " 
	   << this->m_info.datatype << endl;
      return false;
    }
    
    return true;
  }
  
}

#endif // Windows specific
