///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_FUZZYCONNECTEDNESS_FILTER_HH_
#define _WISH_FUZZYCONNECTEDNESS_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishMaxHeap.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  //! Fuzzy Connectedness filter
  //! We have reused the FastMarching code
  template<class InputImage, class OutputImage>
  class FuzzyConnectednessFilter : 
    public FilterObject<InputImage, OutputImage> {
    
  public:
    
    struct SeedStatistics {
      double mean;
      double var;
      double sd;
    
      double bgmean;
      double bgvar;
      double bgsd;
    };
    
    //! pointer typedef
    typedef auto_ptr< FuzzyConnectednessFilter<InputImage,OutputImage> > 
    Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! typedef for heap element
    typedef HeapElement<float> fElem;
    typedef MaxHeap<float> fMaxHeap;
    
    //! typedef for seed iterator
    typedef vector<Vec3i>::iterator SeedsIterator;
    
    //! constructor
    FuzzyConnectednessFilter() :
      m_seeds(vector<Vec3i>(0)),
      m_trial(0),
      m_trial_owner(true),
      m_affinity_function(0),
      m_early_treshold(0.0),
      m_input_is_affinity(false),
      m_invert_affinity(false),
      m_connectedness_log(vector<float>(0)),
      m_front_log(vector<float>(0)),
      m_accepted_log(vector<float>(0)),
      m_log_values(false)
    {}
    
    ~FuzzyConnectednessFilter() {
      if( m_trial && m_trial_owner )
	delete m_trial;
    }
    
    //! set the seeds vector
    void setSeeds(const vector<Vec3i> &_seeds) {
      m_seeds = _seeds;
    }
    
    //! add a seed to the seeds vector
    void addSeed(const Vec3i &_seed) {
      m_seeds.push_back(_seed);
    }
    
    //! use a binary image to set the seeds vector
    void setSeedsFromImage(Image<uint8> *_seed_image);
    
    void setSeedsFromImage(const Image<uint8>::Pointer &_seed_image) {
      initializeFromImage(_seed_image.get());
    }
    
    //! set affinity function
    void setAffinityFunction(double (*f)(const Vec3i &a, const Vec3i &b)) {
      m_affinity_function = f;
    }
    
    //! clear all seeds
    void clearSeeds() {
      m_seeds.resize(0);
    }
    
    //! set the trial image (specify output)
    void setTrialImage(Image<uint8> *_tim) {
      if( m_trial && m_trial_owner )
	delete m_trial;
      
      m_trial = _tim;
      m_trial_owner = false;
    }
    
    void setTrialImage(const Image<uint8>::Pointer &_tim) {
      this->setTrialImage(_tim.get());
    }
    
    //! release trial ownership and return trial (use to create an auto_ptr)
    Image<uint8>* releaseTrial() {
      if( m_trial_owner ) {
	m_trial_owner = false;
	Image<uint8> *tmp = m_trial;
	m_trial = 0;
	return tmp;
      }
      else
	return 0;
    }
    
    //! gets the trial image
    Image<uint8>* getTrial() {
      return m_trial;
    }
    
    //! set early threshold (default 0.0)
    void setEarlyThreshold(const double &et) {
      m_early_treshold = et;
    }
    
    //! set if the input image should be treated as affinity
    void setInputIsAffinity(bool b) {
      m_input_is_affinity = b;
    }
    
    //! set if we should invert affinity (if input is cost)
    void setInvertAffinity(const bool &b) {
      m_invert_affinity = b;
    }
    
    //! get the time log
    const vector<float>& getConnectednessLog() const {
      return m_connectedness_log;
    }
    
    //! get the front log
    const vector<float>& getFrontLog() const {
      return m_front_log;
    }
    
    //! get the accepted log
    const vector<float>& getAcceptedLog() const {
      return m_accepted_log;
    }
    
    //! set whether to log values
    void setLogValues(const bool &b) {
      m_log_values = b;
    }
    
  protected:
    
    //! the connectedness image (contains heap elements)
    //! we also have m_output that contains only the connectedness
    ImageBase<fElem> m_connectedness;

    //! the seed vector
    vector<Vec3i> m_seeds;
    
    //! affinity function
    double (*m_affinity_function)(const Vec3i &a, const Vec3i&b);
    
    //! the trial image
    Image<uint8> *m_trial;
    bool m_trial_owner;
        
    //! log the connectedness, front, and accepted values
    vector<float> m_connectedness_log;
    vector<float> m_front_log;
    vector<float> m_accepted_log;
    bool m_log_values;
    
    //! early threshold
    double m_early_treshold;

    //! input is affininty
    bool m_input_is_affinity;
    
    //! invert affinity (if input is cost)
    bool m_invert_affinity;
    
    //! seed statistics
    SeedStatistics m_seedstats;
    
    //! static constants
    static const uint8 FARAWAY;
    static const uint8 TRIAL;
    static const uint8 ACCEPTED;
    
    virtual void process();

    virtual double computeAffinity(const Vec3i &pos, 
				   const Vec3i &nbpos);
    
  private:
    //! purposely not implemented
    FuzzyConnectednessFilter(const FuzzyConnectednessFilter&);
    FuzzyConnectednessFilter& operator=(const FuzzyConnectednessFilter&);
  };
  
  //! constants
  template<class InputImage, class OutputImage>
  const uint8 FuzzyConnectednessFilter<InputImage, OutputImage>::
  FARAWAY = uint8(0);
  
  template<class InputImage, class OutputImage>
  const uint8 FuzzyConnectednessFilter<InputImage, OutputImage>::
  TRIAL = uint8(255);
  
  template<class InputImage, class OutputImage>
  const uint8 FuzzyConnectednessFilter<InputImage, OutputImage>::
  ACCEPTED = uint8(1);
  
  //! Fill the seeds vector with voxels from the seed image.
  //! So far we consider all voxels > 0 as seed points,
  //! maybe this should be changed later so that we can have several regions
  //! growing simultaneously. Note that the seed vector is cleared! 
  template<class InputImage, class OutputImage>
  void FuzzyConnectednessFilter<InputImage, OutputImage>::
  setSeedsFromImage(Image<uint8> *_seed_image) {
    m_seeds.clear();
    for(int i=0; i<_seed_image->totalSize(); i++) {
      if( (*_seed_image)[i] > 0 ) {
	m_seeds.push_back(_seed_image->indexToPos(i));
      }
    }
  }
  

  //! compute the affinity between spels at locations pos and nbpos
  //!
  //! only face neighbors (6 nb) are sent to this function.
  template<class InputImage, class OutputImage>
  double FuzzyConnectednessFilter<InputImage, OutputImage>::
  computeAffinity(const Vec3i &pos, const Vec3i &nbpos) {
    
    // if an external function is specified
    if( m_affinity_function )
      return m_affinity_function(pos,nbpos);
    
    // if input image is affinity
    else if( m_input_is_affinity ) {
      double a = this->m_input->getNormalized(pos);
      double b = this->m_input->getNormalized(nbpos);
      
      if( !m_invert_affinity ) 
	return util::minOf(a,b);
      else // input is cost, invert it 
	return util::minOf(1.0f-a,1.0f-b);
    }
    
    // use seed statistics and distance between voxels
    else {
      double fa = m_input->getAsDouble( pos );
      double fb = m_input->getAsDouble( nbpos );
      
      // adjacency
      double adj = 1.0;
      
      // use this to punish z-steps
      //if( abs(pos.z-nbpos.z) > 0 ) {
      //	adj = m_input->getSpacing().x/m_input->getSpacing().z;
      //}
      
      // affinity
      double aff=adj;
      
      // homogeneity comp.
      aff *= exp( -0.5*util::square( (fa-fb)/(4.0*m_seedstats.var) ) );
      
      // object comp.
      double ave = 0.5*(fa+fb);
      aff *= exp(-0.5*util::square(ave-m_seedstats.mean)/(4.0*m_seedstats.var));
      return aff; 
    }
    
  }
  
  //! the computation
  template<class InputImage, class OutputImage>
  void FuzzyConnectednessFilter<InputImage, OutputImage>::process() {
    
    if( m_seeds.size()==0 ) {
      cerr << "No seeds specified!" << endl;
      return;
    }
    
    // compute seed statistics
    m_seedstats.mean = 0.0;
    m_seedstats.var = 0.0;
    m_seedstats.sd = 0.0;
    
    for( size_t i=0; i<m_seeds.size(); i++ ) {
      double tmp = m_input->getAsDouble( m_seeds[i] );
      m_seedstats.mean += tmp;
      m_seedstats.var += tmp*tmp;
    }
    m_seedstats.mean /= m_seeds.size();
    m_seedstats.var -= m_seeds.size()*util::square( m_seedstats.mean );
    m_seedstats.var /= m_seeds.size();
    if( m_seedstats.var < 1e-3 ) 
      m_seedstats.var = 1.0;
    
    m_seedstats.sd = sqrt( m_seedstats.var );
    
    cerr << "Mean: " << m_seedstats.mean << endl;
    cerr << "SD:   " << m_seedstats.sd << endl;
    cerr << "Var:  " << m_seedstats.var << endl;
    
    // trial image
    if( !m_trial ) {
      m_trial = new Image<uint8>;
      m_trial_owner = true;
    }
    
    m_trial->resize( *(this->m_input) );
    m_trial->reset( FARAWAY );
    
    // connectedness
    m_connectedness.resize( (*(this->m_input)) );
    m_connectedness.reset( fElem(0.0) );
    
    // logs
    m_connectedness_log.resize(0);
    m_front_log.resize(0);
    m_accepted_log.resize(0);
    
    // dimensions
    int W = this->m_input->width();
    int H = this->m_input->height();
    int D = this->m_input->depth();
    
    // counters
    int nrAccepted = 0;
    int nrFront = 0;
    double sumAccepted = 0.0;
    double sumFront = 0.0;
    
    // the heap
    fMaxHeap Q((W*H*D)/1);
    
    // select neighborhood depending on dimension
    NeighborhoodType nt = C6;
    int dim = this->m_input->getDim();
    if( dim == 2 )
      nt = C4;
    
    // iterator over the neighborhood
    NeighborhoodIterator< Image<uint8> > labelNb(m_trial, nt);
    
    // set the box
    if( this->m_use_box )
      labelNb.setBox(this->m_box);
    else
      labelNb.computeNeighborhoodBox();
    
    cout << "FuzzyConnectednessFilter: Initializing heap..." << endl;
    
    for(SeedsIterator it=m_seeds.begin();
	it!=m_seeds.end();
	it++) {
      
      Vec3i current = *it;
      
      // set the seeds to ACCEPTED
      m_trial->at(current) = ACCEPTED;
      
      // step accepted counter and sum
      nrAccepted++;
      sumAccepted += computeAffinity(current,current);
      
      // set connectedness to one
      m_connectedness.at(current) = fElem( 1.0f );
    }
    
    // set the seeds neighbors to TRIAL and compute affinity
    labelNb.first();
    do {
      
      // if accepted
      if( labelNb.value() == ACCEPTED ) {
	
	Vec3i pos = labelNb.getPosition();

	labelNb.firstnbbox();
	do {
	  
	  // find surface voxels and add them to trial
	  Vec3i nbpos = labelNb.getNeighborPosition();
	  if( labelNb.valuenb() == FARAWAY ) {
	    labelNb.valuenb() = TRIAL;
	    m_connectedness.at(nbpos) = fElem((float)computeAffinity(pos,nbpos));
	    Q.insert(&(m_connectedness.at(nbpos)));
	    
	    // update front counter and sum
	    nrFront++;	    
	    sumFront += computeAffinity(nbpos,nbpos);
	  }
	  
	  // if the voxel already is TRIAL, check the connectedness 
	  // and update if necessary
	  else if( labelNb.valuenb() == TRIAL ) {
	    double fval = util::minOf( m_connectedness.at(pos).value,
				       computeAffinity(pos,nbpos) );
	    if( fval < m_connectedness.at(nbpos).value ) {
	      m_connectedness.at(nbpos).value = (float)fval;
	      Q.upHeap(m_connectedness.at(nbpos).index);
	    }
	  }
	} while( labelNb.nextnbbox() );
	
      }
    } while ( labelNb.next() );
    
    cout << "FuzzyConnectednessFilter: Initializing procedure finished!" 
	 << endl;
    
    double maxconn = 1.0, current_conn = 0.0;
    int nrpops=0;
    
    // main loop
    double ethresh = m_early_treshold;
    if( ethresh < 0.0 ) ethresh = 0.0;
    if( ethresh > 1.0 ) ethresh = 1.0; 
    while ( !(Q.empty()) && maxconn>=ethresh ) { 
      //&& mintime <= m_time_max ) {
      
      // find largest
      fElem *cmax = Q.deleteRoot();
      nrpops++;
      maxconn = cmax->value;
      
      Vec3i pmax = m_connectedness.indexToPos(int(cmax-&m_connectedness[0]));
      
      // set the largest to ACCEPTED
      labelNb.moveTo(pmax);
      labelNb.value() = ACCEPTED;
      
      // update accepted counter
      double tmpaff = computeAffinity(pmax,pmax);
      nrAccepted++;
      sumAccepted += tmpaff;
	    
      // update front counter
      nrFront--;
      sumFront -= tmpaff;

      Vec3i nbpos;
      
      // examine the neighbors
      labelNb.firstnbbox();
      do {
	nbpos = labelNb.getNeighborPosition();
	
	if( labelNb.valuenb() == ACCEPTED )
	  continue;
	
	current_conn = computeAffinity(pmax,nbpos);
	
	double fval = util::minOf( current_conn, 
				   m_connectedness.at(pmax).value );
	if( fval > m_connectedness.at(nbpos).value ) {
	  m_connectedness.at(nbpos).value = (float)fval;
	  
	  if( labelNb.valuenb() == FARAWAY ) {
	    Q.insert(&(m_connectedness.at(nbpos)));
	    labelNb.valuenb() = TRIAL;	  
	    
	    // update front counters
	    nrFront++;
	    sumFront += computeAffinity(nbpos,nbpos);
	  }
	  else if( labelNb.valuenb() == TRIAL )
	    Q.upHeap(m_connectedness.at(nbpos).index);
	  else
	    cerr << "Should not happen!" << endl; // ACCEPTED
	}
	
      } while( labelNb.nextnbbox() );
      
      // log the front and accepted values      
      if( nrFront > 0 && m_log_values ) {
	m_connectedness_log.push_back((float)maxconn);
	m_front_log.push_back(float(sumFront)/nrFront);
	m_accepted_log.push_back(float(sumAccepted)/nrAccepted);
      }
    }
    
    cout << "FuzzyConnectednessFilter: Main loop finished!" << endl;
    cout << "FuzzyConnectednessFilter: nrpops = " << nrpops << endl;
    
    // generate connectivity image (m_output)
    this->m_output->resize( *(this->m_input) );
    this->m_output->setLabel( this->m_input->getLabel() + "_FC" );
    for(int j=0; j<m_connectedness.totalSize(); j++) {
      if( m_trial->at(j) == ACCEPTED )
	this->m_output->setNormalizedToArray( m_connectedness[j].value, j );
      else
	this->m_output->setNormalizedToArray( 0.0, j );
    }
    
    cout << "FuzzyConnectednessFilter: Finished!" << endl;  
  }
  
} // end namespace wish

#endif
