///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_LIGHT_SOURCE_HH_
#define _WISH_LIGHT_SOURCE_HH_

#include "wish.hh"
#include "wishutil.hh"
#include "wishLinalg.hh"
#include "wishColor.hh"

#include <GL/gl.h>

namespace wish {
  
  //! wrapper class for a gl light
  class LightSource {
  protected:
    GLenum  gl_light;
    Color4f ambient;
    Color4f diffuse;
    Color4f specular;
    Vec4f   position;
    Vec4f   direction;      
    float   exponent;
    float   cutoff;
    float   constant_attenuation;
    float   linear_attenuation;
    float   quadratic_attenuation;
    
  public:
    void setLight(GLenum light) {
      gl_light = light;
      reset();
      enable();
    }
    
    void setAmbient(const Color4f &col) {
      ambient = col;
      glLightfv(gl_light, GL_AMBIENT, ambient.getPointer());
    }
      
    void setDiffuse(const Color4f &col) {
      diffuse = col;	
      glLightfv(gl_light, GL_DIFFUSE, diffuse.getPointer());
    }
      
    void setSpecular(const Color4f &col) {
      specular = col;
      glLightfv(gl_light, GL_SPECULAR, specular.getPointer());
    }
      
    void setPosition(const Vec4f &pos) {
      position = pos;
      glLightfv(gl_light, GL_POSITION, position.getPointer());
    }
      
    void setDirection(const Vec4f &d) {
      direction = d;
      glLightfv(gl_light, GL_SPOT_DIRECTION, direction.getPointer());
    }
      
    // valid is [0,128]
    void setExponent(float e) {
      exponent = e;
      glLightf(gl_light, GL_SPOT_EXPONENT, exponent);
    }
      
    // valid is [0,90] and 180
    void setCutoff(float c) {
      cutoff = c;
      glLightf(gl_light, GL_SPOT_CUTOFF, cutoff);
    }
      
    // 1.0 / (ca + la*d + qa*d*d), d = distance between light and vertex
    void setAttenuation(float ca, float la, float qa) {
      setConstantAttenuation(ca);
      setLinearAttenuation(la);
      setQuadraticAttenuation(qa);
    }
      
    void setConstantAttenuation(float ca) {
      constant_attenuation = ca;
      glLightf(gl_light, GL_CONSTANT_ATTENUATION, constant_attenuation);
    }
      
    void setLinearAttenuation(float la) {
      linear_attenuation = la;
      glLightf(gl_light, GL_LINEAR_ATTENUATION, linear_attenuation);
    }
      
    void setQuadraticAttenuation(float qa) {
      quadratic_attenuation = qa;
      glLightf(gl_light, GL_QUADRATIC_ATTENUATION, quadratic_attenuation);
    }
      
    void reset() {
      // these are OpenGL default values
      setAmbient(Color4f(0.0, 0.0, 0.0, 1.0));
      setDiffuse(Color4f(1.0, 1.0, 1.0, 1.0));
      setSpecular(Color4f(1.0, 1.0, 1.0, 1.0));    
      setPosition(Vec4f(0.0, 0.0, 1.0, 0.0));
      setDirection(Vec4f(0.0, 0.0, -1.0, 0.0));	
      setExponent(0.0);
      setCutoff(180.0);
      setAttenuation(1.0, 0.0, 0.0);
    }
      
    // only enables/disables this lightsource
    // you must take care of GL_LIGHTING yourself
    void enable() {
      if( glIsEnabled(gl_light) == GL_FALSE )
	glEnable(gl_light);
    }
      
    void disable() {
      if( glIsEnabled(gl_light) == GL_TRUE )
	glDisable(gl_light);
    }
      
    LightSource() {
    }
      
    LightSource(GLenum light) {
      gl_light = light;
      reset();
      enable();
    }
  };

} // namespace wish

#endif
