///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GL_LIVE_MARCHING_HH_
#define _WISH_GL_LIVE_MARCHING_HH_

#include "wishImage.hh"
#include "wishLiveMarching.hh"
#include "wishLinalg.hh"

#include <GL/glut.h>

namespace wish {
  
  class GLLiveMarching {
    
  private:
    //! constructor
    GLLiveMarching() {}
    
    //! purposely not implemented
    GLLiveMarching(const GLLiveMarching &);
    GLLiveMarching& operator=(const GLLiveMarching &);
    
  public:
    
    //! typedefs
    typedef auto_ptr<GLLiveMarching> Pointer;
    typedef ImageObject::Pointer ImageObjectPointer;
    
    enum MenuEntry {MENU_LOAD, 
		    MENU_SAVE, 
		    MENU_CLOSE, 
		    MENU_RESET_WINDOW, 
		    MENU_QUIT};
    
    //! get the one and only instance
    static GLLiveMarching* getInstance();
    
    //! function for checking errors
    static void errorCheck() {
      GLenum err = glGetError();
      if( err ) cerr << gluErrorString( err ) << endl;
    }
    
    //! add input 
    void setInput(ImageObject *_input) {
      m_input.push_back(_input);
      
      m_slice.push_back(0);
      m_current = (int) m_input.size()-1;
    }
    
    //! add cost 
    void setCostInput(ImageObject *_input) {
      m_lm.setInput(_input);
    }
    
    void render();
    
  protected:
    
    //! the data type (GL_UNSIGNED_BYTE, GL_UNSIGNED_SHORT, ...)
    GLenum m_gl_datatype;
    
    //! the storage format (GL_LUMINANCE, GL_RGB, ...)
    GLenum m_gl_storage;
    
    //! the images
    vector<ImageObject*> m_input;
    
    //! slice
    vector<int> m_slice;
    
    //! current image
    int m_current;
    
    //! buffer pointer
    byte *m_buffer_pointer;
    
    //! image dimensions
    int m_width;
    int m_height;
    
    // window dimensions
    int   m_win_width;
    int   m_win_height;
    float m_win_aspect;
    
    // probe position
    Vec3i m_win_pos;
    Vec3i m_cursor;
    Vec3d m_normalized_cursor;
    
    // the live marching object
    LiveMarching m_lm;
    
    // choose which image to view
    // 1 = original, 2 = cost, 3 = trial
    int m_displayimage;
    
    // contrast and brightness
    float m_contrast;
    float m_brightness;
    float m_scale;
    float m_bias;
    
    //! member functions for callbacks
    void m_display();
    void m_reshape(int w, int h);
    void m_keyboard(unsigned char key, int x, int y);
    void m_special_keys(int key, int x, int y);
    void m_mouse(int button, int state, int x, int y);
    void m_motion(int x, int y);
    void m_passive_motion(int x, int y);
    void m_menu(int value);
    void m_idle();

    //! update
    void m_update();
    
    //! draw the display image
    void m_draw_image();
    
    //! initialize GLLiveMarching
    void m_initialize();
    
    //! static callback functions
    static void display() {
      getInstance() -> m_display();
    }
    static void reshape(int w, int h) {
      getInstance() -> m_reshape(w,h);
    }
    static void keyboard(unsigned char key, int x, int y) {
      getInstance() -> m_keyboard(key,x,y);
    }
    static void special_keys(int key, int x, int y) {
      getInstance() -> m_special_keys(key,x,y);
    }
    static void mouse(int button, int state, int x, int y) {
      getInstance() -> m_mouse(button,state,x,y);
    }
    static void motion(int x, int y) {
      getInstance() -> m_motion(x,y);
    }
    static void passive_motion(int x, int y) {
      getInstance() -> m_passive_motion(x,y);
    }
    static void menu(int value) {
      getInstance() -> m_menu(value);
    }
    static void idle() {
      getInstance() -> m_idle();
    }
    
  };
  
  //! get the one and only instance
  GLLiveMarching* GLLiveMarching::getInstance() {
    static GLLiveMarching the_instance;
    return &the_instance;
  }
  
  void GLLiveMarching::m_update() {
    
    m_gl_storage = GL_LUMINANCE;
    
    m_width = m_input[m_current]->width();
    m_height = m_input[m_current]->height();
    
    if( m_displayimage == 2 ) { // cost
      if( m_lm.getInput()->getDataType() == "uint8" ) {
	m_gl_datatype = GL_UNSIGNED_BYTE;
      }
      else if( m_lm.getInput()->getDataType() == "uint16" )
	m_gl_datatype = GL_UNSIGNED_SHORT;
      else {
	cerr << "GLLiveMarching: Data type error!" << endl;
	exit(1);
      }
      m_buffer_pointer = m_lm.getInput()->getBytePointer();
    }
    
    else if( m_displayimage == 3 ) { // trial
      m_gl_datatype = GL_UNSIGNED_BYTE; // OBS!!! TEMPORARY
      m_buffer_pointer = m_lm.getTrialImage()->getBytePointer();
    }
    
    else { //if( m_displayimage == 1 ) {
      if( m_input[m_current] -> getDataType() == "uint8" ) {
	m_gl_datatype = GL_UNSIGNED_BYTE;
      }
      else if( m_input[m_current] -> getDataType() == "uint16" )
	m_gl_datatype = GL_UNSIGNED_SHORT;
      else {
	cerr << "GLLiveMarching: Data type error!" << endl;
	exit(1);
      }
      m_buffer_pointer = 
	m_input[m_current]->getBytePointer()+
	m_input[m_current]->getVoxelBytes()*
	m_width*m_height*
	m_slice[m_current];
    }
    
    //glutReshapeWindow(m_width,m_height);
  }
  
  // draw the image
  void GLLiveMarching::m_draw_image() {
    
    // draw the image by using the DrawPixels function
    glRasterPos2f(0.0,0.0);
    
    glDrawPixels(m_width,
		 m_height,
		 m_gl_storage,
		 m_gl_datatype,
		 m_buffer_pointer);
  }
  
  // display function
  void GLLiveMarching::m_display() {
    // clear buffer
    glClear(GL_COLOR_BUFFER_BIT);
  
    // draw the image
    m_draw_image();
    
    const vector<Vec3i> &sds = m_lm.getSeeds();
    const vector<Vec3f> &btc = m_lm.getBacktrackCursor();
    const vector< vector<Vec3f> > &bt = m_lm.getBacktrack();
    
    // cursor
    glLineWidth(2.0);    
    glColor3f(0,1,0);
    glBegin(GL_LINE_STRIP);
    for(unsigned int i=0; i<btc.size(); i++) {
      Vec3f v = btc[i];
      glVertex2f(float(v.x)/m_width, float(v.y)/m_height);
    }
    glEnd();
    
    // old
    glColor3f(1,0,0);
    for(unsigned int j=0; j<bt.size(); j++) {
      glBegin(GL_LINE_STRIP);
      for(unsigned int k=0; k<bt[j].size(); k++) {
	Vec3f v = bt[j][k];
	glVertex2f(float(v.x)/m_width, float(v.y)/m_height);
      }
      glEnd();
    }
    
    // seeds
    glPointSize(5.0);
    glColor3f(0,0,1);    
    glBegin(GL_POINTS);
    for(unsigned int l=0; l<sds.size(); l++) {
      Vec3i v = sds[l];
      glVertex2f(float(v.x)/m_width, float(v.y)/m_height);
    }
    glEnd();
    
    // swap buffers
    glutSwapBuffers();
    glFlush();
  }
  
  // reshape function
  // set the window size and the projection
  void GLLiveMarching::m_reshape(int w, int h) {
    glutReshapeWindow(w,h);
    
    m_win_width = w;
    m_win_height = h;
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0,1.0,0.0,1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    // correct coordinate system:
    // top left = (0,0), bottom right = (box_width,box_height)
    glTranslatef(0.5, 0.5, 0.0);
    glRotatef(180.0, 1.0, 0.0, 0.0);
    glTranslatef(-0.5f, -0.5f, 0.0f);
    
    // viewport
    glViewport(0,0,w,h);
    
    // correct drawing
    glPixelZoom( (1.0f*w)/m_width,
		 (-1.0f*h)/m_height );
    
    glutPostRedisplay();
  }
  
  // keyboard callback
  void GLLiveMarching::m_keyboard(unsigned char key, int x, int y) {
    
    x; y;
    
    switch(key) {
      
      // remove seed
    case ' ': // space
      m_lm.removeSeed();
      break;

      // close curve
    case 'c':
    case 'C':
      m_lm.closeCurve();
      break;
      
      // exit
    case 'q':
    case 27: // escape
      exit(0);
      break;
    
    default:
      break;
    }
  
    glutPostRedisplay();  
  }

  
  void GLLiveMarching::m_special_keys(int key, int x, int y) {
    x; y;
    
    switch(key) {
      
      // browse slices if 3D
    case GLUT_KEY_UP:
      m_slice[m_current]++;
      if( m_slice[m_current] >= m_input[m_current]->depth() )
	m_slice[m_current] = m_input[m_current]->depth()-1;
      m_update();
      break;

    case GLUT_KEY_DOWN:
      m_slice[m_current]--;
      if( m_slice[m_current] < 0 )
	m_slice[m_current] = 0;
      m_update();
      break;
      
    case GLUT_KEY_PAGE_UP:
      m_current++;
      if( m_current >= (int) m_input.size() )
	m_current = 0;
      m_update();
      break;
      
    case GLUT_KEY_PAGE_DOWN:
      m_current--;
      if( m_current < 0 )
	m_current = (int) m_input.size()-1;
      m_update();
      break;
      
    case GLUT_KEY_F1:
      cerr << "DISPLAY ORIGINAL" << endl;
      m_displayimage = 1;
      m_update();
      break;
      
    case GLUT_KEY_F2:
      cerr << "DISPLAY COST" << endl;
      m_displayimage = 2;
      m_update();
      break;
      
    case GLUT_KEY_F3:
      cerr << "DISPLAY TRIAL" << endl;
      m_displayimage = 3;      
      m_update();
      break;
      
    default:
      break;
      
    }
    
    glutPostRedisplay();
  }
  
  // mouse callback
  void GLLiveMarching::m_mouse(int button, int state, int x, int y) {
    
    m_win_pos.x = x;
    m_win_pos.y = y;
    
    m_normalized_cursor.x = double(x)/m_win_width;
    m_normalized_cursor.y = double(y)/m_win_height;
    
    m_cursor.x = int( m_normalized_cursor.x*m_width+1 );
    m_cursor.y = int( m_normalized_cursor.y*m_height+1 );
    m_cursor.z = int( m_slice[m_current] );
    
    m_lm.setCursor(m_cursor);
    
    cerr << "MOUSE: " << m_cursor << endl;
    
    if( button == GLUT_LEFT_BUTTON && state == GLUT_DOWN ) {
      m_lm.putSeed();
      glutSetCursor(GLUT_CURSOR_CROSSHAIR);
    }
//     else if( button == GLUT_LEFT_BUTTON && state == GLUT_UP ) {
//       glutSetCursor(GLUT_CURSOR_INHERIT);
//     }
    
    glutPostRedisplay();
  }
  
  void GLLiveMarching::m_idle() {
  }
  
  // mouse motion callback
  void GLLiveMarching::m_motion(int x, int y) {
    x; y;
  }
  
  // passive motion callback
  void GLLiveMarching::m_passive_motion(int x, int y) {
    
    m_win_pos.x = x;
    m_win_pos.y = y;
    
    m_normalized_cursor.x = double(x)/m_win_width;
    m_normalized_cursor.y = double(y)/m_win_height;
    
    m_cursor.x = int( m_normalized_cursor.x*m_width+1 );
    m_cursor.y = int( m_normalized_cursor.y*m_height+1 );
    m_cursor.z = int( m_slice[m_current] );
    
    //cerr << "PASSIVE MOTION: " << m_cursor << endl;
    
    m_lm.setCursor(m_cursor);
    m_lm.update();
    
    glutPostRedisplay();
  }
  
  // menu callback  
  void GLLiveMarching::m_menu(int value) {
    
    switch(value) {
    case MENU_LOAD:
      break;
      
    case MENU_SAVE:	
      break;

    case MENU_CLOSE:
      break;
      
    case MENU_RESET_WINDOW:
      getInstance() -> m_reshape(m_input[m_current]->width(),
				 m_input[m_current]->height());
      break;
      
    case MENU_QUIT:
      exit(0);
      break;
      
    default:
      break;
    }
    
    glutPostRedisplay();
    
  }

  // initialize
  void GLLiveMarching::m_initialize() { 
    
    // set up parameters
    m_win_width = 256;
    m_win_height = 256;
    
    m_displayimage=1;
    
    // clear color
    glClearColor(0,0,0,0);
    
    // tell how the image is stored in memory (for glDrawPixels)
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    
    // init display mode
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    
    // set the window name to the filename
    glutCreateWindow("GLLiveMarching");
    
    // set callbacks
    glutDisplayFunc(GLLiveMarching::display);    
    glutReshapeFunc(GLLiveMarching::reshape);
    glutKeyboardFunc(GLLiveMarching::keyboard);
    glutSpecialFunc(GLLiveMarching::special_keys);
    glutMouseFunc(GLLiveMarching::mouse);
    glutMotionFunc(GLLiveMarching::motion);
    glutPassiveMotionFunc(GLLiveMarching::passive_motion);
    glutIdleFunc(GLLiveMarching::idle);
    
    // init window size
    glutInitWindowSize(256,256);
    
    // set up menu
    glutCreateMenu(menu);
    glutAddMenuEntry("Load image",MENU_LOAD);
    glutAddMenuEntry("Save image",MENU_SAVE);
    glutAddMenuEntry("Close image",MENU_CLOSE);
    glutAddMenuEntry("Reset window size",MENU_RESET_WINDOW);
    glutAddMenuEntry("Quit",MENU_QUIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);


  }
  
  void GLLiveMarching::render() {
    if( m_input[0] ) {
      m_initialize();
      m_update();
      glutMainLoop();
    }
  }
  
} // end namespace wish

#endif
