#ifndef _WISH_GL_TEXTURE_HH_
#define _WISH_GL_TEXTURE_HH_

#include "wishImageObject.hh"

#include <GL/gl.h>

namespace wish {
  
  // texture class
  class Texture {
  protected:
    ImageObject *image;
    GLuint id;
    GLenum target, format, type;
    
    bool repeat_s;
    bool repeat_t;
    bool repeat_r;
    bool linear_interpolation;
    
  public:	
    Texture(ImageObject *_image=0) :
      image(_image),
      id(0),
      repeat_s(false),
      repeat_t(false),
      repeat_r(false),
      linear_interpolation(true)
    {
      initialize();
    }
    
    void setImage(ImageObject *_image) {
      image = _image;
    }
    
    ImageObject* getImage() {
      return image;
    }
    
    const ImageObject* getImage() const {
      return image;
    }
    
    void setRepeatS(bool b) {
      repeat_s = b;
    }

    void setRepeatT(bool b) {
      repeat_t = b;
    }

    void setRepeatR(bool b) {
      repeat_r = b;
    }
    
    void setLinearInterpolation(bool b) {
      linear_interpolation = b;
    }
    
    GLuint getID() const {
      return id;
    }
    
    void bind() const {
      glBindTexture(target, id);
    }
    
    void initialize() {
      
      if(image) {
	// find out data type and dimensions
	int W = image->width();
	int H = image->height();
	int D = image->depth();
	
	if( D>1 )
	  target = GL_TEXTURE_3D;
	
	if( image->getDataType() == "uint8" ) {
	  format = GL_LUMINANCE;
	  type = GL_UNSIGNED_BYTE;
	}
	else if( image->getDataType() == "uint16" ) {
	  format = GL_LUMINANCE;
	  type = GL_UNSIGNED_SHORT;
	}
	else if( image->getDataType() == "RGB8" ) {
	  format = GL_RGB;
	  type = GL_UNSIGNED_BYTE;
	}
	else if( image->getDataType() == "RGBA8" ) {
	  format = GL_RGBA;
	  type = GL_UNSIGNED_BYTE;
	}
	else {
	  cerr << "Unsupported texture format!" << endl;
	}
	
	// generate name
	glGenTextures(1, &id);
	
	// bind
	glBindTexture(target, id);
	
	// prepare environment
	if( repeat_s )
	  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
	else
	  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	if( repeat_t )
	  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
	else
	  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	if( repeat_r )
	  glTexParameteri(target, GL_TEXTURE_WRAP_R, GL_REPEAT);
	else
	  glTexParameteri(target, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	if( linear_interpolation ) {
	  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else {
	  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}
	
	// load
	if( target == GL_TEXTURE_2D ) {
	  glTexImage2D(target,           // target 
		       0,                // level 
		       format,           // internal format
		       W,                // width
		       H,                // height
		       0,                // border width (0 or 1)
		       format,           // pixel format
		       type,             // data type  
		       (GLvoid*) image->getBytePointer()); // pointer
	}
	else {
	  glTexImage3D(target,           // target 
		       0,                // level 
		       format,           // internal format
		       W,                // width
		       H,                // height
		       D,                // depth
		       0,                // border width (0 or 1)
		       format,           // pixel format
		       type,             // data type  
		       (GLvoid*) image->getBytePointer()); // pointer
	}
      }
    }
    
  };
  
}
#endif
  
