///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GL_TRACKBALL_HH_
#define _WISH_GL_TRACKBALL_HH_

#if defined (__WIN32__) || defined (WIN32)
#include <windows.h>
#endif

#include <GL/gl.h>

#include "wishLinalg.hh"

namespace wish {

  class Trackball {
  public:
    Trackball() {
      deactivate();
      reset();
    }
    virtual ~Trackball() {}
    
    void reset() {
      rotation = Matrix4f(1,0,0,0,
			  0,1,0,0,
			  0,0,1,0,
			  0,0,0,1);
    }
    
    void activate() {
      active = true;
    }
  
    void deactivate() {
      active = false;
    }
  
    bool isActive() const {
      return active;
    }
  
    // this one should be called in the mouse callback
    // x is the window x-coordinate
    // y is the window y-coordinate
    // width is the window width
    // height is the window height
    void grab(int x, int y, int width, int height) {
    
      // normalize window coords
      last_point.x = -1.0f + (2.0f*x)/width;
      last_point.y = -1.0f + (2.0f*(height-y))/height;
      last_point.z = 0.0f;
    
      // project to hemisphere
      float d = float( last_point.norm() );
      last_point.z = cos((float(constants::pi)/2.0f)*((d < 1.0f) ? d : 1.0f));
      last_point.normalize();
  
    }
  
    // this one should be called in the motion callback
    // x is the window x-coordinate
    // y is the window y-coordinate
    // width is the window width
    // height is the window height
    void move(int x, int y, int width, int height) {
    
      // normalize window coords
      current_point.x = -1.0f + (2.0f*x)/(width-1);
      current_point.y = -1.0f + (2.0f*(height-y))/(height-1);
      current_point.z = 0.0f;
    
      // project to hemisphere
      float d = float( current_point.norm() );
      current_point.z = cos((float(constants::pi)/2.0f)*((d < 1.0f) ? d:1.0f));
      current_point.normalize();
    
      // compute the rotation axis and angle
      Vec3f delta = current_point-last_point;
      float tdabs = float( delta.norm() );
      if ( tdabs > 1e-9 ) {
	angle = 90.0f*tdabs;
	axis = (last_point.cross(current_point));
	last_point = current_point;
      }
    
      // use the wish rotation, probably slower, but anyway...
      Matrix4f temp = wish::rotation<float>(angle, axis);
    
      // accumulate the rotation
      rotation = temp*rotation;
    
    }
  
    // apply the rotation
    void applyRotation() const {
      glMultMatrixf( rotation.getPointer() );
    }
  
  protected:
    Matrix4f rotation;
    Vec3f axis;
    float angle;
    Vec3f last_point;
    Vec3f current_point;
    bool active;
  };

}
#endif
