///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishGPUVolumeRenderer.hh"
#include "wishShaderUtilities.hh"

namespace wish {
  
  void GPUVolumeRenderer::drawCube(float xp, float yp, float zp) const {
  
    glBegin(GL_QUADS);
  
    // front
    glNormal3f(0,0,1);
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp, zp);
    glTexCoord3f(1,1,0); glVertex3f( xp, -yp, zp);
    glTexCoord3f(1,0,0); glVertex3f( xp,  yp, zp);
    glTexCoord3f(0,0,0); glVertex3f(-xp,  yp, zp);
  
    // back
    glNormal3f(0,0,-1);
    glTexCoord3f(0,0,1); glVertex3f(-xp,  yp, -zp);  
    glTexCoord3f(1,0,1); glVertex3f( xp,  yp, -zp);
    glTexCoord3f(1,1,1); glVertex3f( xp, -yp, -zp);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);
  
    // left
    glNormal3f(-1,0,0);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);  
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp,  zp);
    glTexCoord3f(0,0,0); glVertex3f(-xp,  yp,  zp);
    glTexCoord3f(0,0,1); glVertex3f(-xp,  yp, -zp);
  
    // right
    glNormal3f(1,0,0);
    glTexCoord3f(1,0,1); glVertex3f(xp,  yp, -zp);
    glTexCoord3f(1,0,0); glVertex3f(xp,  yp,  zp);
    glTexCoord3f(1,1,0); glVertex3f(xp, -yp,  zp);
    glTexCoord3f(1,1,1); glVertex3f(xp, -yp, -zp);
  
    // top
    glNormal3f(0,1,0);
    glTexCoord3f(0,0,0); glVertex3f(-xp, yp,  zp);
    glTexCoord3f(1,0,0); glVertex3f( xp, yp,  zp);
    glTexCoord3f(1,0,1); glVertex3f( xp, yp, -zp);
    glTexCoord3f(0,0,1); glVertex3f(-xp, yp, -zp);
  
    // bottom
    glNormal3f(0,-1,0);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);  
    glTexCoord3f(1,1,1); glVertex3f( xp, -yp, -zp);
    glTexCoord3f(1,1,0); glVertex3f( xp, -yp,  zp);
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp,  zp);
  
    glEnd();

  }

  // display callback
  void GPUVolumeRenderer::render() const {
    
    if( !shader_ok )
      return;
    
    Vec3f te = textures[0]->getImage()->getTotalExtent();
    
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    
    // draw outline and text
    glUseProgramObjectARB(0);
    
    // smooth lines
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);    
    
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_3D);
    glPushMatrix();
    glScalef(te.x, te.y, te.z);
    glColor3f(0.7f,0.7f,0.7f);
    glLineWidth(2.0f);
    glutWireCube(1.001f);
    glPopMatrix();
    
    glPopAttrib();
  
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    
    // enable blending
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    // enable culling
    glEnable(GL_CULL_FACE);

    float xp = te.x/2.0f;
    float yp = te.y/2.0f;
    float zp = te.z/2.0f;

    // render volume
    glCullFace(GL_BACK);
    glUseProgramObjectARB(raycast_program);  
    
    // bind texture from data
    glActiveTexture(GL_TEXTURE0);
    textures[0]->bind();
    glActiveTexture(GL_TEXTURE1);
    textures[1]->bind();
    
    //set parameters
    glUniform1iARB(volume_handle, 0);
    glUniform1iARB(volume_handle2, 1);
    glUniform1fARB(blend_handle, blend);
    glUniform3fARB(total_extent_handle, te.x, te.y, te.z);
    
    glUniform1fARB(ray_step_handle, ray_step);
    glUniform1fARB(wl_width_handle, wl_width);
    glUniform1fARB(wl_level_handle, wl_level);
    
    // render a cube
    drawCube(xp, yp, zp);
    
    // turn off shader program
    glUseProgramObjectARB(0);
    
    glPopAttrib();
    
  }

  // reshape
  void GPUVolumeRenderer::reshape(int w, int h) {
  
    // update our variables
    window_width = w;
    window_height = h;
  
  }

  // initialize
  void GPUVolumeRenderer::initialize( bool _shader_ok ) {
    
    shader_ok = _shader_ok;
    if( !shader_ok ) {
      return;
    }
    
    // total extent
    Vec3f te = textures[0]->getImage()->getTotalExtent();
    
    // setup shaders
    string wishroot = string(getenv("WISH_ROOT"));
    string vs = wishroot + "/wish/src/shaders/rayVertex.glsl";
    string fs = wishroot + "/wish/src/shaders/rayFragment.glsl";
    raycast_program = createShaderProgram( vs, fs );
    
    // get the addresses of uniform variables
    volume_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "volumeTexture");
    volume_handle2 = 
      glGetUniformLocationARB(raycast_program, 
			      "volumeTexture2");
    blend_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "blend");
    total_extent_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "totalExtent");
    ray_step_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "rayStep");
    wl_width_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "wl_width");
    wl_level_handle = 
      glGetUniformLocationARB(raycast_program, 
			      "wl_level");
  
    glBindTexture(GL_TEXTURE_2D, 0);  
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  

  }

}
