///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm, Per Sundqvist
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishGVFMultigrid.hh"

namespace wishGVF {

  //
  // THE DEFINITIONS
  //
  //
  int multigrid(float *u, float *u2, float *u3,
		float *rhs, float *rhs2, float *rhs3,
		float *vc, float *vc2, float *vc3,
		float MU, float AR2, 
		int Xp, int Yp, int Zp, float eps, 
		int nu1, int nu2, int nu3, float omega){

    float resnorm = 1.0;
    int iter = 0;
    float *temp = new float[Xp*Yp*Zp];
    memset(temp, 0, sizeof(float)*Xp*Yp*Zp);

    int X2, Y2, Z2, X3, Y3, Z3; 
    X2 = (Xp+1)/2;  Y2 = (Yp+1)/2;  Z2 = (Zp+1)/2;
    X3 = (X2+1)/2;  Y3 = (Y2+1)/2;  Z3 = (Z2+1)/2;

    float MU2 = MU/4;
    float MU3 = MU2/4;

    float omega_coarse = 1.6f; //1.5 works for SSOR, 1.6 for SOR

    float *r = new float[Xp*Yp*Zp];
    float *r2 = new float[X2*Y2*Z2];
    for(int i=0; i<Xp*Yp*Zp; i++) r[i]=0;
    for(int i=0; i<X2*Y2*Z2; i++) r2[i]=0;  

    // FMG: full multigrid V-cycle
    // Creating right hand sides
    restrict(rhs2,rhs,Xp,Yp,Zp);
    restrict(rhs3,rhs2,X2,Y2,Z2);

    // solve on coarsest grid
    coarse_solve(u3, rhs3, vc3, MU3, AR2, X3, Y3, Z3, nu3,omega_coarse,temp);

    // go to second coarsest
    prolongate(u2, u3, X2, Y2, Z2);

    // V-cycle starting from second coarsest grid
    pre_smoother(u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2, nu1,omega,temp);
    residual(r2, u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2);
    restrict(rhs3, r2, X2, Y2, Z2);  
    coarse_solve(u3, rhs3, vc3, MU3, AR2, X3, Y3, Z3, nu3,omega_coarse,temp);
    prolongate(r2, u3, X2, Y2, Z2);
    update(u2, r2, X2, Y2, Z2);

    post_smoother(u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2, nu2,omega,temp);

    // go to third coarsest (= finest here)
    prolongate(u, u2, Xp, Yp, Zp);
    post_smoother(u, rhs, vc, MU, AR2, Xp, Yp, Zp, nu2,omega,temp);

    // V-cycle starting from third coarsest grid
    while( resnorm > eps*4 ) {
      pre_smoother(u, rhs, vc, MU, AR2, Xp, Yp, Zp, nu1,omega,temp);
      residual(r, u, rhs, vc, MU, AR2, Xp, Yp, Zp);
      restrict(rhs2, r, Xp, Yp, Zp);
      pre_smoother(u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2, nu1,omega,temp);
      residual(r2, u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2);
      restrict(rhs3, r2, X2, Y2, Z2);
      coarse_solve(u3, rhs3, vc3, MU3, AR2, X3, Y3, Z3, nu3,omega_coarse,temp);
      prolongate(r2, u3, X2, Y2, Z2);
      update(u2, r2, X2, Y2, Z2);
      post_smoother(u2, rhs2, vc2, MU2, AR2, X2, Y2, Z2, nu2,omega,temp);
      prolongate(r, u2, Xp, Yp, Zp);
      update(u, r, Xp, Yp, Zp);
      post_smoother(u, rhs, vc, MU, AR2, Xp, Yp, Zp, nu2,omega,temp);

      iter ++;
      resnorm = residual_maxnorm(rhs,u,vc,MU,AR2,Xp,Yp,Zp);

      std::cout << "max residual = " << resnorm << std::endl;    
    }  //while-end
    delete r; 
    delete r2; 
    delete temp;  	      
    return iter;
  }

  //
  void mg_setup(float *vc, float *vc2, float *vc3,
		float MU, float AR2, int Xp, int Yp, int Zp){
    int X2, Y2, Z2, X3, Y3, Z3, k; 
    X2 = (Xp+1)/2;  Y2 = (Yp+1)/2;  Z2 = (Zp+1)/2;
    X3 = (X2+1)/2;  Y3 = (Y2+1)/2;  Z3 = (Z2+1)/2;

    float MU2 = MU/4;  float MU3 = MU2/4;
    restrict(vc2, vc, Xp, Yp, Zp);
    restrict(vc3, vc2, X2, Y2, Z2);

    for(int z=1; z<Zp-1; z++) {
      for(int y=1; y<Yp-1; y++) {
	for(int x=1; x<Xp-1; x++) {
	  k=z*Xp*Yp+y*Xp+x;
	  vc[k] = vc[k]+(4.0f+2.0f*AR2)*MU;
	}
      }
    }
    for(int z=1; z<Z2-1; z++) {
      for(int y=1; y<Y2-1; y++) {
	for(int x=1; x<X2-1; x++) {
	  k=z*X2*Y2+y*X2+x;
	  vc2[k] = vc2[k]+(4.0f+2.0f*AR2)*MU2;
	}
      }
    }
    for(int z=1; z<Z3-1; z++) {
      for(int y=1; y<Y3-1; y++) {
	for(int x=1; x<X3-1; x++) {
	  k=z*X3*Y3+y*X3+x;
	  vc3[k] = vc3[k]+(4.0f+2.0f*AR2)*MU3;
	}
      }
    }

  }

  //
  void pre_smoother(float *u, float *rhs, float *vc, float MU,float AR2,
		    int X, int Y, int Z, int nu,float omega, float *unew){

    memset(unew, 0, sizeof(float)*X*Y*Z);
    for(int iter=0; iter<nu; iter++){
      //     jacobi(u,rhs,unew,vc,MU,AR2,X,Y,Z,omega);
      //     temp = u;     u = unew;      unew = temp;
      SOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
      //      SSOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
    }  
  }

  //
  void post_smoother(float *u, float *rhs, float *vc, float MU,float AR2,
		     int X, int Y, int Z, int nu,float omega, float *unew){

    memset(unew, 0, sizeof(float)*X*Y*Z);
    for(int iter=0; iter<nu; iter++){
      //     jacobi(u,rhs,unew,vc,MU,AR2,X,Y,Z,omega);
      //     temp = u;     u = unew;     unew = temp;
      SOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
      //  SSOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
    }
  }
  void coarse_solve(float *u, float *rhs, float *vc, float MU,float AR2,
		    int X, int Y, int Z, int nu,float omega, float *unew){
    for(int iter=0; iter<nu; iter++){
      SSOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
    }
  }

  //
  void restrict(float *coarse, float *fine, int X, int Y, int Z){
    int k, kc;
    int XY = X*Y;
    int Xc = (X+1)/2;
    int XYc = Xc*(Y+1)/2;
    for(int z=2; z<Z-2; z += 2) {
      for(int y=2; y<Y-2; y += 2) {
	for(int x=2; x<X-2; x += 2) {
	  kc = XYc*z/2 + Xc*y/2 + x/2;   // index in coarse vector
	  k = z*XY+y*X+x;                // index in fine vector
	  coarse[kc] = fine[k]/8.0f+
	    (fine[k-1]+fine[k+1]+fine[k-X]+fine[k+X]+fine[k-XY]+fine[k+XY])/16.0f+
	    (fine[k-1-X]+fine[k-1+X]+fine[k+1-X]+fine[k+1+X]+
	     fine[k-X-XY]+fine[k-X+XY]+fine[k+X-XY]+fine[k+X+XY]+
	     fine[k-1-XY]+fine[k-1+XY]+fine[k+1-XY]+fine[k+1+XY])/32.0f+
	    (fine[k-1-X-XY]+fine[k-1-X+XY]+fine[k-1+X-XY]+fine[k-1+X+XY]+
	     fine[k+1-X-XY]+fine[k+1-X+XY]+fine[k+1+X-XY]+fine[k+1+X+XY])/64.0f;
	}
      }
    }
  }

  //
  void prolongate(float *fine, float *coarse, int X, int Y, int Z){
    int k, kc;
    int XY = X*Y;
    int Xc = (X+1)/2;
    int XYc = Xc*(Y+1)/2;
    for(int z=2; z<Z-2; z += 2) {
      for(int y=2; y<Y-2; y += 2) {
	for(int x=2; x<X-2; x +=2) {
	  kc = XYc*z/2 + Xc*y/2 + x/2;
	  k = z*XY+y*X+x;
	  fine[k] = coarse[kc];
	}
      }
    }
    for(int z=2; z<Z-2; z += 2) {
      for(int y=2; y<Y-2; y += 2) {
	for(int x=2; x<X; x +=2) {
	  k = z*XY+y*X+x;
	  fine[k-1] = 0.5f*(fine[k]+fine[k-2]);
	}
      }
    }
    for(int z=2; z<Z-2; z += 2) {
      for(int y=2; y<Y; y += 2) {
	for(int x=1; x<X-1; x ++) {
	  k = z*XY+y*X+x;
	  fine[k-X] = 0.5f*(fine[k]+fine[k-2*X]);  // Potential error
	}
      }
    }
    for(int z=2; z<Z; z += 2) {
      for(int y=1; y<Y-1; y ++) {
	for(int x=1; x<X-1; x ++) {
	  k = z*XY+y*X+x;
	  fine[k-XY] = 0.5f*(fine[k]+fine[k-2*XY]);
	}
      }
    }
  }
 
}
