///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm, Per Sundqvist
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishGVFSolverUtil.hh"

namespace wishGVF {
  
  // iterate stationary
  int iterate_stationary(float *u,float *rhs,float *unew, float *vc,
			 float MU, float AR2, float omega,float eps,
			 int X, int Y, int Z){
    float *temp;
    float res_norm = 1.0;
    int iter = 0;

    while(res_norm > eps*4){
      euler(u,rhs,unew,vc,MU,AR2,X,Y,Z,omega);
      //    jacobi(u,rhs,unew,vc,MU,AR2,X,Y,Z,omega);
      temp = u;    u = unew;    unew = temp;
      //    SOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
      //    SSOR(u,rhs,vc,MU,AR2,X,Y,Z,omega);
      iter++;
      res_norm = residual_maxnorm(rhs,u,vc,MU,AR2,X,Y,Z);
      std::cout << res_norm << std::endl; 
    }
    return iter;
  }
  
  // Setting up right hand side and initial guess
  void rhs_setup_stationary(wish::Image<float> &Fc, 
			    float *vc,   //Var. coeff.(diagonal of matrix) 
			    float *rhs, float *u, float MU, 
			    int X, int Y, int Z, 
			    float AR2) { //square of aspect ratio
    int k, x, y, z;  
    for(z=0; z<Z; z++) {
      for(y=0; y<Y; y++) {
	for(x=0; x<X; x++) {
	  k = z*X*Y+y*X+x;
	  rhs[k] = Fc(x,y,z)*(vc[k]-(4+2*AR2)*MU);
	}
      }
    }
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	for(x=1; x<X-1; x++) {
	  k = z*X*Y+y*X+x;
	  u[k] = Fc(x,y,z);
	}
      }
    }
    // Inhomogeneous Dirichlet boundary conditions at x=0 and x=1
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	x = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k-1];
	rhs[k-1] = 0;
	x = X-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k+1]; 
	rhs[k+1]=0;
      }
    }
    // ... at y=0 and y=1
    for(z=1; z<Z-1; z++) {
      for(x=1; x<X-1; x++) {
	y = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k-X];
	rhs[k-X]=0;
	y = Y-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k+X];
	rhs[k+X]=0;
      }
    }
    // ... and at z=0 and z=1  
    for(y=1; y<Y-1; y++) {
      for(x=1; x<X-1; x++) {
	z = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k-X*Y];
	rhs[k-X*Y]=0;
	z = Z-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k+X*Y];
	rhs[k+X*Y]=0;
      }
    }
    // Right hand side should be zero at corners
    rhs[0]=0;
    x=X-1; 
    y=0; 
    z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=0; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=0; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;

  } // Right hand side ready

  // Setting up right hand side for multigrid
  void rhs_setup_mg(wish::Image<float> &Fc, 
		    float *vc,   //Var. coeff.(diagonal of matrix) 
		    float *rhs, float MU, 
		    int X, int Y, int Z, 
		    int p,                    // padding parameter
		    float AR2                 //square of aspect ratio
		    ) {
    
    int k, x, y, z, Xp, Yp, Zp;
    Xp = pad(X,p); Yp = pad(Y,p); Zp = pad(Z,p);
    
    for(z=0; z<Z; z++) {
      for(y=0; y<Y; y++) {
	for(x=0; x<X; x++) {
	  k = z*Xp*Yp+y*Xp+x;
	  rhs[k] = Fc(x,y,z)*(vc[k]-(4.0f+2.0f*AR2)*MU);
	}
      }
    }
    


    // Inhomogeneous Dirichlet boundary conditions at x=0 and x=1
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	x = 1;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*rhs[k-1];
	rhs[k-1] = 0;
	x = X-2;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*rhs[k+1]; 
	rhs[k+1]=0;
      }
    }
    
    // ... at y=0 and y=1
    for(z=1; z<Z-1; z++) {
      for(x=1; x<X-1; x++) {
	y = 1;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*rhs[k-Xp];
	rhs[k-Xp]=0;
	y = Y-2;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*rhs[k+Xp];
	rhs[k+Xp]=0;
      }
    }
    
    // ... and at z=0 and z=1  
    for(y=1; y<Y-1; y++) {
      for(x=1; x<X-1; x++) {
	z = 1;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k-Xp*Yp];
	rhs[k-Xp*Yp]=0;
	z = Z-2;
	k = z*Xp*Yp+y*Xp+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k+Xp*Yp];
	rhs[k+Xp*Yp]=0;
      }
    }

    // Right hand side should be zero at corners
    rhs[0]=0;
    x=X-1; 
    y=0; 
    z=0;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=0;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=0;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=0; y=0; z=Z-1;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=X-1; y=0; z=Z-1;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=Z-1;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=Z-1;
    k = z*Xp*Yp+y*Xp+x;
    rhs[k] = 0;

  } // Right hand side ready

  // Setting up right hand side for cg
  void rhs_setup_cg(wish::Image<float> &Fc, 
		    float *vc,   //Var. coeff.(diagonal of matrix) 
		    float *rhs, float MU, int X, int Y, int Z, 
		    float AR2){                 //square of aspect ratio
    int k, x, y, z;  

    for(z=0; z<Z; z++) {
      for(y=0; y<Y; y++) {
	for(x=0; x<X; x++) {
	  k = z*X*Y+y*X+x;
	  rhs[k] = Fc(x,y,z)*(vc[k]-(4+2*AR2)*MU);
	}
      }
    }

    // Inhomogeneous Dirichlet boundary conditions at x=0 and x=1
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	x = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k-1];
	rhs[k-1] = 0;
	x = X-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k+1]; 
	rhs[k+1]=0;
      }
    }

    // ... at y=0 and y=1
    for(z=1; z<Z-1; z++) {
      for(x=1; x<X-1; x++) {
	y = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k-X];
	rhs[k-X]=0;
	y = Y-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*rhs[k+X];
	rhs[k+X]=0;
      }
    }

    // ... and at z=0 and z=1  
    for(y=1; y<Y-1; y++) {
      for(x=1; x<X-1; x++) {
	z = 1;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k-X*Y];
	rhs[k-X*Y]=0;
	z = Z-2;
	k = z*X*Y+y*X+x;
	rhs[k] = rhs[k] + MU*AR2*rhs[k+X*Y];
	rhs[k+X*Y]=0;
      }
    }

    // Right hand side should be zero at corners
    rhs[0]=0;
    x=X-1; 
    y=0; 
    z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=0;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=0; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=0; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=0; y=Y-1; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;
    x=X-1; y=Y-1; z=Z-1;
    k = z*X*Y+y*X+x;
    rhs[k] = 0;

  } // Right hand side ready

  //
  void IC(float *u, float *rhs, float *a, 
	  float MU, float AR2, int X, int Y, int Z){
    int XY = X*Y;
    int k, x, y, z;
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	for(x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k] =(rhs[k]+MU*(u[k-1]+u[k-X]+AR2*u[k-XY]))/a[k];
	}
      }
    }
    for(z=Z-2; z>0; z--) {
      for(y=Y-2; y>0; y--) {
	for(x=X-2; x>0; x--) {
	  k = z*XY+y*X+x;
	  u[k] = u[k]+MU*(u[k+1]+u[k+X]+AR2*u[k+XY])/a[k];
	}
      }
    }
  }

  // Setup for incomplete Cholesky
  void IC_setup(float *a, float *vc, float MU, float AR2, 
		float eta, int X, int Y, int Z){

    int XY = X*Y;
    int x, y, z, k;
    float mumu = MU*MU;
    float muar = mumu*AR2*AR2;
    x = 1; y = 1; z = 1;
    k = z*XY+y*X+x;
    a[k] = vc[k];
    for(x=2; x<X-1; x++) {
      k = z*XY+y*X+x;
      a[k] = vc[k]-mumu/a[k-1];
    }
    x = 1; 
    for(y=2; y<Y-1; y++) {
      k = z*XY+y*X+x;
      a[k] = vc[k]-mumu/a[k-X];
    }
    y = 1;
    for(z=2; z<Z-1; z++) {
      k = z*XY+y*X+x;
      a[k] = vc[k]-muar/a[k-XY];
    }

    z = 1;
    for(y=2; y<Y-1; y++) {
      for(x=2; x<X-1; x++) {
	k = z*XY+y*X+x;
	a[k] = vc[k]-mumu/a[k-1]-mumu/a[k-X];
      }
    }
    y = 1;
    for(z=2; z<Z-1; z++) {
      for(x=2; x<X-1; x++) {
	k = z*XY+y*X+x;
	a[k] = vc[k]-mumu/a[k-1]-muar/a[k-XY];
      }
    }
    x = 1;
    for(z=2; z<Z-1; z++) {
      for(y=2; y<Y-1; y++) {
	k = z*XY+y*X+x;
	a[k] = vc[k]-mumu/a[k-X]-muar/a[k-XY];
      }
    }

    for(z=2; z<Z-1; z++) {
      for(y=2; y<Y-1; y++) {
	for(x=2; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  a[k] = vc[k]-mumu/a[k-1]-mumu/a[k-X]-muar/a[k-XY];
	}
      }
    }   

  }   // Incomplete Cholesky setup ready

  //
  void MIC(float *u, float *rhs, float *a, 
	   float MU, float AR2, int X, int Y, int Z){
    int XY = X*Y;
    int k, x, y, z;
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	for(x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k] =(rhs[k]+MU*(u[k-1]+u[k-X]+AR2*u[k-XY]))/a[k];
	}
      }
    }
    for(z=Z-2; z>0; z--) {
      for(y=Y-2; y>0; y--) {
	for(x=X-2; x>0; x--) {
	  k = z*XY+y*X+x;
	  u[k] = u[k]+MU*(u[k+1]+u[k+X]+AR2*u[k+XY])/a[k];
	}
      }
    }
  }

  // Setup for modified incomplete Cholesky
  void MIC_setup(float *a, float *vc, float MU, float AR2, 
		 float eta, int X, int Y, int Z){

    int XY = X*Y;
    int x, y, z, k;
    float mumu = MU*MU*(2+AR2);
    float muar = MU*MU*(AR2+2*AR2*AR2);
    x = 1; y = 1; z = 1;
    k = z*XY+y*X+x;
    a[k] = (1+eta)*vc[k];
    for(x=2; x<X-1; x++) {
      k = z*XY+y*X+x;
      a[k] = (1+eta)*vc[k]-mumu/a[k-1];
    }
    x = 1; 
    for(y=2; y<Y-1; y++) {
      k = z*XY+y*X+x;
      a[k] = (1+eta)*vc[k]-mumu/a[k-X];
    }
    y = 1;
    for(z=2; z<Z-1; z++) {
      k = z*XY+y*X+x;
      a[k] = (1+eta)*vc[k]-muar/a[k-XY];
    }
    z = 1;
    for(y=2; y<Y-1; y++) {
      for(x=2; x<X-1; x++) {
	k = z*XY+y*X+x;
	a[k] = (1+eta)*vc[k]-mumu/a[k-1]-mumu/a[k-X];
      }
    }
    y = 1;
    for(z=2; z<Z-1; z++) {
      for(x=2; x<X-1; x++) {
	k = z*XY+y*X+x;
	a[k] = (1+eta)*vc[k]-mumu/a[k-1]-muar/a[k-XY];
      }
    }
    x = 1;
    for(z=2; z<Z-1; z++) {
      for(y=2; y<Y-1; y++) {
	k = z*XY+y*X+x;
	a[k] = (1+eta)*vc[k]-mumu/a[k-X]-muar/a[k-XY];
      }
    }  
    for(z=2; z<Z-1; z++) {
      for(y=2; y<Y-1; y++) {
	for(x=2; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  a[k] = (1+eta)*vc[k]-mumu/a[k-1]-mumu/a[k-X]-muar/a[k-XY];
	}
      }
    }   
  }   // Midified incomplete Cholesky setup ready




  //
  void euler(float *u, float *rhs, float *unew, float *vc,   
	     float MU, float AR2, int X, int Y, int Z, float omega){
    int k;
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  unew[k]=u[k]+omega*(rhs[k]-appl_op_loc(u,vc,MU,AR2,k,X,XY));
	  //	u = unew;
	}
      }
    }

  }

  //
  void jacobi(float *u, float *rhs, float *unew, float *vc,   
	      float MU, float AR2, int X, int Y, int Z, float omega){
    int k;
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  unew[k]=(1-omega)*u[k]+
	    omega*(rhs[k]+MU*(u[k+1]+u[k-1]+u[k+X]+u[k-X]+
			      AR2*(u[k+XY]+u[k-XY])))/vc[k];
	}
      }
    }
  }

  //
  void jacobi_precond(float *u, float *rhs, float * vc, 
		      int X, int Y, int Z, float omega){
    int k;
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k] = omega*rhs[k]/vc[k];
	}
      }
    }
  }

  //
  void SOR(float *u, float *rhs, float * vc,   
	   float MU, float AR2, int X, int Y, int Z, float omega){
    int k;
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k]=(1-omega)*u[k]+omega*(rhs[k]+MU*(u[k+1]+u[k-1]+u[k+X]+u[k-X]+
						AR2*(u[k+XY]+u[k-XY])))/vc[k];
	}
      }
    }

  }

  //
  void SSOR(float *u, float *rhs, float * vc,   
	    float MU, float AR2, int X, int Y, int Z, float omega){
    int k, x, y, z;
    int XY = X*Y;
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	for(x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k]=(1-omega)*u[k]+omega*(rhs[k]+MU*(u[k+1]+u[k-1]+u[k+X]+u[k-X]+
						AR2*(u[k+XY]+u[k-XY])))/vc[k];
	}
      }
    }
    for(z=Z-2; z>0; z--) {
      for(y=Y-2; y>0; y--) {
	for(x=X-2; x>0; x--) {
	  k = z*XY+y*X+x;
	  u[k] = (1-omega)*u[k]+omega*(rhs[k]+MU*(u[k+1]+u[k-1]+u[k+X]+u[k-X]+
						  AR2*(u[k+XY]+u[k-XY])))/vc[k];
	}
      }
    }
  } 
  //
  void SSOR_precond(float *u, float *rhs, float * vc,   
		    float MU, float AR2, int X, int Y, int Z, float omega){
    int k, x, y, z;
    int XY = X*Y;
    for(z=1; z<Z-1; z++) {
      for(y=1; y<Y-1; y++) {
	for(x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  u[k]=(rhs[k]+omega*MU*(u[k-1]+u[k-X]+AR2*u[k-XY]))/vc[k];
	}
      }
    }
    for(z=Z-2; z>0; z--) {
      for(y=Y-2; y>0; y--) {
	for(x=X-2; x>0; x--) {
	  k = z*XY+y*X+x;
	  u[k] = u[k]+omega*MU*(u[k+1]+u[k+X]+AR2*u[k+XY])/vc[k];
	}
      }
    } 
  }

  //
  void residual(float *r, float *u, float *rhs, float *vc, float MU, float AR2,
		int X, int Y, int Z){
    int k;  
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k = z*XY+y*X+x;
	  r[k] = rhs[k]-appl_op_loc(u,vc,MU,AR2,k,X,XY);
	}
      }
    }
  }

  //
  float residual_maxnorm(float *rhs, float *w, float *vc, float mu, float AR2,
			 int X, int Y, int Z) {
    float res_norm;
    int k;
    int XY = X*Y;
    res_norm = 0;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k=z*XY+y*X+x;
	  res_norm = 
	    wish::util::maxOf(res_norm,
			      float(fabs(rhs[k]-appl_op_loc(w,vc,mu,AR2,k,X,XY))));
	}
      }
    }
    return res_norm;
  }
  
  //
  float appl_op_loc(float *w, float *vc, float mu, float AR2,
		    int k, int X, int XY) {
    return vc[k]*w[k]-mu*(w[k+1]+w[k-1]+w[k+X]+w[k-X]+AR2*(w[k+XY]+w[k-XY]));
  }

  int pad(int X,int p) {
    int tmp = X+1;
    while((tmp-1)%p) tmp++;
    return tmp;
  }

  //
  void update(float *u, float *e, int X , int Y, int Z){
    int k;
    int XY = X*Y;
    for(int z=1; z<Z-1; z++) {
      for(int y=1; y<Y-1; y++) {
	for(int x=1; x<X-1; x++) {
	  k=z*XY+y*X+x;
	  u[k] = u[k]+e[k];
	}
      }
    }
  }

}
