///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm, Per Sundqvist
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GVF_SOLVER_UTIL_HH_
#define _WISH_GVF_SOLVER_UTIL_HH_

#include "wishutil.hh"
#include "wishImage.hh"

namespace wishGVF {

  // setup right hand side for multigrid
  void rhs_setup_mg(wish::Image<float> &Fc, float *vc,
		    float *rhs, float MU, 
		    int X, int Y, int Z, int p, float AR);
  
  // setup right hand side for cg
  void rhs_setup_cg(wish::Image<float> &Fc, float *vc,
		    float *rhs, float MU, int X, int Y, int Z, float AR2);
  
  // setup right hand side for stationary
  void rhs_setup_stationary(wish::Image<float> &Fc, float *vc,
			    float *rhs, float *u, float MU, 
			    int X, int Y, int Z, float AR2);
  
  // iterate stationary
  int iterate_stationary(float *u,float *rhs,float *unew, float *vc,
			 float MU, float AR2, float omega,float epsilon,
			 int X, int Y, int Z);
  
  //
  // One application of incomplete Cholesky preconditioner
  void IC(float *u, float *rhs, float *a, 
	  float MU, float AR2, int X, int Y, int Z);
  void IC_setup(float *a, float *vc, float MU, float AR2, 
		float eta, int X, int Y, int Z);

  //
  // One application of a modified incomplete Cholesky preconditioner
  void MIC(float *u, float *rhs, float *a, 
	   float MU, float AR2, int X, int Y, int Z);
  void MIC_setup(float *a, float *vc, float MU, float AR2, 
		 float eta, int X, int Y, int Z);

  //
  // Performes one forward Euler iteration on u
  // with time step omega 
  void euler(float *u,     // grid function
	     float *rhs,   // right hand side
	     float *unew,  // temporary
	     float * vc,   // variable diagonal of operator
	     float MU,     // coefficient for Laplacian
	     float AR2,    // SQUARE of aspact ratio, used in operator
	     int X, int Y, int Z, // sizes
	     float omega); // relaxationparameter  

  //
  // Performes one weighted Jacobi iteration on u
  // If omega = 1 one obtains standard Jacobi
  void jacobi(float *u,     // grid function
	      float *rhs,   // right hand side
	      float *unew,  // temporary
	      float * vc,   // variable diagonal of operator
	      float MU,     // coefficient for Laplacian
	      float AR2,    // SQUARE of aspact ratio, used in operator
	      int X, int Y, int Z, // sizes
	      float omega); // relaxationparameter  

  //
  // Applies weighted Jacobi as preconditioner
  // If omega = 1 one obtains standard Jacobi
  void jacobi_precond(float *u,     // grid function
		      float *rhs,   // right hand side
		      float * vc,   // variable diagonal of operator
		      int X, int Y, int Z, // sizes
		      float omega); // relaxationparameter  

  //
  // Performes one SOR iteration on u
  // If omega = 1 one obtains Gauss-Seidel 
  void SOR(float *u,     // grid function
	   float *rhs,   // right hand side
	   float * vc,   // variable diagonal of operator
	   float MU,     // coefficient for Laplacian
	   float AR2,    // SQUARE of aspact ratio, used in operator
	   int X, int Y, int Z, // sizes
	   float omega); // relaxationparameter  

  //
  // Performes one SSOR iteration on u
  // If omega = 1 one obtains symmetric Gauss-Seidel 
  void SSOR(float *u,     // grid function
	    float *rhs,   // right hand side
	    float * vc,   // variable diagonal of operator
	    float MU,     // coefficient for Laplacian
	    float AR2,    // SQUARE of aspact ratio, used in operator
	    int X, int Y, int Z, // sizes
	    float omega); // relaxationparameter  

  //
  // Applies SSOR as preconditioner
  // If omega = 1 one obtains symmetric Gauss-Seidel 
  void SSOR_precond(float *u,     // grid function
		    float *rhs,   // right hand side
		    float * vc,   // variable diagonal of operator
		    float MU,     // coefficient for Laplacian
		    float AR2,    // SQUARE of aspact ratio, used in operator
		    int X, int Y, int Z, // sizes
		    float omega); // relaxationparameter  

  //
  // Computes the residual r=A*u-rhs, but only at inner points
  // assuming that there is a border of ghost points around
  void residual(float *r,  // the residual
		float *u,  // gridfunction whose residual is r
		float *rhs,// right hand side 
		float *vc, // variable diagonal of operator
		float MU,  // coefficient for Laplacian
		float AR2,  // aspact ratio, used in operator
		int X, int Y, int Z); // sizes

  //
  // Returns the maximum norm of the residual
  float residual_maxnorm(float *rhs,  // right hand side
			 float *w,    // gridfunction
			 float *vc,   // variable diagonal
			 float mu,    // coefficient for Laplacian
			 float AR2,    // aspact ratio, used in operator
			 int X, int Y, int Z);  // sizes


  //
  // Applies the operator locally, at gridpoint number k
  // and returns the answer
  // The operator is a 3D, seven-point Laplacian plus a 
  // variable coefficient multiplication. The variable 
  // diagonal vc contains the sum of the variable coeff. and
  // the contribution from the Laplacian.
  // The aspect ration AR refers to the grid spacing in dim 3 
  // onl,y according to 1=xd=dy=dz*AR 
  float appl_op_loc(float *w,    // grid function
		    float *vc,   // variable diagonal
		    float mu,    // coefficient for Laplacian
		    float AR2,   // the SQUARE of the aspect ratio 
		    int k,       // one dimensional index
		    int X,       // # gridpoints in dim 1
		    int XY);     // ... times # in dim 2

  //
  // Updates u by adding e to it
  void update(float *u,             // grid function
	      float *e,             // grid function
	      int X, int Y, int Z); // sizes of u and e


  //
  // Adds to the integer X so that (X-1) becomes divisible by p
  int pad(int X,int p);


}

#endif


