///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GAUSSIAN_FILTER_HH_
#define _WISH_GAUSSIAN_FILTER_HH_

#include "wishFilterObject.hh"

namespace wish {

 
  //! The GaussianFilter implements a 1D family of filters  
  //! represented by Gaussians. The order parameter tells
  //! which kind of filter that is used.
  //! 0 => smoothing
  //! 1 => first derivative
  //! 2 => second derivative
  //! It is also needed to specify in which direction the
  //! filter should be applied.
 
  template<class InputImage, class OutputImage>
  class GaussianFilter : public FilterObject<InputImage, OutputImage> {
   
  public:
   
    //! pointer typedef
    typedef auto_ptr< GaussianFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
   
    //! constructor
    GaussianFilter() :
      m_sigma(1.0),
      m_order(0),
      m_direction(0),
      m_max_kernel_radius(20),
      m_offset(0.0)
    {}
   
    //! set sigmas
    void setSigma(double sigma) {
      m_sigma = sigma;
    }
   
    //! set radius
    void setMaxKernelRadius(int radius) {
      m_max_kernel_radius = radius;
    }
  
    //! set order
    void setOrder(int order) {
      m_order = order;
    }

  
    //! set direction
    void setDirection(int _direction) {
      m_direction = _direction;
    }

  
    //! set offset (use for derivative filters to handle negative values)
    void setOffset(int _offset) {
      m_offset = _offset;
    }

  protected:
  
    // generate filter
    vector<double> generateFilter(int dim) const;
   
    double m_sigma;
    int m_order;
    int m_direction;
    int m_max_kernel_radius;
    double m_offset;
   
    virtual void process();
   
  private:

    // purposely not implemented
    GaussianFilter(const GaussianFilter&);
    GaussianFilter& operator=(const GaussianFilter&);
  };

 
  template<class InputImage, class OutputImage>
  vector<double>
  GaussianFilter<InputImage,OutputImage>::generateFilter(int dim) const {
    Vec3f spacing = this->m_input->getSpacing();
    double aspect = spacing.data[dim]/spacing.data[0];
    double sigma = m_sigma/aspect;
   
    int t,n;
    t = util::minOf( int(3.0*sigma+1.0), m_max_kernel_radius );
    n = 2*t+1;
   
    vector<double> filter_data(n);
  
    // squared sigma
    double s2 = util::square(sigma);
   
    // order = 0, i.e. smoothing filter
    if( m_order == 0 ) {
      double fsum = 0.0;
      for(int w=-t; w<=t; w++) {
	filter_data[w+t] = exp(-0.5*w*w/s2);
	fsum += filter_data[w+t];
      }
     
      // normalize
      for(int i=0; i<n; i++) {
	filter_data[i] = filter_data[i]/fsum;
      }
    }

  
    // first derivative
    // normalized so that K*sum_{i=-n}^{n} h(i)*(a+b*i) = b
    // this yields normalization constant K = 1/sum_{i=-n}^{n} h(i)*i
    else if( m_order == 1 ) {
      double fsum = 0.0;
      for(int w=-t; w<=t; w++) {
	filter_data[w+t] = -w*exp(-0.5*w*w/s2);
	fsum += w*filter_data[w+t];
      }

      // normalize
      for(int i=0; i<n; i++) {
	filter_data[i] = filter_data[i]/fsum;
      }
    }
   
    // second derivative
    // normalized so that K*sum_{i=-n}^{n} h(i)*(a+b*i+c*i*i) = 2*c
    // this yields normalization constant K = 2/sum_{i=-n}^{n} h(i)*i*i
    else if( m_order == 2 ) {
      double fsum = 0.0;
      for(int w=-t; w<=t; w++) {
	filter_data[w+t] = (w*w/s2-1)*exp(-0.5*w*w/s2);
	fsum += w*w*filter_data[w+t]/2.0;
      }
     
      // normalize
      for(int i=0; i<n; i++)
	filter_data[i] = filter_data[i]/fsum;
    }
   
    return filter_data;
  }

 
  template<class InputImage, class OutputImage>
  void GaussianFilter<InputImage, OutputImage>::process() {
   
    this->m_output->resize( *(this->m_input) );
    this->m_output->reset( OutputImageVoxelType(0) );
  
    // dimensions
    int W = this->m_input->width();
    int H = this->m_input->height();
    int D = this->m_input->depth();
   
    // temporary images
    InputImage  &in  = *(this->m_input);
    OutputImage &out = *(this->m_output);
   
    // filter image
    vector<double> filter_data;
    int t;
    double local_result;
    int x,y,z;

    // x-direction OPTIMIZED
    if( m_direction == 0 ) {
      filter_data = generateFilter(0);
      t = int( filter_data.size() )/2;
      for(z=0; z<D; z++) {
	for(y=0; y<H; y++) {
	  for(x=t; x<W-t; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in(x+p,y,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }

      // left border
      for(z=0; z<D; z++) {
	for(y=0; y<H; y++) {
	  for(x=0; x<t; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x+p,y,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
      // right border
      for(z=0; z<D; z++) {
	for(y=0; y<H; y++) {
	  for(x=W-t; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x+p,y,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
    }
    
    // y-direction OPTIMIZED
    else if( m_direction == 1 ) {
      
      filter_data = generateFilter(1);
      t = int( filter_data.size() )/2;
      for(z=0; z<D; z++) {
	for(y=t; y<H-t; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in(x,y+p,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
      // upper border
      for(z=0; z<D; z++) {
	for(y=0; y<t; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x,y+p,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
      // lower border
      for(z=0; z<D; z++) {
	for(y=H-t; y<H; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x,y+p,z);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
    }
    
    // z-direction OPTIMIZED
    else if( m_direction == 2 ) {
      
      filter_data = generateFilter(2);
      t = int( filter_data.size() )/2;
      for(z=t; z<D-t; z++) {
	for(y=0; y<H; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in(x,y,z+p);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
      // near border
      for(z=0; z<t; z++) {
	for(y=0; y<H; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x,y,z+p);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
      // far border
      for(z=D-t; z<D; z++) {
	for(y=0; y<H; y++) {
	  for(x=0; x<W; x++) {
	    local_result = 0;
	    for(int p=-t; p<=t; p++) {
	      local_result += filter_data[p+t]*in.getSafeMirror(x,y,z+p);
	    }
	    out(x,y,z) = OutputImageVoxelType(local_result+m_offset);
	  }
	}
      }
      
    }
    
  }
  
} // end namespace wish



#endif

