///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GAUSSIAN_GRADIENT_FILTER_HH_
#define _WISH_GAUSSIAN_GRADIENT_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishGaussianFilter.hh"

namespace wish {
  
  //! GaussianGradientFilter
  
  template<class InputImage>
  class GaussianGradientFilter : 
    public FilterObject<InputImage, Image<Vec3f> > {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< GaussianGradientFilter<InputImage> > 
    Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename Image<Vec3f>::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename Image<Vec3f>::VoxelType OutputImageVoxelType;
    
    //! constructor
    GaussianGradientFilter() : 
      m_sigma(1.0),
      m_max_kernel_radius(20)
    {}
    
    //! set sigmas
    void setSigma(double sigma) {
      m_sigma = sigma;
    }
    
    //! set radius
    void setMaxKernelRadius(int radius) {
      m_max_kernel_radius = radius;
    }
    
  protected:
    
    double m_sigma;
    int m_max_kernel_radius;
    
    virtual void process();
    
  private:
    // purposely not implemented
    GaussianGradientFilter(const GaussianGradientFilter&);
    GaussianGradientFilter& operator=(const GaussianGradientFilter&);
  };
  
  template<class InputImage>
  void GaussianGradientFilter<InputImage>::process() {
    
    // references
    InputImage &in = *(this->m_input);
    Image<Vec3f> &out = *(this->m_output);
    
    out.setLabel( in.getLabel() + "_GRAD" );

    Image<float> sx, sy, sz, dx, dy, dz;
    
    GaussianFilter< InputImage,Image<float> > gaussian;
    GaussianFilter< Image<float>,Image<float> > gaussian2;
    gaussian.setSigma(m_sigma);
    gaussian.setMaxKernelRadius(m_max_kernel_radius);
    gaussian2.setSigma(m_sigma);
    gaussian2.setMaxKernelRadius(m_max_kernel_radius);
    
    // 3D case
    if( in.depth() > 1 ) {
      
      // DX
      
      cout << "Generating DX..." << endl;
      
      // smoothing in z
      gaussian.setOrder(0);
      gaussian.setInput( &in );    
      gaussian.setOutput(&sz);    
      gaussian.setDirection(2);
      gaussian.execute();      
      
      // smoothing in y direction
      gaussian2.setOrder(0);
      gaussian2.setInput(&sz);
      gaussian2.setOutput(&sy);    
      gaussian2.setDirection(1);
      gaussian2.execute();
      
      // differentiating in x direction
      gaussian2.setOrder(1);
      gaussian2.setInput(&sy);
      gaussian2.setOutput(&dx);
      gaussian2.setDirection(0);
      gaussian2.execute();
      
      // DY
      
      cout << "Generating DY..." << endl;
      
      // smoothing in x direction
      gaussian2.setOrder(0);
      gaussian2.setInput(&sz);
      gaussian2.setOutput(&sx);    
      gaussian2.setDirection(0);
      gaussian2.execute();
      
      // differentiating in y direction
      gaussian2.setOrder(1);
      gaussian2.setInput(&sx); 
      gaussian2.setOutput(&dy);   
      gaussian2.setDirection(1);
      gaussian2.execute();
      
      // DZ
      
      cout << "Generating DZ..." << endl;
      
      // smoothing in x direction
      gaussian.setOrder(0);
      gaussian.setInput(&in);
      gaussian.setOutput(&sx);    
      gaussian.setDirection(0);
      gaussian.execute();
      
      // smoothing in y
      gaussian2.setOrder(0);
      gaussian2.setInput(&sx);
      gaussian2.setOutput(&sy);    
      gaussian2.setDirection(1);
      gaussian2.execute();
      
      // differentiating in z
      gaussian2.setOrder(1);
      gaussian2.setInput(&sy); 
      gaussian2.setOutput(&dz);   
      gaussian2.setDirection(2);
      gaussian2.execute();
      
      // free som memory
      sx.allocate(0,0,0);
      sy.allocate(0,0,0);
      sz.allocate(0,0,0);
      
      // output
      out.resize( in );
      for(int i=0; i<out.totalSize(); i++) 
	out[i] = Vec3f(dx[i], dy[i], dz[i]);
    }
    
    // 2D case
    else {
      
      // DX
      
      cout << "Generating DX..." << endl;
      
      // smoothing in y direction
      gaussian.setOrder(0);
      gaussian.setInput(&in);
      gaussian.setOutput(&sy);    
      gaussian.setDirection(1);
      gaussian.execute();
      
      // differentiating in x direction
      gaussian2.setOrder(1);
      gaussian2.setInput(&sy);
      gaussian2.setOutput(&dx);
      gaussian2.setDirection(0);
      gaussian2.execute();
      
      // DY
      
      cout << "Generating DY..." << endl;
      
      // smoothing in x direction
      gaussian.setOrder(0);
      gaussian.setInput(&in);
      gaussian.setOutput(&sx);    
      gaussian.setDirection(0);
      gaussian.execute();
      
      // differentiating in y direction
      gaussian2.setOrder(1);
      gaussian2.setInput(&sx); 
      gaussian2.setOutput(&dy);   
      gaussian2.setDirection(1);
      gaussian2.execute();
      
      // free som memory
      sx.allocate(0,0,0);
      sy.allocate(0,0,0);
      
      // output
      out.resize( in );
      for(int i=0; i<out.totalSize(); i++) 
	out[i] = Vec3f(dx[i], dy[i], 0.0f);
    }
    
  }
  
} // end namespace wish

#endif
