///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GAUSSIAN_SMOOTHING_FILTER_HH_
#define _WISH_GAUSSIAN_SMOOTHING_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishGaussianFilter.hh"

namespace wish {
  
  //! GaussianSmoothingFilter performs gaussian smoothing
  //! by convolution with a gaussian function with standard deviation
  //! sigma. See wishGaussianFilter
  
  template<class InputImage, class OutputImage>
  class GaussianSmoothingFilter : 
    public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< GaussianSmoothingFilter<InputImage,OutputImage> > 
    Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    GaussianSmoothingFilter() : 
      m_sigma(1.0),
      m_max_kernel_radius(20)
    {}
    
    //! set sigmas
    void setSigma(double sigma) {
      m_sigma = sigma;
    }
    
    //! set alpha
    void setMaxKernelRadius(int radius) {
      m_max_kernel_radius = radius;
    }
    
  protected:
    
    double m_sigma;
    int m_max_kernel_radius;
    
    virtual void process();
    
  private:
    // purposely not implemented
    GaussianSmoothingFilter(const GaussianSmoothingFilter&);
    GaussianSmoothingFilter& operator=(const GaussianSmoothingFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void GaussianSmoothingFilter<InputImage, OutputImage>::process() {
    
    // references
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    out.resize( in );
    out.reset( OutputImageVoxelType(0) );
    out.setLabel( in.getLabel() + "_GS" );
	
    Image<float> tmp1, tmp2;
    tmp1 = in;
        
    GaussianFilter< Image<float>,Image<float> > gaussian;
    gaussian.setSigma(m_sigma);
    gaussian.setMaxKernelRadius(m_max_kernel_radius);
    gaussian.setOrder(0);

    // x direction
    gaussian.setInput(&tmp1);
    gaussian.setOutput(&tmp2);    
    gaussian.setDirection(0);
    gaussian.execute();
    
    // y direction
    gaussian.setInput(&tmp2);
    gaussian.setOutput(&tmp1);    
    gaussian.setDirection(1);
    gaussian.execute();
    
    if( this->m_input->depth() > 1 ) {
      // z direction
      gaussian.setInput(&tmp1);
      gaussian.setOutput(&tmp2);
      gaussian.setDirection(2);      
      gaussian.execute();
      *(this->m_output) = tmp2;
    }
    else
      *(this->m_output) = tmp1;
  }
  
} // end namespace wish

#endif
