///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm, Per Sundqvist
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_GRADIENT_VECTOR_FLOW_FILTER_HH_
#define _WISH_GRADIENT_VECTOR_FLOW_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishGVFSolverUtil.hh"
#include "wishGVFConjugateGradients.hh"
#include "wishGVFMultigrid.hh"
 
namespace wish {
  
  //! The gradient vector flow filter computes the GVF field
  //! of the input image which should be an edge map 
  //! (i.e., gradient magnitude)
  //!
  //! This filter implements the GVF for four different solvers
  //! standard, stationary, conjugate gradients, and multigrid.
  //! It replaces the old GVFFilter. The standard solver is implemented
  //! by Erik Vidholm from the paper by Xu and Prince. The other 
  //! solvers are implemented by Per Sundqvist.
  
  template<class InputImage>
  class GradientVectorFlowFilter : 
    public FilterObject<InputImage, Image<Vec3f > > {
    
  public:
    
    //! solver types
    typedef enum { GVF_STANDARD, 
		   GVF_STATIONARY, 
		   GVF_CONJUGATE_GRADIENTS, 
		   GVF_MULTIGRID }
      GVFSolverType;
    
    //! pointer typedef
    typedef auto_ptr< GradientVectorFlowFilter<InputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    
    //! constructor
    GradientVectorFlowFilter() :
      m_mu(.5), 
      m_epsilon(1e-3),
      m_solver_type(GVF_MULTIGRID)
      
    {}
    
    //! set mu
    void setMu(const double &_mu) {
      m_mu = _mu;
    }
    
    //! set epsilon
    void setEpsilon(const double &_epsilon) {
      m_epsilon = _epsilon;
    }
    
    //! set solver type
    void setSolverType(const string &st) {
      if( st == "GVF_STANDARD" )
	m_solver_type = GVF_STANDARD;
      else if( st == "GVF_STATIONARY" )
	m_solver_type = GVF_STATIONARY;
      else if( st == "GVF_CONJUGATE_GRADIENTS" )
	m_solver_type = GVF_CONJUGATE_GRADIENTS;
      else if( st == "GVF_MULTIGRID" )
	m_solver_type = GVF_MULTIGRID;
      else
	cerr << "Unknown solver type, not changing" << endl;
    }
    
    void setSolverType(const GVFSolverType &st) {
      m_solver_type = st;
    }
    
    void setSolverTypeToStandard() {
      m_solver_type = GVF_STANDARD;      
    }
    void setSolverTypeToStationary() {
      m_solver_type = GVF_STATIONARY;
    }
    void setSolverTypeToConjugateGradients() {
      m_solver_type = GVF_CONJUGATE_GRADIENTS;
    }
    void setSolverTypeToMultigrid() {
      m_solver_type = GVF_MULTIGRID;
    }
    
  protected:
    double m_mu;
    double m_epsilon;
    GVFSolverType m_solver_type;
    
    virtual void process();
    
  private:
    
    GradientVectorFlowFilter(const GradientVectorFlowFilter&);
    GradientVectorFlowFilter& operator=(const GradientVectorFlowFilter&);
  };
  
  template<class InputImage>
  void GradientVectorFlowFilter<InputImage>::process() {

    // create references
    InputImage &in = *(this->m_input);
    Image<Vec3f> &out = *(this->m_output);
    
    out.resize( in );
    out.reset( Vec3f(0,0,0) );
    
    out.setLabel( in.getLabel() + "_GVF" );
    
    // dimensions
    int X = in.width();
    int Y = in.height();
    int Z = in.depth();
    
    // compute the gradient of the edgemap and normalize
    cerr << "Generating initial solution..." << endl;
    Image<float> Fx(X,Y,Z), Fy(X,Y,Z), Fz(X,Y,Z);
    InputImageVoxelType maxval = in.getMax();    
    for(int z=0; z<Z; z++) {
      for(int y=0; y<Y; y++) {
	for(int x=0; x<X; x++) {
	  Vec3f grad = in.getGradient(x,y,z);
	  Fx(x,y,z) = grad.x/float(maxval);
	  Fy(x,y,z) = grad.y/float(maxval);
	  Fz(x,y,z) = grad.z/float(maxval);
	}
      }
    }
    cerr << "Finished!" << endl;
    
    // Aspect ratio, dx=dy=dz*AR, AR<1 gives long voxels
    // OBS! It is assumed that dx=dy
    float AR = in.getSpacing().x/in.getSpacing().z; 
    float AR2 = AR*AR;
    
    // laplacian coefficient
    float MU = float( m_mu );
    
    // ###############
    // STANDARD SOLVER
    // ###############    
    if( m_solver_type == GVF_STANDARD ) {
      
      // create images for GVF
      Image<Vec3f> Vnew(X,Y,Z);
      Image<float> B(X,Y,Z);    
      
      // compute squared magnitude
      int k;
      for(k=0; k<B.totalSize(); k++) {
	B[k] = Fx[k]*Fx[k] + Fy[k]*Fy[k] + Fz[k]*Fz[k];
      }
      
      // another alias for the output
      Image<Vec3f> &V = *(this->m_output);
      
      // OBS! have not incorporated non-uniform voxel size here
      float DX = 1.0f;
      float DT = DX*DX/(4.0f*MU)/2.0f;
      float R  = MU*DT/(DX*DX);
      
      // initial solution
      for(k=0; k<V.totalSize(); k++) {
	V[k].x = Fx[k];
	V[k].y = Fy[k];
	V[k].z = Fz[k];
      }
      
      int iter = 0;
      float residual = 1000.0f;
      while( residual > m_epsilon ) {
	iter++;
	
	// 3D
	if( this->m_input->depth() > 1 ) {
	  for(int z=1; z<Z-1; z++) {
	    for(int y=1; y<Y-1; y++) {
	      for(int x=1; x<X-1; x++) {
		Vnew(x,y,z).x = 
		  (1.0f-B(x,y,z)*DT)*V(x,y,z).x + 
		  R*(V(x+1,y,z).x + V(x,y+1,z).x + V(x,y,z+1).x +
		     V(x-1,y,z).x + V(x,y-1,z).x + V(x,y,z-1).x - 
		     6.0f*V(x,y,z).x) + Fx(x,y,z)*B(x,y,z)*DT;
		
		Vnew(x,y,z).y = 
		  (1.0f-B(x,y,z)*DT)*V(x,y,z).y + 
		  R*(V(x+1,y,z).y + V(x,y+1,z).y + V(x,y,z+1).y +
		     V(x-1,y,z).y + V(x,y-1,z).y + V(x,y,z-1).y - 
		     6.0f*V(x,y,z).y) + Fy(x,y,z)*B(x,y,z)*DT;
		
		Vnew(x,y,z).z = 
		  (1.0f-B(x,y,z)*DT)*V(x,y,z).z + 
		  R*(V(x+1,y,z).z + V(x,y+1,z).z + V(x,y,z+1).z +
		     V(x-1,y,z).z + V(x,y-1,z).z + V(x,y,z-1).z - 
		     6.0f*V(x,y,z).z) + Fz(x,y,z)*B(x,y,z)*DT;
	      }
	    }
	  }
	}
	
	// 2D
	else {
	  int z=0;
	  for(int y=1; y<Y-1; y++) {
	    for(int x=1; x<X-1; x++) {
	      Vnew(x,y,z).x = 
		(1.0f-B(x,y,z)*DT)*V(x,y,z).x + 
		R*(V(x+1,y,z).x + V(x,y+1,z).x +
		   V(x-1,y,z).x + V(x,y-1,z).x - 4.0f*V(x,y,z).x) +
		Fx(x,y,z)*B(x,y,z)*DT;
	      
	      Vnew(x,y,z).y = 
		(1.0f-B(x,y,z)*DT)*V(x,y,z).y + 
		R*(V(x+1,y,z).y + V(x,y+1,z).y +
		   V(x-1,y,z).y + V(x,y-1,z).y - 4.0f*V(x,y,z).y) +
		Fy(x,y,z)*B(x,y,z)*DT;
	      
	      Vnew(x,y,z).z = 0;
	    }
	  }
	}
	
	// compute residual
	residual = 0.0f;
	float tmp_res;
	for(k=0; k<V.totalSize(); k++) {
	  tmp_res = float( (Vnew[k]-V[k])*(Vnew[k]-V[k]) );      
	  if( tmp_res > residual ) {
	    residual = tmp_res;
	  }
	}
	
	residual = sqrt(residual);
	
	cerr << "max residual = " << residual << endl;
	
	V = Vnew;
      }
      
      cerr << "GVF_STANDARD finished!" << endl;
      cerr << "iter = " << iter << endl;
      cerr << "max residual = " << residual << endl;
    }
    
    // ###############
    // STATIONARY SOLVER
    // ###############    
    else if( m_solver_type == GVF_STATIONARY ) {
      
      int k;
      
      // grid functions
      float *u = new float[X*Y*Z];    // unknown
      float *rhs = new float[X*Y*Z];  // right hand side
      float *vc = new float[X*Y*Z];   // variable coefficient
    
      // Temporary variables
      float *unew = new float[X*Y*Z];
      // Parameter
      float omega;
      cerr << "Give omega" << endl;
      cin >> omega;
    
      // Some compilers does not do this at some optimization levels
      memset(u, 0, sizeof(float)*X*Y*Z);
      memset(rhs, 0, sizeof(float)*X*Y*Z);
      memset(vc, 0, sizeof(float)*X*Y*Z);
      memset(unew, 0, sizeof(float)*X*Y*Z);

      // Performance results
      int iter_x, iter_y, iter_z;
    
      // Computing the variable coefficient and recasting it into 1D vector
      // Done once since the same coeff. is used for all three components
      for(int z=1; z<Z-1; z++) {
	for(int y=1; y<Y-1; y++) {
	  for(int x=1; x<X-1; x++) {
	    k = z*X*Y+y*X+x;
	    vc[k] = Fx(x,y,z)*Fx(x,y,z)+Fy(x,y,z)*Fy(x,y,z)+Fz(x,y,z)*Fz(x,y,z)
	      +(4+2*AR2)*MU;
	  }
	}
      }
    
    
      //
      // Computation of first component
      //
    
      cerr << "Iterating on first component." << endl; 
      //  cin >> k;  
      wishGVF::rhs_setup_stationary(Fx, vc, rhs, u, MU, X, Y, Z, AR2);
      iter_x =  wishGVF::iterate_stationary(u,rhs,unew,vc,MU,AR2,omega,float(m_epsilon),X,Y,Z);
    
      // Preparing for uriting to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).x=u[k]; 
	  }
	}
      }
    
      cerr << "Iterating on second component." << endl; 
      wishGVF::rhs_setup_stationary(Fy, vc, rhs, u, MU, X, Y, Z, AR2);
      iter_y =  wishGVF::iterate_stationary(u,rhs,unew,vc,MU,AR2,omega,float(m_epsilon),X,Y,Z);
    
    
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).y=u[k]; 
	  }
	}
      }
    
      cerr << "Iterating on third component." << endl; 
      wishGVF::rhs_setup_stationary(Fz, vc, rhs, u, MU, X, Y, Z, AR2);
      iter_z =  wishGVF::iterate_stationary(u,rhs,unew,vc,MU,AR2,omega,float(m_epsilon),X,Y,Z);
    
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).z=u[k]; 
	  }
	}
      }
    
      cerr << "GVF_STATIONARY Finished!" << endl;
      cerr << "Summary: " << endl;
      cerr << "x-iter = " << iter_x << endl;
      cerr << "y-iter = " << iter_y << endl;
      cerr << "z-iter = " << iter_z << endl;
      cerr << "mean-iter = " << (iter_x+iter_y+iter_z)/3.0 << endl;
      cerr << "omega = " << omega << endl;
      //cout << "Iteration time: " << eUtil::tstop()-t0 << " s" << endl;   
      delete u;
      delete vc;
      delete rhs;
      delete unew;
    }
  
  
    // ###############
    // CONJUGATE GRADIENTS SOLVER
    // ###############    
    else if( m_solver_type == GVF_CONJUGATE_GRADIENTS ) {
    
      int k;
        
      // grid functions
      float *u = new float[X*Y*Z];    // unknown
      float *rhs = new float[X*Y*Z];  // right hand side
      float *vc = new float[X*Y*Z];   // variable coefficient
    
      // Some compilers does not do this at some optimization levels
      memset(u, 0, sizeof(float)*X*Y*Z);
      memset(rhs, 0, sizeof(float)*X*Y*Z);
      memset(vc, 0, sizeof(float)*X*Y*Z);

      // PCG-stuff
      float *r = new float[X*Y*Z];
      float *p = new float[X*Y*Z];
      float *v = new float[X*Y*Z];
      float *w = new float[X*Y*Z];
      float *a = new float[X*Y*Z];
    
      float omega = 1.7f;
      
      // For safety and from experience
      memset(r, 0, sizeof(float)*X*Y*Z);
      memset(p, 0, sizeof(float)*X*Y*Z);
      memset(v, 0, sizeof(float)*X*Y*Z);
      memset(w, 0, sizeof(float)*X*Y*Z);
      memset(a, 0, sizeof(float)*X*Y*Z);
      
      // Performance results
      int iter_x, iter_y, iter_z;
    
      // Computing the variable coefficient and recasting it into 1D vector
      // Done once since the same coeff. is used for all three components
      for(int z=1; z<Z-1; z++) {
	for(int y=1; y<Y-1; y++) {
	  for(int x=1; x<X-1; x++) {
	    k = z*X*Y+y*X+x;
	    vc[k] = Fx(x,y,z)*Fx(x,y,z)+Fy(x,y,z)*Fy(x,y,z)+Fz(x,y,z)*Fz(x,y,z)
	      +(4+2*AR2)*MU;
	  }
	}
      }
    
      // Setup of modified incomplete Cholesky preconditioner
      // Also needed only once since matrix does not change
      wishGVF::MIC_setup(a, vc, MU, AR2, omega, X, Y, Z);
      
      //
      // Computation of first component
      //
    
      cerr << "Iterating on first component." << endl;
      wishGVF::rhs_setup_cg(Fx, vc, rhs, MU, X, Y, Z, AR2);
      iter_x =  wishGVF::pcg(u,rhs,vc,r,p,w,v,MU,AR2,X,Y,Z,float(m_epsilon),a,omega);
    
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).x=u[k]; 
	  }
	}
      }
    
      cerr << "Iterating on second component." << endl; 
      wishGVF::rhs_setup_cg(Fy, vc, rhs, MU, X, Y, Z, AR2);
      iter_y =  wishGVF::pcg(u,rhs,vc,r,p,w,v,MU,AR2,X,Y,Z,float(m_epsilon),a,omega);
    
    
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).y=u[k]; 
	  }
	}
      }
    
      cerr << "Iterating on third component." << endl; 
      wishGVF::rhs_setup_cg(Fz, vc, rhs, MU, X, Y, Z, AR2);
      iter_z =  wishGVF::pcg(u,rhs,vc,r,p,w,v,MU,AR2,X,Y,Z,float(m_epsilon),a,omega);
    
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*X*Y+y*X+x;
	    out(x,y,z).z=u[k]; 
	  }
	}
      }
    
      cerr << "GVF_CONJUGATE_GRADIENTS Finished!" << endl;
      cerr << "Summary: " << endl;
      cerr << "x-iter = " << iter_x << endl;
      cerr << "y-iter = " << iter_y << endl;
      cerr << "z-iter = " << iter_z << endl;
      cerr << "mean-iter = " << (iter_x+iter_y+iter_z)/3.0 << endl;
      cerr << "omega = " << omega << endl;
      
      delete u;
      delete vc;
      delete rhs;
      delete v;
      delete w;
      delete r;
      delete p;
      delete a;
    }
  
  
    // ###############
    // MULTIGRID SOLVER
    // ###############    
    else if( m_solver_type == GVF_MULTIGRID ) {
      
      int k;

      // Padding to obtain N+1 gridpoints in each dimension, 
      // where N is divisible by p
      // If a "multi" in multigrid is l, then p must be 2^(l-1)
      int p = 4;
      int Xp, Yp, Zp;      
      Xp = wishGVF::pad(X,p);
      Yp = wishGVF::pad(Y,p);
      Zp = wishGVF::pad(Z,p);
      
      cerr << "wish::GradientVectorFlowFilter - Multigrid" << endl;
      cerr << "X = " << X << '\t' << "Xp = " << Xp << endl;
      cerr << "Y = " << Y << '\t' << "Yp = " << Yp << endl;
      cerr << "Z = " << Z << '\t' << "Zp = " << Zp << endl;
      
      // grid functions
      float *w = new float[Xp*Yp*Zp];    // unknown
      float *rhs = new float[Xp*Yp*Zp];  // right hand side
      float *vc = new float[Xp*Yp*Zp];   // variable coefficient
    
      // Multigrid stuff
      int X2, Y2, Z2, X3, Y3, Z3, nu1, nu2, nu3; 
      X2 = (Xp+1)/2;  Y2 = (Yp+1)/2;  Z2 = (Zp+1)/2;
      X3 = (X2+1)/2;  Y3 = (Y2+1)/2;  Z3 = (Z2+1)/2;
    
      nu1 = 2;  // Presmothing iterations
      nu2 = 2;  // Postsmothing iterations
      nu3 = 7;  // Iterations on coarsest grid (non-standard notation)
      // 5 OK for SSOR and SOR, 7 needed for Jacobi
      float omega = 1.1f;  //1.1 good for SSOR
    
      // Grid vectors for coarser grids
      float *u2 = new float[X2*Y2*Z2];
      float *u3 = new float[X3*Y3*Z3];
    
      float *vc2 = new float[X2*Y2*Z2];
      float *vc3 = new float[X3*Y3*Z3];
    
      float *rhs2 = new float[X2*Y2*Z2];
      float *rhs3 = new float[X3*Y3*Z3];
      
      // Performance results
      int iter_x, iter_y, iter_z;
    
      
      // Some compilers does not do this at some optimization levels
      memset(w,0,sizeof(float)*Xp*Yp*Zp);
      memset(rhs,0,sizeof(float)*Xp*Yp*Zp);
      memset(vc,0,sizeof(float)*Xp*Yp*Zp);
      
      // For safety and from experience
      memset(u2,0,sizeof(float)*X2*Y2*Z2);
      memset(vc2,0,sizeof(float)*X2*Y2*Z2);
      memset(rhs2,0,sizeof(float)*X2*Y2*Z2);
      
      memset(u3,0,sizeof(float)*X3*Y3*Z3);
      memset(vc3,0,sizeof(float)*X3*Y3*Z3);
      memset(rhs3,0,sizeof(float)*X3*Y3*Z3);
      
      // Computing the variable coefficient and recasting it into 1D vector
      // Done once since the same coeff. is used for all three components
      for(int z=1; z<Z-1; z++) {
	for(int y=1; y<Y-1; y++) {
	  for(int x=1; x<X-1; x++) {
	    k = z*Xp*Yp+y*Xp+x;
	    vc[k] = 
	      Fx(x,y,z)*Fx(x,y,z)+
	      Fy(x,y,z)*Fy(x,y,z)+
	      Fz(x,y,z)*Fz(x,y,z);
	  }
	}
      }
      
      cerr << "Setting up multigrid..." << endl;
      // Variable coefficients on coarse grids
      wishGVF::mg_setup(vc, vc2, vc3, MU, AR2, Xp, Yp, Zp);
      cerr << "Finished!" << endl;
      
      //
      // Computation of first component
      //
      cerr << "Iterating on first component." << endl; 
      wishGVF::rhs_setup_mg(Fx, vc, rhs, MU, X, Y, Z, p, AR2);
      iter_x =  wishGVF::multigrid(w, u2, u3, rhs, rhs2, rhs3, vc, vc2, vc3,
				   MU, AR2, Xp, Yp, Zp, float(m_epsilon), nu1, nu2, nu3, omega);
      
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*Xp*Yp+y*Xp+x;
	    out(x,y,z).x=w[k]; 
	  }
	}
      }
      
      cerr << "Iterating on second component." << endl; 
      wishGVF::rhs_setup_mg(Fy, vc, rhs, MU, X, Y, Z, p, AR2);
      iter_y =  wishGVF::multigrid(w, u2, u3, rhs, rhs2, rhs3, vc, vc2, vc3,
				   MU, AR2, Xp, Yp, Zp, float(m_epsilon), nu1, nu2, nu3, omega);
      
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*Xp*Yp+y*Xp+x;
	    out(x,y,z).y=w[k]; 
	  }
	}
      }
      
      cerr << "Iterating on third component." << endl; 
      wishGVF::rhs_setup_mg(Fz, vc, rhs, MU, X, Y, Z, p, AR2);
      iter_z =  wishGVF::multigrid(w, u2, u3, rhs, rhs2, rhs3, vc, vc2, vc3,
				   MU, AR2, Xp, Yp, Zp, float(m_epsilon), nu1, nu2, nu3, omega);
      
      // Preparing for writing to file
      for(int z=0; z<Z; z++) {
	for(int y=0; y<Y; y++) {
	  for(int x=0; x<X; x++) {
	    k = z*Xp*Yp+y*Xp+x;
	    out(x,y,z).z=w[k]; 
	  }
	}
      }
      
      cerr << "GVF_MULTIGRID Finished!" << endl;
      cerr << "Summary: " << endl;
      cerr << "x-iter = " << iter_x << endl;
      cerr << "y-iter = " << iter_y << endl;
      cerr << "z-iter = " << iter_z << endl;
      cerr << "mean-iter = " << (iter_x+iter_y+iter_z)/3.0 << endl;
      cerr << "omega = " << omega << endl;
      
      delete w;
      delete u2;
      delete u3;
      delete vc;
      delete vc2;
      delete vc3;
      delete rhs; 
      delete rhs2;
      delete rhs3;
    }
    
    else {
      // this can not happen 
    }
  
  }

} // end namespace wish

#endif




