///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_IMAGE_HH_
#define _WISH_IMAGE_HH_

#include "wish.hh"
#include "wishutil.hh"
#include "wishLinalg.hh"
#include "wishColor.hh"
#include "wishImageObject.hh"

namespace wish {
  
  // class ImageBase
  // Template class that have information about dimension, spacing, ...
  // Also have a lot of function for accessing data. This class is intended 
  // to be used for any data type, none of the template member functions
  // should contain anything special. Used as a parent class
  // for the Image class that is restricted to scalar values, vectors, 
  // and colors. See 'class Image' below.
  template<class TVoxelType> 
  class ImageBase : public ImageObject {
    
  protected:
    //! image dimensions
    int w, h, d;
    
    //! voxel spacing
    double sx, sy, sz;
    
    //! data buffer
    TVoxelType *data;
    
    //! flag if we own the buffer or not
    bool data_owner;
    
    //! information string
    std::string info;
    
    //! label
    std::string label;
    
  public:
    
    //! typedefs
    typedef std::auto_ptr< ImageBase<TVoxelType> > Pointer;
    typedef TVoxelType VoxelType;
    
    //! constructor
    ImageBase(int _w=0, 
	      int _h=0, 
	      int _d=1,
	      double _sx=1.0, 
	      double _sy=1.0, 
	      double _sz=1.0,
	      TVoxelType *_data=0,
	      bool copy_data=false) : 
      w(_w), 
      h(_h), 
      d(_d), 
      sx(_sx), 
      sy(_sy), 
      sz(_sz),
      data(0),
      data_owner(true)
    {
      // allocate memory      
      if( _data == 0 ) {
	allocate(_w,_h,_d);
      }
      else if( !copy_data ) {
	data=_data;
      }
      else {
	allocate(_w,_h,_d);
	memcpy(data, _data, totalSize()*sizeof(TVoxelType));
      }

      label = "Image";
    }
    
    //! destructor
    virtual ~ImageBase() { 
      if( data_owner )
	delete [] data; 
    }
    
    // purposely not implemented
  private:
    //! copy constructor
    ImageBase(const ImageBase<TVoxelType> &vm);
    
    //! assignment operator
    ImageBase<TVoxelType>& operator= (const ImageBase<TVoxelType> &vm);
            
  public:
    
    //! set information string
    void setInfo(const std::string &_info) {
      info = _info;
    }

    //! get information string
    std::string getInfo() const {
      return info;
    }

    //! set label string
    void setLabel(const std::string &_lbl) {
      label = _lbl;
    }
    
    //! get label string
    std::string getLabel() const {
      return label;
    }
    
    //! print datatype info 
    void printSelf() const;    
    
    //! get datatype info as string
    std::string getDataType() const;
    
    //! allocate memory
    void allocate(int _w=0, int _h=0, int _d=1) {
      w = _w; 
      h = _h; 
      d = _d;
      
      if( data_owner )
	delete [] data;
      
      if( totalSize() > 0 ) {
	data = new TVoxelType[totalSize()];
	memset(data, 0, totalSize()*sizeof(TVoxelType)); // set to zero
      }
      else {
	data = 0;
      }
    }
    
    //! allocate memory
    void allocate(const Vec3i &v) {
      allocate(v.x, v.y, v.z);
    }
    
    //! set if we own data or not
    void setDataOwner(bool _data_owner) {
      data_owner=_data_owner;
    }
    
    //! set data
    void setData(TVoxelType *_data, bool copy_data=false) {
      if( data_owner ) {
	delete [] data;
      }
      if( copy_data ) {
	allocate(w,h,d);
	memcpy(data, _data, totalSize()*sizeof(TVoxelType));
      }
      else {
	data = _data;
      }
    }
    
    //! extract 2D slice from 3D (only z)
    ImageBase<TVoxelType> extractSlice(int _slice) const;
    
    //! insert 2D slice in 3D, replaces the current one
    void insertSlice(int _slice, 
		     const ImageBase<TVoxelType> &im) const;
    
    //! resize according to im, both dims and spacing
    template<class TVoxelType2>
    void resize(const ImageBase<TVoxelType2> &im) {
      allocate(im.width(),im.height(),im.depth());
      setSpacing(im.getSpacing());
    }

    //! copy image
    template<class TVoxelType2>
    void copy(const ImageBase<TVoxelType2> &im);
    
    //! reset to value t
    void reset(const TVoxelType &t);
    
    //! position to index
    int posToIndex(const Vec3i &v) const {
      return w*h*v.z + w*v.y + v.x;
    }
    
    //! index to position
    Vec3i indexToPos(int index) const {
      Vec3i temp;
      temp.z = index/(w*h);
      index = index%(w*h);
      temp.y = index/w;
      temp.x = index%w;
      return temp;
    }
    
    //! clamp coordinates
    void clampCoords(int &x, int &y, int &z) const {
      if( x < 0 )        x = 0;
      else if( x > w-1 ) x = w-1;
      if( y < 0 )        y = 0;
      else if( y > h-1 ) y = h-1;
      if( z < 0 )        z = 0;
      else if( z > d-1 ) z = d-1;
    }
    
    //! clamp coordinates
    void clampCoords(Vec3i &v) const {
      clampCoords(v.x,v.y,v.z);
    }
    
    //! mirror coordinates
    void mirrorCoords(int &x, int &y, int &z) const {
      if(x < 0)         x = -x-1;
      else if( x >= w ) x = 2*w-x-1;
      if(y < 0)         y = -y-1;
      else if( y >= h ) y = 2*h-y-1;
      if(z < 0)         z = -z-1;
      else if( z >= d ) z = 2*d-z-1;
    }
    
    //! mirror coordinates
    void mirrorCoords(Vec3i &v) const {
      mirrorCoords(v.x,v.y,v.z);
    }
    
    //! set voxel spacing
    void setSpacing(double _sx, double _sy, double _sz) {
      sx = _sx; sy = _sy; sz = _sz;
    }
    
    //! set voxel spacing
    void setSpacing(const Vec3f &sp) {
      sx = sp.x; sy = sp.y; sz = sp.z;
    }
    
    //! get voxel spacing
    double getSpacing(int i) const {
      if(i==0)      
	return sx;
      else if(i==1) 
	return sy;
      else if(i==2) 
	return sz;
      else          
	return -1.0;
    }
    
    //! get voxel spacing
    Vec3f getSpacing() const { 
      return Vec3f((WISH_FLOAT)sx,(WISH_FLOAT)sy,(WISH_FLOAT)sz); 
    }
    
    //! get the total extent
    Vec3f totalExtent() const { 
      return Vec3f(WISH_FLOAT(w*sx), WISH_FLOAT(h*sy), WISH_FLOAT(d*sz)); 
    }
    Vec3f getTotalExtent() const { 
      return totalExtent();
    }
    
    //! returns image width
    int width() const { 
      return w; 
    }
    
    int getWidth() const {
      return w;
    }
    
    //! returns image height
    int height() const { 
      return h; 
    }
    
    //! returns image height
    int getHeight() const { 
      return h; 
    }
    
    //! return image depth
    int depth() const { 
      return d; 
    }
    
    //! return image depth
    int getDepth() const { 
      return d; 
    }
    
    //! return dimensions as Vec3i
    Vec3i getSize() const {
      return Vec3i(w,h,d);
    }

    //! return dimensions as Vec3i
    Vec3i getDimensions() const {
      return Vec3i(w,h,d);
    }

    //! return dimensions (1,2, or 3)  
    int getDim() const {
      return int(w > 1) + int(h > 1) + int(d > 1);
    }
    
    //! returns number of voxels
    int totalSize() const { 
      return w*h*d; 
    }
    
    //! returns number of voxels
    int getTotalSize() const { 
      return w*h*d; 
    }
    
    //! tests if (x,y,z) is inside image
    bool isInside(int x, int y, int z) const {
      return (x>=0 && x<w && y>=0 && y<h && z>=0 && z<d);
    }
    
    //! tests if v is inside image
    bool isInside(const Vec3i &v) const {
      return isInside(v.x, v.y, v.z);
    }
    
    //! returns bytes per voxel
    int voxelBytes() const { return sizeof(TVoxelType); }

    //! returns bytes per voxel
    int getVoxelBytes() const { return sizeof(TVoxelType); }
    
    //! returns image size in bytes
    int totalBytes() const { return voxelBytes()*totalSize(); }

    //! returns image size in bytes
    int getTotalBytes() const { return voxelBytes()*totalSize(); }
    
    //! get the buffer pointer
    TVoxelType* getData() const { return data; }

    //! get the buffer pointer as byte
    byte* getBytePointer() const { return (byte*) data; }
    
    //! get element i in buffer
    TVoxelType& operator[] (int i) const { 
      return data[i]; 
    }
    
    //! returns reference to voxel (x,y,z)
    TVoxelType& operator() (int x, int y, int z=0) const { 
      return data[w*h*z + w*y + x];
    }
    
    //! returns reference to voxel v
    TVoxelType& operator() (const Vec3i &v) const { 
      return operator()(v.x,v.y,v.z);
    }
    
    //! returns reference to voxel (x,y,z)
    TVoxelType& at(int x, int y, int z=0) const {
      return data[w*h*z + w*y + x];
    }
    
    //! returns reference to voxel v
    TVoxelType& at(const Vec3i &v) const {
      return at(v.x,v.y,v.z);
    }
    
    //! returns reference to voxel with index i in data
    TVoxelType& at(int i) const {
      return data[i];
    }
    
    //! get value at (x,y,z)
    TVoxelType get(int x, int y, int z=0) const {
      return data[w*h*z + w*y + x];
    }
    
    //! get value at v
    TVoxelType get(const Vec3i &v) const {
      return get(v.x,v.y,v.z);
    }
    
    //! set value at (x,y,z)
    void set(const TVoxelType &t, int x, int y, int z=0) {
      data[w*h*z + w*y + x] = t;
    }
    
    //! set value at v
    void set(const TVoxelType &t, const Vec3i &v) {
      set(t,v.x,v.y,v.z);
    }
    
    //! safe get function
    TVoxelType getSafe(int x, int y, int z=0) const {
      clampCoords(x,y,z);
      return data[w*h*z + w*y + x];
      //return getSafeConstant(TVoxelType(0),x,y,z);
    }
    
    //! safe get function
    TVoxelType getSafe(const Vec3i &v) const {
      return getSafe(v.x,v.y,v.z);
    }
    
    //! safe get function (clamp coordinate first)
    TVoxelType getSafeClamp(int x, int y, int z=0) const {
      clampCoords(x,y,z);
      return data[w*h*z + w*y + x];
    }
    
    //! safe get function (clamp coordinate first)
    TVoxelType getSafeClamp(const Vec3i &v) const {
      return getSafeClamp(v.x,v.y,v.z);
    }
    
    //! safe get function (mirror coordinate first)
    TVoxelType getSafeMirror(int x, int y, int z=0) const {
      mirrorCoords(x,y,z);
      return data[w*h*z + w*y + x];
    }
    
    //! safe get function (mirror coordinate first)
    TVoxelType getSafeMirror(const Vec3i &v) const {
      return getSafeMirror(v.x,v.y,v.z);
    }
    
    //! safe get function (return constant if outside)
    TVoxelType getSafeConstant(const TVoxelType &t, 
			       int x, int y, int z=0) const {
      if( isInside(x,y,z) )
	return data[w*h*z + w*y + x];
      else
	return t;
    }
    
    //! safe get function (return constant if outside)
    TVoxelType getSafeConstant(const TVoxelType &t, 
			       const Vec3i &v) const {
      return getSafeConstant(t,v.x,v.y,v.z);
    }
    
    //! safe set function (set only if inside image)
    void setSafe(const TVoxelType &t, int x, int y, int z=0) {
      if( isInside(x,y,z) )
	data[w*h*z + w*y + x] = t;
    }
    
    //! safe set function (set only if inside image)
    void setSafe(const TVoxelType &t, const Vec3i &v) {
      setSafe(t,v.x,v.y,v.z);
    }
    

    
  };
  
  //! Image
  //! This class is intended to be used for common datatypes such as 
  //! the standard signed and unsigned integers and floats. We also 
  //! implement the functions for Vec3f, RGB8, and RGBA8
  template<class TVoxelType>
  class Image : public ImageBase<TVoxelType> {
    
  public:
    // static functions
    
    //! returns maximum/minimum value that can be stored 
    //! (use only for integer types)
    static double maxValue();
    static double minValue();
    
    //! constructor
    Image(int _w=0, 
	  int _h=0, 
	  int _d=1,
	  double _sx=1.0, 
	  double _sy=1.0, 
	  double _sz=1.0,
	  TVoxelType *_data=0,
	  bool copy_data=false) :
      ImageBase<TVoxelType>(_w, _h, _d, 
			    _sx, _sy, _sz, 
			    _data, copy_data) {}
    
    //! destructor
    ~Image() {}
    
    //! copy constructor
    Image(const Image<TVoxelType> &vm) : 
      ImageBase<TVoxelType>(vm.w, vm.h, vm.d, 
 			    vm.sx, vm.sy, vm.sz, 
 			    vm.data, true) {}
        
    //! assignment operator
    Image<TVoxelType>& operator= (const Image<TVoxelType> &vm);
    
    //! assignment operator (another class)
    template<class TVoxelType2>
    Image<TVoxelType>& operator= (const Image<TVoxelType2> &vm);

    //! extract 2D slice from 3D (only z)
    Image<TVoxelType> extractSlice(int _slice) const;
    
    //! insert 2D slice in 3D, replaces the current one
    void insertSlice(int _slice, 
		     const Image<TVoxelType> &im) const;
    
    //! true if unsigned integer
    bool isUnsignedIntegerType() const;
    
    //! true if signed integer
    bool isSignedIntegerType() const;
    
    //! get maximum value stored in image
    TVoxelType getMax() const;
    
    //! get minimum value in image
    TVoxelType getMin() const;
    
    // access functions for double
    double getAsDoubleFromArray(int i) const;
    
    double getAsDouble(int x, 
		       int y, 
		       int z=0) const;
    double getAsDouble(const Vec3i &v) const {
      return getAsDouble(v.x,v.y,v.z);
    }
    
    double getSafeAsDouble(int x, 
			   int y, 
			   int z=0) const;
    double getSafeAsDouble(const Vec3i &v) const {
      return getSafeAsDouble(v.x,v.y,v.z);
    }
    
    // get normalized values (0-255 -> 0-1, 0-65535 -> 0-1)
    double getNormalizedFromArray(int i) const;
    
    double getNormalized(int x, 
			 int y, 
			 int z=0) const;
    
    double getNormalized(const Vec3i &v) const {
      return getNormalized(v.x,v.y,v.z);
    }
    
    double getSafeNormalized(int x, 
			     int y, 
			     int z=0) const;
    
    double getSafeNormalized(const Vec3i &v) const {
      return getSafeNormalized(v.x,v.y,v.z);
    }
    
    double getNormalizedInterpd(double x, 
				double y, 
				double z=0) const;
    
    double getNormalizedInterpd(const Vec3f &v) const {
      return getNormalizedInterpd(v.x,v.y,v.z);
    }
    
    // set normalized (value in [0,1])
    void setNormalized(double value,
		       int x,
		       int y,
		       int z);
    
    void setNormalized(double value, const Vec3i& v) {
      setNormalized(value, v.x, v.y, v.z);
    }

    void setNormalizedToArray(double value, int i);
    
    //! tri-linear interpolation, returns a double
    double getInterpd(double x, 
		      double y, 
		      double z=0) const;
    
    //! tri-linear interpolation, returns a double
    double getInterpd(const Vec3f &v) const {
      return getInterpd(v.x,v.y,v.z);
    }
    
    //! tri-linear interpolation
    TVoxelType getInterp(double x, 
			 double y, 
			 double z=0) const;
    
    //! tri-linear interpolation
    TVoxelType getInterp(const Vec3f &v) const {
      return getInterp(v.x,v.y,v.z);
    }

    // the arithmetics are intended for floats, be careful
    //! arithmetic operator
    template<class TVoxelType2>
    Image<TVoxelType> operator+(const Image<TVoxelType2> &im) const;
    
    //! arithmetic operator
    template<class TVoxelType2>
    Image<TVoxelType> operator-(const Image<TVoxelType2> &im) const;
    
    //! arithmetic operator
    template<class TVoxelType2>
    Image<TVoxelType> operator*(const Image<TVoxelType2> &im) const;
    
    //! arithmetic operator
    template<class TVoxelType2>
    Image<TVoxelType> operator/(const Image<TVoxelType2> &im) const; 
    
    //! arithmetic operator
    Image<TVoxelType> operator+(double t) const;
    
    //! arithmetic operator
    Image<TVoxelType> operator-(double t) const;
    
    //! arithmetic operator
    Image<TVoxelType> operator*(double t) const;
    
    //! arithmetic operator
    Image<TVoxelType> operator/(double t) const;
    
    // friends
    //! arithmetic operator
    template<class X>
    friend Image<X>
    operator+(double t, const Image<X> &im);
    
    //! arithmetic operator
    template<class X>
    friend Image<X> 
    operator-(double t, const Image<X> &im);
    
    //! arithmetic operator
    template<class X>
    friend Image<X> 
    operator-(const Image<X> &im);
    
    //! arithmetic operator
    template<class X>
    friend Image<X> 
    operator*(double t, const Image<X> &im);
    
    //! arithmetic operator
    template<class X>
    friend Image<X> 
    operator/(double t, const Image<X> &im);

    //! logic operator
    template<class TVoxelType2>
    Image<TVoxelType> AND(const Image<TVoxelType2> &im) const; 
    
    //! logic operator
    template<class TVoxelType2>
    Image<TVoxelType> OR(const Image<TVoxelType2> &im) const; 
    
    //! logic operator
    template<class TVoxelType2>
    Image<TVoxelType> XOR(const Image<TVoxelType2> &im) const; 
    
    //! logic operator
    Image<TVoxelType> NOT() const; 

    //! The following functions are used to 
    //! transform positions and directions between the two coordinate
    //! systems IMAGE and WORLD. In the IMAGE coordinate system
    //! we have coords in the interval
    //!
    //! ([0.0,w], [0.0,h], [0.0,d]); ([0,w-1], [0,h-1], [0,d-1]) for integers) 
    //!
    //! Here we have the x-axis pointing to the right, the y-axis pointing
    //! down, and the z-axis pointing inwards. The origin is located at the top
    //! left corner of the first slice.
    //!
    //! In the WORLD coordinate system we have coords in the interval
    //!
    //! ([-w*sx/2,w*sx/2], [-h*sy/2,h*sy/2], [-d*sz/2,d*sz/2])
    //!
    //! Here we have the x-axis pointing to the right, the y-axis pointing
    //! up, and the z-axis pointing against us 
    //! (a standard CG coordinate system). The origin is located in
    //! the center of the volume. The IMAGE TO WORLD transform
    //! is therefore M = RTS, where S is a non-uniform scaling (sx,sy,sz),
    //! T is the transform (-w*sx/2, -h*sy/2, -d*sz/2), and
    //! R is the 180 degree rotation about the x-axis ( 1, 0, 0,
    //!                                                 0,-1, 0,
    //!                                                 0, 0,-1)
    //! 
    //! The WORLD TO IMAGE transform is just the inverse of M.
    //!
    //! It is also possible to get the 3D texture matrix from
    //! WORLD to TEXTURE coords. The TEXTURE coords are the normalized
    //! IMAGE coords.
    //!
    //! Regarding vectors (normals and gradients) you have to be a bit 
    //! careful. If you compute the gradient with the getGradient function
    //! the spacing is taken into account, so the direction is correct except
    //! for the rotation about the x-axis so you have to 
    //! perform that one with the imageToWorldR that only uses
    //! the rotational part (there is also a 
    //! worldToImageR that is identical).
    //! If you have a vector defined in the IMAGE 
    //! coordinates and want to transform it to WORLD coordinates 
    //! (or vice versa) you need to take both rotation and 
    //! spacing into account, 
    //! this is done in the functions imageToWorldRS and worldToImageRS
    //! that use the formula M_normal = transpose( inverse(M_geometry) ) from
    //! computer graphics. It becomes very simple as you can see.
    //!   
    
    // matrices    
    Matrix4f getImageToWorldMatrix() const {
      Vec3f te = this->totalExtent();
      Vec3f sp = this->getSpacing();
      return 
	rotation<float>( 180.0f, Vec3f(1,0,0) )*
	translation<float>(-te.x/2.0f, -te.y/2.0f, -te.z/2.0f)*
	scaling<float>(sp.x, sp.y, sp.z);
    }
    
    Matrix4f getWorldToImageMatrix() const {
      Vec3f te = this->totalExtent();
      Vec3f sp = this->getSpacing();
      return 
	scaling<float>(1.0f/sp.x, 1.0f/sp.y, 1.0f/sp.z)*
	translation<float>(te.x/2.0f, te.y/2.0f, te.z/2.0f)*      
	rotation<float>( -180.0f, Vec3f(1,0,0) );
      //return getImageToWorldMatrix().transformInverse();
    }
    
    Matrix4f getWorldToTextureMatrix() const {
      Vec3f te = this->getTotalExtent();
      return 
	translation<float>(0.5f,0.5f,0.5f)*
	scaling<float>(1.0f/te.x, 1.0f/te.y, 1.0f/te.z)*
	rotation<float>( 180.0f, Vec3f(1,0,0) );
    }
    
    Matrix4f getImageToTextureMatrix() const {
      return getWorldToTextureMatrix()*getImageToWorldMatrix();
    }
    
    Vec3f worldToImage(const Vec3f &v) const {
      Matrix4f M = this->getWorldToImageMatrix();
      return M*v;
    }
    
    Vec3f imageToWorld(const Vec3f &v) const {
      Matrix4f M = this->getImageToWorldMatrix();
      return M*v;
    }
    
    // RS = rotate and scale (rotational and scaling part)
    Vec3f imageToWorldRS(const Vec3f &v) const {
      Matrix3f M = this->getImageToWorldMatrix().getScaleRotation();
      return M*v;
    }
    
    Vec3f worldToImageRS(const Vec3f &v) const {
      Matrix3f M = this->getWorldToImageMatrix().getScaleRotation();
      return M*v;
    }
    
    // R = rotation (only the rotational part (180 deg about the x-axis))
    Vec3f imageToWorldR(const Vec3f &v) const {
      Matrix3f M = this->getImageToWorldMatrix().getRotation();
      return M*v;
    }
    
    Vec3f worldToImageR(const Vec3f &v) const {
      Matrix3f M = this->getWorldToImageMatrix().getRotation();
      return M*v;
    }
    
    //! forward difference in x-direction
    double getForwardDX(int x, int y, int z=0) const {
      return 
	( getSafeAsDouble(x+1,y,z) - getSafeAsDouble(x,y,z) ) / this->sx;
    }
    
    //! forward difference in y-direction
    double getForwardDY(int x, int y, int z=0) const {
      return 
	( getSafeAsDouble(x,y+1,z) - getSafeAsDouble(x,y,z) ) / this->sy;
    }

    //! forward difference in z-direction
    double getForwardDZ(int x, int y, int z) const {
      return 
	( getSafeAsDouble(x,y,z+1) - getSafeAsDouble(x,y,z) ) / this->sz;
    }

    //! backward difference in x-direction
    double getBackwardDX(int x, int y, int z=0) const {
      return 
	( getSafeAsDouble(x,y,z) - getSafeAsDouble(x-1,y,z) ) / this->sx;
    }

    //! backward difference in y-direction
    double getBackwardDY(int x, int y, int z=0) const {
      return 
	( getSafeAsDouble(x,y,z) - getSafeAsDouble(x,y-1,z) ) / this->sy;
    }

    //! backward difference in z-direction
    double getBackwardDZ(int x, int y, int z) const {
      return 
	( getSafeAsDouble(x,y,z) - getSafeAsDouble(x,y,z-1) ) / this->sz;
    }

    //! centered difference in x-direction
    double getCenteredDX(int x, int y, int z=0) const {
      return 
   	( getSafeAsDouble(x+1,y,z) - getSafeAsDouble(x-1,y,z) ) / (2*this->sx);
    }

    //! centered difference in y-direction
    double getCenteredDY(int x, int y, int z=0) const {
      return 
   	( getSafeAsDouble(x,y+1,z) - getSafeAsDouble(x,y-1,z) ) / (2*this->sy);
    }

    //! centered difference in z-direction
    double getCenteredDZ(int x, int y, int z) const {
      return 
   	( getSafeAsDouble(x,y,z+1) - getSafeAsDouble(x,y,z-1) ) / (2*this->sz);
    }
  
    //! second derivative in xx
    double getDXX(int x, int y, int z=0) const {
      return 
   	( getSafeAsDouble(x+1,y,z)-
   	  2*getSafeAsDouble(x,y,z)+
   	  getSafeAsDouble(x-1,y,z) ) / util::square(this->sx);
    }
    
    double getDXX(const Vec3i &v) const {
      return getDXX(v.x,v.y,v.z);
    }

    //! second derivative in yy
    double getDYY(int x, int y, int z=0) const {
      return 
  	( getSafeAsDouble(x,y+1,z)-
   	  2*getSafeAsDouble(x,y,z)+
   	  getSafeAsDouble(x,y-1,z) ) / util::square(this->sy);
    }

    double getDYY(const Vec3i &v) const {
      return getDYY(v.x,v.y,v.z);
    }

    //! second derivative in zz
    double getDZZ(int x, int y, int z) const {
      return 
   	( getSafeAsDouble(x,y,z+1)-
   	  2*getSafeAsDouble(x,y,z)+
   	  getSafeAsDouble(x,y,z-1) ) / util::square(this->sz);
    }

    double getDZZ(const Vec3i &v) const {
      return getDZZ(v.x,v.y,v.z);
    }
    
    //! second derivative in xy
    double getDXY(int x, int y, int z=0) const {
      return 
   	(0.5*getCenteredDX(x,y+1,z)-0.5*getCenteredDX(x,y-1,z))/this->sy;
    }

    double getDXY(const Vec3i &v) const {
      return getDXY(v.x,v.y,v.z);
    }
  
    //! second derivative in xz
    double getDXZ(int x, int y, int z) const {
      return 
   	(0.5*getCenteredDX(x,y,z+1)-0.5*getCenteredDX(x,y,z-1))/this->sz;
    }

    double getDXZ(const Vec3i &v) const {
      return getDXZ(v.x,v.y,v.z);
    }
    
    //! second derivative in yz
    double getDYZ(int x, int y, int z) const {
      return 
   	(0.5*getCenteredDY(x,y,z+1)-0.5*getCenteredDY(x,y,z-1))/this->sz;
    }

    double getDYZ(const Vec3i &v) const {
      return getDYZ(v.x,v.y,v.z);
    }
    
    //! gradient (centered differences)
    Vec3f getGradient(int x, int y, int z=0) const;
    
    //! gradient (centered differences)
    Vec3f getGradient(const Vec3i &v) const {
      return getGradient(v.x,v.y,v.z);
    }
    
    //! linearly interpolated gradient
    Vec3f getGradientInterp(double x, 
			    double y, 
			    double z=0) const;
    
    //! linearly interpolated gradient
    Vec3f getGradientInterp(const Vec3f &v) const {
      return getGradientInterp(v.x,v.y,v.z);
    }
    
    //! file reader
    bool readImage(const std::string &filename,
		   int _w=0, 
		   int _h=0, 
		   int _d=1,
		   double _sx=1.0, 
		   double _sy=1.0, 
		   double _sz=1.0);
    
    //! VTK file reader
    bool readVTK(const std::string &filename);
    
    //! VTK file reader help function
    bool checkVTKDataType(const std::string &scvstring, 
			  const std::string &typestring) const;
    
    //! VTK file reader help function
    void readVTKAscii(std::ifstream &is);
    
    //! VTK file reader help function
    void readVTKBinary(std::ifstream &is);
    
    //! VTK file writer
    bool writeVTK(const std::string &filename,
		  const std::string &mode="BINARY") const;
    
    //! VTK file writer help function
    void writeVTKDataTypeInfo(std::ofstream &os) const;
    
    //! VTK file writer help function
    void writeVTKAscii(std::ofstream &os) const;

    //! VTK file writer help function
    void writeVTKBinary(std::ofstream &os) const;
    
    //! Pixy file reader
    bool readPixy(const std::string &filename);
    
    //! Pixy file writer, set reg to true if writing a region
    bool writePixy(const std::string &filename, 
		   bool reg=false) const;
    
    //! raw file reader
    bool readRaw(const std::string &filename, 
		 int _w=0, 
		 int _h=0, 
		 int _d=1,
		 double _sx=1.0, 
		 double _sy=1.0, 
		 double _sz=1.0);
    
    //! raw file writer
    bool writeRaw(const std::string &filename) const;
  };
   
  // typedefs
  typedef Image<uint8>  Image8;
  typedef Image<uint16> Image16;
  typedef Image<uint32> Image32;
  typedef Image<float>  Imagef;
  typedef Image<double> Imaged;
  typedef Image<RGB8>   ImageRGB8;
  typedef Image<RGBA8>  ImageRGBA8;
  
} // namespace wish

#include "wishImage.tpp"

#endif
 

