///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_IMAGE_TPP_
#define _WISH_IMAGE_TPP_

#include "wishImage.hh"
#include "wishColor.hh"

namespace wish {
    
  // ############ ImageBase ##############
  
  template<class TVoxelType>
  void ImageBase<TVoxelType>::printSelf() const {
    cout << "Image label: " << this->getLabel() << endl;
    cout << "Image type:  " << this->getDataType() << endl;
    cout << "Dimensions:  " << this->getDimensions() << endl;
    cout << "Spacing:     " << this->getSpacing() << endl;
    cout << "Information: " << this->getInfo() << endl;
  }
  
  // if unknown
  template<class TVoxelType>
  std::string ImageBase<TVoxelType>::getDataType() const {
    return std::string( typeid(TVoxelType).name() );
  }
  
  template<>
  inline std::string ImageBase<uint8>::getDataType() const {
    return "uint8";
  }
  
  template<>
  inline std::string ImageBase<int8>::getDataType() const {
    return "int8";
  }

  template<>
  inline std::string ImageBase<uint16>::getDataType() const {
    return "uint16";
  }

  template<>
  inline std::string ImageBase<int16>::getDataType() const {
    return "int16";
  }

  template<>
  inline std::string ImageBase<uint32>::getDataType() const {
    return "uint32";
  }

  template<>
  inline std::string ImageBase<int32>::getDataType() const {
    return "int32";
  }

  template<>
  inline std::string ImageBase<float>::getDataType() const {
    return "float";
  }

  template<>
  inline std::string ImageBase<double>::getDataType() const {
    return "double";
  }

  template<>
  inline std::string ImageBase<Vec3f >::getDataType() const {
    return "Vec3f";
  }

  template<>
  inline std::string ImageBase<RGB8>::getDataType() const {
    return "RGB8";
  }

  template<>
  inline std::string ImageBase<RGBA8>::getDataType() const {
    return "RGBA8";
  }
  
  // extract slice
  template<class TVoxelType>
  ImageBase<TVoxelType> ImageBase<TVoxelType>::extractSlice(int _slice) const {
    assert(0 <= _slice && _slice < this->depth());
    ImageBase<TVoxelType> slc(this->width(),this->height(),1);
    for(int y=0; y<this->height(); y++)
      for(int x=0; x<this->width(); x++)
	slc(x,y,0) = this->operator()(x,y,_slice);
    
    return slc;
  }

  // extract slice
  template<class TVoxelType>
  void ImageBase<TVoxelType>::insertSlice(int _slice, 
					  const ImageBase<TVoxelType> &im) const {
    assert(0 <= _slice && _slice < this->depth());
    if( im.width() != this->width() ||
	im.height() != this->height() ) {
      cerr << "insertSlice: Dimension mismatch!" << endl;
      return;
    }
    for(int y=0; y<this->height(); y++)
      for(int x=0; x<this->width(); x++)
	this->at(x,y,_slice) = im(x,y);
  }
  
  //! copy image
  template<class TVoxelType>
  template<class TVoxelType2>
  void ImageBase<TVoxelType>::copy(const ImageBase<TVoxelType2> &im) {
    resize(im);
    for(int i=0; i<this->totalSize(); i++)
      (*this)[i] = TVoxelType(im[i]);
  }
  
  //! reset to value t
  template<class TVoxelType>
  void ImageBase<TVoxelType>::reset(const TVoxelType &t) {
    for(int i=0; i<this->totalSize(); i++)
      data[i] = t;
  }
  
  // ############ Image ##############

  // assignment operator  
  template<class TVoxelType>
  Image<TVoxelType>& 
  Image<TVoxelType>::operator=(const Image<TVoxelType> &im) {    
    if( this != &im ) {
      this->resize(im);
      memcpy(this->getData(), im.getData(), this->totalBytes());
    }
    
    return *this;
  }
    
  // assignment operator from a different class
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType>& 
  Image<TVoxelType>::operator=(const Image<TVoxelType2> &im) {
    
    if( static_cast<const void*>(this) != static_cast<const void*>(&im) ) {
      this->resize(im);
      for(int i=0; i<this->totalSize(); i++)
	(*this)[i] = TVoxelType(im[i]);
    }
    
    return *this;
  }
  
  // maxvalue/minvalue
  template<class TVoxelType>
  double Image<TVoxelType>::maxValue() {
    return double( std::numeric_limits<TVoxelType>::max() );
  }
  
  template<class TVoxelType>
  double Image<TVoxelType>::minValue() {
    return double( std::numeric_limits<TVoxelType>::min() );
  }
  
  template<>
  inline double Image<float>::maxValue() {
    return 1.0f;
  }
  
  template<>
  inline double Image<float>::minValue() {
    return 0.0f;
  }
  
  template<>
  inline double Image<double>::maxValue() {
    return 1.0f;
  }
  
  template<>
  inline double Image<double>::minValue() {
    return 0.0f;
  }
  
  // extract slice
  template<class TVoxelType>
  Image<TVoxelType> Image<TVoxelType>::extractSlice(int _slice) const {
    assert(0 <= _slice && _slice < this->depth());
    Image<TVoxelType> slc(this->width(),this->height(),1);
    for(int y=0; y<this->height(); y++)
      for(int x=0; x<this->width(); x++)
	slc(x,y,0) = this->operator()(x,y,_slice);
    
    return slc;
  }
  
  // extract slice
  template<class TVoxelType>
  void Image<TVoxelType>::insertSlice(int _slice, 
				      const Image<TVoxelType> &im) const {
    assert(0 <= _slice && _slice < this->depth());
    if( im.width() != this->width() ||
	im.height() != this->height() ) {
      cerr << "insertSlice: Dimension mismatch!" << endl;
      return;
    }
    for(int y=0; y<this->height(); y++)
      for(int x=0; x<this->width(); x++)
	this->at(x,y,_slice) = im(x,y);
  }
  
  //
  template<class TVoxelType>
  bool Image<TVoxelType>::isUnsignedIntegerType() const {
    return false;
  }
  
  template<>
  inline bool Image<uint8>::isUnsignedIntegerType() const {
    return true;
  }

  template<>
  inline bool Image<uint16>::isUnsignedIntegerType() const {
    return true;
  }
  
  template<>
  inline bool Image<uint32>::isUnsignedIntegerType() const {
    return true;
  }
  
  template<>
  inline bool Image<RGBA8>::isUnsignedIntegerType() const {
    return true;
  }
  
  template<>
  inline bool Image<RGB8>::isUnsignedIntegerType() const {
    return true;
  }
  
  //
  template<class TVoxelType>
  bool Image<TVoxelType>::isSignedIntegerType() const {
    return false;
  }

  template<>
  inline bool Image<int8>::isSignedIntegerType() const {
    return true;
  }

  template<>
  inline bool Image<int16>::isSignedIntegerType() const {
    return true;
  }
  
  template<>
  inline bool Image<int32>::isSignedIntegerType() const {
    return true;
  }
  
  // get as double from array
  template<class TVoxelType>
  double Image<TVoxelType>::getAsDoubleFromArray(int i) const {
    return double( (*this)[i] );
  }
  
  // have to implement for vectors and colors
  template<>
  inline double Image<Vec3f>::getAsDoubleFromArray(int i) const {
    return double( (*this)[i].x );
  }
  
  // return red component
  template<>
  inline double Image<RGB8>::getAsDoubleFromArray(int i) const {
    return double( (*this)[i].r );
  }
  
  // return alpha component
  template<>
  inline double Image<RGBA8>::getAsDoubleFromArray(int i) const {
    return double( (*this)[i].a );
  }
  
  // get as double
  template<class TVoxelType>
  double Image<TVoxelType>::getAsDouble(int x,
					int y,
					int z) const {
    return double( this->at(x,y,z) );
  }
  
  template<>
  inline double Image<Vec3f>::getAsDouble(int x,
					  int y,
					  int z) const {
    return double( this->at(x,y,z).x );
  }
  
  template<>
  inline double Image<RGB8>::getAsDouble(int x,
					 int y,
					 int z) const {
    return double( this->at(x,y,z).r );
  }
  
  template<>
  inline double Image<RGBA8>::getAsDouble(int x,
					  int y,
					  int z) const {
    return double( this->at(x,y,z).a );
  }
  
  // get safe as double
  template<class TVoxelType>
  double Image<TVoxelType>::getSafeAsDouble(int x,
					    int y,
					    int z) const {
    return double( this->getSafe(x,y,z) );
  }
  
  template<>
  inline double Image<Vec3f>::getSafeAsDouble(int x,
					      int y,
					      int z) const {
    return double( this->getSafe(x,y,z).x );
  }
  
  // red component
  template<>
  inline double Image<RGB8>::getSafeAsDouble(int x,
					     int y,
					     int z) const {
    return double( this->getSafe(x,y,z).r );
  }
  
  // alpha component
  template<>
  inline double Image<RGBA8>::getSafeAsDouble(int x,
					      int y,
					      int z) const {
    return double( this->getSafe(x,y,z).a );
  }
  
  // normalized from array
  template<class TVoxelType>
  double Image<TVoxelType>::getNormalizedFromArray(int i) const {
    double c = this->getAsDoubleFromArray(i);
    return 
      (c-Image<TVoxelType>::minValue())/
      (Image<TVoxelType>::maxValue()-Image<TVoxelType>::minValue());
  }
  
  template<>
  inline double Image<RGB8>::getNormalizedFromArray(int i) const {
    return 0.0;
  }

  template<>
  inline double Image<RGBA8>::getNormalizedFromArray(int i) const {
    return 0.0;
  }
  
  template<>
  inline double Image<Vec3f>::getNormalizedFromArray(int i) const {
    return 0.0;
  }
  
  // normalized
  template<class TVoxelType>
  double Image<TVoxelType>::getNormalized(int x, 
					  int y, 
					  int z) const {
    double c = this->getAsDouble(x,y,z);
    return 
      (c-Image<TVoxelType>::minValue())/
      (Image<TVoxelType>::maxValue()-Image<TVoxelType>::minValue());
  }
  
  template<>
  inline double Image<RGB8>::getNormalized(int x, 
					   int y, 
					   int z) const {
    return 0.0;
  }
  
  template<>
  inline double Image<RGBA8>::getNormalized(int x, 
					    int y, 
					    int z) const {
    return 0.0;
  }

  template<>
  inline double Image<Vec3f>::getNormalized(int x, 
					    int y, 
					    int z) const {
    return 0.0;
  }
  
  // normalized safe
  template<class TVoxelType>
  double Image<TVoxelType>::getSafeNormalized(int x, 
					      int y, 
					      int z) const {
    double c = this->getSafeAsDouble(x,y,z);
    return 
      (c-Image<TVoxelType>::minValue())/
      (Image<TVoxelType>::maxValue()-Image<TVoxelType>::minValue());
  }
  
  template<>
  inline double Image<RGB8>::getSafeNormalized(int x, 
					       int y, 
					       int z) const {
    return 0.0;
  }

  template<>
  inline double Image<RGBA8>::getSafeNormalized(int x, 
						int y, 
						int z) const {
    return 0.0;
  }

  template<>
  inline double Image<Vec3f>::getSafeNormalized(int x, 
						int y, 
						int z) const {
    return 0.0;
  }
  
  // normalized interpd
  template<class TVoxelType>
  double Image<TVoxelType>::getNormalizedInterpd(double x, 
						 double y, 
						 double z) const {
    double c = this->getInterpd(x,y,z);
    return 
      (c-Image<TVoxelType>::minValue())/
      (Image<TVoxelType>::maxValue()-Image<TVoxelType>::minValue());
  }
  
  template<>
  inline double Image<RGB8>::getNormalizedInterpd(double x, 
						  double y, 
						  double z) const {
    return 0.0;
  }
  
  template<>
  inline double Image<RGBA8>::getNormalizedInterpd(double x, 
						   double y, 
						   double z) const {
    return 0.0;
  }
  
  template<>
  inline double Image<Vec3f>::getNormalizedInterpd(double x, 
						   double y, 
						   double z) const {
    return 0.0;
  }
  
  template<class TVoxelType>
  void Image<TVoxelType>::setNormalized(double value, int x, int y, int z) {
    double v = 
      value * (Image<TVoxelType>::maxValue()-
	       Image<TVoxelType>::minValue()) +
      Image<TVoxelType>::minValue();
    this->at(x,y,z) = TVoxelType( util::round( v ) );
  }
  
  template<class TVoxelType>
  void Image<TVoxelType>::setNormalizedToArray(double value, int i) {
    double v = 
      value * (Image<TVoxelType>::maxValue()-
	       Image<TVoxelType>::minValue()) +
      Image<TVoxelType>::minValue();
    this->at(i) = TVoxelType( util::round( v ) );
  }
  
  template<>
  inline void Image<float>::setNormalized(double value, int x, int y, int z) {
    this->at(x,y,z) = (float) value;
  }
  
  template<>
  inline void Image<float>::setNormalizedToArray(double value, int i) {
    this->at(i) = (float) value;
  }

  template<>
  inline void Image<double>::setNormalized(double value, int x, int y, int z) {
    this->at(x,y,z) = value;
  }
  
  template<>
  inline void Image<double>::setNormalizedToArray(double value, int i) {
    this->at(i) = value;
  }

  template<>
  inline void Image<RGB8>::setNormalized(double value, int x, int y, int z) {
  }
  
  template<>
  inline void Image<RGB8>::setNormalizedToArray(double value, int i) {
  }

  template<>
  inline void Image<RGBA8>::setNormalized(double value, int x, int y, int z) {
  }
  
  template<>
  inline void Image<RGBA8>::setNormalizedToArray(double value, int i) {
  }
  
  template<>
  inline void Image<Vec3f>::setNormalized(double value, int x, int y, int z) {
  }
  
  template<>
  inline void Image<Vec3f>::setNormalizedToArray(double value, int i) {
  }
  
  // arithmetic ops  
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::operator+(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] + im[i] );
    
    return tmp;
  }
  
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::operator-(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
		    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] - im[i] );
    
    return tmp;
  }

  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::operator*(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
		    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] * im[i] );
    
    return tmp;
  }
  
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::operator/(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] / im[i] );
    
    return tmp;
  }
  
  template<class TVoxelType>
  Image<TVoxelType> Image<TVoxelType>::operator+(double t) const {
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] + t );
    
    return tmp;
  }
  
  template<class TVoxelType>  
  Image<TVoxelType> Image<TVoxelType>::operator-(double t) const {
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] - t );
    
    return tmp;  
  }
  
  template<class TVoxelType>  
  Image<TVoxelType> Image<TVoxelType>::operator*(double t) const {
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] * t );
    
    return tmp;
  }
  
  template<class TVoxelType>  
  Image<TVoxelType> Image<TVoxelType>::operator/(double t) const {
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++)
      tmp[i] = TVoxelType( (*this)[i] / t );
    
    return tmp;
  }
  
  // friends
  template<class X>
  Image<X> operator+(double t, const Image<X> &im) {
    return im+t;
  }
  
  template<class X>
  Image<X> operator*(double t, const Image<X> &im) {
    return im*t;
  }
  
  template<class X>
  Image<X> operator-(const Image<X> &im) {
    return -1.0*im;
  }
  
  template<class X>  
  Image<X> operator/(double t, const Image<X> &im) {
    Image<X> tmp;
    tmp.resize(im);
    
    for(int i=0; i<tmp.totalSize(); i++)
      tmp[i] = X( t / im[i] );
    
    return tmp;
  }
  
  template<class X>  
  Image<X> operator-(double t, const Image<X> &im) {
    Image<X> tmp;
    tmp.resize(im);
    
    for(int i=0; i<tmp.totalSize(); i++)
      tmp[i] = X( t - im[i] );
    
    return tmp;
  }
  
  //! logic function
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::AND(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++) {
      if( (*this)[i] > 0 &&
	  im[i] > 0 )
	tmp[i] = std::numeric_limits<TVoxelType>::max();
      else
	tmp[i] = TVoxelType(0);
    }
    
    return tmp;
  }

  //! logic function
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::OR(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++) {
      if( (*this)[i] > 0 ||
	  im[i] > 0 )
	tmp[i] = std::numeric_limits<TVoxelType>::max();
      else
	tmp[i] = TVoxelType(0);
    }
    
    return tmp;
  }
  
  //! logic function
  template<class TVoxelType>
  template<class TVoxelType2>
  Image<TVoxelType> 
  Image<TVoxelType>::XOR(const Image<TVoxelType2> &im) const {
    assert( this -> w == im.width() &&
	    this -> h == im.height() &&
	    this -> d == im.depth() );
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
    
    for(int i=0; i<this->totalSize(); i++) {
      if( ((*this)[i] > 0 && im[i] == 0) ||
	  ((*this)[i] == 0 && im[i] > 0) )
	tmp[i] = std::numeric_limits<TVoxelType>::max();
      else
	tmp[i] = TVoxelType(0);
    }
    
    return tmp;
  }
  
  //! logic function
  template<class TVoxelType>
  Image<TVoxelType> 
  Image<TVoxelType>::NOT() const {
    
    Image<TVoxelType> tmp; 
    tmp.resize(*this);	
 
    for(int i=0; i<this->totalSize(); i++) {
      if( (*this)[i] == 0 )
	tmp[i] = std::numeric_limits<TVoxelType>::max();
      else
	tmp[i] = TVoxelType(0);
    }
    
    return tmp;
  }
  
  // get max
  template<class TVoxelType>
  TVoxelType Image<TVoxelType>::getMax() const {
    TVoxelType maxval = this->at(0);	
    for(int i=1; i<this->totalSize(); i++)
      maxval = util::maxOf(maxval,this->at(i));
    return maxval;		
  }
  
  // get min
  template<class TVoxelType>
  TVoxelType Image<TVoxelType>::getMin() const {
    TVoxelType minval = this->at(0);
    for(int i=1; i<this->totalSize(); i++)
      minval = util::minOf(minval,this->at(i));
    return minval;
  }
  
  // interpolation
  template<class TVoxelType>
  inline double Image<TVoxelType>::getInterpd(double x, 
					      double y, 
					      double z) const {
    
    // integer part
    int X = (int) floor(x);
    int Y = (int) floor(y);
    int Z = (int) floor(z);
    
    // decimal part
    double ax = x-X;
    double ay = y-Y;
    double az = z-Z;
    
    return 
      util::interpd(
		    util::interpd(
				  util::interpd(getSafeAsDouble(X,Y,Z),
						getSafeAsDouble(X+1,Y,Z), 
						ax),
				  util::interpd(getSafeAsDouble(X,Y+1,Z),
						getSafeAsDouble(X+1,Y+1,Z), 
						ax),
				  ay),
		    util::interpd(
				  util::interpd(getSafeAsDouble(X,Y,Z+1), 
						getSafeAsDouble(X+1,Y,Z+1), 
						ax),
				  util::interpd(getSafeAsDouble(X,Y+1,Z+1), 
						getSafeAsDouble(X+1,Y+1,Z+1), 
						ax),
				  ay),
		    az
		    );
  }
  
  template<class TVoxelType>
  TVoxelType Image<TVoxelType>::getInterp(double x, 
					  double y, 
					  double z) const {
    
    // integer part
    int X = (int) floor(x);
    int Y = (int) floor(y);
    int Z = (int) floor(z);
    
    // decimal part
    double ax = x-X;
    double ay = y-Y;
    double az = z-Z;
    
    return 
      util::interp(
		   util::interp(
				util::interp(this->getSafe(X,Y,Z),
					     this->getSafe(X+1,Y,Z), 
					     ax),
				util::interp(this->getSafe(X,Y+1,Z),
					     this->getSafe(X+1,Y+1,Z), 
					     ax),
				ay),
		   util::interp(
				util::interp(this->getSafe(X,Y,Z+1), 
					     this->getSafe(X+1,Y,Z+1), 
					     ax),
				util::interp(this->getSafe(X,Y+1,Z+1), 
					     this->getSafe(X+1,Y+1,Z+1), 
					     ax),
				ay),
		   az
		   );
  }
  
  //! gradient (centered differences)
  template<class TVoxelType>
  Vec3f Image<TVoxelType>::getGradient(int x, 
				       int y, 
				       int z) const {
    Vec3f g;
    g.x = (WISH_FLOAT) getCenteredDX(x,y,z);
    g.y = (WISH_FLOAT) getCenteredDY(x,y,z);
    g.z = (WISH_FLOAT) getCenteredDZ(x,y,z);
    return g;
  }
  
  // interpolated gradient  
  template<class TVoxelType>
  Vec3f Image<TVoxelType>::getGradientInterp(double x, 
					     double y, 
					     double z) const {
    // integer part
    int X = (int) floor(x);
    int Y = (int) floor(y);
    int Z = (int) floor(z);
    
    // decimal part
    double ax = x-X;
    double ay = y-Y;
    double az = z-Z;
    
    return 
      util::interp(
		   util::interp(
				util::interp(getGradient(X,Y,Z),
					     getGradient(X+1,Y,Z), 
					     ax),
				util::interp(getGradient(X,Y+1,Z),
					     getGradient(X+1,Y+1,Z), 
					     ax),
				ay),
		   util::interp(
				util::interp(getGradient(X,Y,Z+1), 
					     getGradient(X+1,Y,Z+1), 
					     ax),
				util::interp(getGradient(X,Y+1,Z+1), 
					     getGradient(X+1,Y+1,Z+1), 
					     ax),
				ay),
		   az
		   );
    
  }
  
  // I/O 
  template<class TVoxelType>  
  bool Image<TVoxelType>::readImage(const std::string &filename,
				    int _w, 
				    int _h, 
				    int _d,
				    double _sx, 
				    double _sy, 
				    double _sz) {
    
    bool ok;
    ok = readVTK(filename);
    if( !ok )
      ok = readPixy(filename);
    if( !ok )
      ok = readRaw(filename,_w,_h,_d,_sx,_sy,_sz);
    
    return ok;
  }
  
  template<class TVoxelType>  
  bool Image<TVoxelType>::readVTK(const std::string &filename) {
    
    ifstream is(filename.c_str(), ifstream::in | ifstream::binary);
    if( !is ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Reading VTK..." << endl;
    
    //   # vtk DataFile Version x.x
    //   Image::writeVTK
    //   BINARY
    //   DATASET STRUCTURED_POINTS
    //   DIMENSIONS 128 128 128
    //   ORIGIN 0.0 0.0 0.0
    //   SPACING 1.0 1.0 1.0
    //   POINT_DATA 2097152
    //   SCALARS image_data unsigned_char
    //   LOOKUP_TABLE default
    //   DATA
    
    // check if it is a VTK file
    std::string checkvtk;
    is >> checkvtk; // '#'
    is >> checkvtk; // vtk
    if( !(checkvtk == "vtk" || checkvtk =="VTK") )
      return false;
    
    // read the rest of the first line
    char infostr[1000];    
    is.getline(infostr,1000);
    
    // read info
    is.getline(infostr,1000);
    this->setInfo(std::string(infostr));
    
    std::string str;
    // read the BINARY/ASCII field
    do {
      is >> str;
    } while( str != "ASCII" && str != "BINARY" );
    std::string mode = str;
    
    // read the DIMENSIONS field
    int W,H,D;  
    do {
      is >> str;
    } while( str != "DIMENSIONS" );
    is >> W >> H >> D;
    
    // read the SPACING field
    double ssx, ssy, ssz;  
    do {
      is >> str;
    } while( str != "SPACING" );
    is >> ssx >> ssy >> ssz;
    
    // read the SCALARS/VECTORS field
    std::string typestring;
    do {
      is >> str;
    } while( str != "SCALARS" && str != "COLOR_SCALARS" && str != "VECTORS" );
    std::string scv = str;
    is >> str; // the name
    is >> typestring; // the type (or numComp (3 or 4) if COLOR_SCALARS)
    
    // implemented specific for each valid data type
    bool check = checkVTKDataType(scv,typestring);
    
    if( !check )
      return false;
    
    // read lookup if scalars
    if( scv == "SCALARS" ) {
      // read the lookup table and the eol
      do {
	is >> str;
      }	while( str != "LOOKUP_TABLE" );
      is >> str; // name (default)
    }

    // end of line
    is >> ws;
    
    // set spacing
    this->setSpacing(ssx,ssy,ssz);  
    
    // allocate memory
    this->allocate(W,H,D);
    
    // read file
    if( mode == "BINARY" )
      readVTKBinary(is);
    else
      readVTKAscii(is);
    
    is.close();
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  template<>
  inline bool Image<uint8>::checkVTKDataType(const std::string &scvstring, 
				      const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "unsigned_char";
  }

  template<>
  inline bool Image<int8>::checkVTKDataType(const std::string &scvstring, 
				      const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "char";
  }
  
  template<>
  inline bool Image<uint16>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "unsigned_short";
  }

  template<>
  inline bool Image<int16>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "short";
  }
  
  template<>
  inline bool Image<uint32>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "unsigned_int";
  }

  template<>
  inline bool Image<int32>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "int";
  }
  
  template<>
  inline bool Image<RGB8>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    // cheating with "3" in the typestring
    return scvstring == "COLOR_SCALARS" && typestring == "3";
  }

  template<>
  inline bool Image<RGBA8>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    // cheating with "4" in the typestring
    return scvstring == "COLOR_SCALARS" && typestring == "4";
  }
  
  template<>
  inline bool Image<float>::checkVTKDataType(const std::string &scvstring, 
				      const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "float";
  }
  
  template<>
  inline bool Image<double>::checkVTKDataType(const std::string &scvstring, 
				       const std::string &typestring) const {
    return scvstring == "SCALARS" && typestring == "double";
  }
  
  template<>
  inline bool Image<Vec3f >::checkVTKDataType(const std::string &scvstring, 
				      const std::string &typestring) const {
    return scvstring == "VECTORS" && typestring == "float";
  }
  
  // read binary
  template<class TVoxelType>
  void Image<TVoxelType>::readVTKBinary(ifstream &is) {
    // vtk binary is big endian
    util::readBE( is, this->getData(), (uint32) this->totalSize() );
  }
  
  // vectors are read component by component
  template<>
  inline void Image<Vec3f>::readVTKBinary(ifstream &is) {
    util::readBE( is, 
		  (WISH_FLOAT*) this->getData(), 
		  (uint32) this->totalSize()*3 );
  }
  
  // RGB8 and RGBA8 are read byte by byte
  template<>
  inline void Image<RGB8>::readVTKBinary(ifstream &is) {
    is.read( (char *) this->getData(), (uint32) this->totalSize()*3 );
  }
  template<>
  inline void Image<RGBA8>::readVTKBinary(ifstream &is) {
    is.read( (char *) this->getData(), (uint32) this->totalSize()*4 );
  }
  
  // read ascii
  template<class TVoxelType>
  void Image<TVoxelType>::readVTKAscii(ifstream &is) {
    for(int i=0; i<this->totalSize(); i++)
      is >> this->getData()[i];
  }   
  
  // ASCII uint8 and int8 must be treated specially
  template<>
  inline void Image<uint8>::readVTKAscii(ifstream &is) {
    uint8 *ptr = (uint8*) this->getData();
    for(int i=0; i<this->totalSize(); i++) {
      float val;
      is >> val;
      ptr[i] = uint8( val );
    }
  }   
  template<>
  inline void Image<int8>::readVTKAscii(ifstream &is) {
    int8 *ptr = (int8*) this->getData();
    for(int i=0; i<this->totalSize(); i++) {
      float val;
      is >> val;
      ptr[i] = int8( val );
    }
  }   
  
  // ASCII for COLOR_SCALARS only accepts floats in [0.0, 1.0]
  template<>
  inline void Image<RGB8>::readVTKAscii(ifstream &is) {
    RGB8 *ptr = (RGB8*) this->getData();
    for(int i=0; i<this->totalSize(); i++) {
      RGBf col;
      is >> col;
      ptr[i] = RGB8( col );
    }
  }   
  
  template<>
  inline void Image<RGBA8>::readVTKAscii(ifstream &is) {
    RGBA8 *ptr = (RGBA8*) this->getData();
    for(int i=0; i<this->totalSize(); i++) {
      RGBAf col;
      is >> col;
      ptr[i] = RGBA8( col );
    }
  }    
  
  // SHOULD REWRITE THIS ONE!
  // USING WRITE HEADER...
  template<class TVoxelType>  
  bool Image<TVoxelType>::writeVTK(const std::string &filename, 
				   const std::string &mode) const {
    
    ofstream os(filename.c_str(), ofstream::out | ofstream::binary);
    if( !os ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Writing VTK..." << endl;
    
    os << "# vtk DataFile Version 3.0" << endl;
    
    if( this->getInfo() == "" )
      os << "created by wish::Image::writeVTK" << endl;
    else
      os << this->getInfo() << endl;
    
    if( mode == "BINARY" || mode == "ASCII" )
      os << mode << endl;
    else {
      cerr << "Unknown mode: " << mode << endl;
      return false;
    }
    
    os << "DATASET STRUCTURED_POINTS" << endl;
    os << "DIMENSIONS " 
       << this->width() << " " 
       << this->height() << " " 
       << this->depth() << endl;
    os << "ORIGIN 0.0 0.0 0.0" << endl; // use 0.0 
    os << "SPACING " << this->sx << " " << this->sy << " " << this->sz << endl;
    os << "POINT_DATA " << this->totalSize() << endl;
    
    // this is implemented specifically for each valid data type
    writeVTKDataTypeInfo(os);
    
    if( mode == "BINARY" )
      writeVTKBinary(os);
    else      
      writeVTKAscii(os);
    
    os.close();
    
    cerr << "Finished!" << endl;
    
    return true;  
  }
  
  template<>
  inline void Image<uint8>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "unsigned_char" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }

  template<>
  inline void Image<int8>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "char" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }

  template<>
  inline void Image<uint16>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "unsigned_short" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }

  template<>
  inline void Image<int16>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "short" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }

  template<>
  inline void Image<uint32>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "unsigned_int" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }

  template<>
  inline void Image<int32>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "int" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }
  
  template<>
  inline void Image<RGB8>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "COLOR_SCALARS image_data 3" << endl;
  }

  template<>
  inline void Image<RGBA8>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "COLOR_SCALARS image_data 4" << endl;
  }
  
  template<>
  inline void Image<float>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "float" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }
  
  template<>
  inline void Image<double>::writeVTKDataTypeInfo(ofstream &os) const {
    os << "SCALARS image_data " << "double" << endl;
    os << "LOOKUP_TABLE default" << endl;
  }
  
  template<>
  inline void Image<Vec3f >::writeVTKDataTypeInfo(ofstream &os) const {
    os << "VECTORS vector_data " << "float" << endl;
  }
  
  // write binary
  template<class TVoxelType>
  void Image<TVoxelType>::writeVTKBinary(ofstream &os) const {
    // vtk binary is always in big endian
    util::writeBE( os, this->getData(), (uint32) this->totalSize() );    
  }
  
  // vectors are written component by component
  template<>
  inline void Image<Vec3f>::writeVTKBinary(ofstream &os) const {
    util::writeBE( os, 
		   (WISH_FLOAT*) this->getData(), 
		   this->totalSize()*3 );
  }  
  
  // RGB8, RGBA8 are written byte by byte
  template<>
  inline void Image<RGB8>::writeVTKBinary(ofstream &os) const {
    os.write( (const char*) this->getData(), this->totalSize()*3 );
  }
  template<>
  inline void Image<RGBA8>::writeVTKBinary(ofstream &os) const {
    os.write( (const char*) this->getData(), this->totalSize()*4 );
  }
  
  // write ascii
  template<class TVoxelType>
  void Image<TVoxelType>::writeVTKAscii(ofstream &os) const {
    for(int i=0; i<this->totalSize(); i++)
      os << this->getData()[i];
  }
  
  // unsigned char converted to int before printed as ascii
  template<>
  inline void Image<uint8>::writeVTKAscii(ofstream &os) const {
    for(int i=0; i<this->totalSize(); i++)
      os << (uint32) this->getData()[i];
  }

  // unsigned char and char converted to int before printed as ascii
  template<>
  inline void Image<int8>::writeVTKAscii(ofstream &os) const {
    for(int i=0; i<this->totalSize(); i++)
      os << (int32) this->getData()[i];
  }
  
  // RGB8 and RGBA8 are written as floats in [0,1]
  template<>
  inline void Image<RGB8>::writeVTKAscii(ofstream &os) const {
    for(int i=0; i<this->totalSize(); i++)
      os << RGBf(this->getData()[i]) << endl;
  }
  
  // RGBA8 is written as floats in [0,1]
  template<>
  inline void Image<RGBA8>::writeVTKAscii(ofstream &os) const {
    for(int i=0; i<this->totalSize(); i++)
      os << RGBAf(this->getData()[i]) << endl;
  }
  
  template<class TVoxelType>  
  bool Image<TVoxelType>::readPixy(const std::string &filename) {
    
    ifstream is(filename.c_str(), ifstream::in | ifstream::binary);
    if( !is ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
  
    cerr << "Reading Pixy..." << endl;
    
    // pixy/ipad header, ends with form feed followed by raw data
    // IMAGE {
    //  CLASS 1
    //  NAME  'IMP_5'
    //  SIZE { 394 505 1 1 1 }
    //  PIXTYPE UNSIGNED
    //  PIXSIZE 8
    //  PHYSICALPIXSIZE { 0.000000 0.000000 0.000000 0.000000 0.000000 }
    //  RASTER 1
    //  LITTLEENDIAN
    //  LAYERSIZE { 198970 }
    //  Z {   0.0000 }
    //  T { 0 }
    //  B { 0 }
    // }
    // '\f'
    // RAW DATA
    
    // check if it is a IPAD/PIXY file
    std::string checkipad;
    is >> checkipad;
    if( !(checkipad == "IMAGE" || checkipad == "REGION") ) {
      return false;
    }
    // go to first in file
    is.seekg(0,ios::beg);
    
    char c;
    std::string str;
    
    // read everything until '\f' and store it in str
    is.get(c);
    while( c != '\f' ) {
      str+=c;
      is.get(c);
    }
    
    // position of the raw data
    unsigned long raw_pos = is.tellg();
    
    bool le=false;
    // find the BIGENDIAN field if it exists, otherwise use little endian
    std::string::size_type be_pos = str.find("BIGENDIAN");
    if(be_pos == std::string::npos) {
      le = true;
    }
    
    // find the SIZE field
    std::string::size_type size_pos = str.find("SIZE");
    if(size_pos == std::string::npos) {
      cerr << "File error: SIZE not found" << endl;
      return false;
    }
    
    int W,H,D;
    // read the SIZE field
    is.seekg((int) size_pos, ios::beg);
    while(is.get() != '{');
    is >> W >> H >> D;
    
    // find the PIXTYPE field
    std::string::size_type pixtype_pos = str.find("PIXTYPE");
    if(pixtype_pos == std::string::npos) {
      cerr << "File error: PIXTYPE not found" << endl;
      return false;
    }
    
    std::string pt;
    // read the PIXTYPE field
    is.seekg((int) pixtype_pos, ios::beg);
    is >> pt; // read "PIXTYPE"
    is >> pt; // read value of PIXTYPE
    
    if( pt != "UNSIGNED" && pt != "SIGNED" && pt != "REAL" ) {
      cerr << "File error, bad PIXTYPE" << endl;
      return false;
    }
    
    // find the PIXSIZE field
    std::string::size_type pixsize_pos = str.find("PIXSIZE");
    if(pixsize_pos == std::string::npos) {
      cerr << "File error, PIXSIZE not found" << endl;
      return false;
    }
    
    std::string ps_str;
    int ps;
    // read the PIXSIZE field
    is.seekg((int) pixsize_pos, ios::beg);
    is >> ps_str; // read "PIXSIZE"
    is >> ps; // read value of PIXSIZE
    
    if( ps/8 != this->voxelBytes() ) {
      cerr << "File error: PIXSIZE mismatch" << endl;
      return false;
    }
    
    // find the PHYSICALPIXSIZE field
    std::string::size_type physicalpixsize_pos = str.find("PHYSICALPIXSIZE");
    if(physicalpixsize_pos == std::string::npos) {
      cerr << "File warning: PHYSICALPIXSIZE not found, using 1.0" << endl;
      this->setSpacing(1.0,1.0,1.0);
    }
    else {
      double spx,spy,spz;
      // read the PHYSICALPIXSIZE field
      is.seekg((int) physicalpixsize_pos, ios::beg);
      while(is.get() != '{');
      is >> spx >> spy >> spz;
      
      // if some crazy default value...
      if( spx == 0.0 && spy == 0.0 && spz == 0.0 )
	spx = spy = spz = 1.0;
      
      this->setSpacing(spx,spy,spz);
    }
    
    // step to raw
    is.seekg(raw_pos, ios::beg);
    
    // allocate memory
    this->allocate(W,H,D);
    
    // read file
    if( le )
      util::readLE( is, this->getData(), (uint32) this->totalSize() );
    else
      util::readBE( is, this->getData(), (uint32) this->totalSize() );
    
    is.close();
    
    // assign label from filename
    //this->setLabel( ... );
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  template<class TVoxelType>  
  bool Image<TVoxelType>::writePixy(const std::string &filename, bool reg) const {
    
    ofstream os(filename.c_str(), ofstream::out | ofstream::binary);
    if( !os ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Writing Pixy..." << endl;
    
    if( reg )
      os << "REGION {" << endl;
    else
      os << "IMAGE {" << endl;
    
    os << "  SIZE { " 
       << this->width() << " " 
       << this->height() << " " 
       << this->depth() << " " 
       << 1 << " " 
       << 1 << " "
       << "}" << endl;

    if( isUnsignedIntegerType() )
      os << "  PIXTYPE UNSIGNED" << endl;
    else if( this->isSignedIntegerType() )
      os << "  PIXTYPE SIGNED" << endl;
    else if( this->getDataType()=="float" || this->getDataType()=="double" )
      os << "  PIXTYPE REAL" << endl;
    else {
      cerr << "Failed writing pixy, bad PIXTYPE" << endl;
      return false;
    }
    
    os << "  PIXSIZE " << this->voxelBytes()*8 << endl;
    os << setiosflags(ios::fixed) << setprecision(6) << "  PHYSICALPIXSIZE { " << this->sx << " " << this->sy << " " << this->sz << " " << 1.0 << " " << 1.0 << " }" << endl;
    os << "}" << endl;
    
    // always write little endian
    os << "  LITTLEENDIAN" << endl;
    
    // the form feed that preceeds the raw data
    os << '\f';
    
    // write data in little endian
    util::writeLE( os, this->getData(), (uint32) this->totalSize() );
    
    os.close();
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  template<class TVoxelType>  
  bool Image<TVoxelType>::readRaw(const std::string &filename, 
				  int _w, 
				  int _h, 
				  int _d,
				  double _sx, 
				  double _sy, 
				  double _sz) {
    
    ifstream is(filename.c_str(), ifstream::in | ifstream::binary);    
    if( !is ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    // check size of file
    int length;
    is.seekg(0, ifstream::end);
    length = is.tellg();
    
    if( length != _w*_h*_d*this->voxelBytes() ) {
      cerr << "Filesize mismatch!" << endl;
      return false;
    }
    
    is.seekg(0, ios::beg);

    cerr << "Reading raw..." << endl;
    
    this->allocate(_w,_h,_d);
    this->setSpacing(_sx,_sy,_sz);
    
    is.read( (char *) this->getData(), this->totalBytes() );
    is.close();
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  template<class TVoxelType>  
  bool Image<TVoxelType>::writeRaw(const std::string &filename) const {
    
    ofstream os(filename.c_str(), ofstream::out | ofstream::binary);    
    if( !os ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Writing raw..." << endl;
    
    os.write( (const char*) this->getData(), this->totalBytes() );
    os.close();
    
    cerr << "Finished!" << endl;
    
    return true;
  }

} // end namespace wish

#endif
