///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_IMAGE_FILE_READER_HH_
#define _WISH_IMAGE_FILE_READER_HH_

#include "wishImage.hh"

namespace wish {
  
  //! ImageFileReader
  class ImageFileReader {
    
  public:
    
    //! factory, cretes a reader from filename (looks at extension)
    static ImageFileReader* Create(const string &filename);
    
    //! struct for holding data
    struct ImageInformation {
      int headersize;  // header size in bytes
      string infostr;  // information about the file
      string datatype; // uint8, float, ...
      Vec3i dim;       // width,height,depth
      Vec3f spc;       // spacing
      
      ImageInformation() : 
	headersize(0),
	infostr(""), 
	datatype(""),
	dim(Vec3i(0,0,1)),
	spc(Vec3f(1.0f,1.0f,1.0f)) {}
      
      void print() {
	cerr << "Headersize: " << headersize << endl;
	cerr << "Datatype:   " << datatype << endl;
	cerr << "Dimensions: " << dim << endl;
	cerr << "Spacing:    " << spc << endl;
      }

    };
    
  protected:
    string m_filename;
    ImageInformation m_info;
    ImageObject *m_output;
    bool m_output_owner;
    
    //! reads the header and stores the properties in the info struct
    virtual bool readHeader() = 0;
    
    //! reads the data
    virtual bool readData() = 0;
    
  public:
    //! constructor
    ImageFileReader() :
      m_filename(""),
      m_output(0),
      m_output_owner(true)
    {}
    
    //! destructor
    virtual ~ImageFileReader() {
      if( m_output_owner ) {
	delete m_output;
      }
    }
    void setFilename(const string &str) {
      m_filename = str;
    }
    
    ImageObject* getOutput() const {
      return m_output;
    }
    
    //! release output ownership and return output (use to create an auto_ptr)
    ImageObject* releaseOutput() {
      m_output_owner = false;
      return m_output;
    }
    
    //! print self (used mostly for debugging)
    virtual void printSelf() const = 0;
    
    //! the function to call
    void execute() {
      
      // delete old output if we own it 
      if( m_output_owner )
	delete m_output;
      
      // set to zero
      m_output = 0;
      
      // try to read header
      if( !readHeader() ) {
	cerr << "Error reading header for file: " << m_filename << endl;
	return;
      }
            
      // read the image data 
      if( !readData() ) {
	cerr << "Error reading data for file: " << m_filename << endl;
      }
    }
    
  };
  
}

#endif
