///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_IMAGE_OBJECT_CPP_
#define _WISH_IMAGE_OBJECT_CPP_

#include "wishImageObject.hh"
#include "wishImageFileReader.hh"
#include "wishRawFileReader.hh"
#include "wishColor.hh"

namespace wish {
  
  // ############ ImageObject ##############  
  ImageObject*
  ImageObject::createFromFile(const std::string &filename,
			      int W, 
			      int H, 
			      int D,
			      double ssx, 
			      double ssy, 
			      double ssz,
			      const string &voxeltype,
			      bool littleEndian,
			      int headerSize) {
    
    // use the factory
    auto_ptr<ImageFileReader> ifr( ImageFileReader::Create(filename) );
    
    // if the factory did not work, use raw reader
    if( !ifr.get() ) {
      RawFileReader *rfr = new RawFileReader;
      rfr->setFilename( filename );
      rfr->setDimensions( Vec3i(W,H,D) );
      rfr->setSpacing( Vec3f(float(ssx),float(ssy),float(ssz)) );
      rfr->setDatatype( voxeltype );
      rfr->setLittleEndian(littleEndian);
      rfr->setHeaderSize(headerSize);
      ifr.reset( rfr );
    }
    
    // read and return result
    ifr->execute();

    // set label from filename
    string dir, name, ext;
    util::splitFilename(filename, dir, name, ext);
    ifr->getOutput()->setLabel(name);
    
    return ifr->releaseOutput();
        
  }
  
  bool ImageObject::writeImage(const std::string &filename) const {
    // VTK
    string::size_type vtkext = filename.find(".vtk");
    if( vtkext != string::npos )
      return this->writeVTK(filename);
    
    // IPAD / PIXY
    string::size_type ipadext = filename.find(".ipad");
    string::size_type pixyext = filename.find(".pixy");
    if( (ipadext != string::npos) ||
	(pixyext != string::npos) )
      return this->writePixy(filename);
    
    // RAW
    else
      return this->writeRaw(filename);
  }
  
  int ImageObject::voxelBytesFromString(const string &dt) {
    if( dt=="uint8" || dt=="int8" )
      return 1;
    else if( dt=="uint16" || dt=="int16" )
      return 2;
    else if( dt=="uint32" || dt=="int32" || dt=="float" )
      return 4;
    else if( dt=="double" )
      return 8;
    else if( dt=="Vec3f" )
      return 12;
    else if( dt=="RGB8" )
      return 3;
    else if( dt=="RGBA8" )
      return 4;
    else
      return 0;
  }
  
} // end namespace wish

#endif
