///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_IMAGE_OBJECT_HH_
#define _WISH_IMAGE_OBJECT_HH_

#include "wish.hh"
#include "wishutil.hh"
#include "wishLinalg.hh"
#include "wishColor.hh"

namespace wish {
  
  //! ImageObject
  //! This is the class to use for general purposes.
  //! To read a file from disk, use the createFromFile function
  //! that will return the correct image type depending on
  //! whats in the header of the file. All functions in this class 
  //! must be implemented in derived classes. 
  class ImageObject {
  public:
    typedef std::auto_ptr<ImageObject> Pointer;

    // A "factory" like image reader that creates an object
    // from file. Today it is possible to read ipad/pixy, vtk,
    // and raw. If it is a raw file you need to provide the
    // dimensions, the spacing, and the datatype. Valid
    // datatypes are:
    // uint8, int8, uint16, int16, uint32, int32, 
    // float, double, RGB8, RGBA8, and Vec3f.  
    static ImageObject* createFromFile(const std::string &filename,
				       int W=0, 
				       int H=0, 
				       int D=1,
				       double ssx=1.0, 
				       double ssy=1.0, 
				       double ssz=1.0,
				       const std::string &voxeltype="",
				       bool littleEndian=true,
				       int headerSize=0);
    
    // returns bytes per voxel for the datatype described by the string
    static int voxelBytesFromString(const string &dt);
    
    virtual ~ImageObject() {}
    
    // pure virtual members defined in ImageBase
    virtual void setInfo(const std::string &_info) = 0;
    virtual std::string getInfo() const = 0;
    virtual void setLabel(const std::string &_lbl) = 0;
    virtual std::string getLabel() const = 0;
    virtual void printSelf() const = 0;
    virtual std::string getDataType() const = 0;
    virtual void allocate(int _w=0, int _h=0, int _d=1) = 0;
    virtual void allocate(const Vec3i &v) = 0;
    virtual byte* getBytePointer() const = 0;
    virtual int width() const = 0;
    virtual int height() const = 0;
    virtual int depth() const = 0;
    virtual int getWidth() const = 0;
    virtual int getHeight() const = 0;
    virtual int getDepth() const = 0;
    virtual int totalSize() const = 0;
    virtual int getTotalSize() const = 0;
    virtual void setSpacing(double _sx, double _sy, double _sz) = 0;
    virtual void setSpacing(const Vec3f &sp) = 0;
    virtual double getSpacing(int i) const = 0;
    virtual Vec3f getSpacing() const = 0;
    virtual Vec3f totalExtent() const = 0;
    virtual Vec3f getTotalExtent() const = 0;
    virtual Vec3i getDimensions() const = 0;
    virtual int getDim() const = 0;
    virtual int voxelBytes() const = 0;
    virtual int getVoxelBytes() const = 0;
    virtual int totalBytes() const = 0;
    virtual int getTotalBytes() const = 0;
    virtual bool isInside(int x, int y, int z) const = 0;
    virtual bool isInside(const Vec3i &v) const = 0;
    
    // virtual members defined in Image
    // these return values are just dummies, should never be returned since
    // these functions are defined in Image
    virtual bool isUnsignedIntegerType() const { return false; }
    virtual bool isSignedIntegerType() const { return false; }
    
    // access functions for scalar volumes
    virtual double getAsDoubleFromArray(int i) const { return 0.0; }
    
    virtual double getAsDouble(int x, 
			       int y, 
			       int z=0) const { return 0.0; }
    virtual double getAsDouble(const Vec3i &v) const { return 0.0; }
    
    virtual double getSafeAsDouble(int x, 
				   int y, 
				   int z=0) const { return 0.0; }
    virtual double getSafeAsDouble(const Vec3i &v) const { return 0.0; }
    
    // normalized values in [0,1], only valid for uint8, uint16, and uint32
    virtual double getNormalizedFromArray(int i) const { return 0.0; }
    virtual double getNormalized(int x, 
				 int y, 
				 int z=0) const { return 0.0; }
    virtual double getNormalized(const Vec3i &v) const { return 0.0; }
    
    virtual double getSafeNormalized(int x, 
				     int y, 
				     int z=0) const { return 0.0; }
    virtual double getSafeNormalized(const Vec3i &v) const { return 0.0; }
    
    virtual double getNormalizedInterpd(double x, 
					double y, 
					double z=0) const { return 0.0;}
    
    virtual double getNormalizedInterpd(const Vec3f &v) const { return 0.0; }
    
    virtual void setNormalized(double value, int x, int y, int z) {}
    virtual void setNormalized(double value, const Vec3i &v) {}
    virtual void setNormalizedToArray(double value, int i) {}
    
    // interpolation
    virtual double getInterpd(double x, 
			      double y, 
			      double z=0) const { return 0.0; }
    virtual double getInterpd(const Vec3f &v) const { return 0.0; }
    
    // gradient
    virtual Vec3f getGradient(int x, 
			      int y, 
			      int z=0) const { return Vec3f(0,0,0); }
    virtual Vec3f getGradient(const Vec3i &v) const { return Vec3f(0,0,0); }
    
    // gradient interpolation
    virtual Vec3f getGradientInterp(double x, 
				    double y, 
				    double z=0) const { return Vec3f(0,0,0); }
    virtual Vec3f getGradientInterp(const Vec3f &v) const { return Vec3f(0,0,0); }
    
    // coordinate transformations
    virtual Vec3f worldToImage(const Vec3f &v) const { return Vec3f(0,0,0); }
    virtual Vec3f imageToWorld(const Vec3f &v) const { return Vec3f(0,0,0); }
    virtual Vec3f imageToWorldRS(const Vec3f &v) const { return Vec3f(0,0,0); }
    virtual Vec3f worldToImageRS(const Vec3f &v) const { return Vec3f(0,0,0); }
    virtual Vec3f imageToWorldR(const Vec3f &v) const { return Vec3f(0,0,0); }
    virtual Vec3f worldToImageR(const Vec3f &v) const { return Vec3f(0,0,0); }
    
    virtual Matrix4f getWorldToImageMatrix() const { return Matrix4f(); }
    virtual Matrix4f getImageToWorldMatrix() const { return Matrix4f(); }
    virtual Matrix4f getWorldToTextureMatrix() const { return Matrix4f(); }
    virtual Matrix4f getImageToTextureMatrix() const { return Matrix4f(); }
    
    // file writing
    virtual bool writePixy(const std::string &filename, bool reg=false) const { return false; }
    virtual bool writeVTK(const std::string &filename, 
			  const std::string &mode="BINARY") const { return false; }
    virtual bool writeRaw(const std::string &filename) const { return false; }
    virtual bool writeImage(const std::string &filename) const;
  };
  
} // namespace wish

#endif


