///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_LABELING_FILTER_HH_
#define _WISH_LABELING_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  //! connected components labeling
  //! The method labels the components according to the given connectivity
  //! The labels range from 1 to the number of objects
  template<class InputImage, class OutputImage>
  class LabelingFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< LabelingFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    LabelingFilter() :
      m_number_labels(0),
      connectivity(0)
    {}
    
    //! get the number of generated labels 
    int getNrOfLabels() {
      return m_number_labels;
    }
    
    //! set connectivity
    //! Valid values are 4, 8 for 2D and for 6, 18, 26 for 3D.
    //! If connectivity is 0, it will automatically select 4-connectivity
    //! in 2D and 6-connectivity in 3D. 
    void setConnectivity(int _connectivity) {
      if( _connectivity != 0  &&
	  _connectivity != 4  &&
	  _connectivity != 8  &&
	  _connectivity != 6  &&
	  _connectivity != 18 &&
	  _connectivity != 26 ) {
	cerr << "LabelingFilter: bad connectivity, setting to 0 (default)" 
	     << endl;
	connectivity = 0;
      }
      else {
	connectivity = _connectivity;
      }
    }
    
  protected:
    int m_number_labels;
    int connectivity;
    
    virtual void process();
    
  private:
    
    LabelingFilter(const LabelingFilter&);
    LabelingFilter& operator=(const LabelingFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void LabelingFilter<InputImage, OutputImage>::process() {
    
    this -> m_output -> resize( *(this->m_input) );
    this -> m_output -> reset( OutputImageVoxelType(0) );
    
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    // set the appropriate neighborhood
    NeighborhoodType nt;
    if( connectivity == 0 ) {
      if( in.getDim() == 2 ) 
	nt = CHFWD4;
      else 
	nt = CHFWD6;
    }
    else {
      if( connectivity == 4 )
	nt = CHFWD4;
      else if( connectivity == 8 ) 
	nt = CHFWD8;
      else if( connectivity == 6 ) 
	nt = CHFWD6;
      else if( connectivity == 18 )
	nt = CHFWD18;
      else if( connectivity == 26 )
	nt = CHFWD26;
    }
    
    // neighborhoods
    NeighborhoodIterator<OutputImage> nbout(&out,nt);
    nbout.setDefaultValue(0); // return 0 if outside
    nbout.first();
    
    NeighborhoodIterator<InputImage> nbin(&in,nt);
    nbin.setDefaultValue(0); // return 0 if outside
    nbin.first();
  
    // the value of the current label
    uint32 currentlabel = 0;
    
    // help variables
    uint32 result, found;
    uint32 count;
    
    // associations table (maximum number of labels is 9/25)
    vector<uint32> labelmap(nbout.getTotalBoxSize()/3);
    for(uint32 i=0; i<labelmap.size(); i++)
      labelmap[i] = i;
    
    do { // image loop
      
      // result is the label set at the current pixel
      result=0;
      
      if( nbin.value() > 0 ) { // if object 
	count = 0;
	found = 999999;
	nbout.firstnb();
	do { // neighbor loop
	  if( nbout.valuenbsafe() > 0 ) {
	    // select the smallest label in the nb 
	    found = util::minOf(found, nbout.valuenbsafe());
	    count++;
	  }
	} while( nbout.nextnb() );
	
	if( count == 0 ) // no neighbor label, create new
	  result = ++currentlabel;
	
	else if( count == 1 ) // one neighbor label 
	  result = labelmap[found];
	
	else { // need to associate labels
	  result = labelmap[found];
	  nbout.firstnb();
	  uint32 nbval;
	  do { // neighbor loop
	    nbval = nbout.valuenbsafe();
	    if( nbval > 0 && nbval != result ) {
	      labelmap[nbval] = result;
	    }
	  } while( nbout.nextnb() );
	  
	}
	
      }
      
      // set the label
      nbout.value() = result;
      
    } while( nbin.next() && nbout.next() );
    
    // now we need to update the label map
    for(uint32 i=currentlabel; i>0; i--) {
      uint32 k = labelmap[i];
      while( labelmap[k] != k ) {
	k = labelmap[k];
      }
      labelmap[i] = k;
    }
        
    // update output values according to the label map
    nbout.first();
    do {      
      nbout.value() = labelmap[nbout.value()];
    } while( nbout.next() );
    
    // condense down, i.e., if labels are 1, 7, 9 we want 1, 2, 3 
    vector<uint32> condensed(currentlabel+1,0);
    uint32 current = 0;
    for (int i=0; i<out.totalSize(); i++) {
      if( out[i] > 0 && condensed[out[i]] == 0 ) {
	condensed[out[i]] = ++current;
	out[i] = current;
      }
      else {
	out[i] = condensed[out[i]];
      }
    }
    
    m_number_labels = current;
    
  }
  
} // end namespace wish

#endif
