///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_LINALG_HH_
#define _WISH_LINALG_HH_

#include <cmath>
#include <iostream>

#include "wishutil.hh"

namespace wish {
  // simple vector classes
  
  template<class T> class Vec2;
  template<class T> class Vec3;
  template<class T> class Vec4;
  template<class T> class Matrix3;
  template<class T> class Matrix4;
  
  //! 2D vector
  template<class T>
  class Vec2 {
  public:
    
    //! unions to be able to use different names
    union {
      struct {
	T x; T y;
      };
      struct {
	T u; T v;
      };
      struct {
	T s; T t;
      };
      T data[2];
    };
    
    //! pointer
    const T* getPointer() const {
      return &data[0];
    }
    
    //! pointer
    T* getPointer() {
      return &data[0];
    }
    
    //! access operator
    const T& operator[](int i) const {
      return data[i];
    }
    
    T& operator[](int i) {
      return data[i];
    }

    //! convenience operator, glVertex??v( v )
    operator T*() {
      return &data[0];
    }
    
    //! default constructor
    Vec2(T _x=0, T _y=0) :
      x(_x), y(_y) {}
    
    //! copy constructor
    template<class U>
    Vec2(const Vec2<U> &v) {
      x = T(v.x); y = T(v.y);
    }
    
    //! assignment operator
    template<class U>
    Vec2<T>& operator=(const Vec2<U> &v) {
      if( this != &v ) {
	x = T(v.x);
	y = T(v.y);
      }
      return *this;
    }
    
    //! construct from Vec3
    template<class U>
    Vec2(const Vec3<U> &v);

    //! construct from Vec4
    template<class U>
    Vec2(const Vec4<U> &v);
    
    //! vector addition
    Vec2<T> operator+(const Vec2<T> &v) const {
      return Vec2<T>(x+v.x,y+v.y);
    }
    //! vector subtraction
    Vec2<T> operator-(const Vec2<T> &v) const {
      return Vec2<T>(x-v.x,y-v.y);
    }
    //! scalar product
    double operator*(const Vec2<T> &v) const {
      return x*v.x + y*v.y;
    }
    double dot(const Vec2<T> &v) const {
      return (*this)*v;
    }
    
    //! cross product
    template<class U>
    Vec3<T> cross(const Vec2<U> &v) const;
    
    //! multiplication with double
    Vec2<T> operator*(const double &d) const {
      return Vec2<T>(T(x*d),T(y*d));
    }
    //! division with double
    Vec2<T> operator/(const double &d) const {
      return Vec2<T>(T(x/d),T(y/d));
    }
    
    //! vector addition
    Vec2<T>& operator+=(const Vec2<T> &v) {
      x += v.x;
      y += v.y;
      return (*this);
    }
    //! vector subtraction
    Vec2<T>& operator-=(const Vec2<T> &v) {
      x -= v.x;
      y -= v.y;
      return (*this);
    }
    //! multiplication with double
    Vec2<T>& operator*=(const double &d) {
      x *= d;
      y *= d;
      return (*this);
    }
    //! division with double
    Vec2<T>& operator/=(const double &d) {
      x /= d;
      y /= d;
      return (*this);
    }
    
    //! vector norm
    double norm() const {
      return sqrt(x*x+y*y);
    }

    //! vector norm squared
    double norm2() const {
      return x*x+y*y;
    }
    
    //! normalize
    Vec2<T>& normalize() {
      double n = norm();
      (*this) /= n;
      return (*this);
    }
    
    //! normalize safe, check norm before division
    Vec2<T>& normalizeSafe() {
      double n = norm();
      if( fabs(n) > constants::eps ) {
	(*this) /= n;
      }
      return (*this);
    }
    
    //! multiplication with double
    friend Vec2<T> operator*(const double &d, const Vec2<T> &v) {
      return v*d;
    }
    //! unary minus
    friend Vec2<T> operator-(const Vec2<T> &v) {
      return -1.0*v;
    }
    //! input stream
    friend std::istream& operator>>(std::istream &is, Vec2<T> &v) {
      is >> v.x >> v.y;
      return is;
    }
    //! output stream
    friend std::ostream& operator<<(std::ostream &os, const Vec2<T> &v) {
      os << v.x << " " << v.y;
      return os;
    }
  };

  //! 3D vector
  /*!
    Template class for 3D vectors. Intended for float and double.
  */
  template<class T>
  class Vec3 {
  public:
        
    //! unions to be able to use different names
    union {
      struct {
	T x; T y; T z;
      };
      struct {
	T u; T v; T w;
      };
      struct {
	T s; T t; T r;
      };
      struct {
	T rho; T theta; T phi;
      };
      T data[3];
    };
    
    //! pointer (maybe not nice code...)
    T* getPointer() {
      return &data[0];
    }
    const T* getPointer() const {
      return &data[0];
    }
    
    //! convenience operator, glVertex??v( v )
    operator T*() {
      return &data[0];
    }
    
    const T& operator[](int i) const {
      return data[i];
    }
    
    T& operator[](int i) {
      return data[i];
    }
    
    T getX() const {
      return x;
    }

    T getY() const {
      return y;
    }

    T getZ() const {
      return z;
    }
    
    void set(T _x, T _y, T _z) {
      x=_x; y=_y; z=_z;
    }
    
    //! default constructor
    Vec3(T _x=0, T _y=0, T _z=0) :
      x(_x), y(_y), z(_z) {}
    
    //! copy constructor
    template<class U>
    Vec3(const Vec3<U> &v) {
      x = T(v.x); y = T(v.y); z = T(v.z);
    }
    
    //! assignment operator
    template<class U>
    Vec3<T>& operator=(const Vec3<U> &v) {
      if( static_cast<const void*>(this) != static_cast<const void*>(&v) ) {
	x = T(v.x);
	y = T(v.y);
	z = T(v.z);
      }
      return *this;
    }
    
    //! construct from Vec2
    template<class U>
    Vec3(const Vec2<U> &v);
    
    //! construct from Vec4
    template<class U>
    Vec3(const Vec4<U> &v);
    
    //! vector addition
    Vec3<T> operator+(const Vec3<T> &v) const {
      return Vec3<T>(x+v.x,y+v.y,z+v.z);
    }
    //! vector subtraction
    Vec3<T> operator-(const Vec3<T> &v) const {
      return Vec3<T>(x-v.x,y-v.y,z-v.z);
    }
    //! scalar product
    double operator*(const Vec3<T> &v) const {
      return x*v.x + y*v.y + z*v.z;
    }
    //! multiplication with double
    Vec3<T> operator*(const double &d) const {
      return Vec3<T>(T(x*d),T(y*d),T(z*d));
    }
    //! division with double
    Vec3<T> operator/(const double &d) const {
      return Vec3<T>(T(x/d),T(y/d),T(z/d));
    }
    //! cross product
    Vec3<T> cross(const Vec3<T> &v) const {
      Vec3<T> tmp;
      tmp.x=y*v.z-z*v.y;
      tmp.y=z*v.x-x*v.z;
      tmp.z=x*v.y-y*v.x;
      return tmp;
    }
    //! cross product
    T dot(const Vec3<T> &v) const {
      return (*this)*v;
    }
    
    //! vector addition
    Vec3<T>& operator+=(const Vec3<T> &v) {
      x += v.x;
      y += v.y;
      z += v.z;
      return (*this);
    }
    //! vector subtraction
    Vec3<T>& operator-=(const Vec3<T> &v) {
      x -= v.x;
      y -= v.y;
      z -= v.z;
      return (*this);
    }
    //! multiplication with double
    Vec3<T>& operator*=(const double &d) {
      x *= d;
      y *= d;
      z *= d;
      return (*this);
    }
    //! division with double
    Vec3<T>& operator/=(const double &d) {
      x = T(x/d);
      y = T(y/d);
      z = T(z/d);
      return (*this);
    }
    //! equal (used for integer position vectors)
    bool operator==(const Vec3<T> &v) const {
      return (x==v.x && y==v.y && z==v.z);
    }
    //! not equal (used for integer position vectors)
    bool operator!=(const Vec3<T> &v) const {
      return !((*this)==v);
    }
    
    //! vector norm
    double norm() const {
      return std::sqrt(x*x+y*y+z*z);
    }
    
    //! vector norm squared
    double norm2() const {
      return x*x+y*y+z*z;
    }
    
    //! normalize
    Vec3<T>& normalize() {
      double n = norm();
      (*this) /= n;
      return (*this);
    }
    
    //! normalize safe, check norm before division
    Vec3<T>& normalizeSafe() {
      double n = norm();
	  if( std::fabs(n) > constants::eps ) {
	(*this) /= n;
      }
      return (*this);
    }
    
    //! set length
    Vec3<T>& setLength(double l) {
      normalizeSafe();
      (*this) = l*(*this);
      return (*this);
    }
    
    //! orthogonalize ourself against v (does not modify the object)
    //! OBS! No normalization is performed!
    Vec3<T> orthogonalizeAgainst(const Vec3<T> &v) const {
      Vec3<T> ret = (*this)-((*this)*v)*v;
      return ret;
    }
    
    //! find a vector that is orthogonal to ourselves 
    //! (no normalization is done)
    Vec3<T> findOrthogonal() const {
      
      Vec3<T> ret = this->cross(Vec3<T>(1,0,0));
      if( ret.norm() < constants::eps9 )
	ret = this->cross(Vec3<T>(0,1,0));
      
      return ret;
      
    }
    
    // spherical to cartesian
    // rho [0,inf) is the radial part
    // theta [0,2pi) is the rotation (azimuth angle) among the z-axis,
    // phi [-pi/2,pi/2] is the latitude angle relative to the x-y plane
    // the definition is fetched from mathworld
    Vec3<T> sph2cart() const {
		float xx = rho * std::cos(phi) * std::cos(theta);
		float yy = rho * std::cos(phi) * std::sin(theta);
		float zz = rho * std::sin(phi);
      return Vec3<T>(xx,yy,zz);
    }
    
    // cartesian to spherical
    Vec3<T> cart2sph() const {
		float r = std::sqrt(x*x+y*y+z*z);
		float t = std::atan2(y,x);
		float p = std::atan2(z,std::sqrt(x*x+y*y));
      return Vec3<T>(r,t,p);
    }
    
    //! multiplication with double
    friend Vec3<T> operator*(const double &d, const Vec3<T> &v) {
      return v*d;
    }
    //! unary minus
    friend Vec3<T> operator-(const Vec3<T> &v) {
      return -1.0*v;
    }
    //! input stream
    friend std::istream& operator>>(std::istream &is, Vec3<T> &v) {
      is >> v.x >> v.y >> v.z;
      return is;
    }
    //! output stream
    friend std::ostream& operator<<(std::ostream &os, const Vec3<T> &v) {
      os << v.x << " " << v.y << " " << v.z;
      return os;
    }
  };

  //! 4D vector
  /*!
    Template class for 4D vectors. Intended for float and double.
  */
  template<class T>
  class Vec4 {
  public:
        
    //! unions to be able to use different names
    union {
      struct {
	T x; T y; T z; T w;
      };
      T data[4];
    };
    
    //! pointer (maybe not nice code...)
    T* getPointer() {
      return &data[0];
    }
    const T* getPointer() const {
      return &data[0];
    }
    
    //! convenience operator, glVertex??v( v )
    operator T*() {
      return &data[0];
    }
    
    const T& operator[](int i) const {
      return data[i];
    }
    
    T& operator[](int i) {
      return data[i];
    }
    
    //! default constructor
    Vec4(T _x=0, T _y=0, T _z=0, T _w=1) :
      x(_x), y(_y), z(_z), w(_w) {}
    
    //! copy constructor
    template<class U>
    Vec4(const Vec4<U> &v) {
      x = T(v.x); y = T(v.y); z = T(v.z); w = T(v.w);
    }
    
    //! assignment operator
    template<class U>
    Vec4<T>& operator=(const Vec4<U> &v) {
      if( this != &v ) {
	x = T(v.x);
	y = T(v.y);
	z = T(v.z);
	w = T(v.w);
      }
      return *this;
    }
    
    //! construct from Vec3
    template<class U>
    Vec4(const Vec3<U> &v, const U &_w=1.0);
    
    //! input stream
    friend std::istream& operator>>(std::istream &is, Vec4<T> &v) {
      is >> v.x >> v.y >> v.z;
      return is;
    }
    //! output stream
    friend std::ostream& operator<<(std::ostream &os, const Vec4<T> &v) {
      os << v.x << " " << v.y << " " << v.z;
      return os;
    }
  };
  

  typedef Vec2<WISH_FLOAT>  Vec2f;
  typedef Vec2<double>      Vec2d;
  typedef Vec2<int>         Vec2i;
  typedef Vec3<WISH_FLOAT>  Vec3f;
  typedef Vec3<double>      Vec3d;
  typedef Vec3<int>         Vec3i;
  typedef Vec4<WISH_FLOAT>  Vec4f;
  typedef Vec4<double>      Vec4d;
  
  // Matrix3
  template<class T>
  class Matrix3 {
    
  protected:
    T data[3*3];
    
  public:
    
    // default and init constructor
    Matrix3(T a00=1, T a01=0, T a02=0,
	    T a10=0, T a11=1, T a12=0,
	    T a20=0, T a21=0, T a22=1) {
      
      data[0 ] = a00; data[3 ] = a01; data[6 ] = a02;
      data[1 ] = a10; data[4 ] = a11; data[7 ] = a12;
      data[2 ] = a20; data[5 ] = a21; data[8 ] = a22;
    }
    
    // init constructor (vectors are ROWS)
    // changed this 2007-01-05, before vectors were columns
    Matrix3(const Vec3<T> &v1,
	    const Vec3<T> &v2,
	    const Vec3<T> &v3) {
      data[0 ] = v1.x; data[3 ] = v1.y; data[6 ] = v1.z;
      data[1 ] = v2.x; data[4 ] = v2.y; data[7 ] = v2.z;
      data[2 ] = v3.x; data[5 ] = v3.y; data[8 ] = v3.z;
    }
    
    // construct from pointer
    Matrix3(T *_data) {
      memcpy(data, _data, 9*sizeof(T));
    }
    
    // copy constructor
    template<class U>
    Matrix3(const Matrix3<U> &B) {
      Matrix3<T> &m = *this;
      m(0,0)=T(B(0,0)); m(0,1)=T(B(0,1)); m(0,2)=T(B(0,2));
      m(1,0)=T(B(1,0)); m(1,1)=T(B(1,1)); m(1,2)=T(B(1,2));
      m(2,0)=T(B(2,0)); m(2,1)=T(B(2,1)); m(2,2)=T(B(2,2));
    }
    
    // explicit constructor from Matrix4, returns upper 3x3 part
    explicit Matrix3(const Matrix4<T> &B);
    
    // assignment operator
    const Matrix3<T>& operator=(const Matrix3<T> &B) {
      if( this != &B ) {
	memcpy(data, B.data, 9*sizeof(T));
      }
      return *this;
    }
    
    // access array
    const T* getPointer() const {
      return data;
    }
    T* getPointer() {
      return data;
    }
    
    // access operator
    T& operator()(int i, int j) {
      return data[3*j + i];
    }
    const T& operator()(int i, int j) const {
      return data[3*j + i];
    }
    
    // convenient way of obtaining a matrix in OpenGL
    operator T*() {
      return data;
    }
    
    // algebra
    Matrix3<T> operator* (const Matrix3<T> &B) const;
    
    Matrix3<T> operator* (double c) const;        
    
    Matrix3<T> operator/ (double c) const {
      return (*this)*(1.0/c);
    }
    
    Matrix3<T> operator+ (const Matrix3<T> &B) const;
    
    Matrix3<T> operator- (const Matrix3<T> &B) const {
      return -B + *this;
    }
    
    Vec3<T> operator* (const Vec3<T> &v) const;
    
    friend Matrix3<T> operator* (double c, const Matrix3<T> &A) {
      return A*c;
    }

    friend Matrix3<T> operator- (const Matrix3<T> &A) {
      return -1.0*A;
    }
    
    Matrix3<T> transpose() const;
    
    Matrix3<T> inverse() const {
      Matrix4<T> m(*this);
      return Matrix3<T>( m.transformInverse() );
    }
    
    Matrix3<T> eig(Matrix3<T> &V, double tol=constants::eps) const;
    
    Matrix3<T> diagonal() const {
      return Matrix3<T>((*this)(0,0),0,0,
			0,(*this)(1,1),0,
			0,0,(*this)(2,2));
    }
    
    friend std::ostream& operator<< (std::ostream &os, const Matrix3 &A) {
      os << A(0,0) << " " << A(0,1) << " " << A(0,2) << std::endl
	 << A(1,0) << " " << A(1,1) << " " << A(1,2) << std::endl
	 << A(2,0) << " " << A(2,1) << " " << A(2,2);
      
      return os;
    }
    
  };
  
  typedef Matrix3<WISH_FLOAT> Matrix3f;
  typedef Matrix3<double> Matrix3d;

  // Matrix4, note that elemnts are stored in column-major order.
  // Just like in OpenGL => glMultMatrix4fv( mat )
  template<class T>
  class Matrix4 {
    
  protected:
    T data[4*4];
    
  public:
    
    // default and init constructor
    Matrix4(T a00=1, T a01=0, T a02=0, T a03=0,
	    T a10=0, T a11=1, T a12=0, T a13=0,
	    T a20=0, T a21=0, T a22=1, T a23=0,
	    T a30=0, T a31=0, T a32=0, T a33=1) {
      data[0 ] = a00; data[4 ] = a01; data[8 ] = a02; data[12] = a03; 
      data[1 ] = a10; data[5 ] = a11; data[9 ] = a12; data[13] = a13;
      data[2 ] = a20; data[6 ] = a21; data[10] = a22; data[14] = a23;
      data[3 ] = a30; data[7 ] = a31; data[11] = a32; data[15] = a33;
    }

    // init constructor (vectors are ROWS)
    // changed this 2007-01-05, before vectors were columns
    Matrix4(const Vec4<T> &v1,
	    const Vec4<T> &v2,
	    const Vec4<T> &v3,
	    const Vec4<T> &v4) {
      data[0 ] = v1.x; data[4 ] = v1.y; data[8 ] = v1.z; data[12] = v1.w;
      data[1 ] = v2.x; data[5 ] = v2.y; data[9 ] = v2.z; data[13] = v2.w;
      data[2 ] = v3.x; data[6 ] = v3.y; data[10] = v3.z; data[14] = v3.w;
      data[3 ] = v4.x; data[7 ] = v4.y; data[11] = v4.z; data[15] = v4.w;
    }
    
    // construct from pointer
    Matrix4(T *_data) {
      memcpy(data, _data, 16*sizeof(T));
    }

    // copy constructor
    template<class U>
    Matrix4(const Matrix4<U> &B) {
      Matrix4<T> &m = *this;
      m(0,0)=T(B(0,0)); m(0,1)=T(B(0,1)); m(0,2)=T(B(0,2)); m(0,3)=T(B(0,3));
      m(1,0)=T(B(1,0)); m(1,1)=T(B(1,1)); m(1,2)=T(B(1,2)); m(1,3)=T(B(1,3));
      m(2,0)=T(B(2,0)); m(2,1)=T(B(2,1)); m(2,2)=T(B(2,2)); m(2,3)=T(B(2,3));
      m(3,0)=T(B(3,0)); m(3,1)=T(B(3,1)); m(3,2)=T(B(3,2)); m(3,3)=T(B(3,3));
    }

    // explicit constructor from Matrix3
    explicit Matrix4(const Matrix3<T> &B) {
      Matrix4<T> &m = *this;
      m(0,0)=B(0,0); m(0,1)=B(0,1); m(0,2)=B(0,2); m(0,3)=0;
      m(1,0)=B(1,0); m(1,1)=B(1,1); m(1,2)=B(1,2); m(1,3)=0;
      m(2,0)=B(2,0); m(2,1)=B(2,1); m(2,2)=B(2,2); m(2,3)=0;
      m(3,0)=0;      m(3,1)=0;      m(3,2)=0;      m(3,3)=1;
    }
    
    // assignment operator
    const Matrix4<T>& operator=(const Matrix4<T> &B) {
      if( this != &B ) {
	memcpy(data, B.data, 16*sizeof(T));
      }
      return *this;
    }
    
    // access array
    const T* getPointer() const {
      return data;
    }
    T* getPointer() {
      return data;
    }
    
    // access operator
    T& operator()(int i, int j) {
      return data[4*j + i];
    }
    const T& operator()(int i, int j) const {
      return data[4*j + i];
    }
    
    // convenient way of obtaining a matrix in OpenGL
    operator T*() {
      return data;
    }
    
    // algebra
    Matrix4<T> operator* (const Matrix4<T> &B) const;
    
    Matrix4<T> operator* (double c) const;        
    
    Matrix4<T> operator/ (double c) const {
      return (*this)*(1.0/c);
    }
    
    Matrix4<T> operator+ (const Matrix4<T> &B) const;
    
    Matrix4<T> operator- (const Matrix4<T> &B) const {
      return -B + *this;
    }
    
    Vec3<T> operator* (const Vec3<T> &v) const;
    Vec4<T> operator* (const Vec4<T> &v) const;
    
    friend Matrix4<T> operator* (double c, const Matrix4<T> &A) {
      return A*c;
    }
    
    friend Matrix4<T> operator- (const Matrix4<T> &A) {
      return -1.0*A;
    }
    
    // return 3x3 part
    Matrix3<T> getScaleRotation() const;

    // return normalized 3x3 part
    Matrix3<T> getRotation() const;
    
    // return translation vector
    Vec3<T> getTranslation() const;
    
    // transpose
    Matrix4<T> transpose() const;
    
    // affine transform inverse
    Matrix4<T> transformInverse() const;

    // general inverse
    Matrix4<T> inverse() const;
        
    template<class U>
    Matrix4<U> rotation(float angle, const Vec3f &axis);
    
    template<class U>
    Matrix4<U> scaling(float sx, float sy, float sz);
    
    template<class U>
    Matrix4<U> translation(float dx, float dy, float dz);
    
    template<class U>
    Matrix4<U> scaling(const Vec3f &s);
    
    template<class U>
    Matrix4<U> translation(const Vec3f &t);
    
    friend std::ostream& operator<< (std::ostream &os, const Matrix4 &A) {
      os << A(0,0) << " " << A(0,1) << " " << A(0,2) << " " << A(0,3) << std::endl
	 << A(1,0) << " " << A(1,1) << " " << A(1,2) << " " << A(1,3) << std::endl
	 << A(2,0) << " " << A(2,1) << " " << A(2,2) << " " << A(2,3) << std::endl
	 << A(3,0) << " " << A(3,1) << " " << A(3,2) << " " << A(3,3);
      
      return os;
    }
    
  };
  
  typedef Matrix4<WISH_FLOAT> Matrix4f;
  typedef Matrix4<double> Matrix4d;
  
} // end namespace wish

#include "wishLinalg.tpp"

#endif
