///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_LINALG_TPP_
#define _WISH_LINALG_TPP_

#include "wishLinalg.hh"

namespace wish {
  
  template<class T>
  template<class U> 
  Vec2<T>::Vec2(const Vec3<U> &v) {
    x = T(v.x); y = T(v.y);
  }

  template<class T>
  template<class U> 
  Vec2<T>::Vec2(const Vec4<U> &v) {
    x = T(v.x); y = T(v.y);
  }

  template<class T>
  template<class U> 
  Vec3<T> Vec2<T>::cross(const Vec2<U> &v) const {
    return Vec3<T>(T(0), T(0), T(x*v.y-v.x*y));
  }
  
  template<class T>
  template<class U> 
  Vec3<T>::Vec3(const Vec2<U> &v) {
    x = T(v.x); y = T(v.y); z = T(0);
  }
  
  template<class T>
  template<class U> 
  Vec3<T>::Vec3(const Vec4<U> &v) {
    x = T(v.x); y = T(v.y); z = T(v.z);
  }
  
  template<class T>
  template<class U>
  Vec4<T>::Vec4(const Vec3<U> &v, const U &_w) {
    x = T(v.x); y = T(v.y); z = T(v.z); w = T(_w);
  }
  
  // explicit constructor from Matrix4
  template<class T>
  Matrix3<T>::Matrix3(const Matrix4<T> &B) {
    Matrix3<T> &m = *this;
    m(0,0)=B(0,0); m(0,1)=B(0,1); m(0,2)=B(0,2);
    m(1,0)=B(1,0); m(1,1)=B(1,1); m(1,2)=B(1,2);
    m(2,0)=B(2,0); m(2,1)=B(2,1); m(2,2)=B(2,2);
  }
  
  template<class T>
  Matrix3<T> Matrix3<T>::operator* (const Matrix3<T> &B) const {
    
    Matrix3<T> AB;
    
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
	AB(i,j) = 0;
	for(int k=0; k<3; k++) {
	  AB(i,j) += (*this)(i,k)*B(k,j);
	}
      }
    }
    
    return AB;
  }
  
  template<class T>
  Matrix3<T> Matrix3<T>::operator* (double c)  const {
    
    Matrix3<T> Ac;
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
	Ac(i,j) = Ac(i,j)*c;
      }
    }
    return Ac;
  }
  
  template<class T>
  Matrix3<T> Matrix3<T>::operator+ (const Matrix3<T> &B) const {
    
    Matrix3<T> S;
    
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
	S(i,j) = (*this)(i,j) + B(i,j);
      }
    }
    return S;
  }
  
  template<class T>
  Vec3<T> Matrix3<T>::operator* (const Vec3<T> &v) const {
    
    Vec3<T> vv;
    
    vv.x = (*this)(0,0)*v.x + (*this)(0,1)*v.y + (*this)(0,2)*v.z;
    vv.y = (*this)(1,0)*v.x + (*this)(1,1)*v.y + (*this)(1,2)*v.z;
    vv.z = (*this)(2,0)*v.x + (*this)(2,1)*v.y + (*this)(2,2)*v.z;
    
    return vv;
  }
  
  template<class T>
  Matrix3<T> Matrix3<T>::transpose() const {
    
    Matrix3<T> TR;
    
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
	TR(i,j) = (*this)(j,i);
      }
    }
    return TR;
  }   
  
  // eig
  // OBS! NOT IMPLEMENTED FOR MATRIX4 YET
  // Computes the eigenvalues of the square and symmetric matrix A 
  // using Jacobis iterative method.
  // The function returns a diagonal Matrix3<T> with the
  // eigenvalues on the diagonal, and the reference Matrix3<T> &V
  // will contain the eigenvectors.
  // The algorithm was developed by Ed Wilson, 1995
  template<class T>
  Matrix3<T> Matrix3<T>::eig(Matrix3<T> &V, double tol) const {
    
    int N=3, iter=0;
    double theta, st, ct, TT, sum, ssum, lmax, ll;
    Matrix3<T> L = *this;
    
    // compute sum of all elements
    sum = 0;
    for(int i=0; i<N; i++) {
      for(int j=0; j<N; j++) {
	sum += fabs(L(i,j));
      }
    }
    
    // trivial case (watch out for V, not true)
    if(sum <= 0)
      return L;
    
    // mean value
    sum = sum/(N*N);
    
    do { // do-loop
      iter++;
      ssum = 0;
      lmax = 0;
      for(int j=1; j<N; j++) {
	for(int i=0; i<j; i++) {
	  // check if L(i,j) is to be reduced
	  ll = fabs(L(i,j));
	  if(ll > lmax) 
	    lmax = ll;
	  ssum += ll;
	  if(ll >= 0.1*lmax) { // go reduce L(i,j)
	    // compute rotation angle
	    theta = atan2(2*L(i,j),L(i,i)-L(j,j)+constants::eps)/2.0;
	    st = sin(theta);
	    ct = cos(theta);
	    
	    // modify i and j columns	
	    for(int k=0; k<N; k++) {
	      TT = L(k,i);
	      L(k,i) =  ct*TT+st*L(k,j);
	      L(k,j) = -st*TT+ct*L(k,j);
	      TT = V(k,i);
	      V(k,i) =  ct*TT+st*V(k,j);
	      V(k,j) = -st*TT+ct*V(k,j);
	    }
	    
	    // modify diagonal
	    L(i,i) =  ct*L(i,i)+st*L(j,i);
	    L(j,j) = -st*L(i,j)+ct*L(j,j);
	    L(i,j) = 0;
	    
	    // make symmetric
	    for(int kk=0; kk<N; kk++) {
	      L(i,kk) = L(kk,i);
	      L(j,kk) = L(kk,j);
	    }
	  }
	}
      }
    } while( (ssum/sum) > tol && iter < 100 );
    
    return L.diagonal();
  }
  
  // Matrix4
  
  template<class T>
  Matrix4<T> Matrix4<T>::operator* (const Matrix4<T> &B) const {
    
    Matrix4<T> AB;
    
    for(int i=0; i<4; i++) {
      for(int j=0; j<4; j++) {
	AB(i,j) = 0;
	for(int k=0; k<4; k++) {
	  AB(i,j) += (*this)(i,k)*B(k,j);
	}
      }
    }
    
    return AB;
  }
  
  template<class T>
  Matrix4<T> Matrix4<T>::operator* (double c)  const {
    
    Matrix4<T> Ac;
    for(int i=0; i<4; i++) {
      for(int j=0; j<4; j++) {
	Ac(i,j) = Ac(i,j)*c;
      }
    }
    return Ac;
  }
  
  template<class T>
  Matrix4<T> Matrix4<T>::operator+ (const Matrix4<T> &B) const {
    
    Matrix4<T> S;
    
    for(int i=0; i<4; i++) {
      for(int j=0; j<4; j++) {
	S(i,j) = (*this)(i,j) + B(i,j);
      }
    }
    return S;
  }
  
  template<class T>
  Vec3<T> Matrix4<T>::operator* (const Vec3<T> &v) const {
    
    Vec3<T> vv;
    
    vv.x = 
      (*this)(0,0)*v.x + 
      (*this)(0,1)*v.y + 
      (*this)(0,2)*v.z +
      (*this)(0,3);
    
    vv.y = 
      (*this)(1,0)*v.x + 
      (*this)(1,1)*v.y + 
      (*this)(1,2)*v.z +
      (*this)(1,3);
    
    vv.z = 
      (*this)(2,0)*v.x + 
      (*this)(2,1)*v.y + 
      (*this)(2,2)*v.z +
      (*this)(2,3);
    
    return vv;
  }

  template<class T>
  Vec4<T> Matrix4<T>::operator* (const Vec4<T> &v) const {
    
    Vec4<T> vv;
    
    vv.x = 
      (*this)(0,0)*v.x + 
      (*this)(0,1)*v.y + 
      (*this)(0,2)*v.z +
      (*this)(0,3)*v.w;
    
    vv.y = 
      (*this)(1,0)*v.x + 
      (*this)(1,1)*v.y + 
      (*this)(1,2)*v.z +
      (*this)(1,3)*v.w;
    
    vv.z = 
      (*this)(2,0)*v.x + 
      (*this)(2,1)*v.y + 
      (*this)(2,2)*v.z +
      (*this)(2,3)*v.w;

    vv.w = 
      (*this)(3,0)*v.x + 
      (*this)(3,1)*v.y + 
      (*this)(3,2)*v.z +
      (*this)(3,3)*v.w;
    
    return vv;
  }
  
  template<class T>
  Matrix3<T> Matrix4<T>::getScaleRotation() const {
    const Matrix4<T> &m = *this;
    return Matrix3<T>(m(0,0), m(0,1), m(0,2),
		      m(1,0), m(1,1), m(1,2),
		      m(2,0), m(2,1), m(2,2));
  }
  
  template<class T>
  Matrix3<T> Matrix4<T>::getRotation() const {
    Matrix3<T> m = getScaleRotation();
    Vec3<T> x_axis = m * Vec3<T>(1,0,0);
    Vec3<T> y_axis = m * Vec3<T>(0,1,0);
    Vec3<T> z_axis = m * Vec3<T>(0,0,1);
    
    x_axis.normalizeSafe();
    y_axis.normalizeSafe();
    z_axis.normalizeSafe();
    
    return Matrix3<T>( x_axis.x, y_axis.x, z_axis.x,
		       x_axis.y, y_axis.y, z_axis.y,
		       x_axis.z, y_axis.z, z_axis.z );
  }
  
  template<class T>
  Vec3<T> Matrix4<T>::getTranslation() const {
    const Matrix4<T> &m = *this;    
    return Vec3<T>(m(0,3), m(1,3), m(2,3));
  }
  
  template<class T>
  Matrix4<T> Matrix4<T>::transpose() const {
    
    Matrix4<T> TR;
    
    for(int i=0; i<4; i++) {
      for(int j=0; j<4; j++) {
	TR(i,j) = (*this)(j,i);
      }
    }
    return TR;
  }   

  template<class T>
  Matrix4<T> Matrix4<T>::transformInverse() const {
    
    const Matrix4<T> &m = *this;
    
    T m00 = m(0,0);
    T m01 = m(0,1);
    T m02 = m(0,2);
    T m03 = m(0,3);
    
    T m10 = m(1,0);
    T m11 = m(1,1);
    T m12 = m(1,2);
    T m13 = m(1,3);
    
    T m20 = m(2,0);
    T m21 = m(2,1);
    T m22 = m(2,2);
    T m23 = m(2,3);
    
    T t4 = m00*m11;
    T t6 = m00*m21;
    T t8 = m10*m01;
    T t10 = m10*m21;
    T t12 = m20*m01;
    T t14 = m20*m11;
    T d = (t4*m22-t6*m12-t8*m22+t10*m02+t12*m12-t14*m02);
    
    T t17 = 1/d;
    
    T t20 = m21*m02;
    T t23 = m01*m12;
    T t24 = m11*m02;
    T t43 = m20*m02;
    T t46 = m00*m12;
    T t47 = m10*m02;
    T t51 = m00*m13;
    T t54 = m10*m03;
    T t57 = m20*m03;
    T inv00 = (m11*m22-m21*m12)*t17;
    T inv01 = -(m01*m22-t20)*t17;
    T inv02 = (t23-t24)*t17;
    T inv03 = 
      -(t23*m23-m01*m13*m22-t24*m23+m11*m03*m22+t20*m13-m21*m03* m12)*t17;
    T inv10 = -(m10*m22-m20*m12)*t17;
    T inv11 = (m00*m22-t43)*t17;
    T inv12 = -(t46-t47)*t17;
    T inv13 = (t46*m23-t51*m22-t47*m23+t54*m22+t43*m13-t57*m12)*t17;
    T inv20 = (t10-t14)*t17;
    T inv21 = -(t6-t12)*t17;
    T inv22 = (t4-t8)*t17;
    T inv23 = -(t4*m23-t51*m21-t8*m23+t54*m21+t12*m13-t57*m11)*t17;
    
    return Matrix4<T>( inv00, inv01, inv02, inv03,
		       inv10, inv11, inv12, inv13,
		       inv20, inv21, inv22, inv23,
		       0, 0, 0, 1 );  
  }
  
  template<class T>
  Matrix4<T> Matrix4<T>::inverse() const {
    
    const Matrix4<T> &m = *this;
    
    T m00 = m(0,0);
    T m01 = m(0,1);
    T m02 = m(0,2);
    T m03 = m(0,3);
    
    T m10 = m(1,0);
    T m11 = m(1,1);
    T m12 = m(1,2);
    T m13 = m(1,3);
    
    T m20 = m(2,0);
    T m21 = m(2,1);
    T m22 = m(2,2);
    T m23 = m(2,3);
    
    T m30 = m(3,0);
    T m31 = m(3,1);
    T m32 = m(3,2);
    T m33 = m(3,3);

    T t14 = m00*m11;
    T t15 = m22*m33;
    T t17 = m23*m32;
    T t19 = m00*m21;
    T t20 = m12*m33;
    T t22 = m13*m32;
    T t24 = m00*m31;
    T t25 = m12*m23;
    T t27 = m13*m22;
    T t29 = m10*m01;
    T t32 = m10*m21;
    T t33 = m02*m33;
    T t35 = m03*m32;
    T t37 = m10*m31;
    T t38 = m02*m23;
    T t40 = m03*m22;
    T t42 =	
      t14*t15-t14*t17-t19*t20+t19*t22+t24*t25-t24*t27-
      t29*t15+t29*t17+t32*t33-t32*t35-t37*t38+t37*t40;
    T t43 = m20*m01;
    T t46 = m20*m11;
    T t49 = m20*m31;
    T t50 = m02*m13;
    T t52 = m03*m12;
    T t54 = m30*m01;
    T t57 = m30*m11;
    T t60 = m30*m21;
    T t63 = 
      t43*t20-t43*t22-t46*t33+t46*t35+t49*t50-t49*t52-
      t54*t25+t54*t27+t57*t38-t57*t40-t60*t50+t60*t52;
    
    T d = t42+t63;
    
    T t65 = 1/d;
    T t71 = m21*m02;
    T t73 = m03*m21;
    T t75 = m02*m31;
    T t77 = m03*m31;
    T t81 = m01*m12;
    T t83 = m01*m13;
    T t85 = m11*m02;
    T t87 = m03*m11;
    T t101 = m10*m22;
    T t103 = m10*m23;
    T t105 = m20*m12;
    T t107 = m20*m13;
    T t111 = m30*m13;
    T t115 = m00*m22;
    T t117 = m00*m23;
    T t119 = m20*m02;
    T t121 = m20*m03;
    T t123 = m30*m02;
    T t125 = m30*m03;
    T t129 = m00*m12;
    T t131 = m00*m13;
    T t133 = m10*m02;
    T t135 = m10*m03;
    T inv00 = 
      (m11*m22*m33-m11*m23*m32-m21*m12*m33+
       m21*m13*m32+m31*m12*m23-m31*m13*m22)*t65;
    T inv01 = 
      -(m01*m22*m33-m01*m23*m32-t71*m33+
	t73*m32+t75*m23-t77*m22)*t65;
    T inv02 = (t81*m33-t83*m32-t85*m33+t87*m32+t75*m13-t77*m12)*t65;
    T inv03 = -(t81*m23-t83*m22-t85*m23+t87*m22+t71*m13-t73*m12)*t65;
    T inv10 = 
      -(t101*m33-t103*m32-t105*m33+t107*m32+
	m30*m12*m23-t111*m22)*t65;
    T inv11 = 
      (t115*m33-t117*m32-t119*m33+t121*m32+
       t123*m23-t125*m22)*t65;
    T inv12 = 
      -(t129*m33-t131*m32-t133*m33+t135*m32+
	t123*m13-t125*m12)*t65;
    T inv13 = 
      (t129*m23-t131*m22-t133*m23+t135*m22+
       t119*m13-t121*m12)*t65;
    T inv20 = (t32*m33-t103*m31-t46*m33+t107*m31+t57*m23-t111*m21)*t65;
    T inv21 = -(t19*m33-t117*m31-t43*m33+t121*m31+t54*m23-t125*m21)*t65;
    T inv22 = (t14*m33-t131*m31-t29*m33+t135*m31+t54*m13-t125*m11)*t65;
    T inv23 = -(t14*m23-t131*m21-t29*m23+t135*m21+t43*m13-t121*m11)*t65;
    T inv30 = -(t32*m32-t101*m31-t46*m32+t105*m31+t57*m22-t60*m12)*t65;
    T inv31 = (t19*m32-t115*m31-t43*m32+t119*m31+t54*m22-t60*m02)*t65;
    T inv32 = -(t14*m32-t129*m31-t29*m32+t133*m31+t54*m12-t57*m02)*t65;
    T inv33 = (t14*m22-t19*m12-t29*m22+t32*m02+t43*m12-t46*m02)*t65;
    
    return( Matrix4<T>( inv00, inv01, inv02, inv03,
			inv10, inv11, inv12, inv13,
			inv20, inv21, inv22, inv23,
			inv30, inv31, inv32, inv33 ) );  
  }
  
  template<class U>
  Matrix4<U> rotation(float angle_degrees, const Vec3f &axis) {
    
    Vec3f f1 = axis;
    f1.normalize();
    
    Vec3f f2 = f1.cross(Vec3f(.5f,.5f,.5f));
    if( f2.norm() < 1e-6 )
      f2 = f1.cross(Vec3f(1,0,0));
    f2.normalize();
    
    Vec3f f3 = f1.cross(f2);

    Matrix4<U> m(f1.x, f1.y, f1.z, 0.0,
                 f2.x, f2.y, f2.z, 0.0,
                 f3.x, f3.y, f3.z, 0.0,
                 0.0,  0.0,  0.0,  1.0);
    
    float theta = angle_degrees*float(constants::pi/180.0f);
    float ct = cos(theta);
    float st = sin(theta);
    
    Matrix4<U> Rx(1,  0,   0, 0,
                  0, ct, -st, 0,
                  0, st,  ct, 0,
                  0,  0,   0, 1);

    return m.transpose()*Rx*m;
    
  }

  template<class U>
  Matrix4<U> scaling(float sx, float sy, float sz) {
    return Matrix4<U>(sx, 0.0, 0.0, 0.0,
		      0.0, sy, 0.0, 0.0,
		      0.0, 0.0, sz, 0.0,
		      0.0, 0.0, 0.0, 1.0);
  }
  
  template<class U>
  Matrix4<U> translation(float dx, float dy, float dz) {
    return Matrix4<U>(1.0, 0.0, 0.0, dx,
		      0.0, 1.0, 0.0, dy,
		      0.0, 0.0, 1.0, dz,
		      0.0, 0.0, 0.0, 1.0);
  }

  template<class U>
  Matrix4<U> scaling(const Vec3f &s) {
    return Matrix4<U>(s.x, 0.0, 0.0, 0.0,
		      0.0, s.y, 0.0, 0.0,
		      0.0, 0.0, s.z, 0.0,
		      0.0, 0.0, 0.0, 1.0);
  }
  
  template<class U>
  Matrix4<U> translation(const Vec3f &t) {
    return Matrix4<U>(1.0, 0.0, 0.0, t.x,
		      0.0, 1.0, 0.0, t.y,
		      0.0, 0.0, 1.0, t.z,
		      0.0, 0.0, 0.0, 1.0);
  }
  
} // end namespace wish

#endif
