///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include <exception>
#include <cassert>

#include "wishLiveMarching.hh"

#include "wishRasterization.hh"

namespace wish {
  
  // constants
  const uint8 LiveMarching::ACCEPTED = uint8(128);
  const uint8 LiveMarching::FARAWAY = uint8(0);
  const uint8 LiveMarching::TRIAL = uint8(255);
  
  // compute the time at a certain position
  float LiveMarching::computeTime(const Vec3i &pos) {
    
    int dim = this->m_input->getDim();
    
    // compute the differences and find min  
    double a[3];
    a[0] = util::minOf(m_time.getSafe(pos.x-1,pos.y,pos.z).value, 
		       m_time.getSafe(pos.x+1,pos.y,pos.z).value);
    a[1] = util::minOf(m_time.getSafe(pos.x,pos.y-1,pos.z).value, 
		       m_time.getSafe(pos.x,pos.y+1,pos.z).value);
    
    if( dim == 3 ) 
      a[2] = util::minOf(m_time.getSafe(pos.x,pos.y,pos.z-1).value, 
			 m_time.getSafe(pos.x,pos.y,pos.z+1).value);
    
    // sort them
    util::bubblesort(a,dim);
    
    // risk for overflow if not casting
    // we add 1.0, because we can not have a cost of zero
    double f2 = util::square( this->m_input->getSafeAsDouble(pos)+1.0f );

    double m;    
    double asum, sum, ssum, sqsum, t; 
    
    if( dim == 3 ) {
      // update step
      asum = (a[2]-a[0])*(a[2]-a[0]) + (a[2]-a[1])*(a[2]-a[1]);
      if( asum <= f2 ) {
	m = 3;
	sum   = a[0] + a[1] + a[2];
	ssum  = sum*sum;
	sqsum = a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
	t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
      }
      else {
	m = 2;
	asum = (a[1]-a[0])*(a[1]-a[0]);
	if( asum <= f2 ) {
	  sum   = a[0] + a[1];
	  ssum  = sum*sum;
	  sqsum = a[0]*a[0] + a[1]*a[1];
	  t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
	}
	else {
	  t = a[0] + m_input->getSafeAsDouble(pos) + 1.0f;
	}
      }
    }
    
    else { // dim = 2   
      // update step
      asum = (a[1]-a[0])*(a[1]-a[0]);
      if( asum <= f2 ) {
	m = 2;
	sum   = a[0] + a[1];
	ssum  = sum*sum;
	sqsum = a[0]*a[0] + a[1]*a[1];
	t     = (sum+sqrt(m*f2+ssum-m*sqsum))/m;
      }
      else {
	t = a[0] + m_input->getSafeAsDouble(pos) + 1.0f;
      }
    }
    
    return float(t);
  }
  
  Vec3f LiveMarching::GodunovGradient(const Vec3i &pos) const {
    
    Vec3f g;
    
    // compute differences
    float diff1, diff2;
    float tp = m_time.getSafe(pos).value;
    
    // diff1 should be <= 0
    diff1 = util::minOf(0.0f, m_time.getSafe(pos.x+1,pos.y,pos.z).value-tp);
    
    // diff2 should be >= 0
    diff2 = util::maxOf(0.0f, tp-m_time.getSafe(pos.x-1,pos.y,pos.z).value);
    
    // put largest in diff1
    if( fabs(diff2) > fabs(diff1) )
      util::swap(diff1,diff2);
    
    // set as gradient
    g.x = diff1;
    
    // same for y
    diff1 = util::minOf(0.0f, m_time.getSafe(pos.x,pos.y+1,pos.z).value-tp);
    diff2 = util::maxOf(0.0f, tp-m_time.getSafe(pos.x,pos.y-1,pos.z).value);
    if( fabs(diff2) > fabs(diff1) )
      util::swap(diff1,diff2);
    g.y = diff1;
    
    // ...and z
    diff1 = util::minOf(0.0f, m_time.getSafe(pos.x,pos.y,pos.z+1).value-tp);
    diff2 = util::maxOf(0.0f, tp-m_time.getSafe(pos.x,pos.y,pos.z-1).value);
    if( fabs(diff2) > fabs(diff1) )
      util::swap(diff1,diff2);
    g.z = diff1;
    
    return g;
  }
  
  Vec3f LiveMarching::GradientInterp(const Vec3f &pos) const {
    
#define G(_argument_) GodunovGradient(_argument_)

    // integer part
    int X = int(pos.x);
    int Y = int(pos.y);
    int Z = int(pos.z);
    
    // decimal part
    double ax = pos.x-X;
    double ay = pos.y-Y;
    double az = pos.z-Z;
    
    Vec3f g =  
      util::interp(
		   util::interp(
				util::interp(G(Vec3i(X,Y,Z)),
					     G(Vec3i(X+1,Y,Z)), 
					     ax),
				util::interp(G(Vec3i(X,Y+1,Z)),
					     G(Vec3i(X+1,Y+1,Z)), 
					     ax),
				ay),
		   util::interp(
				util::interp(G(Vec3i(X,Y,Z+1)), 
					     G(Vec3i(X+1,Y,Z+1)), 
					     ax),
				util::interp(G(Vec3i(X,Y+1,Z+1)), 
					     G(Vec3i(X+1,Y+1,Z+1)), 
					     ax),
				ay),
		   az
		   );
    
    return g;
  }
  
  void LiveMarching::backtrack() {
        
    m_backtrack_cursor.clear();
    
    // initialize variables
    Vec3i p = m_cursor;
    Vec3f p1 = getLastSeed();
    Vec3f pos = p;
    float diff = 1e9f;
    
    m_backtrack_cursor.push_back(pos);
    
    // dont backtrack if distance too small, then just trust the user
    if( (pos-p1).norm() < m_minbacktrackdistance ) {
      m_backtrack_cursor.push_back(p1);
      return;
    }
    
    int iter=0;
    const int MAXITER=10000;
    while( diff > 1.0 && iter < MAXITER ) {
      iter++;
      
      Vec3f g = GradientInterp(pos);
      g.normalizeSafe();
      
      // step in negative gradient direction
      pos = pos - 0.5f*g;
      diff = float( (pos-p1).norm() );
      
      m_backtrack_cursor.push_back(pos);
    }
    
    // cheat a little bit...
    // we always reach our goal 
    if( iter == MAXITER ) 
      m_backtrack_cursor.push_back(p1);

  }
  
  void LiveMarching::initializeFM() {

    cerr << "(Re)initializing, heap size: " << m_heap.size() << endl;
    
    // images
    m_trial.reset(FARAWAY);
    m_time.reset( fElem(1.0e9f) );
    
    // heap
    m_heap.clear();
    
    cerr << "Clearing heap, heap size: " << m_heap.size() << endl;
    cerr << "Clearing heap, heap capacity: " << m_heap.getItems().capacity() << endl;
    
    // at the seed
    m_trial(getLastSeed()) = ACCEPTED;
    m_time(getLastSeed()) = fElem(0.0f);
    
    // move there
    labelNb.moveTo(getLastSeed());
    
    // loop neighbors and compute times
    labelNb.firstnbbox();
    do {
      Vec3i pos = labelNb.getNeighborPosition();
      labelNb.valuenb() = TRIAL;
      m_time.at(pos) = fElem(computeTime(pos));
      m_heap.insert(&(m_time.at(pos)));
    } while( labelNb.nextnbbox() );
  }
  
  void LiveMarching::FM() {
    
    float mintime = 0.0f, current_time = 0.0f;

    // main loop
    while ( ( !(m_heap.empty()) ) && 
	    ( m_trial(m_cursor) != TRIAL ) && 
	    ( m_trial(m_cursor) != ACCEPTED ) ) {
	
      // find smallest
      fElem *tmin = m_heap.deleteRoot();

      mintime = tmin->value;
	
      Vec3i pmin = m_time.indexToPos(int(tmin-&m_time[0]));
	
      // set the smallest to ACCEPTED
      labelNb.moveTo(pmin);
      labelNb.value() = ACCEPTED;
      
      Vec3i pos;
	
      // examine the neighbors
      labelNb.firstnbbox();
      do {
	pos = labelNb.getNeighborPosition();
	current_time = computeTime(pos);

	// FARAWAY neighbor, move to TRIAL set
	if( labelNb.valuenb() == FARAWAY ) {
	  labelNb.valuenb() = TRIAL;
	  m_time.at(pos) = fElem(current_time);
	  m_heap.insert(&(m_time.at(pos)));
	}
	
	// TRIAL neighbor, update time and restore heap
	else if( labelNb.valuenb() == TRIAL ) {
	  fElem *t = &(m_time.at(pos));

	  // update only if current_time is smaller
	  if( current_time < t->value ) {
	    t->value = current_time;
	    m_heap.upHeap(t->index);
	  }
	  
	}
	
      } while( labelNb.nextnbbox() );
      
    }
    
    
  }
  
  // rasterize contour to image
  void LiveMarching::rasterize(wish::Image8* im, 
			       int slice,
			       bool fill) const {
    
    if( im ) {
      cerr << "Rasterizing contour on slice " << m_slice << endl;
            
      // the point arrays
      const vector< vector<wish::Vec3f> > &bt = this->getBacktrack();
      
      // only rasterize the completed segments, not the backtrack from cursor
      for(int j=0; j<int(bt.size()); j++) {
	for(int k=0; k<int(bt[j].size())-1; k++) {
	  Vec3f v0 = bt[j][k];
	  Vec3f v1 = bt[j][k+1];
	  v0.z = float(slice);
	  v1.z = float(slice);
	  
	  Line l( v0, v1 );
	  l.setColor(255);
	  l.draw(im);
	}
      }
      
      if( fill ) { 
	
      }
      
      cerr << "Finished!" << endl;
    }
    
  }
  
  
}
