///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_LIVE_MARCHING_HH_
#define _WISH_LIVE_MARCHING_HH_

#include "wishMinHeap.hh"
#include "wishNeighborhoodIterator.hh"

namespace wish {
  
  class LiveMarching {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< LiveMarching > Pointer;
    
    //! typedef for heap element
    typedef HeapElement<float> fElem;
    typedef MinHeap<float> fMinHeap;
    
    //! typedef for seed iterator
    typedef vector<Vec3i>::iterator SeedsIterator;
    
  protected:
    
    // the cost image
    ImageObject *m_input;
    
    // the slice
    int m_slice;
    
    // the time image (contains heap elements)
    ImageBase<fElem> m_time;
    
    // the trial image
    Image<uint8> m_trial;
    
    // the heap 
    fMinHeap m_heap;
    
    // the seed vector
    vector<Vec3i> m_seeds;
    
    // the cursor position
    Vec3i m_cursor;
    
    // minimum distance for backtracking, if distance between cursor and seed
    // is smaller, we just use a line segment between them
    float m_minbacktrackdistance;
    
    // backtrack positions
    vector< vector<Vec3f> > m_backtrack;
    // backtrack from cursor
    vector<Vec3f> m_backtrack_cursor;
    
    // static constants
    static const uint8 ACCEPTED;
    static const uint8 FARAWAY;
    static const uint8 TRIAL;
    
  public: 
    //! constructor
    LiveMarching(int _capacity=100000) : m_heap(_capacity) {
      m_input = 0;
      m_heap.clear();
      m_seeds.resize(0);
      m_backtrack.resize(0);
      m_backtrack_cursor.resize(0);
      m_slice = 0;
      m_minbacktrackdistance = 5.0f;
    }
    
    //! initialize
    void initialize() {
	
      // clear the seeds
      clearSeeds();
      
      if( m_input ) {
	m_time.allocate(m_input->getDimensions());
	m_time.setSpacing(m_input->getSpacing());
	m_trial.resize(m_time);

	m_trial.reset(FARAWAY);
	m_time.reset(0.0f);

	// neighborhood iterator
	NeighborhoodType nt = C6;
	int dim = this->m_input->getDim();
	if( dim == 2 )
	  nt = C4;
	labelNb.setImagePointer(&m_trial);
	labelNb.setNeighborhood(nt);
	labelNb.computeNeighborhoodBox();
      }
      else {
	cerr << "LiveMarching::initialize() - No input!" << endl;
      }
      
    }
    
    //! set the input image
    void setInput(ImageObject *_input) {
      m_input = _input;
      initialize();
    }

    //! get the input image
    ImageObject* getInput() const {
      return m_input;
    }
    
    //! set slice
    void setSlice(int s) {
      m_slice = s;
      if( m_slice < 0 ) m_slice = 0;
      if( m_slice > m_input->depth()-1 ) m_slice = m_input->depth()-1;
    }
    
    //! get slice
    int getSlice() const {
      return m_slice;
    }
    
    //! set minimum backtrack dist
    void setMinimumBacktrackDistance(const float &f) {
      m_minbacktrackdistance = f;
    }
    
    //! set cursor pos
    void setCursor(const Vec3i &v) {
      m_cursor = v;
    }
    void setCursor(int x, int y, int z=0) {
      setCursor(Vec3i(x,y,z));
    }
    
    //! get cursor
    const Vec3i getCursor() const {
      return m_cursor;
    }
    
    //! get last seed
    const Vec3i& getLastSeed() const {
      return *(m_seeds.rbegin());
    }
    
    //! get first seed
    const Vec3i& getFirstSeed() const {
      return *(m_seeds.begin());
    }
    
    //! get seed vector
    const vector<Vec3i>& getSeeds() const {
      return m_seeds;
    }
    
    //! get cursor backtrack
    const vector<Vec3f>& getBacktrackCursor() const {
      return m_backtrack_cursor;
    }
    
    //! get cursor backtrack
    const vector<Vec3f>& getBacktrack(int i) const {
      return m_backtrack[i];
    }

    //! get cursor backtrack
    const vector< vector<Vec3f> >& getBacktrack() const {
      return m_backtrack;
    }
    
    //! get trial image
    Image8* getTrialImage() {
      return &m_trial;
    }
    
    //! put seed at cursor
    void putSeed() {
      if( m_input->isInside(m_cursor) ) {
	// add the seed and save the cursor backtrack
	m_seeds.push_back(m_cursor);
	m_backtrack.push_back(m_backtrack_cursor);      
	
	// initialize fastmarching
	initializeFM();
      }
    }
    
    //! close curve (use first seed again)
    void closeCurve() {
      if( getNrOfSeeds() <= 1 )
	return;
      
      setCursor(getFirstSeed());
      initializeFM();
      update();
      m_backtrack.push_back(m_backtrack_cursor);
    }
    
    //! remove a seed
    void removeSeed() {
      
      // remove last seed and last backtrack
      m_seeds.pop_back();
      m_backtrack.pop_back();
      
      if( getNrOfSeeds() > 0 ) {	
	// initialize fastmarching
	initializeFM();
	
	// update (backtrack)
	update();
      }
      else {
	// remove the old cursor backtrack
	m_backtrack_cursor.resize(0);
      }
    }
    
    //! clear all seeds
    void clearSeeds() {
      m_seeds.resize(0);
      m_backtrack_cursor.resize(0);
      m_backtrack.resize(0);
    }
    
    //! get nr of seeds
    int getNrOfSeeds() const {
      return int( m_seeds.size() ); 
    }
    
    //! the interaction function, put in a mouse callback for example
    virtual void update() {
      if( m_seeds.size() > 0 ) {
	if( m_trial(m_cursor) != ACCEPTED ) 
	  FM();

	backtrack();

      }

    }
    
    //! rasterize contour to image
    virtual void rasterize(wish::Image8* im, 
			   int slice,
			   bool fill=false) const;
    
  protected:    
    void FM();
    void initializeFM();
    float computeTime(const Vec3i &pos);
    void backtrack();
    Vec3f GodunovGradient(const Vec3i &pos) const;
    Vec3f GradientInterp(const Vec3f &pos) const;
    
    NeighborhoodIterator< Image<uint8> > labelNb;
  
  private:
    // purposely not implemented
    LiveMarching(const LiveMarching&);
    LiveMarching& operator=(const LiveMarching&);
  };
  
} // end namespace wish

#endif
