///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_MIN_HEAP_HH_
#define _WISH_MIN_HEAP_HH_

#include "wish.hh"

namespace wish {
  
  //! a heap element has a value and an index, i.e., we know where
  //! in the heap it is stored. This construction allows for faster 
  //! heap restore 
  template<class X>
  struct HeapElement {
  public:
    X value;
    int index;
    HeapElement(const X &_value=0, int _index=-1) : 
      value(_value), index(_index) {}
  };
  
  //! the minimum heap class
  //!
  //! Implements a minimum heap data structure where each element in the heap
  //! also contains an index that tells where in the heap it is located. This
  //! allows for faster "upHeap" operations. The min heap is used in 
  //! wishFastMarching. For more general priority queues, use stl:s 
  //! priority_queue or heap. 
  template<class T>
  class MinHeap {
  public:
    typedef HeapElement<T>* pElem;
    
    vector<pElem> items;
    int count;
    
    MinHeap(int _capacity=100000) {
      items.resize(0);
      items.reserve(_capacity);
      count = 0;
    }
    
    void clear() {
      items.resize(0);
      count = 0;
    }
    
    virtual bool compare(const int &a, const int &b) {
      return getElement(a)->value < getElement(b)->value;
    }

    virtual bool compare2(const int &a, const int &b) {
      return getElement(a)->value <= getElement(b)->value;
    }
    
    int size() const {
      return count;
    }
    
    bool empty() const { 
      return count == 0; 
    }
    
    bool full() const {
      return false; //count >= capacity;
    }
    
    pElem& operator[] (const int &i) {
      return items[i-1];
    }
    
    pElem& getElement(const int &i) {
      return items[i-1];
    }
    
    void setElement(const int &i, pElem elem) {
      items[i-1] = elem;
      items[i-1]->index = i;
    }
    
    const vector<pElem>& getItems() const {
      return items;
    }
    
    // inserts an element and maintains heap structure
    // ( corresponds to priority_queue::push() )
    void insert(pElem elem);
    
    // restore heap structure from element with index k
    void upHeap(const int &k);
    
    // remove root ( corresponds to priority_queue::pop() )
    pElem deleteRoot();
    
    // get root ( correponds to priority_queue::top() )
    pElem getRoot() {
      if( !empty() )
	return items[0];
      else {
	cerr << "HEAP EMPTY!!!" << endl;
	return 0;
      }
    }
    
    // returns true if array is a heap
    bool isHeap(int k=1) {
      if(2*k >= count)
	return true;
      
      if( compare2( k, 2*k ) && compare2( k, 2*k+1 ) )
	return isHeap(2*k) && isHeap(2*k+1);
      else 
	return false;
    }
    
    // returns parent index
    int parentIndex(const int &i) const {
      return i/2;
    }
    
    // return index of left child
    int leftIndex(const int &i) const {
      return 2*i;
    }
    
    // return index of right child
    int rightIndex(const int &i) const {
      return 2*i+1;
    }
    
    // swap elements at index a and index b
    void swapElements(const int &a, const int &b) {
      pElem tmp = items[a-1];
      items[a-1] = items[b-1];
      items[a-1]->index = a;
      items[b-1] = tmp;
      items[b-1]->index = b;
    }
    
    void print();
    void printSorted();
    
  private:
    
    // purposely not implemented
    MinHeap(const MinHeap &mh);
    MinHeap& operator=(const MinHeap &mh);
  };
  
  template<class T>
  void MinHeap<T>::insert(HeapElement<T>* elem) {
    items.push_back(elem);
    elem->index = ++count;
    upHeap(count);
  }
  
  template<class T>
  void MinHeap<T>::upHeap(const int &k) {
    int current = k;
    int parent = parentIndex(current);
    
    while( (current>1) && compare(current,parent) ) {
      swapElements(current,parent);
      current = parent;
      parent = parentIndex(current);
    }
  }
  
  template<class T>
  HeapElement<T>* MinHeap<T>::deleteRoot() {
    pElem ret = items[0];
    items[0] = items[count-1];
    items[0]->index = 1;
    items.pop_back();
    count--;

    int current = 1;    
    while( leftIndex(current) <= count ) {
      int left = leftIndex(current);
      int right = rightIndex(current);
      int move = left;
      
      if( (right <= count) && !compare(left,right) ) {
	move = right;
      }
      if( !compare(current,move) )
	swapElements(current,move);
      
      current = move;
    }
    
    return ret;
  }
  
  template<class T>
  void MinHeap<T>::print() {
    cout << endl;
    for(int i=1; i<=count; i++)
      cout << "(" 
	   << getElement(i)->index << ", " 
	   << getElement(i)->value 
	   << ")" << endl;
  }
  
  template<class T>
  void MinHeap<T>::printSorted() {
    cout << endl;
    while( !empty() ) {
      HeapElement<T> *elem = deleteRoot();
      cout << elem->value << ", ";
    }
    cout << endl;
  }
  
} // end namespace wish

#endif

  
