///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_NEIGHBORHOOD_ITERATOR_HH_
#define _WISH_NEIGHBORHOOD_ITERATOR_HH_

#include "wishVoxelIterator.hh"

namespace wish {
  
  enum NeighborhoodType 
    { 
      CUserDefined, // user defined neighborhood
      
      C0,      //  no neighbors
      
      C4,      //  4 connectedness 2D 
      C8,      //  8 connectedness 2D
      
      C6,      //  6 connectedness 3D
      C18,     // 18 connectedness 3D
      C26,     // 26 connectedness 3D
      C124,    // 124 connectedness 3D
      
      CHFWD4,  //  4 chamfer forward 2D
      CHFWD8,  //  8 chamfer forward 2D
      CHBWD4,  //  4 chamfer backward 2D
      CHBWD8,  //  8 chamfer backward 2D

      CHFWD6,  //  6 chamfer forward 3D 
      CHFWD18, // 18 chamfer forward 3D 
      CHFWD26, // 26 chamfer forward 3D 
      CHBWD6,  //  6 chamfer backward 3D 
      CHBWD18, // 18 chamfer backward 3D
      CHBWD26  // 26 chamfer backward 3D
    };
  
  static Vec3i NeighborPos2D[8] = 
    {
      // 4 face neighbors (0-3)
      Vec3i( 1, 0), 
      Vec3i( 0, 1), 
      Vec3i( 0,-1), 
      Vec3i(-1, 0),
      
      // 4 vertex neighbors (4-7)
      Vec3i( 1, 1),
      Vec3i( 1,-1),
      Vec3i(-1, 1),
      Vec3i(-1,-1)
    };
  
  static Vec3i NeighborPos3D[26] =
    { 
      // 6 face neighbors (0-5)
      Vec3i( 1, 0, 0), // EAST    
      Vec3i( 0, 1, 0), // DOWN
      Vec3i( 0, 0, 1), // NORTH
      Vec3i( 0, 0,-1), // SOUTH
      Vec3i( 0,-1, 0), // UP
      Vec3i(-1, 0, 0), // WEST
      
      // 12 edge neighbors (6-17)
      Vec3i( 1, 1, 0), // EASTDOWN
      Vec3i( 1, 0, 1), // NORTHEAST
      Vec3i( 1, 0,-1), // SOUTHEAST
      Vec3i( 1,-1, 0), // EASTUP
      Vec3i( 0, 1, 1), // NORTHDOWN        
      Vec3i( 0, 1,-1), // SOUTHDOWN
      Vec3i( 0,-1, 1), // NORTHUP
      Vec3i( 0,-1,-1), // SOUTHUP
      Vec3i(-1, 1, 0), // WESTDOWN
      Vec3i(-1, 0, 1), // NORTHWEST
      Vec3i(-1, 0,-1), // SOUTHWEST
      Vec3i(-1,-1, 0), // WESTUP
    
      // 8 vertex neighbors (17-25)
      Vec3i( 1, 1, 1), // NORTHEASTDOWN    
      Vec3i( 1, 1,-1), // SOUTHEASTDOWN
      Vec3i( 1,-1, 1), // NORTHEASTUP
      Vec3i( 1,-1,-1), // SOUTHEASTUP
      Vec3i(-1, 1, 1), // NORTHWESTDOWN    
      Vec3i(-1, 1,-1), // SOUTHWESTDOWN
      Vec3i(-1,-1, 1), // NORTHWESTUP
      Vec3i(-1,-1,-1)  // SOUTHWESTUP
    };
  
  static Vec3i ChamferForward3D[13] = 
    {
      // 3 FACE
      Vec3i(-1, 0, 0),
      Vec3i( 0,-1, 0),
      Vec3i( 0, 0,-1),
    
      // 6 EDGE
      Vec3i( 1,-1, 0),
      Vec3i(-1,-1, 0),
      Vec3i( 0, 1,-1),
      Vec3i( 1, 0,-1),
      Vec3i(-1, 0,-1),
      Vec3i( 0,-1,-1),
    
      // 4 VERTEX
      Vec3i( 1, 1,-1),
      Vec3i(-1, 1,-1),
      Vec3i( 1,-1,-1),
      Vec3i(-1,-1,-1)
    };

  static Vec3i ChamferBackward3D[13] = 
    {
      // 3 FACE
      Vec3i( 1, 0, 0),
      Vec3i( 0, 1, 0),
      Vec3i( 0, 0, 1),
    
      // 6 EDGE
      Vec3i(-1, 1, 0),
      Vec3i( 1, 1, 0),
      Vec3i( 0,-1, 1),
      Vec3i(-1, 0, 1),
      Vec3i( 1, 0, 1),
      Vec3i( 0, 1, 1),
    
      // 4 VERTEX
      Vec3i(-1,-1, 1),
      Vec3i( 1,-1, 1),
      Vec3i(-1, 1, 1),
      Vec3i( 1, 1, 1)
    };
  
  static Vec3i ChamferForward2D[4] = 
    {
      // 2 EDGE
      Vec3i(-1, 0, 0),
      Vec3i( 0,-1, 0),
      
      // 2 VERTEX
      Vec3i( 1,-1, 0),
      Vec3i(-1,-1, 0)
    };

  static Vec3i ChamferBackward2D[4] = 
    {
      // 2 EDGE
      Vec3i( 1, 0, 0),
      Vec3i( 0, 1, 0),
    
      // 2 VERTEX
      Vec3i(-1, 1, 0),
      Vec3i( 1, 1, 0)
    };
  
  template<class ImageType> class NeighborhoodIterator : 
    public VoxelIterator<ImageType> {
  
  public:
    
    //! typedefs
    typedef auto_ptr< NeighborhoodIterator<ImageType> > Pointer;
    typedef typename ImageType::Pointer ImagePointer;
    typedef typename ImageType::VoxelType ImageVoxelType;
    
    //! constructor
    // the default constructor does not compuet a safe box for the 
    // neighborhood, you need to call computeNeighborhoodBox
    NeighborhoodIterator(ImageType *im=0, 
			 NeighborhoodType nt=C0);
    
    //! destructor
    ~NeighborhoodIterator() {
      offset = 0;
    }
    
    // size of the neighborhood
    int getNeighborhoodSize() {
      return num;
    }
    
    // compute a bounding box that guarantees that the whole 
    // neighborhood is inside the image
    void computeNeighborhoodBox();
    
    // set the neighborhood type
    void setNeighborhood(NeighborhoodType _nt) {
      nt = _nt;
      setupNeighborhood();
    }
    
    // set the neighborhood from user defined array
    void setNeighborhood(Vec3i *user_offset, int _n) {
      nt = CUserDefined;
      num = _n;
      offset = user_offset;
    }
    
    //! get the neighborhood type
    NeighborhoodType getNeighborhoodType() {
      return nt;
    }
    
    // sets up the neighborhood (num and offset)
    void setupNeighborhood();
    
    // set the default value
    void setDefaultValue(const ImageVoxelType &val) {
      defaultnb = val;
    }
    
    // the first, last, next, prev, and moveto
    // functions sets current to 0 before invoking the
    // parents corresponding function
    void first() {
      current=0;
      this->VoxelIterator<ImageType>::first();
    }
    void last() {
      current=0;
      this->VoxelIterator<ImageType>::last();
    }
    bool next() {
      current=0;
      return this->VoxelIterator<ImageType>::next();
    }
    bool prev() {
      current = 0;
      return this->VoxelIterator<ImageType>::prev();
    }
    bool moveto(int x, int y, int z) {
      current=0;
      return this->VoxelIterator<ImageType>::moveto(x,y,z);
    }
    bool moveto(const Vec3i &ep) {
      current=0;
      return this->VoxelIterator<ImageType>::moveto(ep);
    }
  
    // resets the neighbor counter
    void firstnb() {
      current = 0;
    }
    void lastnb() {
      current = num-1;
    }
    
    // increments the neighbor counter
    bool nextnb() {
      if(++current < num)
	return true;
      current = 0;
      return false;
    }
    
    // like nextnb() and firstnb(), but checks if inside image
    bool nextnbsafe();
    void firstnbsafe();
    
    // like nextnb() and firstnb(), but checks if inside box
    bool nextnbbox();
    void firstnbbox();
        
    // reference to the value at the current neighbor position
    ImageVoxelType& valuenb() const {
      return (this->img -> operator()(this->pos+offset[current]));
    }
    ImageVoxelType& valuenb(const int &i) const {
      return (this->img -> operator()(this->pos+offset[i]));
    }
  
    // safe neighbor value, returns defaultnb if outside image
    ImageVoxelType valuenbsafe() const {
      if( this->img -> isInside(this->pos+offset[current]) )      
	return (this->img -> operator()(this->pos+offset[current]));
      return defaultnb;
    }
    ImageVoxelType valuenbsafe(const int &i) const {
      if( this->img -> isInside(this->pos+offset[i]) )      
	return (this->img -> operator()(this->pos+offset[i]));
      return defaultnb;
    }
    
    // safe neighbor value, returns defaultnb if outside box
    ImageVoxelType valuenbbox() const {
      if( isInsideBox(this->pos+offset[current]) )      
	return (this->img -> operator()(this->pos+offset[current]));
      return defaultnb;
    }
    ImageVoxelType valuenbbox(const int &i) const {
      if( isInsideBox(this->pos+offset[i]) )      
	return (this->img -> operator()(this->pos+offset[i]));
      return defaultnb;
    }
    
    // neighbor position
    Vec3i getNeighborPosition() const {
      return this->pos+offset[current];
    }
    
    // neighbor position
    Vec3i getNeighborPosition(const int &i) const {
      return this->pos+offset[i];
    }
    
    
  protected:
    NeighborhoodType nt;      // neighborhood type
    int num;                  // number of neighbors
    int current;	      // index of the current neighbor
    Vec3i *offset;            // array of neighbor offsets
    ImageVoxelType defaultnb; // default neighbor value if outside
    
  };
  
  template<class ImageType>
  NeighborhoodIterator<ImageType>::
  NeighborhoodIterator(ImageType *_img, NeighborhoodType _nt):
    VoxelIterator<ImageType>(_img),
    nt(_nt),
    num(0), 
    current(0), 
    offset(0), 
    defaultnb(ImageVoxelType(0)) {
    
    if( this -> img ) {
      if( nt == C0 ) {
	if( this -> img -> getDim() == 2 )
	  nt = C8;
	else
	  nt = C26;
      }
      setupNeighborhood();
      first();    
    }
    
  }
  
  template<class ImageType> 
  void NeighborhoodIterator<ImageType>::
  setupNeighborhood() {
    
    switch(nt) {
    case C0:
    case CUserDefined:
      num =  0;
      offset = 0;
      break;
    case C4:  
      num =  4; 
      offset = NeighborPos2D; 
      break;
    case C8:  
      num =  4 + 4; 
      offset = NeighborPos2D; 
      break;
    case C6:  
      num =  6; 
      offset = NeighborPos3D; 
      break;
    case C18: 
      num = 6 + 12; 
      offset = NeighborPos3D; 
      break;
    case C26: 
      num = 6 + 12 + 8; 
      offset = NeighborPos3D; 
      break;
    case CHFWD4: 
      num = 2; 
      offset = ChamferForward2D; 
      break;
    case CHFWD8: 
      num = 2 + 2; 
      offset = ChamferForward2D; 
      break;
    case CHBWD4: 
      num = 2; 
      offset = ChamferBackward2D; 
      break;
    case CHBWD8: 
      num = 2 + 2; 
      offset = ChamferBackward2D; 
      break;
    case CHFWD6: 
      num = 3; 
      offset = ChamferForward3D; 
      break;
    case CHFWD18: 
      num = 3 + 6; 
      offset = ChamferForward3D; 
      break;
    case CHFWD26: 
      num = 3 + 6 + 4; 
      offset = ChamferForward3D; 
      break;
    case CHBWD6: 
      num = 3; 
      offset = ChamferBackward3D; 
      break;
    case CHBWD18: 
      num = 3 + 6; 
      offset = ChamferBackward3D; 
      break;
    case CHBWD26: 
      num = 3 + 6 + 4; 
      offset = ChamferBackward3D; 
      break;
    default: 
      cerr << "Warning: Bad neighborhood type" << endl; 
      num = 0; 
      offset = 0; 
      break;
    }
    
  }
  
  template<class ImageType>
  void NeighborhoodIterator<ImageType>::computeNeighborhoodBox() {
    
    Vec3i bmax = offset[0];
    Vec3i bmin = offset[0];
    
    // find the furthest neighbors
    for(int i=1; i<num; i++) {
      bmax.x = util::maxOf(bmax.x, offset[i].x);
      bmax.y = util::maxOf(bmax.y, offset[i].y);
      bmax.z = util::maxOf(bmax.z, offset[i].z);
      bmin.x = util::minOf(bmin.x, offset[i].x);
      bmin.y = util::minOf(bmin.y, offset[i].y);
      bmin.z = util::minOf(bmin.z, offset[i].z);
    }
    
    // set the box
    this->setBox( Vec3i(abs(bmin.x), abs(bmin.y), abs(bmin.z)),
		  Vec3i(this->img->width() -1-bmax.x, 
			this->img->height()-1-bmax.y, 
			this->img->depth() -1-bmax.z) ); 
  }
  
  template<class ImageType>
  void NeighborhoodIterator<ImageType>::firstnbsafe() {
    current=0;
    while( !(this->img->isInside(this->pos+offset[current])) && current < num )
      current++;
    
    if( current >= num )
      cerr << "firstnbsafe(): Didn't find first neighbor" << endl;
  
  }

  template<class ImageType>
  bool NeighborhoodIterator<ImageType>::nextnbsafe() {
    current++;
    while( !(this->img->isInside(this->pos+offset[current])) )
      current++;
  
    if( current < num )
      return true;

    current = 0;
    return false;
  }

  template<class ImageType>
  void NeighborhoodIterator<ImageType>::firstnbbox() {
    current=0;
    while( !(isInsideBox(this->pos+offset[current])) && current < num )
      current++;
  
    if( current >= num )
      cerr << "firstnbbox(): Didn't find first neighbor" << endl;
  
  }

  template<class ImageType>
  bool NeighborhoodIterator<ImageType>::nextnbbox() {
    current++;
    while( !(isInsideBox(this->pos+offset[current])) )
      current++;
  
    if( current < num )
      return true;

    current = 0;
    return false;
  }

} // end namespace wish

#endif

