///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_PIXEL_WISE_GAUSSIAN_FILTER_HH_
#define _WISH_PIXEL_WISE_GAUSSIAN_FILTER_HH_

#include "wishFilterObject.hh"

namespace wish {
  
  //! PixelWiseGaussianFilter
  //! Maps input values to the interval [minval,maxval] through the use
  //! of a gaussian function. 
  //! The output is computed as:
  //! <tt>out(x) = (max-min)*exp(-0.5*in(x)^2/sigma^2)+min</tt>.
  template<class InputImage, class OutputImage>
  class PixelWiseGaussianFilter : 
    public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< PixelWiseGaussianFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    PixelWiseGaussianFilter() : 
      m_sigma2(1.0),
      m_mean(0.0),
      m_max(255.0),
      m_min(0.0)
    {}
    
    //! set sigma
    void setSigma(double _sigma) {
      m_sigma2 = _sigma*_sigma;
    }
    
    //! set mean
    void setMean(double _mean) {
      m_mean = _mean;
    }
    
    //! set max
    void setMax(double _max) {
      m_max = _max;
    }

    //! set min
    void setMin(double _min) {
      m_min = _min;
    }
    
  protected:
    
    double m_sigma2, m_mean, m_max, m_min;
    double fgauss(double arg);
    virtual void process();
    
  private:
    // purposely not implemented
    PixelWiseGaussianFilter(const PixelWiseGaussianFilter&);
    PixelWiseGaussianFilter& operator=(const PixelWiseGaussianFilter&);
  };
  
  template<class InputImage, class OutputImage>
  double PixelWiseGaussianFilter<InputImage, OutputImage>::fgauss(double arg) {
    return 
      double(m_max-m_min)*exp(-0.5*util::square(arg-m_mean)/m_sigma2)+m_min;
  }
  
  template<class InputImage, class OutputImage>
  void PixelWiseGaussianFilter<InputImage, OutputImage>::process() {
    
    this->m_output->resize( *(this->m_input) );
    this->m_output->reset( OutputImageVoxelType(0) );    
    
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    for(int i=0; i<this->m_output->totalSize(); i++) 
      out[i] = OutputImageVoxelType( this->fgauss(double(in[i])) );
  }
  
} // end namespace wish

#endif
