///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishPixyFileReader.hh"

namespace wish {
  
  bool PixyFileReader::readHeader() {
     
    // open file, use binary mode (necessary on windows)
    ifstream is;
    is.open( this->m_filename.c_str(), ios::in | ios::binary );
    if( !is ) {
      cerr << "Unable to open file: " << this->m_filename << endl;
      return false;
    }
    
    // parse the header, ends with formfeed
    // IMAGE {
    //  CLASS 1
    //  NAME  'IMP_5'
    //  SIZE { 394 505 1 1 1 }
    //  PIXTYPE UNSIGNED
    //  PIXSIZE 8
    //  PHYSICALPIXSIZE { 0.000000 0.000000 0.000000 0.000000 0.000000 }
    //  RASTER 1
    //  LITTLEENDIAN
    //  LAYERSIZE { 198970 }
    //  Z {   0.0000 }
    //  T { 0 }
    //  B { 0 }
    // }
    // '\f'
    // RAW DATA
    
    // check if it is a IPAD/PIXY file
    std::string checkipad;
    is >> checkipad;
    if( !(checkipad == "IMAGE" || checkipad == "REGION") ) {
      return false;
    }
    // go to first in file
    is.seekg(0,ios::beg);
    
    char c;
    std::string str;
    
    // read everything until '\f' and store it in str
    is.get(c);
    while( c != '\f' ) {
      str+=c;
      is.get(c);
    }
    
    // headersize in bytes, info struct
    this->m_info.headersize = (int) is.tellg();
    
    // find the BIGENDIAN field if it exists, otherwise use little endian
    std::string::size_type be_pos = str.find("BIGENDIAN");
    if(be_pos == std::string::npos) {
      this->m_little_endian = true;
    }
    else {
      this->m_little_endian = false;
    }
    
    // find the SIZE field
    std::string::size_type size_pos = str.find("SIZE");
    if(size_pos == std::string::npos) {
      this->printSelf();
      cerr << ": SIZE not found" << endl;
      return false;
    }
    
    int W,H,D;
    // read the SIZE field
    is.seekg((int) size_pos, ios::beg);
    while(is.get() != '{');
    is >> W >> H >> D;
    
    // info struct
    this->m_info.dim = Vec3i(W,H,D);
    
    // find the PIXTYPE field
    std::string::size_type pixtype_pos = str.find("PIXTYPE");
    if(pixtype_pos == std::string::npos) {
      this->printSelf();
      cerr << ": PIXTYPE not found" << endl;
      return false;
    }
    
    std::string pt;
    // read the PIXTYPE field
    is.seekg((int) pixtype_pos, ios::beg);
    is >> pt; // read "PIXTYPE"
    is >> pt; // read value of PIXTYPE
    
    if( pt != "UNSIGNED" && pt != "SIGNED" && pt != "REAL" ) {
      printSelf();
      cerr << ": Bad PIXTYPE" << endl;
      return false;
    }
    
    // find the PIXSIZE field
    std::string::size_type pixsize_pos = str.find("PIXSIZE");
    if(pixsize_pos == std::string::npos) {
      this->printSelf();
      cerr << ": PIXSIZE not found" << endl;
      return false;
    }
    
    std::string ps_str;
    int ps;
    // read the PIXSIZE field
    is.seekg((int) pixsize_pos, ios::beg);
    is >> ps_str; // read "PIXSIZE"
    is >> ps; // read value of PIXSIZE
    
    // info struct
    if( pt=="UNSIGNED" ) {
      if( ps==8 ) this->m_info.datatype="uint8";
      else if( ps==16 ) this->m_info.datatype="uint16";
      else if( ps==32 ) this->m_info.datatype="uint32";
      else { this->printSelf(); cerr << ": PIXTYPE/PIXSIZE error" << endl; }
    }
    else if( pt=="SIGNED" ) {
      if( ps==8 ) this->m_info.datatype="int8";
      else if( ps==16 ) this->m_info.datatype="int16";
      else if( ps==32 ) this->m_info.datatype="int32";
      else { this->printSelf(); cerr << ": PIXTYPE/PIXSIZE error" << endl; }
    }
    else { // if( pt=="REAL" )
      if( ps==32 ) this->m_info.datatype="float";
      else if( ps==64 ) this->m_info.datatype="double";
      else { this->printSelf(); cerr << ": PIXTYPE/PIXSIZE error" << endl; }
    }
    
    // find the PHYSICALPIXSIZE field
    std::string::size_type physicalpixsize_pos = str.find("PHYSICALPIXSIZE");
    if(physicalpixsize_pos == std::string::npos) {
      printSelf();
      cerr << ": PHYSICALPIXSIZE not found, using 1.0" << endl;
      
      // info
      this->m_info.spc = Vec3f(1.0f,1.0f,1.0f);
    }
    else {
      float spx,spy,spz;
      // read the PHYSICALPIXSIZE field
      is.seekg((int) physicalpixsize_pos, ios::beg);
      while(is.get() != '{');
      is >> spx >> spy >> spz;
      
      // if some crazy default value...
      if( spx == 0.0f && spy == 0.0f && spz == 0.0f )
	spx = spy = spz = 1.0f;
      
      // info
      this->m_info.spc = Vec3f(spx,spy,spz);
    }
    
    is.close();
    
    return true;
  }
  
  bool PixyFileReader::readData() {
    
    // open file, use binary mode (necessary on windows)
    ifstream is;
    is.open( this->m_filename.c_str(), ios::in | ios::binary );
    if( !is ) {
      cerr << "Unable to open file: " << this->m_filename << endl;
      return false;
    }
    
    // jump to data
    is.seekg(this->m_info.headersize);

    // use header info to allocate a new image
    if( this->m_info.datatype == "uint8" ) {
      Image<uint8> *tmp = new Image<uint8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int8" ) {
      Image<int8> *tmp = new Image<int8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint16" ) {
      Image<uint16> *tmp = new Image<uint16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int16" ) {
      Image<int16> *tmp = new Image<int16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint32" ) {
      Image<uint32> *tmp = new Image<uint32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int32" ) {
      Image<int32> *tmp = new Image<int32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "float" ) {
      Image<float> *tmp = new Image<float>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "double" ) {
      Image<double> *tmp = new Image<double>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else { // should not happen
      this->printSelf();
      cerr << ": Error reading file " << m_filename << ", bad datatype: " 
	   << this->m_info.datatype << endl;
      return false;
    }
    
    is.close();
    
    return true;
  }
  
}
