///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishImage.hh"

namespace wish {
  
  // abstract base class
  class VoxelObject {
  public:
    virtual void draw(Image8 *im) = 0;
  };
  
  // line data structure
  class Line : public VoxelObject {
  public:
    Vec3f p1; // first point
    Vec3f p2; // last point
    
    int color;
    
    Line(const Vec3f &_p1=Vec3f(0,0,0), 
	 const Vec3f &_p2=Vec3f(0,0,0)) : 
      p1(_p1), 
      p2(_p2),
      color(255) {
    }
    
    void setColor(int col) {
      color = col;
    }
    
    // DDA
    virtual void draw(Image8 *im) {
      
      // sorting
      int ind[3] = {0,1,2};
      float diff[3];
      diff[0] = abs(p2[0] - p1[0]);
      diff[1] = abs(p2[1] - p1[1]);
      diff[2] = abs(p2[2] - p1[2]);
      
      // axis with largest difference will be in ind[2]
      util::bubblesort(diff, ind, 3);
      
      // reverse index array to have largest difference in ind[0]
      util::swap(ind[0], ind[2]);
      
      // setup local start and endpoints and
      // check that the difference is positive
      Vec3f v1 = p1;
      Vec3f v2 = p2;
      if( v2[ind[0]] < v1[ind[0]] )
	util::swap(v1, v2);
      
      // slope
      float delta_x = v2[ind[0]]-v1[ind[0]];
      float delta_y = v2[ind[1]]-v1[ind[1]];
      float delta_z = v2[ind[2]]-v1[ind[2]];
      float dydx = 0.0f;
      float dzdx = 0.0f;
      if( fabs(delta_x) > 1e-9 ) { 
	dydx = delta_y/delta_x;
	dzdx = delta_z/delta_x;
      }
      
      // raster pos
      Vec3i pos;
      
      pos[ind[0]] = int(v1[ind[0]]+0.5f);
      float xskip = pos[ind[0]]-v1[ind[0]];
      float y = v1[ind[1]] + xskip*dydx;
      float z = v1[ind[2]] + xskip*dzdx;
      
      // line algorithm
      for(; pos[ind[0]] <= int(v2[ind[0]]+0.5f); pos[ind[0]]++) {
	pos[ind[1]] = int(y+0.5f);
	pos[ind[2]] = int(z+0.5f);
	im->at( pos ) = color;
	y += dydx;
	z += dzdx;
      }
      
    }
    
    // 3D Bresenham line rasterization
    void drawBresenham(Image8 *im) {
      
      // sorting
      int ind[3] = {0,1,2};
      float diff[3];
      diff[0] = abs(p2[0] - p1[0]);
      diff[1] = abs(p2[1] - p1[1]);
      diff[2] = abs(p2[2] - p1[2]);
      
      // axis with largest difference will be in ind[2]
      util::bubblesort(diff, ind, 3);
      
      // reverse index array to have largest difference in ind[0]
      util::swap(ind[0], ind[2]);
      
      // setup local start and endpoints and
      // check that the difference is positive
      Vec3i v1( int(p1.x+0.5f), int(p1.y+0.5f), int(p1.z+0.5f) );
      Vec3i v2( int(p2.x+0.5f), int(p2.y+0.5f), int(p2.z+0.5f) );
      if( v2[ind[0]] < v1[ind[0]] )
	util::swap(v1, v2);
      
      // here we use x,y,z according to the notation in the paper,
      // i.e., x is the coordinate with the biggest difference and
      // x2-x1 > 0
      int delta_x = v2[ind[0]]-v1[ind[0]];
      
      // initialization in y
      int delta_y = abs(v2[ind[1]]-v1[ind[1]]);
      int ysign = (v2[ind[1]] >= v1[ind[1]]) ? 1:-1;
      int dy = 2*delta_y-delta_x;
      int yinc1 = 2*delta_y;
      int yinc2 = 2*(delta_y-delta_x);
      
      // initialization in z
      int delta_z = abs(v2[ind[2]]-v1[ind[2]]);
      int zsign = (v2[ind[2]] >= v1[ind[2]]) ? 1:-1;
      int dz = 2*delta_z-delta_x;
      int zinc1 = 2*delta_z;
      int zinc2 = 2*(delta_z-delta_x);
      
      // first point
      im->at( v1 ) = color;
      
      // line algorithm
      while( v1[ind[0]] < v2[ind[0]] ) {
	
	v1[ind[0]]++;
	
	if( dy < 0 )
	  dy += yinc1;
	else {
	  dy += yinc2;
	  v1[ind[1]] += ysign;
	}
	
	if( dz < 0 )
	  dz += zinc1;
	else {
	  dz += zinc2;
	  v1[ind[2]] += zsign;
	}
	
	im->at( v1 ) = color;
	
      }
      
    }
    
  };
  
  // polygon data structure
  class Triangle : public VoxelObject {
  public:
    
    // used in the drawing alg.
    struct LineElement {
      Line l;
      float X, Z, DX, DZ;
      
      LineElement(Line _l=Line()) {
	l = _l;
      }
      
    };
    
    Vec3f v[3]; // vertices
    Line edges[3]; // edges 
    float plane[4]; // plane equation    
    bool filled;
    int color;
    int line_color;
    
    Triangle(const Vec3f &_p1=Vec3f(0,0,0),
	     const Vec3f &_p2=Vec3f(0,0,0),
	     const Vec3f &_p3=Vec3f(0,0,0)) {
      v[0] = _p1;
      v[1] = _p2;
      v[2] = _p3;
      filled = true;
      color = line_color = 255;
      initialize();
    }
    
    void setColor(int col) {
      color = col;
    }
    
    void setLineColor(int col) {
      line_color = col;
      edges[0].setColor(line_color);
      edges[1].setColor(line_color);
      edges[2].setColor(line_color);
    }
    
    void initialize() {
      
      // compute plane equation
      Vec3f n = (v[1]-v[0]).cross(v[2]-v[0]);
      plane[0] = n.x;
      plane[1] = n.y;
      plane[2] = n.z;
      plane[3] = -(plane[0]*v[0].x+plane[1]*v[0].y+plane[2]*v[0].z);
      
      // add edges
      edges[0] = Line(v[0], v[1]);
      edges[1] = Line(v[1], v[2]);
      edges[2] = Line(v[2], v[0]);
    }
    
    // sort the three edges with ascending value at pos index
    void sortLineElements(LineElement le[], int index) const {
      if(le[0].l.p1[index] > le[1].l.p1[index])
	util::swap(le[0], le[1]);
      if(le[1].l.p1[index] > le[2].l.p1[index])
	util::swap(le[1], le[2]);
      if(le[0].l.p1[index] > le[1].l.p1[index])
	util::swap(le[0], le[1]);

      assert( le[0].l.p1[index] <= le[1].l.p1[index] );
      assert( le[1].l.p1[index] <= le[2].l.p1[index] );
    }
    
    virtual void draw(Image8 *im) {
      
      // rasterize
      if( filled ) {
	
	// find the largest coefficient
	int ind[3] = {0,1,2};
	float diff[3];
	
	diff[0] = fabs(plane[0]);
	diff[1] = fabs(plane[1]);
	diff[2] = fabs(plane[2]);
	
	// coord with smallest difference 
	// (largest coeff) will have index ind[2]
	util::bubblesort(diff, ind, 3);
	
	// create line elements, sort endpoints, and compute parameters
	LineElement edge_list[3];	
	for(int i=0; i<3; i++) {
	  
	  // create line elements
	  edge_list[i] = LineElement(edges[i]);
	  
	  // sort endpoints according to y-value
	  if( edge_list[i].l.p1[ind[1]] > edge_list[i].l.p2[ind[1]] ) {
	    util::swap(edge_list[i].l.p1, edge_list[i].l.p2);
	  }
	  
	  // compute parameters
	  
	  // DX/DY and DZ/DY
	  float dy = edge_list[i].l.p2[ind[1]]-edge_list[i].l.p1[ind[1]];
	  if( fabs(dy) > 1e-9 ) {
	    edge_list[i].DX =
	      float(edge_list[i].l.p2[ind[0]]-edge_list[i].l.p1[ind[0]])/dy;
	    edge_list[i].DZ =
	      float(edge_list[i].l.p2[ind[2]]-edge_list[i].l.p1[ind[2]])/dy;
	  }
	  else {
	    edge_list[i].DX = 0.0f;
	    edge_list[i].DZ = 0.0f;
	  }
	  
	  // current X and Z
	  edge_list[i].X = edge_list[i].l.p1[ind[0]];
	  edge_list[i].Z = edge_list[i].l.p1[ind[2]];
	}
	
	// sort edge_list with ascending p1.y (p1[ind[1]])
	sortLineElements(edge_list, ind[1]);
	
	// compute min-max in y
	float y0=1e9f, yf=-1e9f;
	for(unsigned int i=0; i<3; i++) {
	  if( edges[i].p1[ind[1]] < y0 )
	    y0 = edges[i].p1[ind[1]];
	  if( edges[i].p1[ind[1]] > yf )
	    yf = edges[i].p1[ind[1]];
	  
	  if( edges[i].p2[ind[1]] < y0 )
	    y0 = edges[i].p2[ind[1]];
	  if( edges[i].p2[ind[1]] > yf )
	    yf = edges[i].p2[ind[1]];
	}
	
	// construct active list
        LineElement active_edge_list[2];
	active_edge_list[0] = edge_list[0];
	active_edge_list[1] = edge_list[1];
	
	// our position
	Vec3i pos;
	
	// main loop
	pos[ind[1]] = int(y0+1.0f);
	float yskip = pos[ind[1]]-y0;
	
	// step X and Z
	active_edge_list[0].X += yskip*active_edge_list[0].DX;
	active_edge_list[0].Z += yskip*active_edge_list[0].DZ;
	active_edge_list[1].X += yskip*active_edge_list[1].DX;
	active_edge_list[1].Z += yskip*active_edge_list[1].DZ;
	
	bool edge_change = false;
	for(; pos[ind[1]]<=int(yf); pos[ind[1]]++) {
	  
	  // resort
	  if(active_edge_list[0].X > active_edge_list[1].X) {
	    util::swap(active_edge_list[0], active_edge_list[1]); 
	  }	 
	  
	  // if the third edge becomes active we update active edge list
	  if( !edge_change &&
	      edge_list[2].l.p1[ind[1]] <= pos[ind[1]] &&
	      edge_list[2].l.p2[ind[1]] > pos[ind[1]] ) {
	    
	    edge_change = true;
	    yskip = pos[ind[1]]-edge_list[2].l.p1[ind[1]];
	    edge_list[2].X += yskip*edge_list[2].DX;
	    edge_list[2].Z += yskip*edge_list[2].DZ;
	    
	    // find out which edge to replace
	    if( active_edge_list[0].l.p2[ind[1]] <= pos[ind[1]] ) {
	      active_edge_list[0] = edge_list[2];
	    }
	    else {
	      active_edge_list[1] = edge_list[2];
	    }
	    
	    // resort
	    if(active_edge_list[0].X > active_edge_list[1].X) {
	      util::swap(active_edge_list[0], active_edge_list[1]); 
	    }
	    
	  }
	  
	  float z = active_edge_list[0].Z;
	  float dx = (active_edge_list[1].X-active_edge_list[0].X); 
	  float dzdx = 0.0f;
	  if( fabs(dx) > 1e-9 )
	    dzdx = (active_edge_list[1].Z-z)/dx;
	  
	  pos[ind[0]] = int(active_edge_list[0].X+1.0f);
	  float xskip = pos[ind[0]]-active_edge_list[0].X;
	  z += xskip*dzdx;
	  for(; pos[ind[0]]<=int(active_edge_list[1].X); pos[ind[0]]++) {
	    
	    pos[ind[2]] = int( z+0.5f );		

	    im->at( pos ) = color;
	    
	    z += dzdx;
	    
	  }
	  
	  // increment X and Z
	  active_edge_list[0].X += active_edge_list[0].DX;
	  active_edge_list[1].X += active_edge_list[1].DX;
	  
	  active_edge_list[0].Z += active_edge_list[0].DZ;
	  active_edge_list[1].Z += active_edge_list[1].DZ;
	  
	}
	
      } // if filled
      
      // draw edges
      edges[0].draw(im);
      edges[1].draw(im);
      edges[2].draw(im);
      
    }
 
    
  };
  
}
