///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishRawFileReader.hh"

namespace wish {
  
  bool RawFileReader::readHeader() {
    int fs = util::fileSize( this->m_filename );
    int nvoxels = this->m_info.dim.x*this->m_info.dim.y*this->m_info.dim.z;
    int ds = 
      nvoxels*ImageObject::voxelBytesFromString( this->m_info.datatype );
    if( (fs-this->m_info.headersize) != ds ) {
      cerr << "Filesize does not match specified dimensions" << endl;
      return false;
    }
    return true;
  }
  
  bool RawFileReader::readData() {
    
    // open file, use binary mode (necessary on windows)
    ifstream is;
    is.open( this->m_filename.c_str(), ios::in | ios::binary );
    if( !is ) {
      cerr << "Unable to open file: " << this->m_filename << endl;
      return false;
    }
    
    // jump to data
    is.seekg(this->m_info.headersize);
    
    // use header info to allocate a new image
    if( this->m_info.datatype == "uint8" ) {
      Image<uint8> *tmp = new Image<uint8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int8" ) {
      Image<int8> *tmp = new Image<int8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint16" ) {
      Image<uint16> *tmp = new Image<uint16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int16" ) {
      Image<int16> *tmp = new Image<int16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint32" ) {
      Image<uint32> *tmp = new Image<uint32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int32" ) {
      Image<int32> *tmp = new Image<int32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "float" ) {
      Image<float> *tmp = new Image<float>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "double" ) {
      Image<double> *tmp = new Image<double>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "Vec3f" ) {
      Image<Vec3f> *tmp = new Image<Vec3f>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "RGB8" ) {
      Image<RGB8> *tmp = new Image<RGB8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "RGBA8" ) {
      Image<RGBA8> *tmp = new Image<RGBA8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_little_endian )
	util::readLE( is, tmp->getData(), tmp->totalSize() );
      else
	util::readBE( is, tmp->getData(), tmp->totalSize() );
      this->m_output = tmp;
    }
    else { // should not happen
      this->printSelf();
      cerr << ": Error reading file " << m_filename << ", bad datatype: " 
	   << this->m_info.datatype << endl;
      return false;
    }
    
    is.close();
    
    return true;
  }
  
}
