///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_RESAMPLE_FILTER_HH_
#define _WISH_RESAMPLE_FILTER_HH_

#include "wishFilterObject.hh"

namespace wish {
  
  //! ResampleFilter implements a simple filter for resampling
  //! a 3D image with different factors in each dimension.
  //! You can choose between downsampling and upsampling.
  //! This filter replaces the old ResampleXYFilter
  //!  
  //!   Input dimensions:  w,h,d
  //!   Factors:           2 2 1
  //!   Output dimensions: w/2, h/2, d
  //!   
  //!   For each point in the output we here use the average 
  //!   of a 2x2 neighborhood.
  //!
  //! OBS! IMPLEMENTED FOR INTEGERS 
  //! ( out = OutputImageVoxelType( double(...) + 0.5 ) ) 
  template<class InputImage, class OutputImage>
  class ResampleFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< ResampleFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    ResampleFilter() : 
      m_xfactor(2),
      m_yfactor(2),
      m_zfactor(2),
      m_downsample(true)
    {}
    
    //! set sigmas
    void setFactors(int _xfactor, int _yfactor, int _zfactor=1) {
      m_xfactor = _xfactor;
      m_yfactor = _yfactor;
      m_zfactor = _zfactor;  
    }
    
    //! set downsample/upsample
    void downsample() {
      m_downsample = true;
    }
    
    void upsample() {
      m_downsample = false;
    }
    
  protected:
    
    int m_xfactor;
    int m_yfactor;
    int m_zfactor;
    bool m_downsample;
    
    virtual void process();
    
  private:
    // purposely not implemented
    ResampleFilter(const ResampleFilter&);
    ResampleFilter& operator=(const ResampleFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void ResampleFilter<InputImage, OutputImage>::process() {

    // reference images
    InputImage  &in  = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    // dimensions
    int W = in.width();
    int H = in.height();
    int D = in.depth();
    Vec3f spc = in.getSpacing();
    
    // output image
    if( m_downsample ) {
      out.allocate(W/m_xfactor, H/m_yfactor, D/m_zfactor);
      out.setSpacing(m_xfactor*spc.x, m_yfactor*spc.y, m_zfactor*spc.z);
      out.reset( OutputImageVoxelType(0) );    
      
      // scale factor
      double sc = 1.0/(m_xfactor*m_yfactor*m_zfactor);
      
      for(int z=0; z<D; z+=m_zfactor) {  
	for(int y=0; y<H; y+=m_yfactor) {
	  for(int x=0; x<W; x+=m_xfactor) {
	    
	    double tmp=0.0;
	    for(int k=0; k<m_zfactor; k++)
	      for(int j=0; j<m_yfactor; j++)
		for(int i=0; i<m_xfactor; i++)
		  tmp += double( in(x+i, y+j, z+k) );
	    tmp *= sc;
	    out(x/m_xfactor, y/m_yfactor, z/m_zfactor) = 
	      OutputImageVoxelType(tmp+0.5);
	  }
	}
      }
    }
    else {
      out.allocate(W*m_xfactor, H*m_yfactor, D*m_zfactor);
      out.setSpacing((1.0/m_xfactor)*spc.x, 
		     (1.0/m_yfactor)*spc.y, 
		     (1.0/m_zfactor)*spc.z);
      out.reset( OutputImageVoxelType(0) );
      
      for(int z=0; z<D; z++) {  
	for(int y=0; y<H; y++) {
	  for(int x=0; x<W; x++) {
	    InputImageVoxelType val = in(x,y,z);
	    for(int k=0; k<m_zfactor; k++)
	      for(int j=0; j<m_yfactor; j++)
		for(int i=0; i<m_xfactor; i++)
		  out(m_xfactor*x+i, 
		      m_yfactor*y+j, 
		      m_zfactor*z+k) = OutputImageVoxelType(val);
	  }
	}
      }
    }
    
  }
  
} // end namespace wish

#endif
