///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishShaderUtilities.hh"

namespace wish {
  
  // load a shader program from file
  // parameter type is GL_VERTEX_SHADER_ARB or GL_FRAGMENT_SHADER_ARB
  GLhandleARB buildShaderFromFile(const string &filename, 
				  GLenum type) {
    
    string typestr="unknown";
    
    // check shader type
    if( type == GL_VERTEX_SHADER_ARB )
      typestr = "vertex";
    else if( type == GL_FRAGMENT_SHADER_ARB )
      typestr = "fragment";
    
    cerr << "Reading " << typestr << " shader from file: " << filename << endl;
    
    // shader handle
    GLhandleARB shader_handle = 0;
  
    // read shader file
    ifstream is( filename.c_str(), ios::binary );
    
    if( is.good() ) {
      
      shader_handle = glCreateShaderObjectARB(type);
 
      int length;
      char *buffer;
      
      // get length of file:
      is.seekg (0, ios::end);
      length = is.tellg();
      is.seekg (0, ios::beg);
    
      // allocate memory:
      buffer = new char[length+1];
    
      // read data as a block:
      is.read (buffer,length);
      is.close();
      buffer[length] = '\0';
    
      // upload shader and compile it
      glShaderSourceARB( shader_handle, 1, (const GLcharARB **) &buffer, 0 );
      glCompileShaderARB( shader_handle );
    
      delete [] buffer; // clean up
    }
    else {
      cerr << "Could not read " << typestr << " shader" << endl;
      return 0;
    }
  
    // check compilation status
    GLint compile_success;
    glGetObjectParameterivARB( shader_handle,
			       GL_OBJECT_COMPILE_STATUS_ARB,
			       &compile_success );
    if( compile_success == GL_FALSE ) {
      GLint nr_characters;
      glGetObjectParameterivARB( shader_handle,
				 GL_OBJECT_INFO_LOG_LENGTH_ARB,
				 &nr_characters );
      GLcharARB *log = new GLcharARB[nr_characters];
      glGetInfoLogARB( shader_handle,
		       nr_characters,
		       NULL,
		       log );
      cerr << "Error when compiling " << typestr << " shader:" << endl;
      cerr << log << endl;
    
      delete [] log; // clean up
    
      glDeleteObjectARB( shader_handle );
      shader_handle = 0;
    }
    
    return shader_handle;
    
  }
  
  // create shader program
  GLhandleARB createShaderProgram(const string &vs, 
				  const string &fs) {
    
    // create program handle
    GLhandleARB program_handle = 0;
    
    //create vertex shader
    GLhandleARB vs_handle = buildShaderFromFile(vs, 
						GL_VERTEX_SHADER_ARB);
    if( vs_handle == 0 ) {
      return 0;
    } 
    
    //create fragment shader
    GLhandleARB fs_handle = buildShaderFromFile(fs, 
						GL_FRAGMENT_SHADER_ARB);
    if( fs_handle == 0 ) {
      return 0;
    } 
    
    //create shader object and attach
    program_handle = glCreateProgramObjectARB();
    glAttachObjectARB(program_handle, vs_handle); glDeleteObjectARB(vs_handle);
    glAttachObjectARB(program_handle, fs_handle); glDeleteObjectARB(fs_handle);
    
    //link program
    glLinkProgramARB(program_handle);
    
    // check linking
    GLint link_success;
    glGetObjectParameterivARB(program_handle, 
			      GL_OBJECT_LINK_STATUS_ARB, 
			      &link_success);
    if( link_success == GL_FALSE ) {
      GLint nr_characters;
      glGetObjectParameterivARB( program_handle,
				 GL_OBJECT_INFO_LOG_LENGTH_ARB,
				 &nr_characters );
      GLcharARB *log = new GLcharARB[nr_characters];
      glGetInfoLogARB( program_handle,
		       nr_characters,
		       NULL,
		       log );
      cerr << "Error when linking shader program:" << endl;
      cerr << log << endl;
      
      delete [] log; // clean up
      
      glDeleteObjectARB( program_handle );
      program_handle = 0;
    }
    
    return program_handle;
  
  }
  
}
