///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_SIGMOID_FILTER_HH_
#define _WISH_SIGMOID_FILTER_HH_

#include "wishFilterObject.hh"

namespace wish {
  
  //! SigmoidFilter
  //! Maps input values to the interval [minval,maxval] through the use
  //! of a sigmoid function. Can be used to create a cost image
  //! from a gradient magnitude image. The output is computed as:
  //! <tt>out(x) = (max-min)*1.0/(1+exp(-(in(x)-beta)/alpha))+min</tt>.
  //!
  //! If we want the value lambda*max (lambda in [0,1]) at |x-beta| = delta
  //! we can compute alpha = -delta/ln( (1-lambda)/lambda ).
  //! To reverse the function we can just swap min and max 
  template<class InputImage, class OutputImage>
  class SigmoidFilter : 
    public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< SigmoidFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    SigmoidFilter() : 
      m_alpha(1.0),
      m_beta(0.0),
      m_max(1.0),
      m_min(0.0)
    {}
    
    //! set alpha
    void setAlpha(double _alpha) {
      m_alpha = _alpha;
    }

    //! set beta
    void setBeta(double _beta) {
      m_beta = _beta;
    }

    //! set max
    void setMax(double _max) {
      m_max = _max;
    }

    //! set min
    void setMin(double _min) {
      m_min = _min;
    }
    
    // normalized versions, i.e., input is in [0,1]
    
    //! set alpha
    void setNormalizedAlpha(double _alpha) {
      m_alpha = _alpha*(double(InputImage::maxValue())-
			double(InputImage::minValue()));
    }
    
    //! set beta
    void setNormalizedBeta(double _beta) {
      m_beta = 
	double(InputImage::minValue()) +
	_beta*(double(InputImage::maxValue())-
	       double(InputImage::minValue()));
    }
    
    //! set max
    void setNormalizedMax(double _max) {      
      m_max = 
	double(InputImage::minValue()) +
	_max*(double(InputImage::maxValue())-
	      double(InputImage::minValue()));
    }
    
    //! set min
    void setNormalizedMin(double _min) {      
      m_min = 
	double(InputImage::minValue()) +
	_min*(double(InputImage::maxValue())-
	      double(InputImage::minValue()));
    }

  protected:
    
    double m_alpha, m_beta, m_max, m_min;
    double fsig(double arg);
    virtual void process();
    
  private:
    // purposely not implemented
    SigmoidFilter(const SigmoidFilter&);
    SigmoidFilter& operator=(const SigmoidFilter&);
  };
  
  template<class InputImage, class OutputImage>
  double SigmoidFilter<InputImage, OutputImage>::fsig(double arg) {
    return 
      double(this->m_max-this->m_min) /
      (1.0+exp(-(arg-this->m_beta)/this->m_alpha))+this->m_min;
  }
  
  template<class InputImage, class OutputImage>
  void SigmoidFilter<InputImage, OutputImage>::process() {
    
    this->m_output->resize( *(this->m_input) );
    this->m_output->reset( OutputImageVoxelType(0) );    
    
    InputImage &in = *(this->m_input);
    OutputImage &out = *(this->m_output);
    
    out.setLabel( in.getLabel() + "_SIG" );
   
    for(int i=0; i<this->m_output->totalSize(); i++) 
      out[i] = OutputImageVoxelType( this->fsig(double(in[i])) );
  }
  
} // end namespace wish

#endif
