///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include "wishSimplexMesh.hh"
#include "wishRasterization.hh"
#include "wishLabelingFilter.hh"

namespace wish {
    
  /// set spatial weight
  void SimplexMesh::setSpatialWeight(const double &_alpha) {
    SimplexVertexIterator it = vbegin();
    while( it != vend() ) {
      (*it)->setSpatialWeight(_alpha);
      it++;
    }
  }
  
  /// set external weight
  void SimplexMesh::setExternalWeight(const double &_beta) {
    SimplexVertexIterator it = vbegin();
    while( it != vend() ) {
      (*it)->setExternalWeight(_beta);
      it++;
    }
  }
  
  // delete mesh (called in the destructor)
  void SimplexMesh::deleteMesh() {
    while(!faces.empty()) {
      delete faces.back();
      faces.pop_back();
    }
    while(!edges.empty()) {
      delete edges.back();
      edges.pop_back();
    }
    while(!vertices.empty()) {
      delete vertices.back();
      vertices.pop_back();
    }
    
    iterations = -1;
  }
  
  /// refine mesh
  void SimplexMesh::refine() {
    TriangleMesh tm;
    tm.createFromSimplexMesh(this);
    tm.refine();
    this->createFromTriangleMesh(&tm);
    cerr << "Refined SimplexMesh has " 
	 << this->nVertices() << " vertices, " 
	 << this->nEdges() << " edges, and " 
	 << this->nFaces() << " faces." << endl;
  }
  
  bool SimplexMesh::readMesh( const string &filename ) {
    ifstream is(filename.c_str(), ifstream::in | ifstream::binary);
    if( !is ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Reading " << filename << endl;
    
    // read vertices
    int nvert;
    string tmp="";
    while( tmp != "#VERTICES" && is ) {
      is >> tmp;
    }
    
    if( !is ) {
      cerr << "Failed to read mesh from file: " << filename << endl;
      return false;
    }
    
    // number of vertices
    is >> nvert;
    
    vector<SimplexVertexPointer> v(nvert);
    Vec3d pos;
    for(int i=0; i<nvert; i++) {
      // read vertex position
      is >> pos;
      
      // create vertex
      v[i] = this->newVertex(pos);
    }
    
    // read edges
    int nedges;
    while( tmp != "#EDGES" && is ) {
      is >> tmp;
    }
    
    if( !is ) {
      cerr << "Failed to read mesh from file: " << filename << endl;
      while(!v.empty()) {
	delete v.back();
	v.pop_back();
      }
      return false;
    }
    
    // number of edges
    is >> nedges;
    
    vector<SimplexEdgePointer> e(nedges);
    for(int i=0; i<nedges; i++) {
      // read vertex indices
      int v0, v1;
      is >> v0 >> v1;

      // create edge from vertices
      e[i] = this->newEdge(v[v0], v[v1]);
    }
    
    // read faces
    int nfaces;
    while( tmp != "#FACES" && is ) {
      is >> tmp;
    }
    
    if( !is ) {
      cerr << "Failed to read mesh from file: " << filename << endl;
      while(!v.empty()) {
	delete v.back();
	v.pop_back();
      }
      while(!e.empty()) {
	delete e.back();
	e.pop_back();
      }
      return false;
    }
    
    // number of faces
    is >> nfaces;
    
    vector<SimplexFacePointer> f(nfaces);
    for(int i=0; i<nfaces; i++) {
      vector<SimplexEdgePointer> etmp(0);
      // read edge indices
      do {
	int eind;
	is >> eind;
	etmp.push_back(e[eind]);
      } while( is.peek() != '\n' );
      
      // create face from edges
      f[i] = this->newFace(etmp);
      
      // for each edge in the new face, add the face
      for(int j=0; j<(int)etmp.size(); j++) {
	etmp[j]->addFace(f[i]);
      }
    }
    
    // ...and we're finished
    is.close();
    
    // delete current mesh and add the new
    this->deleteMesh();
    this->setVertices(v);
    this->setEdges(e);
    this->setFaces(f);

    //this->sortEdges(); // no need, only 3 edges in a simplex
    this->sortNeighbors();
    this->computeFaceCenters();
    this->computeNormals();
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  // simple write function
  bool SimplexMesh::writeMesh( const string &filename ) const {
    ofstream os(filename.c_str(), ofstream::out | ofstream::binary);
    if( !os ) {
      cerr << "Unable to open file: " << filename << endl;
      return false;
    }
    
    cerr << "Writing " << filename << endl;
    
    // write vertices
    os << "#VERTICES " << nVertices() << endl;
    ConstSimplexVertexIterator vit = vbegin();
    int i=0;
    while( vit != vend() ) {
      (*vit)->flag = i++;
      os << (*vit)->getPosition() << endl;
      vit++;
    }
    
    // write edges
    os << "#EDGES " << nEdges() << endl;
    ConstSimplexEdgeIterator eit = ebegin();
    i=0;
    while( eit != eend() ) {
      (*eit)->flag = i++;
      os << (*eit)->getFirstVertex()->flag << " "
	 << (*eit)->getSecondVertex()->flag << endl;
      eit++;
    }
    
    // write faces
    os << "#FACES " << nFaces() << endl;
    ConstSimplexFaceIterator fit = fbegin();
    while( fit != fend() ) {
      i=0;
      for(; i<(*fit)->nEdges()-1; i++) {
	os << (*fit)->getEdge(i)->flag << " ";
      }
      os << (*fit)->getEdge(i)->flag << endl;
      fit++;
    }
    
    os.close();
    
    cerr << "Finished!" << endl;
    
    return true;
  }
  
  // create simplex mesh from triangle mesh
  // ONLY WORKS FOR CLOSED SURFACES SO FAR
  void SimplexMesh::createFromTriangleMesh(TriangleMeshPointer tmesh) {
    
    // clear
    this->deleteMesh();
    
    // for each triangle face we create a simplex vertex
    vector<SimplexVertexPointer> v(tmesh->nFaces());
    Vec3d pos;
    for(int i=0; i<tmesh->nFaces(); i++) {
      
      // the current triangle face
      TriangleFacePointer tf = tmesh->getFace(i);
      tf->flag = i; // store index in triangle, needed for edge creation
      
      // compute simplex vertex position (centroid of triangle)
      // here we could use getCenter() !?
      TriangleFace::VertexIterator vi = tf->vbegin();
      pos = Vec3d(0,0,0);
      do {
	pos = pos+vi.current()->getPosition(); 
      } while( vi.next() );
      pos /= double(tf->nVertices());
      
      // create vertex and add it to this mesh
      v[i] = this->newVertex(pos);
      this->addVertex(v[i]);
    }
    
    // for each triangle edge we create a simplex edge
    vector<SimplexEdgePointer> e(tmesh->nEdges());
    for(int i=0; i<tmesh->nEdges(); i++) {
      
      // the current triangle edge 
      TriangleEdgePointer te = tmesh->getEdge(i);
      te->flag = i; // store index in edge, needed for face creation

      // create edge with vertex indices fetched from the faces
      e[i] = this->newEdge(v[te->getFirstFace()->flag],
			   v[te->getSecondFace()->flag]);

      // add it
      this->addEdge(e[i]);
    }
    
    // for each triangle vertex we create a simplex face
    vector<SimplexFacePointer> f(tmesh->nVertices());
    for(int i=0; i<tmesh->nVertices(); i++) {
      
      // the current triangle vertex
      TriangleVertexPointer tv = tmesh->getVertex(i);
      
      // get the edges from the triangle vertex and collect
      // the corresponding simplex edges given by the indices stored
      // in the triangle edges
      // this should work since the edges are ordered with sortEdges() 
      
      vector<SimplexEdgePointer> etmp(tv->nEdges());
      for(int j=0; j<tv->nEdges(); j++) {
	etmp[j] = e[tv->getEdge(j)->flag];
      }
      
      // create and add the new face
      f[i] = this->newFace(etmp);
       
      // check orientation
      TriangleEdgePointer e1 = tv->getEdge(0);
      TriangleEdgePointer e2 = tv->getEdge(1);
      TriangleFacePointer t = (e1->getFirstFace() == e2->getFirstFace() ||
			       e1->getFirstFace() == e2->getSecondFace())
	? e1->getFirstFace() : e1->getSecondFace();
      bool invert = false;
      int kk=0;
      TriangleEdgePointer tep = t->getEdge(kk);
      if(tep != e1 && tep != e2) 
	tep = t->getEdge(++kk);
      if(tep == e2) {
	tep = t->getEdge(++kk);
	if(tep != e1) 
	  invert = true;
      }
      else {
	tep = t->getEdge(++kk);
	if(tep == e2) invert = true;
      }
      if(invert) {
	int nn = tv->nEdges();
	vector<SimplexEdgePointer> etmp2(nn);
	for(int j=0; j<nn; j++) {
	  etmp2[j] = etmp[nn-j-1];
	}
	f[i]->reset(etmp2);
      }
      
      this->addFace(f[i]);
      
      // for each edge in the new face, add the face
      for(int j=0; j<tv->nEdges(); j++) {
	etmp[j]->addFace(f[i]);
      }
    }
    // ...and we're finished
    
    this->sortNeighbors();
    this->computeFaceCenters();
    this->computeNormals();
  }
  
  /// create a sphere
  void SimplexMesh::createSphere(double radius, int refinements) {
    TriangleMesh tm;
    tm.createSphere(radius, refinements);
    this->createFromTriangleMesh(&tm);
  }
  
  /// set selected flag for all vertices
  void SimplexMesh::selectVertices(const bool &s) {
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++) {
      (*vi)->setSelected(s);
    }
  }
  
  void SimplexMesh::sortNeighbors() {
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++) {
      (*vi)->sortNeighbors();
    }
  }
  
  /// set selected flag for all faces
  void SimplexMesh::selectFaces(const bool &s) {
    for(SimplexFaceIterator fi=fbegin(); fi!=fend(); fi++) {
      (*fi)->setSelected(s);
    }
  }
  
  /// compute centers for all faces
  void SimplexMesh::computeFaceCenters() {
    for(SimplexFaceIterator fi=fbegin(); fi!=fend(); fi++) {
      (*fi)->computeCenter();
    }
  }
  
  /// compute area for all faces
  void SimplexMesh::computeFaceArea() {
    for(SimplexFaceIterator fi=fbegin(); fi!=fend(); fi++) {
      (*fi)->computeArea();
    }
  }
  
  /// compute normals for all faces and vertices
  void SimplexMesh::computeNormals() {
    // vertex normals
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++) {
      (*vi)->computeNormal();
    } 
    // face normals
    for(SimplexFaceIterator fi=fbegin(); fi!=fend(); fi++) {
      (*fi)->computeNormal();
    } 
  }
  
  /// deform engine
  /// should be rewritten to loop over faces, add a tag to each vertex
  void SimplexMesh::deform() {
    
    // compute angle
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
      (*vi)->computeSimplexAngle();
    
    // compute spatial force
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
      //(*vi)->computeC1Force();
      (*vi)->computeC2Force();
    
    // compute external forces
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++) {
      Vec3d ef(0,0,0);    
      for(ExternalForceIterator efi=efbegin(); efi!=efend(); efi++) {
	ef += (*efi)->computeForce((*vi)->getPosition(), 
				   (*vi)->getNormal(),
				   (*vi)->getSelected());
      }
      (*vi)->setExternalForce(ef);
    }

    // curvature/area adaption
    if( iterations>0 && (iterations%10) == 0 ) {
      // compute nb mean curvature
      for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++) {
	(*vi)->computeNeighborMeanCurvature();
      }
      
      // compute face area
      computeFaceArea();
      
      for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
	(*vi)->curvatureAdaption(0.1);
    }
    
    // move vertex
    if( iterations<=0 ) {
      for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
	(*vi)->setPosition( (*vi)->getPosition() );
    } else {
      double gamma = 0.8; // 0.35
      for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
	(*vi)->move(gamma);
    }
    
    computeNormals();
    computeFaceCenters();
    
    // counter for deformation iterations
    iterations++;
  }
  
  // compute center of mass
  Vec3d SimplexMesh::computeCenterOfMass() const {
    Vec3d cm(0,0,0);
    for(ConstSimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
      cm += (*vi)->getPosition();
    cm /= double(this->nVertices());
    return cm;
  } 
  
  // the volume is computed by adding the volumes of the tetrahedrons  
  // composed from the face center and two neighboring vertices  
  double SimplexMesh::computeVolume() const {    
    // not implemented yet
    return 0.0; 
  }
  
  // compute area
  double SimplexMesh::computeArea() const {
    double area=0.0;
    for(int i=0; i<this->nFaces(); i++) {
      area += this->getFace(i)->computeArea();
    }
    return area;
  }
  
  /// transformation
  void SimplexMesh::transform(const Matrix4d &M) {
    for(SimplexVertexIterator vi=vbegin(); vi!=vend(); vi++)
      (*vi)->transform(M);
  }
  
  // rasterize mesh to image
  void SimplexMesh::rasterizeMesh(wish::Image8* im, bool fill) const {
    if( im ) {
      cerr << "Rasterizing mesh" << endl;
      im->reset(0);
      Matrix4d Mw2i = Matrix4d( im->getWorldToImageMatrix() );
      
      for(int i=0; i<this->nFaces(); i++) {
	SimplexFacePointer f = this->getFace(i);
	SimplexFace::VertexIterator vi = f->vbegin();
	Vec3d p0 = Mw2i*(f->computeCenter());
	if( !(im->isInside(p0)) ) {
	  cerr << "SimplexMesh::rasterizeMesh - vertex outside image" << endl;
	  return;
	}
	vector<Vec3d> pvec(0);
	do {
	  Vec3d p = Mw2i*(vi.current()->getPosition());
	  if( !(im->isInside(p)) ) {
	    cerr << "SimplexMesh::rasterizeMesh - vertex outside image" 
		 << endl;
	    return;
	  }
	  pvec.push_back( p );
	} while( vi.next() );
	int j=0;
	for(; j<(int)pvec.size()-1; j++) {
	  wish::Triangle t(p0, pvec[j], pvec[j+1]);
	  t.draw(im);
	}
	wish::Triangle t(p0, pvec[j], pvec[0]);
	t.draw(im);
      }
      
      if( fill ) { 
	// fill holes algorithm
	
	// complement
	wish::Image8 imtmp;
	imtmp.resize(*im);
	imtmp = im->NOT();
	
	// label
	wish::LabelingFilter<wish::Image8,wish::Image16> lf;
	lf.setConnectivity(6);
	lf.setInput(&imtmp);
	lf.execute();
	
	Image16 &lbl = *(lf.getOutput());
	
	// add the filled hole
	for(int k=0; k<im->totalSize(); k++) {
	  if( lbl[k] > 1 )
	    (*im)[k] = 255;
	}
	
	// end fill holes
      }
      
      cerr << "Finished!" << endl;
    }
    
  }
  
  // render with OpenGL
  void SimplexMesh::renderGL(int mode) const {
    
    glPushAttrib( GL_ALL_ATTRIB_BITS );
        
    // wireframe rendering
    if( mode == 0 ) { 
      glEnable(GL_COLOR_MATERIAL);
            
      glEnable(GL_LINE_SMOOTH);
      glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
      glLineWidth(2.0);
      
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
      glBegin(GL_LINES);	
      for(int i=0; i<this->nEdges(); i++) {
	SimplexEdgePointer e = this->getEdge(i);
	
	// render selected
	if( e->getFirstVertex()->getSelected() ) {
	  glColor3f(1,1,0);
	}
	else {
	  glColor3f(0,1,0);
	}
	glNormal3dv(e->getFirstVertex()->getNormal());
	glVertex3dv(1.001*e->getFirstVertex()->getPosition());
	
	// temporary for checking iterators 
	if( e->getSecondVertex()->getSelected() ) {
	  glColor3f(1,1,0);
	}
	else {
	  glColor3f(0,1,0);
	}
	glNormal3dv(e->getSecondVertex()->getNormal());
	glVertex3dv(1.001*e->getSecondVertex()->getPosition());
	
      }
      glEnd();
      
    }
    
    // flat rendering
    if( mode == 1 ) {
      glEnable(GL_COLOR_MATERIAL); // TEMPORARY
      
      for(int i=0; i<this->nFaces(); i++) {
	// render polygon	
	SimplexFacePointer f = this->getFace(i);
	// render selected
	if( f->getSelected() ) {
	  glColor3f(1,1,0);
	}
	else {
	  glColor3f(0,1,0);
	}
	SimplexFace::VertexIterator vi = f->vbegin();
	SimplexVertexPointer first = vi.current();	
	glBegin(GL_TRIANGLE_FAN); // more efficient than polygon
	glNormal3dv(f->getNormal());
	glVertex3dv(f->getCenter());
	do {
	  glVertex3dv(vi.current()->getPosition());	  
	} while( vi.next() );
	glVertex3dv(first->getPosition());
	glEnd();
      }

    }
    
    // smooth rendering
    if( mode == 2 ) {
      glEnable(GL_COLOR_MATERIAL); // TEMPORARY

      for(int i=0; i<this->nFaces(); i++) {
	// render polygon	
	SimplexFacePointer f = this->getFace(i);
	// render selected
	if( f->getSelected() ) {
	  glColor3f(1,1,0);
	}
	else {
	  glColor3f(0,1,0);
	}
	SimplexFace::VertexIterator vi = f->vbegin();
	SimplexVertexPointer first = vi.current();
	glBegin(GL_TRIANGLE_FAN); // more efficient than polygon
	glNormal3dv(f->getNormal());
	glVertex3dv(f->getCenter());
	do {
	  glNormal3dv(vi.current()->getNormal());	
	  glVertex3dv(vi.current()->getPosition());	  
	} while( vi.next() );
	glNormal3dv(first->getNormal());
	glVertex3dv(first->getPosition());
	glEnd();
      }
      
           
//       // DRAW WIRE ON TOP OF SMOOTH
//       glEnable(GL_LINE_SMOOTH);
//       glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
//       glLineWidth(2.0);
//       glEnable(GL_BLEND);
//       glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//       glPolygonMode(GL_FRONT, GL_LINE);
//       for(int i=0; i<this->nFaces(); i++) {
// 	// render polygon	
// 	SimplexFacePointer f = this->getFace(i);
// 	glColor3f(0,0,1);
// 	glLineWidth(1.0f);
// 	SimplexFace::VertexIterator vi = f->vbegin();
// 	glBegin(GL_POLYGON);
// 	do {
// 	  Vec3d n = vi.current()->getNormal();	
// 	  Vec3d p = vi.current()->getPosition();
// 	  glNormal3dv( n );
// 	  glVertex3dv( p+1.0f*n );
// 	} while( vi.next() );
// 	glEnd();
//       }
//       // END DRAW WIRE ON TOP OF SMOOTH
      
    }
    
    glPopAttrib();        
    
  }
  
}
