///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_SIMPLEX_MESH_HH_
#define _WISH_SIMPLEX_MESH_HH_

#include "wishSimplexVertex.hh"
#include "wishTriangleMesh.hh"
#include "wishExternalForce.hh"

namespace wish {
    
  // ### SimplexMesh ###
  
  class SimplexMesh {
  protected:
    
    /// vertices
    vector<SimplexVertexPointer> vertices;
    
    /// edges
    vector<SimplexEdgePointer> edges;
    
    /// faces
    vector<SimplexFacePointer> faces;
    
    /// external data
    vector<ExternalForcePointer> external_forces; 
 
    /// iteration counter
    int iterations;
        
  public:
    
    /// constructor
    SimplexMesh(int size_hint=0) :
      vertices( vector<SimplexVertexPointer>(0) ),
      edges( vector<SimplexEdgePointer>(0) ),
      faces( vector<SimplexFacePointer>(0) ),
      iterations(-1)
    {
      if( size_hint > 0 ) {
	vertices.reserve(size_hint);
	edges.reserve(size_hint);
	faces.reserve(size_hint);
      }
    }
    
    /// destructor
    virtual ~SimplexMesh() { deleteMesh(); }
    
    /// get iterations
    int getIterations() const {
      return iterations;
    }
    
    /// delete mesh, called in destructor
    void deleteMesh();
    
    /// refine mesh
    void refine();

    /// save mesh to file
    bool writeMesh( const string &filename ) const;
    
    /// read mesh from file
    bool readMesh( const string &filename );
    
    /// create simplex mesh from triangle mesh
    void createFromTriangleMesh(TriangleMeshPointer tmesh);
    
    /// create a sphere
    void createSphere(double radius, int refinements=1);
    
    /// number of vertices
    int nVertices() const { 
      return (int) vertices.size(); 
    }
    /// number of edges
    int nEdges() const { 
      return (int) edges.size(); 
    }
    /// returns number of faces
    int nFaces() const { 
      return (int) faces.size(); 
    }
    
    /// add a vertex
    void addVertex(SimplexVertexPointer v) { 
      vertices.push_back(v); 
    }
    /// add an edge
    void addEdge(SimplexEdgePointer e) { 
      edges.push_back(e); 
    }
    /// add a face
    void addFace(SimplexFacePointer f) { 
      faces.push_back(f); 
    }
    
    /// set vertices
    void setVertices(const vector<SimplexVertexPointer> &v) {
      vertices = v;
    }
    
    /// set edges
    void setEdges(const vector<SimplexEdgePointer> &e) {
      edges = e;
    }
    
    /// set faces
    void setFaces(const vector<SimplexFacePointer> &f) {
      faces = f;
    }
    
    /// set spatial weight
    void setSpatialWeight(const double &_alpha);
    
    /// set external weight
    void setExternalWeight(const double &_beta);
    
    /// get vertex
    SimplexVertexPointer getVertex(int i) const {
      if( 0 <= i && i < nVertices() )
	return vertices[i];
      else
	return 0;
    }
    
    /// get edge
    SimplexEdgePointer getEdge(int i) const {
      if( 0 <= i && i < nEdges() )
	return edges[i];
      else
	return 0;
    }
    
    /// get face
    SimplexFacePointer getFace(int i) const {
      if( 0 <= i && i < nFaces() )
	return faces[i];
      else
	return 0;
    }
    
    /// iterators
    typedef vector<SimplexVertexPointer>::iterator SimplexVertexIterator;
    typedef vector<SimplexVertexPointer>::const_iterator ConstSimplexVertexIterator;
    SimplexVertexIterator vbegin() {
      return vertices.begin();
    }
    ConstSimplexVertexIterator vbegin() const {
      return vertices.begin();
    }
    SimplexVertexIterator vend() {
      return vertices.end();
    }
    ConstSimplexVertexIterator vend() const {
      return vertices.end();
    }
    
    typedef vector<SimplexEdgePointer>::iterator SimplexEdgeIterator;
    typedef vector<SimplexEdgePointer>::const_iterator ConstSimplexEdgeIterator;
    SimplexEdgeIterator ebegin() {
      return edges.begin();
    }
    ConstSimplexEdgeIterator ebegin() const {
      return edges.begin();
    }
    SimplexEdgeIterator eend() {
      return edges.end();
    }
    ConstSimplexEdgeIterator eend() const {
      return edges.end();
    }
    
    typedef vector<SimplexFacePointer>::iterator SimplexFaceIterator;
    typedef vector<SimplexFacePointer>::const_iterator ConstSimplexFaceIterator;
    SimplexFaceIterator fbegin() {
      return faces.begin();
    }
    ConstSimplexFaceIterator fbegin() const {
      return faces.begin();
    }
    SimplexFaceIterator fend() {
      return faces.end();
    }
    ConstSimplexFaceIterator fend() const {
      return faces.end();
    }
    
    typedef vector<ExternalForcePointer>::iterator ExternalForceIterator;
    typedef vector<ExternalForcePointer>::const_iterator ConstExternalForceIterator;
    ExternalForceIterator efbegin() {
      return external_forces.begin();
    }
    ConstExternalForceIterator efbegin() const {
      return external_forces.begin();
    }
    ExternalForceIterator efend() {
      return external_forces.end();
    }
    ConstExternalForceIterator efend() const {
      return external_forces.end();
    }
    
    void sortNeighbors();
    
    /// set selected flag for all vertices
    void selectVertices(const bool &s);
    
    /// set selected flag for all faces
    void selectFaces(const bool &s);
    
    /// compute centers for all faces
    void computeFaceCenters();

    /// compute area for all faces
    void computeFaceArea();
    
    /// compute normals for all faces and vertices
    void computeNormals();
    
    /// update geometry when model was transformed
    void updateGeometry() { 
      computeFaceCenters(); 
      computeNormals();
    }
    
    /// deform engine
    void deform();
    
    /// add external data (have to decide how to delete)
    void addExternalForce(ExternalForcePointer ef) {
      external_forces.push_back(ef);
    }
    
    /// compute centerOfMass (old center)
    Vec3d computeCenterOfMass() const;
    
    /// compute enclosed volume (old volume)
    double computeVolume() const;
    
    /// compute surface area (old area)
    double computeArea() const;
    
    /// transformation
    void transform(const Matrix4d &M);
    
    /// rasterize mesh
    void rasterizeMesh(wish::Image8* im, bool fill=true) const;
    
    /// render mesh with OpenGL
    void renderGL(int mode=1) const;
    
  protected:
    /// internal vertex generators that generates a new vertex
    /// belonging to this mesh
    SimplexVertexPointer newVertex(double _x=0.0, double _y=0.0, double _z=0.0) {
      return new SimplexVertex(Vec3d(_x, _y, _z), this);
    }
    SimplexVertexPointer newVertex(const Vec3d& _position) {
      return new SimplexVertex(_position, this);
    }
    
    // internal edge generator
    SimplexEdgePointer newEdge(SimplexVertexPointer v1, 
			       SimplexVertexPointer v2,
			       SimplexFacePointer f1 = 0,
			       SimplexFacePointer f2 = 0) {
      return new SimplexEdge(v1, v2, f1, f2);
    }
    
    /// internal face generator
    SimplexFacePointer newFace(vector<SimplexEdgePointer> _edges,
			       bool _empty=false) {
      return new SimplexFace(_edges, _empty);
    }
  };
}

#endif
