///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishSimplexVertex.hh"

namespace wish {
  
  // ### SimplexVertex ###
  
  /// compute normal
  Vec3d SimplexVertex::computeNormal() {
    Vec3d p01 = (neighbors[1]->position-neighbors[0]->position);
    Vec3d p02 = (neighbors[2]->position-neighbors[0]->position);
    this->normal = p01.cross(p02);
    this->normal.normalize();
    return this->normal;
  }
  
  /// add edge
  void SimplexVertex::addEdge(SimplexEdgePointer _edge) {
    
    if( edges[0] == 0 ) 
      edges[0] = _edge;
    else if( edges[1] == 0 )
      edges[1] = _edge;
    else if( edges[2] == 0 ) 
      edges[2] = _edge;
    else
      cerr << "SimplexVertex::addEdge: Error, all edges already added!" 
	   << endl;
    
    // add neighbors
    // we do not select ourself as a neighbor
    SimplexVertexPointer v = _edge->getVertex(0);
    if( v == this )
      v = _edge->getVertex(1);
    
    if( neighbors[0] == 0 ) {
      neighbors[0] = v;
    }
    else if( neighbors[1] == 0 ) { 
      neighbors[1] = v;
    }
    else if( neighbors[2] == 0 ) {
      neighbors[2] = v;
      
    }
    else {
      cerr << "SimplexVertex::addEdge: Error, all neighbors already added!" 
	   << endl;
    }
    
  }
  
  // sort neighbors
  // sort neighbors so that normal is properly oriented
  void SimplexVertex::sortNeighbors() {
    FaceIterator fit = fbegin();
    Vec3d n(0,0,0);
    do {
      n += fit.current()->computeNormal();
    } while( fit.next() );
    n.normalize();
    if( this->computeNormal().dot(n) < 0.0 )
      util::swap(neighbors[1],neighbors[2]);
  }
  
  void SimplexVertex::move(const double &damping) {
    Vec3d newpos = 
      position + 
      (1.0-damping)*(position-pposition)+alpha*fspatial+beta*fexternal; 
    setPosition( newpos );
  }
  
  /// compute the simplex angle (inefficient)
  /// This should really be performed with inversion as suggested in
  /// Delingette, H.: Simplex Meshes: a General Representation for 3D
  /// Shape Reconstruction. Research report 2214, INRIA, 1994.
  void SimplexVertex::computeSimplexAngle() {
    
    // the neighbors
    Vec3d P = getPosition();
    Vec3d N0 = neighbors[0]->getPosition();
    Vec3d N1 = neighbors[1]->getPosition();
    Vec3d N2 = neighbors[2]->getPosition();
    Vec3d N01 = N1-N0;
    Vec3d N02 = N2-N0;
    Vec3d N12 = N2-N1;
    Vec3d N0P = P-N0;
    Vec3d N1P = P-N1;
    Vec3d N2P = P-N2;
    double N01_l = N01.norm();
    double N02_l = N02.norm();
    double N12_l = N12.norm();
    double N0P_l = N0P.norm();
    double N1P_l = N1P.norm();
    double N2P_l = N2P.norm();
    
    // the normal
    Vec3d N = N01.cross(N02);
    double area2 = N.norm();
    N.normalizeSafe();
    
    // the circumcircle radius ( 2*r = opposite side / sin angle )
    double _tmp = N01_l*N02_l;
    double sinA = area2/_tmp;
    
    double r = N12_l/sinA/2.0;
    
    // computation of the circumsphere radius R by
    // using the formula 
    //
    //   6RV = A
    //
    // where A is the area of the triangle defined by the sides
    //    
    //   |N12|*|N0P|, |N1P|*|N02|, |N2P|*|N01|
    //
    // and V is the tetrahedron volume
    
    // 6 x volume of tetrahedron
    double V6 = std::fabs( N02.dot( N01.cross(N0P) ) );
    
    // the triangle sides
    double a = std::fabs( N12_l*N0P_l );
    double b = std::fabs( N1P_l*N02_l );
    double c = std::fabs( N2P_l*N01_l );
    double per = (a+b+c)/2.0;
    double A = sqrt(per*(per-a)*(per-b)*(per-c));
    
    // the circumsphere radius
    double R = A/V6;
    
    // sine of simplex angle
    double sinPhi = ( (N1P.dot(N) > 0.0) ? 1.0:-1.0 )*r/R;
    
    // the angle, whohoo!
    this->phi = std::asin( sinPhi );
    
    // compute the curvature and set this vertex scalar value
    this->H = sinPhi/r;

  }
  
  /// compute metric parameters
  void SimplexVertex::computeMetricParameters() {
    
    // the positions
    Vec3d P = getPosition();
    Vec3d N0 = neighbors[0]->getPosition();
    Vec3d N1 = neighbors[1]->getPosition();
    Vec3d N2 = neighbors[2]->getPosition();
    Vec3d N01 = N1-N0;
    Vec3d N02 = N2-N0;
    Vec3d N12 = N2-N1;
    
    // the normal
    Vec3d N = N01.cross(N02);
    double area2 = N.norm(); // used later
    N.normalize();
    
    // the projection
    double height = (P-N0).dot(N);
    Vec3d F = P-height*N;
    
    // compute metric params by dividing area of subtriangle with
    // area of whole triangle
    double m0 = ((N2-N1).cross(F-N1)).norm()/area2;
    double m1 = ((N0-N2).cross(F-N2)).norm()/area2;
    double m2 = 1.0-m0-m1;
    
    // security
    if( m0 < 0.0 || m0 > 1.0 ||
	m1 < 0.0 || m1 > 1.0 ||
	m2 < 0.0 || m2 > 1.0 ) {
      m0 = m1 = m2 = 1.0/3.0;
    }
    
    this->mparam = Vec3d(m0,m1,m2);
    
  }
    
  /// compute the mean curvature
  void SimplexVertex::computeNeighborMeanCurvature() {
    // the neighbors
    double H0 = neighbors[0]->H;
    double H1 = neighbors[1]->H;
    double H2 = neighbors[2]->H;
    
    // compute the curvature
    this->Hbar = (std::fabs(H0) + std::fabs(H1) + std::fabs(H2))/3.0;
  }
  
  /// compute new reference metric params to adapt face area to curvature
  void SimplexVertex::curvatureAdaption(double weight) {
    
    // area equalization stuff
    // we sort the neighboring faces so that tmpfaces[i] corresponds
    // to the face that is opposite to neighbors[i]
    SimplexEdgePointer tmpedges[3];
    SimplexFacePointer tmpfaces[3];
    
    // find edges
    for(int k=0; k<3; k++) {
      for(int j=0; j<3; j++) {
	tmpedges[k] = neighbors[k]->edges[j];
	if( (tmpedges[k]->getFirstVertex() == this) ||
	    (tmpedges[k]->getSecondVertex() == this) )
	  break;
      }
    }
    
    // find faces
    tmpfaces[0] = tmpedges[1]->getFirstFace();
    if( (tmpedges[0]->getFirstFace() == tmpfaces[0]) ||
	(tmpedges[0]->getSecondFace() == tmpfaces[0]) )
      tmpfaces[0] = tmpedges[1]->getSecondFace();
    
    tmpfaces[1] = tmpedges[2]->getFirstFace();
    if( (tmpedges[1]->getFirstFace() == tmpfaces[1]) ||
	(tmpedges[1]->getSecondFace() == tmpfaces[1]) )
      tmpfaces[1] = tmpedges[2]->getSecondFace();

    tmpfaces[2] = tmpedges[0]->getFirstFace();
    if( (tmpedges[2]->getFirstFace() == tmpfaces[2]) ||
	(tmpedges[2]->getSecondFace() == tmpfaces[2]) )
      tmpfaces[2] = tmpedges[0]->getSecondFace();
    
    // get the area of neighboring faces
    Vec3d Anb;
    Anb[0] = std::fabs(tmpfaces[0]->getArea());
    Anb[1] = std::fabs(tmpfaces[1]->getArea());
    Anb[2] = std::fabs(tmpfaces[2]->getArea());
    
    // compute the mean
    double Amean = (Anb[0] + Anb[1] + Anb[2]) / 3.0;
    
    // get the mean curvature of the neighbors
    Vec3d Hnb;
    Hnb[0] = std::fabs(neighbors[0]->Hbar);
    Hnb[1] = std::fabs(neighbors[1]->Hbar);
    Hnb[2] = std::fabs(neighbors[2]->Hbar);
    
    // compute the mean
    double Hmean = (Hnb[0] + Hnb[1] + Hnb[2]) / 3.0;
    
    Vec3d Adelta, Hdelta;
    Vec3d Agamma, Hgamma;
    Vec3d eopt;
    
    // weights
    if( this->getSelected() ) {
      Agamma[0] = Agamma[1] = Agamma[2] = 0.3; //0.125;
      Hgamma[0] = Hgamma[1] = Hgamma[2] = 0.2;
    }
    else {
      Agamma[0] = Agamma[1] = Agamma[2] = 0.3; //0.125;
      Hgamma[0] = Hgamma[1] = Hgamma[2] = 0.2;
    }
    
    for(int i=0; i<3; i++) {
      // compute the relative mean face area deviation
      Adelta[i] = 1.0-Anb[i]/Amean;
      
      // compute the relative mean curvature deviation
      Hdelta[i] = (Hnb[i]-Hmean)/Hmean;
            
      // we weight the movement between curvature and face area
      eopt[i] = 
	1.0/3.0 + (Agamma[i]*Adelta[i] + Hgamma[i]*Hdelta[i]);
    }
    
    // update metric parameters
    double sum=0.0;
    for(int i=0; i<3; i++) {
      refmparam[i] += weight*(eopt[i]-refmparam[i]);
      sum += refmparam[i];
    }
    
    if( refmparam[0] < 0.0 || refmparam[0] > 1.0 ||
	refmparam[1] < 0.0 || refmparam[1] > 1.0 ||
	refmparam[2] < 0.0 || refmparam[2] > 1.0 )
      cerr << "WARNING!!! curvatureAdaption" << endl;
      
  } 
  
  /// C0 spatial force
  void SimplexVertex::computeC0Force() {
    Vec3d f = computeSimplexForce(this->phi);  
    setSpatialForce(f);
  }
  
  /// C1 spatial force
  void SimplexVertex::computeC1Force() {
    Vec3d f = computeSimplexForce(0.0);
    setSpatialForce(f);
  }    
  
  /// C2 spatial force
  void SimplexVertex::computeC2Force() {
    int nbsize=2; // nbsize = 1 gives unstability
    NeighborhoodIterator nbit = this->nbegin(nbsize);
    double angle = this->phi;
    int n=1;
    
    do {
      double anglen = nbit.current()->getSimplexAngle(); 
      angle += anglen; 
      n++;
    } while(nbit.next());
    
    angle /= n;
    
    Vec3d f = computeSimplexForce(angle);
    setSpatialForce(f);
  }    
  
  /// shape spatial force
  void SimplexVertex::computeShapeForce() {
    Vec3d f = computeSimplexForce(this->refphi);
    setSpatialForce(f);
  }    
  
  /// simplex force
  Vec3d SimplexVertex::computeSimplexForce(double _phi) const {
    // the neighbors
    Vec3d P = getPosition();
    Vec3d N0 = neighbors[0]->getPosition();
    Vec3d N1 = neighbors[1]->getPosition();
    Vec3d N2 = neighbors[2]->getPosition();
    Vec3d N01 = N1-N0;
    Vec3d N02 = N2-N0;
    Vec3d N12 = N2-N1;
    double N12_l = N12.norm();
    
    // the normal
    Vec3d N = N01.normalize().cross(N02.normalize());
    double sinA = N.norm(); // used later
    N.normalize();
    
    // the projection
    double height = (P-N0).dot(N);
    Vec3d F = P-height*N;
    
    // the radius ( 2*r = opposite side / sin angle )
    double r = N12_l/sinA/2.0;
    
    // find the center of the circumscribed circle by
    // finding the intersection of two perpendicular bisectors
    Vec3d p0 = 0.5*N0 + 0.5*N1;
    Vec3d p0d = -(N01.cross(N).normalize());
    Vec3d p1 = 0.5*N0 + 0.5*N2;
    Vec3d p1d = N02.cross(N).normalize();
    Vec3d p0p1 = p1-p0;
    double s = (p0p1.cross(p1d))*(p0d.cross(p1d))/(p0d.cross(p1d).norm2());
    Vec3d C = p0 + p0d*s;
    
    // the distance between F and C
    double d = (F-C).norm();
    
    // reference elevation
    double Lref, tanphi;
    double r2_d2 = r*r - d*d;
    if(r2_d2 < 0.0) 
      r2_d2 = 0.0;
    if(fabs(_phi) >= constants::pi / 2.0) {
      if (_phi > 0.0) 
	Lref = r;
      else 
	Lref = -r;
    }
    else {
      tanphi = tan(_phi);
      if (fabs(_phi) < constants::pi / 2.0)
	Lref = r2_d2 * tanphi /
	  (sqrt(r*r + r2_d2*tanphi*tanphi) + r);
      else
	Lref = r2_d2 * tanphi /
	  (-sqrt(r*r + r2_d2*tanphi*tanphi) + r);
    }
    
    // normal component
    Vec3d Fnormal = (Lref-height)*N;
    
    // tangential component
    Vec3d Fref = refmparam[0]*N0 + refmparam[1]*N1 + refmparam[2]*N2;
    Vec3d Ftangent = Fref-F;
    
    return Ftangent+Fnormal;
  }
  
  // ### SimplexVertex::FaceIterator ###
  // first face
  SimplexFacePointer SimplexVertex::FaceIterator::first() {
    edge_index = 0;	
    cur = vertex->getEdge(edge_index)->getFace(0);
    if( cur == vertex->getEdge(edge_index+1)->getFace(0) ||
	cur == vertex->getEdge(edge_index+1)->getFace(1) )
      cur = vertex->getEdge(edge_index)->getFace(1);
    return cur;
  }
  
  // next face
  SimplexFacePointer SimplexVertex::FaceIterator::next() {
    if( edge_index >= 2 ) //if( edge_index >= vertex->nEdges()-1 )
      cur = 0;
    else if( vertex->getEdge(edge_index)->getFace(0) == cur )
      cur = vertex->getEdge(edge_index)->getFace(1);
    else
      cur = vertex->getEdge(edge_index)->getFace(0);
    edge_index++;	
    return cur;
  }
  
  // ### SimplexVertex::NeighborhoodIterator ###
  // first neighbor
  SimplexVertexPointer SimplexVertex::NeighborhoodIterator::first() {
    // clear roots and levels
    roots.clear();
    levels.clear();
    
    // add roots, this one and its neighbors
    // also store the corresponding level for each root in levels
    roots.push_back(vertex);                 levels.push_back(1);
    roots.push_back(vertex->getNeighbor(0)); levels.push_back(2);
    roots.push_back(vertex->getNeighbor(1)); levels.push_back(2);
    roots.push_back(vertex->getNeighbor(2)); levels.push_back(2);
    
    // reset counters
    currootind = 0;
    curind = 0;
    
    // the first one
    cur = roots[currootind]->getNeighbor(curind);
    
    // clear neighbor table
    nbmap.clear();
    
    // add the root and the first neighbor
    // we store the accumulated distance in the map
    nbmap[roots[currootind]] = 0.0;    
    nbmap[cur] = 
      nbmap[roots[currootind]] + 
      (cur->getPosition()-roots[currootind]->getPosition()).norm();
    
    // clear root table 
    rootmap.clear();
    
    // add the roots
    for(int i=0; i<int(roots.size()); i++)
      rootmap[roots[i]] = true;
    
    // return the current neighbor (the first one)
    return cur;
  }
  
  // next neighbor
  SimplexVertexPointer SimplexVertex::NeighborhoodIterator::next() {
    if( currootind >= int(roots.size()) ||
	levels[currootind] > size ) {
      cur = 0;
    }
    else if( curind < 2 ) {
      cur = roots[currootind]->getNeighbor(++curind);
      if( nbmap.find(cur) != nbmap.end() )
	cur = next();
      else {
	// set as visited and store accumulated distance
	nbmap[cur] = 
	  nbmap[roots[currootind]] + 
	  (cur->getPosition()-roots[currootind]->getPosition()).norm();
      }
    }
    else {
      // next root
      currootind++;
      if( currootind >= int(roots.size()) )
	cur = 0;
      else {
	for(int i=0; i<3; i++) {
	  SimplexVertexPointer newroot = roots[currootind]->getNeighbor(i);
	  if( rootmap.find(newroot) == rootmap.end() ) {
	    roots.push_back(newroot);
	    levels.push_back(levels[currootind]+1);
	    rootmap[newroot] = true;
	  }
	}
	curind = -1;
	cur = next();
      }
    }
    return cur;
  }
  
  // ### SimplexEdge ###
  
  /// add face
  void SimplexEdge::addFace(SimplexFacePointer _face) {
    if( faces[0] == 0 )
      faces[0] = _face;
    else if( faces[1] == 0 ) {
      faces[1] = _face;
    }
    else {
      cerr << "SimplexEdge::addFace: Error, both faces already set!" << endl;
    }
  }

  // ### SimplexFace ###
  
  /// compute center
  Vec3d SimplexFace::computeCenter() {
    int n = this->nVertices();
    SimplexFace::VertexIterator vi = this->vbegin();
    this->center = Vec3d(0,0,0);
    do {
      this->center += vi.current()->getPosition();
    } while( vi.next() );
    this->center /= double(n);
    return this->center;
  }
  
  /// compute normal
  Vec3d SimplexFace::computeNormal() {
    Vec3d c = this->computeCenter();
    SimplexFace::VertexIterator vi = this->vbegin();
    Vec3d v1 = vi.current()->getPosition()-c;
    Vec3d v2 = vi.next()->getPosition()-c;
    this->normal = v1.cross(v2);
    this->normal.normalize();
    
    return this->normal;
  }
  
  /// compute area
  double SimplexFace::computeArea() {
    Vec3d ctr = this->computeCenter();
    SimplexFace::VertexIterator vi = this->vbegin();
    this->area = 0.0;
    Vec3d first = vi.current()->getPosition()-ctr;
    Vec3d v = first;
    vi.next();
    do {
      this->area += (v.cross(vi.current()->getPosition()-ctr)).norm();
      v = vi.current()->getPosition()-ctr;
    } while( vi.next() );
    this->area += (v.cross(first)).norm()/2.0;
    
    return this->area;
  }
  
  // ### SimplexFace::VertexIterator ###
  
  // the first vertex
  SimplexVertexPointer SimplexFace::VertexIterator::first() {
    edge_index = 0;	
    cur = face->getEdge(edge_index)->getVertex(0);
    if( cur == face->getEdge(edge_index+1)->getVertex(0) ||
	cur == face->getEdge(edge_index+1)->getVertex(1) )
      cur = face->getEdge(edge_index)->getVertex(1);
    return cur;
  }
  
  // the next vertex
  SimplexVertexPointer SimplexFace::VertexIterator::next() {
    if( edge_index >= face->nEdges()-1 )
      cur = 0;
    else if( face->getEdge(edge_index)->getVertex(0) == cur )
      cur = face->getEdge(edge_index)->getVertex(1);
    else
      cur = face->getEdge(edge_index)->getVertex(0);
    edge_index++;	
    return cur;
  }
  
  // ### SimplexFace::NeighborhoodIterator ###
  // first neighbor
  SimplexFacePointer SimplexFace::NeighborhoodIterator::first() {
    // clear roots and levels
    roots.clear();
    levels.clear();
    
    // add roots, this one and its neighbors
    // also store the corresponding level for each root in levels
    roots.push_back(face); 
    levels.push_back(1);
    
    for(int i=0; i<face->nNeighbors(); i++) {
      roots.push_back(face->getNeighbor(i)); 
      levels.push_back(2);
    }
    
    // reset counters
    currootind = 0;
    curind = 0;
    
    // the first one
    cur = roots[currootind]->getNeighbor(curind);
    
    // clear neighbor table
    nbmap.clear();
    
    // add the root and the first neighbor
    // we store the accumulated distance in the map
    nbmap[roots[currootind]] = 0.0;    
    nbmap[cur] = 0.0;
    
    // clear root table 
    rootmap.clear();
    
    // add the roots
    for(int i=0; i<int(roots.size()); i++)
      rootmap[roots[i]] = true;
    
    // return the current neighbor (the first one)
    return cur;
  }
  
  // next neighbor
  SimplexFacePointer SimplexFace::NeighborhoodIterator::next() {
    //cerr << "Current root is number " << currootind << endl;
    if( currootind >= int(roots.size()) ||
	levels[currootind] > size ) {
       cur = 0;
    }
    else if( curind < roots[currootind]->nNeighbors()-1 ) {
      cur = roots[currootind]->getNeighbor(++curind);
      if( nbmap.find(cur) != nbmap.end() ) {
	cur = next();
      }
      else {
	// set as visited and store accumulated distance
	nbmap[cur] = 0.0;
// 	  nbmap[roots[currootind]] + 
// 	  (cur->getPosition()-roots[currootind]->getPosition()).norm();
      }
    }
    else {
       // next root
      currootind++;
      if( currootind >= int(roots.size()) ) {
	cur = 0;
      }
      else {
	for(int i=0; i<roots[currootind]->nNeighbors(); i++) {
	  SimplexFacePointer newroot = roots[currootind]->getNeighbor(i);
	  if( rootmap.find(newroot) == rootmap.end() ) {
	    roots.push_back(newroot);
	    levels.push_back(levels[currootind]+1);
	    rootmap[newroot] = true;
	  }
	}
	curind = -1;
	cur = next();
      }
    }
    return cur;
  }
  
}
