///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_SIMPLEX_VERTEX_HH_
#define _WISH_SIMPLEX_VERTEX_HH_

#include "wish.hh"
#include "wishImage.hh"
#include "wishLinalg.hh"

namespace wish {
  
  // predefinitions
  class SimplexMesh;
  class SimplexVertex;
  class SimplexEdge;
  class SimplexFace;
  
  // typedefs
  typedef SimplexMesh*   SimplexMeshPointer;
  typedef SimplexVertex* SimplexVertexPointer;
  typedef SimplexEdge*   SimplexEdgePointer;
  typedef SimplexFace*   SimplexFacePointer;
    
  // ### SimplexVertex ###
  
  class SimplexVertex {
  protected:
    
    /// the mesh
    SimplexMeshPointer mesh;
    
    // ### Vertex parameters
    /// edges
    vector<SimplexEdgePointer> edges;
    /// neighbors
    vector<SimplexVertexPointer> neighbors;
    /// position
    Vec3d position;
    /// position at previous iteration for damping computation
    Vec3d pposition;
    /// normal
    Vec3d normal;
    /// selected
    bool selected;
    
    // ### Simplex parameters
    /// vertex simplex angles at all time steps
    double phi;
    /// reference shape simplex angles
    double refphi;
    /// vertex metric parameters at all time steps
    Vec3d mparam;
    /// reference shape metric parameters
    Vec3d refmparam;
    /// the mean curvature
    double H;
    /// neighbor mean curvature
    double Hbar;
    
    // ### Deformable parameters
    /// spatial force weight
    double alpha;
    /// external force weight
    double beta;
    /// spatial force
    Vec3d fspatial;
    /// external force
    Vec3d fexternal;
    /// global force
    Vec3d fglobal;
    /// is this vertex fixed
    bool fixed;
    
  public:
    
    /// flag used in topology algorithms
    mutable int flag;
    
    /// constructor
    SimplexVertex(const Vec3d &_position=Vec3d(0,0,0), 
		  SimplexMeshPointer _mesh=0) :
      position(_position),
      pposition(_position),
      mesh(_mesh),
      
      edges( vector<SimplexEdgePointer>(3,(SimplexEdgePointer)0) ),
      neighbors( vector<SimplexVertexPointer>(3,(SimplexVertexPointer)0) ),
      normal( Vec3d(0,0,0) ),
      selected( false ),
      
      phi(0.0),
      refphi(0.0),
      mparam( Vec3d(1.0/3.0,1.0/3.0,1.0/3.0) ),
      refmparam( Vec3d(1.0/3.0,1.0/3.0,1.0/3.0) ),
      H(0.0),
      Hbar(0.0),
      
      alpha(1.0),
      beta(0.1),
      fspatial( Vec3d(0,0,0) ),
      fexternal( Vec3d(0,0,0) ),
      fglobal( Vec3d(0,0,0) ),
      fixed( false ) 
    {
    }
    
    /// destructor
    ~SimplexVertex() {}
    
  private:
    /// unimplemented
    SimplexVertex(const SimplexVertex& _v) {}   
    SimplexVertex& operator=(const SimplexVertex &_v) {}
    
  public:
    
    /// get mesh
    SimplexMeshPointer getMesh() const {
      return mesh;
    }
    
    /// add edge
    void addEdge(SimplexEdgePointer _edge);
        
    /// get edge
    SimplexEdgePointer getEdge(int i) const {
      if( 0 <= i && i < 3 ) 
	return edges[i];
      else
	return 0;
    }
    
    /// get neighbor
    /// see also NeigborhoodIterator for larger neighborhoods than 3
    SimplexVertexPointer getNeighbor(int i) const {
      if( 0 <= i && i < 3 ) 
	return neighbors[i];
      else
	return 0;
    }

    // sort neighbors to obtain properly oriented normal
    void sortNeighbors();
    
    /// get/set the position
    Vec3d getPosition() const {
      return position; 
    }
    void setPosition(const Vec3d &_pos) { 
      pposition = position;
      position = _pos;
    }
    void setPosition(double x, double y, double z) { 
      setPosition(Vec3d(x,y,z)); 
    }
    void addPosition(const Vec3d &_pos) { 
      setPosition(position + _pos); 
    }
    void addPosition(double x, double y, double z) { 
      addPosition(Vec3d(x,y,z)); 
    }
    
    // compute normal
    Vec3d computeNormal();
    
    // get normal
    Vec3d getNormal() const {
      return normal;
    }
    
    /// get selected
    bool getSelected() const {
      return selected;
    }
    
    /// set selected
    void setSelected(const bool &s) {
      selected = s;
    }
    
    /// get velocity
    Vec3d getVelocity() const {
      return position-pposition;
    }
    
    /// get speed
    double getSpeed() const {
      return getVelocity().norm();
    }
    
    /// get/set spatial force weight
    double getSpatialWeight() const { 
      return alpha; 
    }
    void setSpatialWeight(double _w) {
      alpha = _w;
    }
    
    /// get/set external force weight
    double getExternalWeight() const { 
      return beta; 
    }
    void setExternalWeight(double _w) {
      beta = _w;
    }
    
    /// get/set spatial force
    Vec3d getSpatialForce() { 
      return fspatial; 
    }
    void setSpatialForce(const Vec3d &_f) { 
      fspatial = _f; 
    }
    void addSpatialForce(const Vec3d &_f) { 
      fspatial += _f; 
    }
    
    /// get/set external force
    Vec3d getExternalForce() { 
      return fexternal; 
    }
    void setExternalForce(const Vec3d &_f) { 
      fexternal = _f; 
    }
    void addExternalForce(const Vec3d &_f) { 
      fexternal += _f; 
    }
    
    /// get/set global force
    Vec3d getGlobalForce() { 
      return fglobal; 
    }
    void setGlobalForce(const Vec3d &_f) { 
      fglobal = _f; 
    }
    void addGlobalForce(const Vec3d &_f) { 
      fglobal += _f; 
    }
    
    /// get total force
    Vec3d getTotalForce() { 
      return fspatial + fexternal + fglobal; 
    }
    
    /// move vertex according to newtonian law of motion
    void move(const double &damping);
    
    /// compute the simplex angle
    void computeSimplexAngle();
    
    // get simplex angle
    double getSimplexAngle() const {
      return phi;
    }
    
    /// compute metric parameters
    void computeMetricParameters();
    
    /// compute the mean curvature
    void computeNeighborMeanCurvature();
    
    /// compute new reference metric params to adapt face area to curvature
    void curvatureAdaption(double weight=0.5); 
    
    /// C0 spatial force
    void computeC0Force();    
    
    /// C1 spatial force
    void computeC1Force();    
    
    /// C2 spatial force
    void computeC2Force();    
    
    /// shape spatial force
    void computeShapeForce();    
    
    /// compute simplex force
    Vec3d computeSimplexForce(double _phi) const;
    
    /// get/set fixed 
    bool getFixed() const {
      return fixed;
    }
    void setFixed(const bool &_f) {
      fixed = _f;
    }
    
    /// transformation
    void transform(const Matrix4d &M) {
      pposition = position;  // not deformable
      position = M*position;
      //normal = Vec3d(M*Vec4d(normal,0.0));
    }
    
    // ### SimplexVertex::FaceIterator ###
    struct FaceIterator {
      // the vertex
      SimplexVertexPointer vertex;
      
      // where we are
      SimplexFacePointer cur;
      int edge_index;
      
      // constructor 
      FaceIterator(SimplexVertexPointer _vertex) : vertex(_vertex) {
	first();
      }
      // the first one
      // here we check that we select the vertex that not belong
      // to the next edge
      SimplexFacePointer first();

      // the next one
      SimplexFacePointer next();
      
      // the current one
      SimplexFacePointer current() {
	return cur;
      }
    };
    
    /// iterator starting at first vertex
    FaceIterator fbegin() {
      return FaceIterator(this);
    }
    
    /// ### SimplexVertex::NeighborhoodIterator ###
    struct NeighborhoodIterator {
      // the vertex
      SimplexVertexPointer vertex;
      
      // the size
      int size;
      
      // where we are
      SimplexVertexPointer cur;
      vector<SimplexVertexPointer> roots;
      vector<int> levels;
      int curind;
      int currootind;
      
      map<SimplexVertexPointer, double> nbmap;
      map<SimplexVertexPointer, bool> rootmap;
      
      // constructor 
      NeighborhoodIterator(SimplexVertexPointer _vertex=0, int _size=1) : 
	vertex(_vertex),
	size(_size)
      {
	cur = 0;
	if( vertex != 0 )
	  first();
      }
      
      // the first one
      SimplexVertexPointer first();

      // the next one
      SimplexVertexPointer next();
      
      // the current one
      SimplexVertexPointer current() {
	return cur;
      }
    };
    
    /// iterator starting at first neighbor
    NeighborhoodIterator nbegin(int size=1) {
      return NeighborhoodIterator(this, size);
    }
    
  };
  
  // ### SimplexEdge ###
  
  class SimplexEdge {
  protected:
    /// vertices
    vector<SimplexVertexPointer> vertices;
    
    /// faces
    vector<SimplexFacePointer> faces;
        
  public:

    /// flag used in topology algorithms  
    mutable int flag;

    // build an edge between two vertices
    SimplexEdge(SimplexVertexPointer _v1, 
		SimplexVertexPointer _v2,
		SimplexFacePointer _f1 = 0, 
		SimplexFacePointer _f2 = 0) :
      vertices( vector<SimplexVertexPointer>(2,(SimplexVertexPointer)0) ),
      faces( vector<SimplexFacePointer>(2,(SimplexFacePointer)0) )
    {
      vertices[0] = _v1;
      vertices[1] = _v2;
      vertices[0]->addEdge(this);
      vertices[1]->addEdge(this);
    }
    
    /// get vertex
    SimplexVertexPointer getVertex(int i) const {
      if( 0 <= i && i <= 1 )
	return vertices[i];
      else
	return 0;
    }
    SimplexVertexPointer getFirstVertex() const {
      return vertices[0];
    }
    SimplexVertexPointer getSecondVertex() const {
      return vertices[1];
    }
    
    /// get face
    SimplexFacePointer getFace(int i) { 
      if( 0 <= i && i <= 1 )
	return faces[i];
      else
	return 0;
    }
    SimplexFacePointer getFirstFace() const {
      return faces[0];
    }
    SimplexFacePointer getSecondFace() const {
      return faces[1];
    }

    /// add face
    void addFace(SimplexFacePointer _face);
  };
  
  // ### SimplexFace ###
  
  class SimplexFace {
  protected:
    /// edges (should be ordered around the face)
    vector<SimplexEdgePointer> edges;
    
    /// empty or not
    bool empty;
    
    /// center
    Vec3d center;
    
    /// normal
    Vec3d normal;
    
    /// selected
    bool selected;
    
    /// area
    double area;
        
  public:

    /// flag used in topology algorithms
    mutable int flag;
    
    /// constructor
    SimplexFace(const vector<SimplexEdgePointer> &_edges, 
		bool _empty=false) :
      edges(_edges),
      empty(_empty),
      center( Vec3d(0,0,0) ),
      normal( Vec3d(0,0,0) ),
      selected(false),
      area(0.0) 
    {
    }
    
    /// reset face to a new set of edges
    void reset(vector<SimplexEdgePointer> _edges, bool _empty=false) {
      empty = _empty;
      edges= _edges;
    }
    
    /// compute center
    Vec3d computeCenter();
    
    /// get center
    Vec3d getCenter() const { 
      return center; 
    }
    
    /// compute normal
    Vec3d computeNormal();
    
    /// get normal
    Vec3d getNormal() const {
      return normal;
    }
    
    // get selected
    bool getSelected() const {
      return selected;
    }
    
    // set selected
    void setSelected(const bool &s) {
      selected = s;
    }
    
    /// compute area
    double computeArea();
    
    /// get area
    double getArea() const {
      return area;
    }
    
    /// get number of edges
    int nEdges() const { 
      return (int) edges.size(); 
    }

    /// get number of neighbors (same as number of edges)
    int nNeighbors() const { 
      return nEdges();
    }
    
    /// get number of vertices (same as number of edges)
    int nVertices() const { 
      return nEdges();
    }
    
    /// get/set empty
    bool getEmpty() const { 
      return empty; 
    }
    void setEmpty(bool _h) {
      empty = _h;
    }
    
    /// get edge
    SimplexEdgePointer getEdge(int i) const {
      if( 0<=i && i<nEdges() )
	return edges[i];
      else
	return 0;
    }
    
    /// get face neighbor
    SimplexFacePointer getNeighbor(int i) const {
      SimplexFacePointer sfp = getEdge(i)->getFirstFace();
      if( sfp == this )
	sfp = getEdge(i)->getSecondFace();
      return sfp;
    }
    
    // ### SimplexFace::VertexIterator ###
    struct VertexIterator {
      // the face
      SimplexFacePointer face;
      
      // where we are
      SimplexVertexPointer cur;
      int edge_index;
      
      // constructor 
      VertexIterator(SimplexFacePointer _face) : face(_face) {
	first();
      }
      // the first one
      // here we check that we select the vertex that not belong
      // to the next edge
      SimplexVertexPointer first();
      
      // the next one
      SimplexVertexPointer next();
      
      // the current one
      SimplexVertexPointer current() {
	return cur;
      }
    };
    
    /// iterator starting at first vertex
    VertexIterator vbegin() {
      return VertexIterator(this);
    }
    
    // ### SimplexFace::NeighborhoodIterator ###
    struct NeighborhoodIterator {
      // the face
      SimplexFacePointer face;
      
      // the size
      int size;
      
      // where we are
      SimplexFacePointer cur;
      vector<SimplexFacePointer> roots;
      vector<int> levels;
      int curind;
      int currootind;
      
      map<SimplexFacePointer, double> nbmap;
      map<SimplexFacePointer, bool> rootmap;
      
      // constructor 
      NeighborhoodIterator(SimplexFacePointer _face=0, int _size=1) : 
	face(_face),
	size(_size)
      {
	cur = 0;
	if( face != 0 )
	  first();
      }
      
      // the first one
      SimplexFacePointer first();

      // the next one
      SimplexFacePointer next();
      
      // the current one
      SimplexFacePointer current() {
	return cur;
      }
      
    };
    
    /// iterator starting at first face
    NeighborhoodIterator nbegin(int size=1) {
      return NeighborhoodIterator(this, size);
    }
    
  };
  
}

#endif
