///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_SLICE_VIEWER_HH_
#define _WISH_SLICE_VIEWER_HH_

#if defined (__WIN32__) || defined (WIN32)
#include <windows.h>
#endif

#include <GL/glew.h>
#include <GL/glut.h>

#include "wishImage.hh"
#include "wishGLTexture.hh"
#include "wishShaderUtilities.hh"

namespace wish {
  
  class SliceViewer {
    
  protected:
    // the textures
    vector<Texture*> textures;
    
    // slice
    int z_slice; // axial
    int y_slice; // coronal
    int x_slice; // sagittal

    // window dimensions
    int win_w;
    int win_h;
    
    // shader stuff
    GLuint shader_program;
    GLint volume_handle;
    GLint volume_handle2;
    GLint blend_handle;
    GLint wl_width_handle;
    GLint wl_level_handle;
    bool shader_ok;
    
    // blending
    float blend;
    
    // window/level
    float wl_width;
    float wl_level;
    
//     // useful help function for checking OpenGL errors
//     // use after suspicios function call
//     static void errorCheck() {
//       GLenum err = glGetError();
//       cout << gluErrorString( err ) << endl;
//     }
    
  public:
    
    SliceViewer() : 
      z_slice(0),
      y_slice(0),
      x_slice(0),
      shader_program(0),
      volume_handle(0),
      volume_handle2(0),
      blend_handle(0),
      wl_width_handle(0),
      wl_level_handle(0),
      blend(0.0),
      wl_width(1.0f),
      wl_level(0.5f),
      shader_ok( false )
    {}
    
    // set/get texture
    void addTexture(Texture *_texture) {
      textures.push_back(_texture);
    }
    
    // get image
    ImageObject* getImage(int i=0) {
      return textures[i]->getImage();
    }
    
    // set/get window/level parameters
    void setWL(float _w, float _l) {
      wl_width = _w; wl_level = _l;
    }
    void setWLWidth(float _w) {
      wl_width = _w;
    }
    void setWLLevel(float _l) {
      wl_level = _l;
    }
    float getWLWidth() const {
      return wl_width;
    }
    float getWLLevel() const {
      return wl_level;
    }
    
    // initialize
    void initialize( bool _shader_ok ) {
      
      shader_ok = _shader_ok;
      if( !shader_ok ) {
	return;
      }
      
      // setup shaders
      string wishroot = string(getenv("WISH_ROOT"));
      string vs = wishroot + "/wish/src/shaders/slicerVertex.glsl";
      string fs = wishroot + "/wish/src/shaders/slicerFragment.glsl";
      shader_program = createShaderProgram( vs, fs );
         
      // get the addresses of uniform variables
      volume_handle = 
	glGetUniformLocationARB(shader_program, 
				"volumeTexture");
      volume_handle2 = 
	glGetUniformLocationARB(shader_program, 
				"volumeTexture2");
      blend_handle = 
	glGetUniformLocationARB(shader_program, 
				"blend");
      wl_width_handle = 
	glGetUniformLocationARB(shader_program, 
				"wl_width");
      wl_level_handle = 
	glGetUniformLocationARB(shader_program, 
				"wl_level");
    }
    
    // render slices
    void render( bool renderz=true, 
		 bool rendery=false, 
		 bool renderx=false ) {
      
      glPushAttrib(GL_TEXTURE_BIT);
      
      glMatrixMode(GL_MODELVIEW);
      
      // draw the image by using texture mapping
      // enable texturing
      glEnable(GL_TEXTURE_3D);
      
      // use shader
      if( shader_ok )
	glUseProgramObjectARB(shader_program);  
      
      // texture units
      for( size_t i=0; i<textures.size(); i++ ) { 
	glActiveTexture( GL_TEXTURE0+GLenum(i) );
	textures[i]->bind();
      }
      
      // set uniforms
      if( shader_ok ) {
	glUniform1iARB(volume_handle, 0);
	glUniform1iARB(volume_handle2, 1);
	glUniform1fARB(blend_handle, blend);
	glUniform1fARB(wl_width_handle, wl_width);
	glUniform1fARB(wl_level_handle, wl_level);
      }
      
      Vec3i sz = textures[0]->getImage()->getDimensions();
      Matrix4f im2world = textures[0]->getImage()->getImageToWorldMatrix();
      Matrix4f im2tex = textures[0]->getImage()->getImageToTextureMatrix();
      
      if( renderz ) {
	Vec3f p0 = Vec3f(0.0f, 0.0f, (float)z_slice);
	Vec3f p1 = Vec3f(0.0f, (float)sz.y-1, (float)z_slice);
	Vec3f p2 = Vec3f((float)sz.x-1, (float)sz.y-1, (float)z_slice);
	Vec3f p3 = Vec3f((float)sz.x-1, 0.0f, (float)z_slice);
	
	glBegin(GL_QUADS);
	glTexCoord3fv( im2tex*p0 );
	glVertex3fv( im2world*p0 );
	
	glTexCoord3fv( im2tex*p1 );
	glVertex3fv( im2world*p1 );
	
	glTexCoord3fv( im2tex*p2 );
	glVertex3fv( im2world*p2 );
	
	glTexCoord3fv( im2tex*p3 );
	glVertex3fv( im2world*p3 );
	glEnd();
      }
      
      if( rendery ) {
	Vec3f p0 = Vec3f(0.0f, (float)y_slice, (float)sz.z-1);
	Vec3f p1 = Vec3f(0.0f, (float)y_slice, 0.0f);
	Vec3f p2 = Vec3f((float)sz.x-1, (float)y_slice, 0.0f);
	Vec3f p3 = Vec3f((float)sz.x-1, (float)y_slice, (float)sz.z-1);
	
	glBegin(GL_QUADS);
	glTexCoord3fv( im2tex*p0 );
	glVertex3fv( im2world*p0 );
	
	glTexCoord3fv( im2tex*p1 );
	glVertex3fv( im2world*p1 );
	
	glTexCoord3fv( im2tex*p2 );
	glVertex3fv( im2world*p2 );
	
	glTexCoord3fv( im2tex*p3 );
	glVertex3fv( im2world*p3 );
	glEnd();
      }
      
      if( renderx ) {
	Vec3f p0 = Vec3f((float)x_slice, 0.0f, (float)sz.z-1);
	Vec3f p1 = Vec3f((float)x_slice, (float)sz.y-1, (float)sz.z-1);
	Vec3f p2 = Vec3f((float)x_slice, (float)sz.y-1, 0.0);
	Vec3f p3 = Vec3f((float)x_slice, 0.0f, 0.0);
	
	glBegin(GL_QUADS);
	glTexCoord3fv( im2tex*p0 );
	glVertex3fv( im2world*p0 );
	
	glTexCoord3fv( im2tex*p1 );
	glVertex3fv( im2world*p1 );
	
	glTexCoord3fv( im2tex*p2 );
	glVertex3fv( im2world*p2 );
	
	glTexCoord3fv( im2tex*p3 );
	glVertex3fv( im2world*p3 );
	glEnd();
      }
      
      if( shader_ok )
	glUseProgramObjectARB(0);
      
      glPopAttrib();
      
    }
    
    // reshape callback
    //  set the window size and the projection
    void reshape(int w, int h) {
      win_w = w;
      win_h = h;
    }
    
    // step z slice
    void nextZSlice() {
      z_slice++;
      if(z_slice > textures[0]->getImage()->depth()-1) 
	z_slice = textures[0]->getImage()->depth()-1;
    }
    
    void prevZSlice() {
      z_slice--;
      if(z_slice < 0) 
	z_slice = 0;
    }
    
    void setZSlice(int z) {
      z_slice = z;
      if(z_slice < 0) z_slice = 0;
      if(z_slice > textures[0]->getImage()->depth()-1) 
	z_slice = textures[0]->getImage()->depth()-1;
    }
    
    int getZSlice() const {
      return z_slice;
    }

    // step y slice
    void nextYSlice() {
      y_slice++;
      if(y_slice > textures[0]->getImage()->height()-1) 
	y_slice = textures[0]->getImage()->height()-1;
    }
    
    void prevYSlice() {
      y_slice--;
      if(y_slice < 0) 
	y_slice = 0;
    }
    
    void setYSlice(int y) {
      y_slice = y;
      if(y_slice < 0) y_slice = 0;
      if(y_slice > textures[0]->getImage()->height()-1) 
	y_slice = textures[0]->getImage()->height()-1;
    }
    
    int getYSlice() const {
      return y_slice;
    }

    // step x slice
    void nextXSlice() {
      x_slice++;
      if(x_slice > textures[0]->getImage()->width()-1) 
	x_slice = textures[0]->getImage()->width()-1;
    }
    
    void prevXSlice() {
      x_slice--;
      if(x_slice < 0) 
	x_slice = 0;
    }
    
    void setXSlice(int x) {
      x_slice = x;
      if(x_slice < 0) x_slice = 0;
      if(x_slice > textures[0]->getImage()->width()-1) 
	x_slice = textures[0]->getImage()->width()-1;
    }
    
    int getXSlice() const {
      return x_slice;
    }

    
  };

}

#endif
