///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_THRESHOLD_FILTER_HH_
#define _WISH_THRESHOLD_FILTER_HH_

#include "wishFilterObject.hh"
#include "wishVoxelIterator.hh"

namespace wish {
  
  template<class InputImage, class OutputImage>
  class ThresholdFilter : public FilterObject<InputImage, OutputImage> {
    
  public:
    
    //! pointer typedef
    typedef auto_ptr< ThresholdFilter<InputImage,OutputImage> > Pointer;
    typedef typename InputImage::Pointer InputImagePointer;
    typedef typename OutputImage::Pointer OutputImagePointer;
    typedef typename InputImage::VoxelType InputImageVoxelType;
    typedef typename OutputImage::VoxelType OutputImageVoxelType;
    
    //! constructor
    ThresholdFilter() :
      m_min(0.0), 
      m_max(1.0),
      m_value(numeric_limits<OutputImageVoxelType>::max())
    {}
    
    //! set low threshold
    void setMin(const double &_min) {
      m_min = _min;
    }
    
    //! set high threshold
    void setMax(const double &_max) {
      m_max = _max;
    }
    
    //! set threshold limits
    void thresholdBetween(const double &_min,
			  const double &_max) {
      m_min = _min;
      m_max = _max;
    }
    
    //! set output value
    void setOutputValue(const OutputImageVoxelType &_value) {
      m_value = _value;
    }
    
  protected:
    double m_min;
    double m_max;
    OutputImageVoxelType m_value;
    
    virtual void process();
    
  private:
    
    ThresholdFilter(const ThresholdFilter&);
    ThresholdFilter& operator=(const ThresholdFilter&);
  };
  
  template<class InputImage, class OutputImage>
  void ThresholdFilter<InputImage, OutputImage>::process() {
    
    this -> m_output -> resize( *(this->m_input) );
    this -> m_output -> reset( OutputImageVoxelType(0) );
    
    this -> m_output ->setLabel( this->m_input->getLabel() + "_Thresh" );
    
    VoxelIterator<InputImage> in(this->m_input);
    VoxelIterator<OutputImage> out(this->m_output);
    
    in.setBox(this->m_box);
    out.setBox(this->m_box);
    in.first();
    out.first();
    
    do {
      if( m_min <= double(in.value()) && double(in.value()) <= m_max )
	out.value() = m_value;
      else
	out.value() = OutputImageVoxelType(0);
    } while( in.next() && out.next() );
    
  }
  
} // end namespace wish

#endif
