///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishTransferFunction.hh"

namespace wish {

  // evaluate function
  double CBTransferFunction::evaluate(const double &arg) const {
  
    // Remember that we adjust 1/deltaX, where deltaX is the contrast
    // It should be (maxvalue-minvalue)/contrast, but we use 
    // (maxvalue-minvalue)*contrast
  
    double f = 
      (arg-(minvalue+(maxvalue-minvalue)/2.0)+brightness)*
      (brightness-minvalue)*contrast+(minvalue+(maxvalue-minvalue)/2.0);
    
    f = util::minOf(f,maxvalue);
    f = util::maxOf(f,minvalue);
   
    return f;
  }
  
  // evaluate function
  double SigmoidTransferFunction::evaluate(const double &arg) const {
    return
      minvalue + (maxvalue-minvalue)/
      (1.0+std::exp(-(arg-beta)/(alpha+constants::eps6)));
  }
  
  // evaluate function
  // this function is not safe, you must provide valid ranges and values
  double PiecewiseLinearTransferFunction::evaluate(const double &arg) const {
    
    double f;
    
    if( sample_count == 0 ) // no samples
      f = 0.0;
    
    else if( sample_count == 1 ) // constant function
      f = ysamples[0];
    
    else if( arg < xsamples[0] ) // clamping
      f = ysamples[0];
    
    else if( arg > xsamples[sample_count-1] ) 
      f = ysamples[sample_count-1];
    
    else { // interpolate (linearly)
      for(unsigned int i=0; i<sample_count-1; i++) {
	if( xsamples[i] <= arg && arg <= xsamples[i+1] ) {
	  double alpha = arg-xsamples[i];
	  f = 
	    ysamples[i] + alpha*(ysamples[i+1]-ysamples[i])/
	    (xsamples[i+1]-xsamples[i]);
	}
      }
    }
    
    return f;
  }
  
  // evaluate function
  double BWTransferFunction::evaluate(const double &arg) const {
    return 
      minvalue + (maxvalue-minvalue)*
      (1.0/(1.0+std::pow((arg-offset)/(base+constants::eps6), 2.0*degree)));
  }
  
  // evaluate function
  double GaussianTransferFunction::evaluate(const double &arg) const {
    return 
      minvalue + (maxvalue-minvalue)*
      std::exp(-0.5*util::square(arg-mean)/sigma2);
  }
  
}
