///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_TRANSFER_FUNCTION_HH_
#define _WISH_TRANSFER_FUNCTION_HH_

#include "wishutil.hh"

namespace wish {
  
  // base class
  class TransferFunction {
  public:
    TransferFunction() {}  
    virtual double evaluate(const double &arg) const = 0;
  };
  
  // linear contrast and brightness function
  class CBTransferFunction : public TransferFunction {
    
  protected:
    
    // parameters
    double contrast;
    double brightness;
    double minvalue;
    double maxvalue;
    
  public:

    // set/get parameters
    void setContrast(double c) {
      contrast = c;
    }
    void setBrightness(double b) {
      brightness = b;
    }
    void setMin(double mn) {
      minvalue = mn;
    }
    void setMax(double mx) {
      maxvalue = mx;
    }
    double getContrast() const {
      return contrast;
    }
    double getBrightness() const {
      return brightness;
    }
    double getMin() const {
      return minvalue;
    }
    double getMax() const {
      return maxvalue;
    }
    
    // constructor
    CBTransferFunction() :
      contrast(1.0),
      brightness(0.0),
      minvalue(0.0),
      maxvalue(1.0) {}
    
    // evaluate function
    double evaluate(const double &arg) const;
  };
  
  class SigmoidTransferFunction : public TransferFunction {
    
  protected:
    
    // parameters
    double alpha;
    double beta;
    double minvalue;
    double maxvalue;    
    
  public:
    
    // set/get parameters
    void setAlpha(double a) {
      alpha = a;
    }
    void setBeta(double b) {
      beta = b;
    }
    void setMin(double mn) {
      minvalue = mn;
    }
    void setMax(double mx) {
      maxvalue = mx;
    }
    double getAlpha() const {
      return alpha;
    }
    double getBeta() const {
      return beta;
    }
    double getMin() const {
      return minvalue;
    }
    double getMax() const {
      return maxvalue;
    }
    
    // constructor
    SigmoidTransferFunction() :
      alpha(1.0),
      beta(0.0),
      minvalue(0.0),
      maxvalue(1.0) {}
    
    // evaluate function
    // out(x) = (max-min)*1.0/(1+exp(-(in(x)-beta)/alpha))+min    
    double evaluate(const double &arg) const;
    
  };
  
  // constant function
  class ConstantTransferFunction : public TransferFunction {
    
  protected:
    // parameters
    double value;
    
  public:
    
    // set parameters
    void setValue(double v) {
      value = v;
    }
    double getValue() const {
      return value;
    }
    
    // constructor
    ConstantTransferFunction() : value(0.0) {}
    
    // evaluate function
    double evaluate(const double &arg) const {
      return value;
    }
  };
  
  class PiecewiseLinearTransferFunction : public TransferFunction {

  protected:
    
    // samples
    vector<double> xsamples;
    vector<double> ysamples;
    unsigned int sample_count;
    
  public:
    
    // set/get parameters
    void clearSamples() {
      xsamples.clear();
      ysamples.clear();
      sample_count = 0;
    }
    
    // add a point
    // OBS! No checking is done that x_i is larger than x_{i-1}  
    void addPoint(double x, double y) {
      xsamples.push_back(x);
      ysamples.push_back(y);
      sample_count++;
    }
    
    const vector<double>& getXSamples() const {
      return xsamples;
    }
    const vector<double>& getYSamples() const {
      return ysamples;
    }
    
    // constructor
    PiecewiseLinearTransferFunction() {
      clearSamples();
    }
    
    // evaluate function
    double evaluate(const double &arg) const;
    
  };
  
  // Transfer function based on the function of a Butterworth filter
  class BWTransferFunction : public TransferFunction {
    
  protected:
    // parameters
    double offset;    // x-x0
    int degree;       // n
    double base;      // w0
    double minvalue;  // minvalue < f(x) < maxvalue 
    double maxvalue;  //
    
  public:
    
    // set/get parameters
    void setOffset(double o) {
      offset = o;
    }
    void setDegree(int d) {
      degree = d;
    }
    void setBase(double b) {
      base = b;
    }
    void setMin(double mn) {
      minvalue = mn;
    }
    void setMax(double mx) {
      maxvalue = mx;
    }
    double getOffset() const {
      return offset;
    }
    int getDegree() const {
      return degree;
    }
    double getBase() const {
      return base;
    }
    double getMin() const {
      return minvalue;
    }
    double getMax() const {
      return maxvalue;
    }
    
    // constructor
    BWTransferFunction() :
      offset(0.0),
      degree(3),
      base(0.5),
      minvalue(0.0),
      maxvalue(1.0) {}
    
    // evaluate function
    double evaluate(const double &arg) const;
  };
  
  // Transfer function based on a Gaussian
  class GaussianTransferFunction : public TransferFunction {
    
  protected:
    // parameters
    double mean;     // x-x0
    double sigma;    // sigma
    double sigma2;   // sigma squared
    double minvalue; // minvalue < f(x) < maxvalue
    double maxvalue; // 
    
  public:
    
    // set/get parameters
    void setMean(double m) {
      mean = m;
    }
    void setSigma(double s) {
      sigma = s;
      sigma2 = sigma*sigma;
    }
    void setMin(double mn) {
      minvalue = mn;
    }
    void setMax(double mx) {
      maxvalue = mx;
    }
    double getMean() const {
      return mean;
    }
    double getSigma() const {
      return sigma; 
    }
    double getMin() const {
      return minvalue;
    }
    double getMax() const {
      return maxvalue;
    }
    
    // constructor
    GaussianTransferFunction() :
      mean(0.0),
      sigma(1.0),
      sigma2(1.0),
      minvalue(0.0),
      maxvalue(1.0) {}
    
    // evaluate function
    double evaluate(const double &arg) const;
  };
}

#endif
