///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include "wishTriangleMesh.hh"

namespace wish {
  
  // sort edges so that two edges that share a face are consecutive
  void TriangleMesh::sortEdges() {
    for(int i=0; i<this->nVertices(); i++) 
      this->getVertex(i)->sortEdges();
  }
  
  // delete mesh (called in the destructor)
  void TriangleMesh::deleteMesh() {
    while(!faces.empty()) {
      delete faces.back();
      faces.pop_back();
    }
    while(!edges.empty()) {
      delete edges.back();
      edges.pop_back();
    }
    while(!vertices.empty()) {
      delete vertices.back();
      vertices.pop_back();
    }
  }
  
  /// refine mesh
  void TriangleMesh::refine() {
    int NV = this->nVertices();
    int NE = this->nEdges();
    int NF = this->nFaces();
    
    // one new vertex for each edge => NV+NE vertices in the refined mesh
    vector<TriangleVertexPointer> v(NV+NE);
    
    // one new edge for each edge and three new edges for each face => 
    // NE+NE+3*NF = 2*NE+3*NF edges in the refined mesh
    vector<TriangleEdgePointer> e; e.reserve(NE+NE+3*NF); 
    
    // three new faces for each face => 
    // NF+3*NF = 4*NF faces in the refined mesh
    vector<TriangleFacePointer> f; f.reserve(NF+3*NF); 
    
    // first we add all old vertices
    for(int i=0; i<NV; i++) {
      v[i] = this->getVertex(i);
      
      // clear old edges and neighbors
      v[i]->clearEdges();
      v[i]->clearNeighbors();
    }
    
    // for each edge we create a new vertex
    for(int i=0; i<NE; i++) {
      TriangleEdgePointer te = this->getEdge(i);
      
      // position of the new vertex is the midpoint of the edge
      Vec3d pos = te->getFirstVertex()->getPosition();
      pos += te->getSecondVertex()->getPosition();
      pos /= 2.0;
      
      // create the new vertex
      v[NV+i] = this->newVertex(pos);
      te->flag = NV+i; // save the index in the edge    
    }
    
    // for each face we add the new edges and new faces
    // we use an edge map to avoid redundant edges
    typedef pair<TriangleVertexPointer, TriangleVertexPointer> TVPair;
    typedef map<TVPair, TriangleEdgePointer> Edgemap;
    Edgemap edgemap;
    for(int i=0; i<NF; i++) {
      
      // the current (old) face
      TriangleFacePointer ftmp = this->getFace(i);
      
      // to store the new faces
      vector<TriangleFacePointer> fnew(4);
      
      // to store the new edges
      vector< vector<TriangleEdgePointer> > 
	ee(4,vector<TriangleEdgePointer>(3));
      
      // ### FIRST
      // create edges for first face (the new edges in the center)
      ee[0][0] = this->newEdge(v[ftmp->getEdge(0)->flag], 
			    v[ftmp->getEdge(1)->flag]);
      e.push_back(ee[0][0]);
      ee[0][1] = this->newEdge(v[ftmp->getEdge(0)->flag], 
			    v[ftmp->getEdge(2)->flag]);
      e.push_back(ee[0][1]);
      ee[0][2] = this->newEdge(v[ftmp->getEdge(1)->flag], 
			    v[ftmp->getEdge(2)->flag]);
      e.push_back(ee[0][2]);
      
      // create and add first face
      fnew[0] = this->newFace(ee[0]);
      ee[0][0]->addFace(fnew[0]);
      ee[0][1]->addFace(fnew[0]);
      ee[0][2]->addFace(fnew[0]);
            
      // ### SECOND
      ee[1][0] = ee[0][0]; // already generated
      
      // find the next vertex, must be first or second from edge 0
      // we store it in the variable vp1
      TriangleVertexPointer vp1 = ftmp->getEdge(0)->getFirstVertex();
      TriangleVertexPointer vp2 = ftmp->getEdge(0)->getSecondVertex();
      if( (vp1 != ftmp->getEdge(1)->getFirstVertex()) &&
	  (vp1 != ftmp->getEdge(1)->getSecondVertex()) )
	util::swap(vp1,vp2);
      
      // search for this edge in the edge map
      TVPair tvp = make_pair(v[ftmp->getEdge(0)->flag], vp1);
      Edgemap::iterator it = edgemap.find( tvp );
      if( it == edgemap.end() ) { // check the other ordering as well
	tvp = make_pair(vp1, v[ftmp->getEdge(0)->flag]);
	it = edgemap.find( tvp );
      }
      if( it == edgemap.end() ) { // it did not exist, create new
	ee[1][1] = this->newEdge(v[ftmp->getEdge(0)->flag], vp1);
	e.push_back(ee[1][1]);	
	edgemap[tvp] = ee[1][1];
      }
      else { // already exists, get it from edgemap
	ee[1][1] = edgemap[tvp];
      }
      
      // same procedure again...
      tvp = make_pair(v[ftmp->getEdge(1)->flag], vp1);
      it = edgemap.find( tvp );
      if( it == edgemap.end() ) {
	tvp = make_pair(vp1, v[ftmp->getEdge(1)->flag]);
	it = edgemap.find( tvp );      
      }
      if( it == edgemap.end() ) {
	ee[1][2] = this->newEdge(v[ftmp->getEdge(1)->flag], vp1);
	e.push_back(ee[1][2]);
	edgemap[tvp] = ee[1][2];
      }
      else {
	ee[1][2] = edgemap[tvp];
      }
      
      // create and add second face
      fnew[1] = this->newFace(ee[1]);
      ee[1][0]->addFace(fnew[1]);
      ee[1][1]->addFace(fnew[1]);
      ee[1][2]->addFace(fnew[1]);
      
      // ### THIRD
      ee[2][0] = ee[0][1];
      
      tvp = make_pair(v[ftmp->getEdge(0)->flag], vp2);
      it = edgemap.find( tvp );
      if( it == edgemap.end() ) {
	tvp = make_pair(vp2, v[ftmp->getEdge(0)->flag]);
	it = edgemap.find( tvp );      
      }
      if( it == edgemap.end() ) {
	ee[2][1] = this->newEdge(v[ftmp->getEdge(0)->flag], vp2);
	e.push_back(ee[2][1]);
	edgemap[tvp] = ee[2][1];
      }
      else {
	ee[2][1] = edgemap[tvp];
      }
      
      tvp = make_pair(v[ftmp->getEdge(2)->flag], vp2);
      it = edgemap.find( tvp );
      if( it == edgemap.end() ) {
	tvp = make_pair(vp2, v[ftmp->getEdge(2)->flag]);
	it = edgemap.find( tvp );      
      }
      if( it == edgemap.end() ) {
	ee[2][2] = this->newEdge(v[ftmp->getEdge(2)->flag], vp2);
	e.push_back(ee[2][2]);
	edgemap[tvp] = ee[2][2];
      }
      else {
	ee[2][2] = edgemap[tvp];
      }
            
      // create and add third face
      fnew[2] = this->newFace(ee[2]);
      ee[2][0]->addFace(fnew[2]);
      ee[2][1]->addFace(fnew[2]);
      ee[2][2]->addFace(fnew[2]);
            
      // ### FOURTH
      ee[3][0] = ee[0][2];
      
      // need to find next vertex, must belong to edge 1, and it is not v1
      vp2 = ftmp->getEdge(1)->getFirstVertex(); 
      if( vp2 == vp1 )
	vp2 = ftmp->getEdge(1)->getSecondVertex(); 
      
      tvp = make_pair(v[ftmp->getEdge(1)->flag], vp2);
      it = edgemap.find( tvp );
      if( it == edgemap.end() ) {
	tvp = make_pair(vp2, v[ftmp->getEdge(1)->flag]);
	it = edgemap.find( tvp );      
      }
      if( it == edgemap.end() ) {
	ee[3][1] = this->newEdge(v[ftmp->getEdge(1)->flag], vp2);
	e.push_back(ee[3][1]);
	edgemap[tvp] = ee[3][1];
      }
      else {
	ee[3][1] = edgemap[tvp];
      }
      
      tvp = make_pair(v[ftmp->getEdge(2)->flag], vp2);
      it = edgemap.find( tvp );
      if( it == edgemap.end() ) {
	tvp = make_pair(vp2, v[ftmp->getEdge(2)->flag]);
	it = edgemap.find( tvp );      
      }
      if( it == edgemap.end() ) {
	ee[3][2] = this->newEdge(v[ftmp->getEdge(2)->flag], vp2);
	e.push_back(ee[3][2]);
	edgemap[tvp] = ee[3][2];
      }
      else {
	ee[3][2] = edgemap[tvp];
      }
      
      // create and add fourth face
      fnew[3] = this->newFace(ee[3]);
      ee[3][0]->addFace(fnew[3]);
      ee[3][1]->addFace(fnew[3]);
      ee[3][2]->addFace(fnew[3]);
      
      // check orientation by comparing the normal of the new face
      // to the old face, if needed we swap edge order
      // finally we add the faces
      for(int kf=0; kf<4; kf++) {
	if( fnew[kf]->computeNormal().dot(ftmp->getNormal()) < 0.0 ) {
	  util::swap(ee[kf][0], ee[kf][2]);
	  fnew[kf]->reset(ee[kf]);
	}
	f.push_back(fnew[kf]);
      }
      
    }
    // ...puh, finished
    
    // clean up old stuff and add everything
    while(!faces.empty()) {
      delete faces.back();
      faces.pop_back();
    }
    while(!edges.empty()) {
      delete edges.back();
      edges.pop_back();
    }

    this->faces = f;
    this->edges = e;
    this->vertices = v;
    
    // sort vertices edges
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
    
    cerr << "Refined TriangleMesh has " 
	 << this->nVertices() << " vertices, " 
	 << this->nEdges() << " edges, and " 
	 << this->nFaces() << " faces." << endl;
  }

  // create triangle mesh from simplex mesh
  void TriangleMesh::createFromSimplexMesh(SimplexMeshPointer smesh) {
    
    // clear
    this->deleteMesh();
    
    // for each simplex mesh face, create a triangulation vertex
    vector<TriangleVertexPointer> v(smesh->nFaces());
    for(int i=0; i<smesh->nFaces(); i++) {
      // the current simplex face
      SimplexFacePointer sfp = smesh->getFace(i);
      // store index in face      
      sfp->flag = i; 
            
      // compute the position, centroid 
      Vec3d pos(0.0, 0.0, 0.0);
      SimplexFace::VertexIterator vi = sfp->vbegin();      
      do {
	pos += vi.current()->getPosition();  
      } while( vi.next() );
      pos /= sfp->nVertices();
      
      // create and add vertex
      v[i] = this->newVertex(pos);
      this->addVertex(v[i]);
    }
    
    // for each simplex mesh edge, create a triangulation edge
    vector<TriangleEdgePointer> e(smesh->nEdges());
    for(int i=0; i<smesh->nEdges(); i++) {
      // the current simplex edge
      SimplexEdgePointer sep = smesh->getEdge(i);
      // store index in edge
      sep->flag = i;
      
      // create and add edge, the indices of the triangle edge 
      // vertices are stored in the simplex faces belonging to
      // the simplex edge   
      e[i] = this->newEdge(v[sep->getFirstFace()->flag],
			   v[sep->getSecondFace()->flag]);
      this->addEdge(e[i]);
    }
    
    // for each simplex mesh vertex, create a triangulation face
    for(int i=0; i<smesh->nVertices(); i++) {
      // the current simplex vertex
      SimplexVertexPointer svp = smesh->getVertex(i);
      
      // get face edges from vertex
      // the indices of the triangle edges are stored in the simplex
      // edges belonging to the simplex vertex 
      vector<TriangleEdgePointer> fe(3);
      for(int j=0; j<3; j++) {
	fe[j] = e[svp->getEdge(j)->flag];
      }
      TriangleFacePointer tfp = this->newFace(fe);
      
      // check orientation, if wrong, swap edge order 
      SimplexEdgePointer e1 = svp->getEdge(0);
      SimplexEdgePointer e2 = svp->getEdge(1);
      SimplexFacePointer t = (e1->getFirstFace() == e2->getFirstFace() ||
			       e1->getFirstFace() == e2->getSecondFace())
	? e1->getFirstFace() : e1->getSecondFace();
      int nn=0;
      SimplexEdgePointer etmp = t->getEdge(nn);
      while( etmp != e1 )
	etmp = t->getEdge(++nn);
      if(nn==t->nEdges()-1) 
	nn=0;
      else
	nn++;
      etmp = t->getEdge(nn);
      if( etmp == e2 ) {
	util::swap(fe[0], fe[2]);
	tfp->reset(fe);
      }
      
      // add face
      this->addFace(tfp);
      
      // add face to edges
      for(int j=0; j<3; j++) 
	fe[j]->addFace(tfp);
    }
    
    // finish
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
  }
  
  // construct mesh from an indexed triangle set
  // WORKS ONLY FOR CLOSED SURFACES SO FAR
  void TriangleMesh::createFromITS(const vector<Vec3d> &_v, 
				   const vector<int> &_ind) {
    
    // first we create a triangle mesh
    int nv = (int) _v.size();     // # vertices
    int nt = (int) _ind.size()/3; // # triangles
    
    // create vertices
    vector<TriangleVertexPointer> v(nv);
    for(int i=0; i<nv; i++)
      v[i] = this->newVertex(_v[i]);
    
    // create edges and faces
    // use an edge map to avoid redundant edges
    typedef pair<int, int> intpair;
    typedef map<intpair, TriangleEdgePointer> edgemap;
    edgemap emap;
    vector<TriangleEdgePointer> tmp(3);
    vector<TriangleFacePointer> f(nt);
    
    for(int i=0, j=0; i<nt; i++, j+=3) {
      // edge 0
      int i1 = _ind[j];
      int i2 = _ind[j+1];
      util::sort2(i1,i2); // smallest is i1
      intpair cur = make_pair(i1,i2);
      edgemap::iterator it = emap.find( cur );
      
      if( it == emap.end() ) { // new edge
	tmp[0] = this->newEdge(v[i1], v[i2]);
        emap[ cur ] = tmp[0];
      }
      else { // already added
	tmp[0] = emap[ cur ];
      }

      // edge 1
      i1 = _ind[j+1];
      i2 = _ind[j+2];
      util::sort2(i1,i2);
      cur = make_pair(i1,i2);
      it = emap.find( cur );
      
      if( it == emap.end() ) {
	tmp[1] = this->newEdge(v[i1], v[i2]);
        emap[ cur ] = tmp[1];
      }
      else {
	tmp[1] = emap[ cur ];
      }
      
      // edge 2
      i1 = _ind[j+2];
      i2 = _ind[j];
      util::sort2(i1,i2);
      cur = make_pair(i1,i2);
      it = emap.find( cur );
      
      if( it == emap.end() ) {
	tmp[2] = this->newEdge(v[i1], v[i2]);
        emap[ cur ] = tmp[2];
      }
      else {
	tmp[2] = emap[ cur ];
      }
      
      // the face
      f[i] = this->newFace(tmp);
      tmp[0]->addFace(f[i]);
      tmp[1]->addFace(f[i]);
      tmp[2]->addFace(f[i]);
    }
    
    // add all vertices, faces, and edges
    
    // vertices
    for(int i=0; i<nv; i++) {
      this->addVertex(v[i]);
    }
    
    // faces
    for(int i=0; i<nt; i++) {
      this->addFace(f[i]);
    }
    
    // edges
    for(edgemap::iterator it=emap.begin(); it != emap.end(); it++) {
      this->addEdge(it->second);
    }
    
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
  }
  
  /// create sphere
  void TriangleMesh::createSphere(double radius, int refinements) {
    
    int elongation = 1;
    
    double hsize = radius*elongation*sin(constants::pi/7.0);
    double cpi6 = radius*cos(constants::pi/7.0);
    double a = 0.0;
    double b = constants::pi/5.0;
    double d = 2.0 * b;
    
    unsigned int i, j;
    
    // create 12 vertices
    vector<TriangleVertexPointer> v(12);
    
    // 2 poles
    v[0] = this->newVertex(0.0, 0.0,  elongation*radius);
    v[6] = this->newVertex(0.0, 0.0, -elongation*radius);
    
    // 10 other vertices
    for(i=1; i<6; i++, a+=d, b+=d) {
      v[i] = this->newVertex(cpi6*cos(a), cpi6*sin(a), hsize);
      v[i+6] = this->newVertex(cpi6*cos(b), cpi6*sin(b), -hsize);
    }
    
    // create 30 edges
    vector<TriangleEdgePointer> e(30);
    for(i=0, j=2; i<5; i++, j=(j==5) ? 1:(j+1)) {
      e[i] = this->newEdge(v[0], v[i+1]);
      e[i+5] = this->newEdge(v[6], v[i+7]);
      e[i+10] = this->newEdge(v[i+1], v[j]);
      e[i+15] = this->newEdge(v[i+7], v[j+6]);
      e[2*(i+10)] = this->newEdge(v[i+1], v[i+7]);
      e[2*(i+10)+1] = this->newEdge(v[i+7], v[j]);
    }
    
    // create 20 faces (icosahedron)
    vector<TriangleFacePointer> f(20);
    vector<TriangleEdgePointer> ef(3);
    for(i=0, j=1; i<5; i++, j=(j==4) ? 0:(j+1)) {
      ef[0] = e[i]; ef[1] = e[i+10]; ef[2] = e[j];
      f[i] = this->newFace(ef);
      e[i]->addFace(f[i]);
      e[i+10]->addFace(f[i]);
      e[j]->addFace(f[i]);
      
      ef[0] = e[j+5]; ef[1] = e[i+15]; ef[2] = e[i+5];
      f[i+5] = this->newFace(ef);
      e[i+5]->addFace(f[i+5]);
      e[i+15]->addFace(f[i+5]);
      e[j+5]->addFace(f[i+5]);

      ef[0] = e[i+10]; ef[1] = e[2*i+20]; ef[2] = e[2*i+21];
      f[i+10] = this->newFace(ef);
      e[i+10]->addFace(f[i+10]);
      e[2*i+20]->addFace(f[i+10]);
      e[2*i+21]->addFace(f[i+10]);
      
      ef[0] = e[2*i+21]; ef[1] = e[i+15]; ef[2] = e[2*j+20];
      f[i+15] = this->newFace(ef);
      e[i+15]->addFace(f[i+15]);
      e[2*j+20]->addFace(f[i+15]);
      e[2*i+21]->addFace(f[i+15]);
    }
    
    // add everything
    for(i=0; i<12; i++) this->addVertex(v[i]);
    for(i=0; i<30; i++) this->addEdge(e[i]);
    for(i=0; i<20; i++) this->addFace(f[i]);

    cerr << "Created sphere with " 
	 << this->nVertices() << " vertices, " 
	 << this->nEdges() << " edges, and " 
	 << this->nFaces() << " faces." << endl;
    
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
    
    // refine mesh
    for(int R=0; R<refinements; R++)
      refine();
    
    // project all vertices on sphere
    for(int i=0; i<this->nVertices(); i++) {
      Vec3d v = this->getVertex(i)->getPosition();
      double a = sqrt(radius*radius /
		      (util::square(v.x) + util::square(v.y) +
		       util::square(v.z)/(elongation*elongation)));
      this->getVertex(i)->setPosition(a*v);
    }

    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
  }
  
  /// create tetrahedron
  void TriangleMesh::createTetrahedron(double size, int refinements) {
    
    int elongation = 2;
    
    // create 4 vertices 
    TriangleVertexPointer v1 = this->newVertex(0.0, 0.0, 0.0);
    TriangleVertexPointer v2 = this->newVertex(size, 0.0, 0.0);
    TriangleVertexPointer v3 = this->newVertex(0.0, size, 0.0);
    TriangleVertexPointer v4 = this->newVertex(0.0, 0.0, size*elongation);
    
    // create 6 edges
    TriangleEdgePointer e1 = this->newEdge(v1, v2);
    TriangleEdgePointer e2 = this->newEdge(v1, v3);
    TriangleEdgePointer e3 = this->newEdge(v1, v4);
    TriangleEdgePointer e4 = this->newEdge(v2, v3);
    TriangleEdgePointer e5 = this->newEdge(v2, v4);
    TriangleEdgePointer e6 = this->newEdge(v3, v4);
    
    // create 4 faces
    vector<TriangleEdgePointer> e(3);
    e[0] = e1; e[1] = e4; e[2] = e2;
    TriangleFacePointer f1 = this->newFace(e);
    e[0]->addFace(f1);
    e[1]->addFace(f1);
    e[2]->addFace(f1);
    e[0] = e3; e[1] = e5; e[2] = e1;
    TriangleFacePointer f2 = this->newFace(e);
    e[0]->addFace(f2);
    e[1]->addFace(f2);
    e[2]->addFace(f2);
    e[0] = e3; e[1] = e2; e[2] = e6;
    TriangleFacePointer f3 = this->newFace(e);
    e[0]->addFace(f3);
    e[1]->addFace(f3);
    e[2]->addFace(f3);
    e[0] = e4; e[1] = e5; e[2] = e6;
    TriangleFacePointer f4 = this->newFace(e);
    e[0]->addFace(f4);
    e[1]->addFace(f4);
    e[2]->addFace(f4);
    
    // add everything
    this->addVertex(v1);
    this->addVertex(v2);
    this->addVertex(v3);
    this->addVertex(v4);
    this->addEdge(e1);
    this->addEdge(e2);
    this->addEdge(e3);
    this->addEdge(e4);
    this->addEdge(e5);
    this->addEdge(e6);
    this->addFace(f1);
    this->addFace(f2);
    this->addFace(f3);
    this->addFace(f4);
    
    cerr << "Created tetrahedron with " 
	 << this->nVertices() << " vertices, " 
	 << this->nEdges() << " edges, and " 
	 << this->nFaces() << " faces." << endl;
    
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
    
    for(int R=0; R<refinements; R++) 
      refine();
    
    this->sortEdges();
    this->computeFaceCenters();
    this->computeNormals();
  }
    
  /// compute centers for all faces
  void TriangleMesh::computeFaceCenters() {
    for(int i=0; i<this->nFaces(); i++) {
      this->getFace(i)->computeCenter();
    }
  }
  
  /// compute normals for all faces and vertices
  void TriangleMesh::computeNormals() {
    // face normals computed first since the vertex normal is
    // the average of them 
    for(int i=0; i<this->nFaces(); i++) {
      this->getFace(i)->computeNormal();
    }
    
    // vertex normals
    for(int i=0; i<this->nVertices(); i++) {
      this->getVertex(i)->computeNormal();
    }
    
  }
  
  // compute center of mass
  Vec3d TriangleMesh::computeCenterOfMass() const {
    Vec3d cm(0,0,0);
    for(int i=0; i<this->nVertices(); i++)
      cm += this->getVertex(i)->getPosition();
    cm /= double(this->nVertices());
    return cm;
  } 
  
  // the volume is computed by adding the volumes of the tetrahedrons  
  // composed from the face center and two neighboring vertices  
  // NOT IMPLEMENTED YET
  double TriangleMesh::computeVolume() const {    
    return 0.0;
  }
  
  // compute area
  double TriangleMesh::computeArea() const {
    double area=0.0;
    for(int i=0; i<this->nFaces(); i++) {
      area += this->getFace(i)->computeArea();
    }
    return area;
  }
  
  /// transformation
  void TriangleMesh::transform(const Matrix4d &M) {
    for(int i=0; i<nVertices(); i++)
      getVertex(i)->transform(M);
  }
  
  // render with OpenGL
  void TriangleMesh::renderGL(int mode) const {
    
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    
    // wireframe rendering
    if( mode == 0 ) { 
      glLineWidth(1.0);
      
      glBegin(GL_LINES);	
      for(int i=0; i<this->nEdges(); i++) {
	TriangleEdgePointer e = this->getEdge(i);
	glNormal3dv(e->getFirstVertex()->getNormal());
	glVertex3dv(e->getFirstVertex()->getPosition());
	glNormal3dv(e->getSecondVertex()->getNormal());
	glVertex3dv(e->getSecondVertex()->getPosition());
      }
      glEnd();
    }
    
    // flat rendering
    else if( mode == 1 ) {
      for(int i=0; i<this->nFaces(); i++) {
	// render polygon	
	TriangleFacePointer f = this->getFace(i);
	TriangleFace::VertexIterator vi = f->vbegin();
	glBegin(GL_TRIANGLES);
	glNormal3dv(f->getNormal());
	do {
	  glVertex3dv(vi.current()->getPosition());	  
	} while( vi.next() );
	glEnd();
      }
    }
    
    // smooth rendering
    else if( mode == 2 ) {
      for(int i=0; i<this->nFaces(); i++) {
	// render polygon	
	TriangleFacePointer f = this->getFace(i);
	TriangleFace::VertexIterator vi = f->vbegin();
	glBegin(GL_TRIANGLES);
	do {
	  glNormal3dv(vi.current()->getNormal());	
	  glVertex3dv(vi.current()->getPosition());	  
	} while( vi.next() );
	glEnd();
      }
    }
    
    glPopAttrib();        
    
  }

}
