///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_TRIANGLE_MESH_HH_
#define _WISH_TRIANGLE_MESH_HH_

#include "wishTriangleVertex.hh"
#include "wishSimplexMesh.hh"

namespace wish {
    
  // ### TriangleMesh ###
  
  class TriangleMesh {

  protected:
    
    /// vertices
    vector<TriangleVertexPointer> vertices;
    
    /// edges
    vector<TriangleEdgePointer> edges;
    
    /// faces
    vector<TriangleFacePointer> faces;
    
  public:
    
    /// constructor
    TriangleMesh(int size_hint=0) :
      vertices( vector<TriangleVertexPointer>(0) ),
      edges( vector<TriangleEdgePointer>(0) ),
      faces( vector<TriangleFacePointer>(0) )
    {
      if( size_hint > 0 ) {
	vertices.reserve(size_hint);
	edges.reserve(size_hint);
	faces.reserve(size_hint);
      }
    }
    
    /// deletes all associated edges, vertices, and faces
    virtual ~TriangleMesh() { deleteMesh(); }
    
    /// the function called in the destructor
    void deleteMesh();
    
    /// refine mesh
    void refine();
    
    // create triangle mesh from simplex mesh
    void createFromSimplexMesh(SimplexMeshPointer tmesh);
    
    // construct mesh from an indexed triangle set
    // WORKS ONLY FOR CLOSED SURFACES SO FAR
    void createFromITS(const vector<Vec3d> &_v, 
		       const vector<int> &_ind);
    
    /// create a sphere
    void createSphere(double radius, int refinements=1);
    
    /// create a tetrahedron
    void createTetrahedron(double size, int refinements=1);
    
    /// number of vertices
    int nVertices() const { 
      return (int) vertices.size(); 
    }
    /// number of edges
    int nEdges() const { 
      return (int) edges.size(); 
    }
    /// returns number of faces
    int nFaces() const { 
      return (int) faces.size(); 
    }
    
    /// add a vertex
    void addVertex(TriangleVertexPointer v) { 
      vertices.push_back(v); 
    }
    /// add an edge
    void addEdge(TriangleEdgePointer e) { 
      edges.push_back(e); 
    }
    /// add a face
    void addFace(TriangleFacePointer f) { 
      faces.push_back(f); 
    }
    
    /// get vertex
    TriangleVertexPointer getVertex(int i) const {
      if( 0 <= i && i < nVertices() )
	return vertices[i];
      else
	return 0;
    }
    
    /// get edge
    TriangleEdgePointer getEdge(int i) const {
      if( 0 <= i && i < nEdges() )
	return edges[i];
      else
	return 0;
    }
    
    /// get face
    TriangleFacePointer getFace(int i) const {
      if( 0 <= i && i < nFaces() )
	return faces[i];
      else
	return 0;
    }
    
    /// sort edges so that two edges that share a face are consecutive
    void sortEdges();
    
    /// compute centers for all faces
    void computeFaceCenters();
    
    /// compute normals for all faces and vertices
    void computeNormals();
    
    /// update geometry when model was transformed
    void updateGeometry() { 
      computeFaceCenters();       
      computeNormals();
    }
    
    /// compute centerOfMass (old center)
    Vec3d computeCenterOfMass() const;
    
    /// compute enclosed volume (old volume)
    double computeVolume() const;
    
    /// compute surface area (old area)
    double computeArea() const;
    
    /// transformation
    void transform(const Matrix4d &M);
    
    // render mesh with OpenGL
    void renderGL(int mode=1) const;
    
  protected:
    /// internal vertex generators that generates a new vertex
    /// belonging to this mesh
    TriangleVertexPointer newVertex(double _x=0.0, double _y=0.0, double _z=0.0) {
      return new TriangleVertex(Vec3d(_x, _y, _z), this);
    }
    TriangleVertexPointer newVertex(const Vec3d& _position) {
      return new TriangleVertex(_position, this);
    }
    
    // internal edge generator
    TriangleEdgePointer newEdge(TriangleVertexPointer v1, 
				TriangleVertexPointer v2,
				TriangleFacePointer f1 = 0,
				TriangleFacePointer f2 = 0) {
      return new TriangleEdge(v1, v2, f1, f2);
    }
    
    /// internal face generator
    TriangleFacePointer newFace(vector<TriangleEdgePointer> _edges,
				bool _empty=false) {
      return new TriangleFace(_edges, _empty);
    }
    
  };
}

#endif
