///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishTriangleVertex.hh"

namespace wish {
  
  // ### TriangleVertex ###
  
  /// compute normal
  Vec3d TriangleVertex::computeNormal() {
    TriangleVertex::FaceIterator fi = this->fbegin();
    this->normal = Vec3d(0,0,0);
    do {
      this->normal += fi.current()->computeNormal();
    } while( fi.next() );
    this->normal.normalize();
    return this->normal;
  }
  
  /// add edge
  void TriangleVertex::addEdge(TriangleEdgePointer _edge) {
    
    edges.push_back(_edge);
    
    // add neighbors
    // we do not select ourself as a neighbor
    TriangleVertexPointer v = _edge->getVertex(0);
    if( v == this )
      v = _edge->getVertex(1);
    neighbors.push_back(v);
  }
  
  // sort edge list so that two edges belonging to the 
  // same face are consecutive in the list  
  void TriangleVertex::sortEdges() {
    int n = this->nEdges();
    vector<TriangleEdgePointer> temp_edges = edges;
    edges.clear();  
    edges.push_back(temp_edges[0]);
    
    TriangleFacePointer first = temp_edges[0]->getFirstFace();
    TriangleFacePointer second = temp_edges[0]->getSecondFace();
    temp_edges[0] = 0;
    for(int i=1; i<n; i++) {
      for(int j=0; j<n; j++) {
	if((temp_edges[j] != 0) &&
	   (temp_edges[j]->getFirstFace() == first  || 
	    temp_edges[j]->getSecondFace() == first || 
	    temp_edges[j]->getFirstFace() == second || 
	    temp_edges[j]->getSecondFace() == second)) {
	  first = temp_edges[j]->getFirstFace();
	  second = temp_edges[j]->getSecondFace();
	  edges.push_back(temp_edges[j]);
	  temp_edges[j] = 0;	  
	  break;
	}
      }
    }
  }
  
  // ### TriangleVertex::FaceIterator ###
  // first face
  TriangleFacePointer TriangleVertex::FaceIterator::first() {
    edge_index = 0;	
    cur = vertex->getEdge(edge_index)->getFace(0);
    if( cur == vertex->getEdge(edge_index+1)->getFace(0) ||
	cur == vertex->getEdge(edge_index+1)->getFace(1) )
      cur = vertex->getEdge(edge_index)->getFace(1);
    return cur;
  }
  
  // next face
  TriangleFacePointer TriangleVertex::FaceIterator::next() {
    if( edge_index >= vertex->nEdges()-1 )
      cur = 0;
    else if( vertex->getEdge(edge_index)->getFace(0) == cur )
      cur = vertex->getEdge(edge_index)->getFace(1);
    else
      cur = vertex->getEdge(edge_index)->getFace(0);
    edge_index++;	
    return cur;
  }
  
  // ### TriangleEdge ###
  
  /// add face
  void TriangleEdge::addFace(TriangleFacePointer _face) {
    if( faces[0] == 0 )
      faces[0] = _face;
    else if( faces[1] == 0 ) {
      faces[1] = _face;
    }
    else {
      cerr << "TriangleEdge::addFace: Error, both faces already set!" << endl;
    }
  }

  // ### TriangleFace ###
  
  /// compute center
  Vec3d TriangleFace::computeCenter() {
    int n = this->nVertices(); // should be 3, fix later
    TriangleFace::VertexIterator vi = this->vbegin();
    this->center = Vec3d(0,0,0);
    do {
      this->center += vi.current()->getPosition();
    } while( vi.next() );
    this->center /= double(n);
    return this->center;
  }
  
  /// compute normal
  Vec3d TriangleFace::computeNormal() {
    TriangleFace::VertexIterator vi = this->vbegin();
    Vec3d p0 = vi.current()->getPosition();
    vi.next();
    Vec3d p1 = vi.current()->getPosition();
    vi.next();
    Vec3d p2 = vi.current()->getPosition();
    
    this->normal = (p1-p0).cross(p2-p0);
    this->normal.normalize();
    return this->normal;
  }
  
  /// compute area
  double TriangleFace::computeArea() {
    TriangleFace::VertexIterator vi = this->vbegin();
    Vec3d p0 = vi.current()->getPosition();
    vi.next();
    Vec3d p1 = vi.current()->getPosition();
    vi.next();
    Vec3d p2 = vi.current()->getPosition();
    
    this->area = ((p1-p0).cross(p2-p0)).norm()/2.0;
    return this->area;
  }
  
  // ### TriangleFace::VertexIterator ###
  
  // the first vertex
  TriangleVertexPointer TriangleFace::VertexIterator::first() {
    edge_index = 0;	
    cur = face->getEdge(edge_index)->getVertex(0);
    if( cur == face->getEdge(edge_index+1)->getVertex(0) ||
	cur == face->getEdge(edge_index+1)->getVertex(1) )
      cur = face->getEdge(edge_index)->getVertex(1);
    return cur;
  }
  
  // the next vertex
  TriangleVertexPointer TriangleFace::VertexIterator::next() {
    if( edge_index >= face->nEdges()-1 )
      cur = 0;
    else if( face->getEdge(edge_index)->getVertex(0) == cur )
      cur = face->getEdge(edge_index)->getVertex(1);
    else
      cur = face->getEdge(edge_index)->getVertex(0);
    edge_index++;	
    return cur;
  }
  
}
