///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000-2004 CREATIS/I3S, CNRS, France
// 
// The copyright information above is included because this implementation 
// was initially inspired by the CREA library designed by Johan Montagnat,
// also licenced with GNU GPL.
//  
// The major part of the code has been rewritten.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_TRIANGLE_VERTEX_HH_
#define _WISH_TRIANGLE_VERTEX_HH_

#include "wish.hh"
#include "wishImage.hh"
#include "wishLinalg.hh"

namespace wish {
  
  // predefinitions
  class TriangleMesh;
  class TriangleVertex;
  class TriangleEdge;
  class TriangleFace;
  
  // typedefs
  typedef TriangleMesh*   TriangleMeshPointer;
  typedef TriangleVertex* TriangleVertexPointer;
  typedef TriangleEdge*   TriangleEdgePointer;
  typedef TriangleFace*   TriangleFacePointer;
  
  // ### TriangleVertex ###
  
  class TriangleVertex {
  protected:
    
    /// the mesh
    TriangleMeshPointer mesh;
    
    // ### Vertex parameters
    /// edges
    vector<TriangleEdgePointer> edges;
    /// neighbors
    vector<TriangleVertexPointer> neighbors;
    /// position
    Vec3d position;
    /// normal
    Vec3d normal;
    
  public:
    /// constructor
    TriangleVertex(const Vec3d &_position=Vec3d(0,0,0), 
		  TriangleMeshPointer _mesh=0) :
      position(_position),
      mesh(_mesh),
      
      edges( vector<TriangleEdgePointer>(0) ),
      neighbors( vector<TriangleVertexPointer>(0) ),
      normal( Vec3d(0,0,0) )
    {
    }
    
    /// destructor
    ~TriangleVertex() {}
    
  private:
    /// unimplemented
    TriangleVertex(const TriangleVertex& _v) {}   
    TriangleVertex& operator=(const TriangleVertex &_v) {}
    
  public:
    
    /// get mesh
    TriangleMeshPointer getMesh() const {
      return mesh;
    }
    
    /// get nr of edges
    int nEdges() const {
      return (int) edges.size();
    }
    
    /// get nr of neighbors
    int nNeighbors() const {
      return (int) neighbors.size();
    }
    
    /// add edge
    void addEdge(TriangleEdgePointer _edge);
    
    /// clear edges
    void clearEdges() {
      edges.clear();
    }
    
    /// clear neighbors
    void clearNeighbors() {
      neighbors.clear();
    }
    
    // sort edge list so that two edges belonging to the 
    // same face are consecutive in the list  
    void sortEdges();
    
    /// get edge
    TriangleEdgePointer getEdge(int i) const {
      if( 0 <= i && i < this->nEdges() ) 
	return edges[i];
      else
	return 0;
    }
    
    /// get neighbor
    TriangleVertexPointer getNeighbor(int i) const {
      if( 0 <= i && i < this->nNeighbors() ) 
	return neighbors[i];
      else
	return 0;
    }

    /// get/set the position
    Vec3d getPosition() const {
      return position; 
    }
    void setPosition(const Vec3d &_pos) { 
      position = _pos; 
    }
    void setPosition(double x, double y, double z) { 
      position = Vec3d(x,y,z); 
    }
    void addPosition(const Vec3d &_pos) { 
      position += _pos; 
    }
    void addPosition(double x, double y, double z) { 
      position += Vec3d(x,y,z); 
    }
    
    // compute normal
    Vec3d computeNormal();
    
    // get normal
    Vec3d getNormal() const {
      return normal;
    }
    
    /// transformation
    void transform(const Matrix4d &M) {
      position = M*position;
      normal = Vec3d(M*Vec4d(normal,0.0));
    }
    
    // ### TriangleVertex::FaceIterator ###
    struct FaceIterator {
      // the vertex
      TriangleVertexPointer vertex;
      
      // where we are
      TriangleFacePointer cur;
      int edge_index;
      
      // constructor 
      FaceIterator(TriangleVertexPointer _vertex) : vertex(_vertex) {
	first();
      }
      // the first one
      // here we check that we select the vertex that not belong
      // to the next edge
      TriangleFacePointer first();

      // the next one
      TriangleFacePointer next();
      
      // the current one
      TriangleFacePointer current() {
	return cur;
      }
    };
    
    /// iterator starting at first vertex
    FaceIterator fbegin() {
      return FaceIterator(this);
    }
    
    
  };
  
  // ### TriangleEdge ###
  
  class TriangleEdge {
  protected:
    /// vertices
    vector<TriangleVertexPointer> vertices;
    
    /// faces
    vector<TriangleFacePointer> faces;
    
  public:
    
    /// flag used in topology algorithms  
    mutable int flag;
    
    // build an edge between two vertices
    TriangleEdge(TriangleVertexPointer _v1, 
		 TriangleVertexPointer _v2,
		 TriangleFacePointer _f1 = 0, 
		 TriangleFacePointer _f2 = 0) :
      vertices( vector<TriangleVertexPointer>(2,(TriangleVertexPointer)0) ),
      faces( vector<TriangleFacePointer>(2,(TriangleFacePointer)0) )
    {
      vertices[0] = _v1;
      vertices[1] = _v2;
      vertices[0]->addEdge(this);
      vertices[1]->addEdge(this);
    }
    
    /// get vertex
    TriangleVertexPointer getVertex(int i) const {
      if( 0 <= i && i <= 1 )
	return vertices[i];
      else
	return 0;
    }
    TriangleVertexPointer getFirstVertex() const {
      return vertices[0];
    }
    TriangleVertexPointer getSecondVertex() const {
      return vertices[1];
    }
    
    /// get face
    TriangleFacePointer getFace(int i) { 
      if( 0 <= i && i <= 1 )
	return faces[i];
      else
	return 0;
    }
    TriangleFacePointer getFirstFace() const {
      return faces[0];
    }
    TriangleFacePointer getSecondFace() const {
      return faces[1];
    }

    /// add face
    void addFace(TriangleFacePointer _face);

  };
  
  // ### TriangleFace ###
  
  class TriangleFace {
  protected:
    /// edges (should be ordered around the face)
    vector<TriangleEdgePointer> edges;
    
    /// empty or not
    bool empty;
    
    /// center
    Vec3d center;
    
    /// normal
    Vec3d normal;
    
    /// area
    double area;
    
  public:
    
    /// flag used in topology algorithms
    mutable int flag;
    
    /// constructor
    TriangleFace(const vector<TriangleEdgePointer> &_edges, 
		 bool _empty=false) :
      edges(_edges),
      empty(_empty),
      center( Vec3d(0,0,0) ),
      normal( Vec3d(0,0,0) ),
      area(0.0) 
    {
    }
    
    /// reset face to a new set of edges
    void reset(vector<TriangleEdgePointer> _edges, bool _empty=false) {
      empty = _empty;
      edges = _edges;
    }
    
    /// compute center
    Vec3d computeCenter();
    
    /// get center
    Vec3d getCenter() const { 
      return center; 
    }
    
    /// compute normal
    Vec3d computeNormal();
    
    /// get normal
    Vec3d getNormal() const {
      return normal;
    }
    
    /// compute area
    double computeArea();
    
    /// get area
    double getArea() const {
      return area;
    }
    
    /// get number of edges
    int nEdges() const { 
      return (int) edges.size(); 
    }
    
    /// get number of vertices (same as number of edges)
    int nVertices() const { 
      return nEdges();
    }
    
    /// get/set empty
    bool getEmpty() const { 
      return empty; 
    }
    void setEmpty(bool _h) {
      empty = _h;
    }
    
    /// get edge
    TriangleEdgePointer getEdge(int i) const {
      if( 0<=i && i<this->nEdges() )
	return edges[i];
      else
	return 0;
    }
    
    // ### TriangleFace::VertexIterator ###
    struct VertexIterator {
      // the face
      TriangleFacePointer face;
      
      // where we are
      TriangleVertexPointer cur;
      int edge_index;
      
      // constructor 
      VertexIterator(TriangleFacePointer _face) : face(_face) {
	first();
      }
      // the first one
      // here we check that we select the vertex that not belong
      // to the next edge
      TriangleVertexPointer first();
      
      // the next one
      TriangleVertexPointer next();
      
      // the current one
      TriangleVertexPointer current() {
	return cur;
      }
    };
    
    /// iterator starting at first vertex
    VertexIterator vbegin() {
      return VertexIterator(this);
    }
    
  };
  
}

#endif
