///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishVTKFileReader.hh"

namespace wish {
  
  bool VTKFileReader::readHeader() {
    
    //   # vtk DataFile Version x.x
    //   Some information about the file
    //   BINARY
    //   DATASET STRUCTURED_POINTS
    //   DIMENSIONS 128 128 128
    //   ORIGIN 0.0 0.0 0.0
    //   SPACING 1.0 1.0 1.0
    //   POINT_DATA 2097152
    //   SCALARS image_data unsigned_char
    //   LOOKUP_TABLE default
    //   raw data........
    
    // open file, use binary mode (necessary on windows)
    ifstream is;
    is.open( this->m_filename.c_str(), ios::in | ios::binary );
    if( !is ) {
      cerr << "Unable to open file: " << this->m_filename << endl;
      return false;
    }
    
    // check if it is a VTK file
    string checkvtk;
    is >> checkvtk; // '#'
    is >> checkvtk; // vtk
    if( !(checkvtk == "vtk" || checkvtk =="VTK") )
      return false;
    
    // read the rest of the first line
    char info[1000];    
    is.getline(info,1000);
    
    // read info
    is.getline(info,1000);
    this->m_info.infostr = string(info);
    
    string str;
    // read the BINARY/ASCII field
    do {
      is >> str;
    } while( str != "ASCII" && str != "BINARY" );
    if( str=="ASCII" )
      this->m_binary = false;
    else
      this->m_binary = true;
    
    // read the DIMENSIONS field
    do {
      is >> str;
    } while( str != "DIMENSIONS" );
    is >> this->m_info.dim;
    
    // read the SPACING field
    do {
      is >> str;
    } while( str != "SPACING" );
    is >> this->m_info.spc;
    
    // read the SCALARS/VECTORS field
    string typestring;
    do {
      is >> str;
    } while( str != "SCALARS" && str != "COLOR_SCALARS" && str != "VECTORS" );
    string scv = str;
    is >> str; // the name
    is >> typestring; // the type (or numComp (3 or 4) if COLOR_SCALARS)
    
    if( scv == "SCALARS" ) {
      if( typestring == "unsigned_char" ) 
	this->m_info.datatype = "uint8";
      else if( typestring == "char" ) 
	this->m_info.datatype = "int8";
      else if( typestring == "unsigned_short" ) 
	this->m_info.datatype = "uint16";
      else if( typestring == "short" ) 
	this->m_info.datatype = "int16";
      else if( typestring == "unsigned_int" ) 
	this->m_info.datatype = "uint32";
      else if( typestring == "int" ) 
	this->m_info.datatype = "int32";
      else if( typestring == "float" ) 
	this->m_info.datatype = "float";
      else if( typestring == "double" ) 
	this->m_info.datatype = "double";
      else
	return false;
      
      // read the lookup table
      do {
	is >> str;
      }	while( str != "LOOKUP_TABLE" );
      is >> str; // name (default)
    }
    else if( scv == "COLOR_SCALARS" ) {
      if( typestring == "3" ) 
	this->m_info.datatype = "RGB8";
      else if( typestring == "4" ) 
	this->m_info.datatype = "RGBA8";
      else
	return false;
    }
    else { // VECTORS
      if( typestring == "float" ) 
	this->m_info.datatype = "Vec3f";
      else
	return false;
    }
    
    // read end of line
    is >> ws;
    
    // header size in bytes
    this->m_info.headersize = (int) is.tellg();
    
    is.close();
    
    return true;
  }
  
  // help functions
  template<class T> 
  void readVTKASCII(ifstream &is, T *array, int n) {
    for(int i=0; i<n; i++)
      is >> array[i];
  }
  
  template<> 
  void readVTKASCII<uint8>(ifstream &is, uint8 *array, int n) {
    for(int i=0; i<n; i++) {
      float val;
      is >> val;
      array[i] = uint8( val );
    }
  }
  
  template<> 
  void readVTKASCII<int8>(ifstream &is, int8 *array, int n) {
    for(int i=0; i<n; i++) {
      float val;
      is >> val;
      array[i] = int8( val );
    }  
  }
  
  template<> 
  void readVTKASCII<RGB8>(ifstream &is, RGB8 *array, int n) {
    for(int i=0; i<n; i++) {
      RGBf col;
      is >> col;
      array[i] = RGB8( col );
    }  
  }
  
  template<> 
  void readVTKASCII<RGBA8>(ifstream &is, RGBA8 *array, int n) {
    for(int i=0; i<n; i++) {
      RGBAf col;
      is >> col;
      array[i] = RGBA8( col );
    }    
  }
  
  template<class T> 
  void readVTKBinary(ifstream &is, T *array, int n) {
    util::readBE( is, array, n );
  }
  
  template<> 
  void readVTKBinary<Vec3f>(ifstream &is, Vec3f *array, int n) {
    util::readBE( is, (WISH_FLOAT*) array, n*3 );
  }
  
  template<> 
  void readVTKBinary<RGB8>(ifstream &is, RGB8 *array, int n) {
    is.read( (char *) array, n*3 );
  }
  
  template<> 
  void readVTKBinary<RGBA8>(ifstream &is, RGBA8 *array, int n) {
    is.read( (char *) array, n*4 );
  }
  
  bool VTKFileReader::readData() {
    
    // open file, use binary mode (necessary on windows)
    ifstream is;
    is.open( this->m_filename.c_str(), ios::in | ios::binary );
    if( !is ) {
      cerr << "Unable to open file: " << this->m_filename << endl;
      return false;
    }
    
    // jump to data
    is.seekg(this->m_info.headersize);
    
    // use header info to allocate a new image
    if( this->m_info.datatype == "uint8" ) {
      Image<uint8> *tmp = new Image<uint8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int8" ) {
      Image<int8> *tmp = new Image<int8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint16" ) {
      Image<uint16> *tmp = new Image<uint16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int16" ) {
      Image<int16> *tmp = new Image<int16>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "uint32" ) {
      Image<uint32> *tmp = new Image<uint32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "int32" ) {
      Image<int32> *tmp = new Image<int32>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "float" ) {
      Image<float> *tmp = new Image<float>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "double" ) {
      Image<double> *tmp = new Image<double>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "Vec3f" ) {
      Image<Vec3f> *tmp = new Image<Vec3f>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "RGB8" ) {
      Image<RGB8> *tmp = new Image<RGB8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else if( this->m_info.datatype == "RGBA8" ) {
      Image<RGBA8> *tmp = new Image<RGBA8>;
      tmp->allocate(this->m_info.dim);
      tmp->setSpacing(this->m_info.spc);
      tmp->setInfo(this->m_info.infostr);
      if( m_binary )
	readVTKBinary(is, tmp->getData(), tmp->totalSize());
      else
	readVTKASCII(is, tmp->getData(), tmp->totalSize());
      this->m_output = tmp;
    }
    else { // should not happen
      this->printSelf();
      cerr << ": Error reading file " << m_filename << ", bad datatype: " 
	   << this->m_info.datatype << endl;
      return false;
    }
    
    is.close();
    
    return true;
  }
  
}
