///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_VOLUME_HAPTICS_HH_
#define _WISH_VOLUME_HAPTICS_HH_

#include "wish.hh"
#include "wishutil.hh"
#include "wishLinalg.hh"

namespace wish {
  
  //! VolumeHaptics implements a version of the haptic algorithm described
  //! in Ikits et al., 
  //! "A constraint based technique for haptic volume exploration",
  //! IEEE Visualization 2003, pp. 263--269.
  //!
  //! One change is the addition of the directed force which 
  //! is based on the idea in Lundins WHC'05 paper. That algorithm
  //! is more general but the numerical solver that is needed is 
  //! a bit hard to implement so we stick to this for now. Some 
  //! initial work has been done in wishPrimitiveHaptics.hh
  class VolumeHaptics {
    
  public:
    
    typedef float Float;
    typedef Vec3<Float> vec3;
    typedef Matrix3<Float> mat3;
    
    // UNILATERAL means that the proxy is allowed to move freely in 
    // the positive directon of the axis but is controlled by the strength
    // (threshold) for movements in the negative direction
    //
    // BILATERAL means that proxy movement is controlled by the strength
    // for movements in any direction
    //
    // DIRECTED_FORCE means that the proxy is moved along the
    // positive axis if the strength is stronger than the one that
    // the user supplies
    //
    // NO_FORCE means that the proxy is moved to the probe
    // and therefore no force will be rendered
    typedef enum {UNILATERAL = 0, 
		  BILATERAL,
		  DIRECTED_FORCE,
		  NO_FORCE} MotionRuleType;
    
  protected:
    
    // probe and proxy 
    // ---------------
    // Local representation in world/local coordinates that are 
    // NOT scaled relative to the device coordinates. The things
    // you need to consider is
    // 
    //   1) the scale factor m_device_scale between the coordinate
    //      system you work in to device coordinates 
    //      in meters, i.e., m_proxy = your_proxy * m_device_scale;
    //    
    //   2) getForceFeedback() returns the force in
    //      the same coordinate space as the probe and proxy
    //      so you need to apply the correct transformation
    //      before sending te force to the device.
    // 
    vec3 m_probe;         
    vec3 m_old_probe;    
    vec3 m_proxy;        
    vec3 m_old_proxy;
    vec3 m_velocity;      // given in the real-time loop of the API 
    mat3 m_orientation;   
    Float m_device_scale;
    
    vec3  m_lrf[3];
    MotionRuleType m_mrl[3];
    Float m_str[3];
    
    Float m_stiffness;
    Float m_damping;
    Float m_time;
    Float m_old_time;
    
    // these should not be called explicitly,
    // the proxy is moved within updateProxy()
    
    // set proxy position
    void setProxy(const vec3 &px) {
      m_old_proxy = m_proxy;
      m_proxy = px;
    }
    
    // set proxy position
    void setProxy(Float x, Float y, Float z) {
      setProxy(vec3(x,y,z));
    }
    
    // move proxy
    void moveProxy(const vec3 &dp) {
      setProxy(m_proxy+dp);
    }
    
    // move proxy
    void moveProxy(Float dx, Float dy, Float dz) {
      moveProxy(vec3(dx,dy,dz));
    }
    
    // move proxy
    void moveProxy(const vec3 &dir, Float dist) {
      moveProxy(dist*dir);
    }

    // help function for computing proxy movement dp along lrf axis i
    // dx is the current displacement along axis i
    void computeDP(int i, Float *dx, Float *dp) const {
      
      // component i
      if( m_mrl[i] == UNILATERAL ) {
	dp[i] = (dx[i]>0.0) ? dx[i] : -util::maxOf(-dx[i]-m_str[i],Float(0.0));
      }
      else if( m_mrl[i] == BILATERAL ) {
	dp[i] = 
	  Float(util::sign(dx[i]))*
	  util::maxOf(std::fabs(dx[i])-m_str[i], Float(0.0));
      }

      else if( m_mrl[i] == DIRECTED_FORCE ) {
	dp[i]  = m_str[i]+dx[i];
      }
      else { // NO_FORCE
	dp[i] = dx[i];
      }
    }
    
  public:
    
    VolumeHaptics() : 
      m_probe( vec3(0,0,0) ),
      m_old_probe( vec3(0,0,0) ),
      m_proxy( vec3(0,0,0) ),
      m_old_proxy( vec3(0,0,0) ),
      m_velocity( vec3(0,0,0) ),
      //m_orientation( vec3(0,0,0) ),
      m_stiffness( Float(400.0) ),
      m_device_scale( Float(1.0) ),
      m_damping( Float(0.0) ), 
      m_time( Float(0.0) ),
      m_old_time( Float(0.0) )
    {
      setLocalReferenceFrame(vec3(1,0,0), vec3(0,1,0), vec3(0,0,1));
      setMotionRuleTypes(NO_FORCE, NO_FORCE, NO_FORCE);
      setMotionRuleStrengths(Float(0.0), Float(0.0), Float(0.0));
    }
    
    // set probe position (in world coordinates)
    void setProbe(const vec3 &pb) {
      m_old_probe = m_probe;
      m_probe = m_device_scale*pb;
    }
    
    // set probe position (in world coordinates)
    void setProbe(Float x, Float y, Float z) {
      setProbe(vec3(x,y,z));
    }
    
    // set time
    void setTime(Float _time) {
      m_old_time = m_time;
      m_time = _time;
    }
    
    // get time
    Float getTime() {
      return m_time;
    }
    
    // get probe position
    vec3 getProbe() const {
      return (1/m_device_scale)*m_probe;
    }
    
    // get proxy position in world coordinates
    vec3 getProxy() const {
      return (1/m_device_scale)*m_proxy;
    }
    
    // set velocity
    void setVelocity(const vec3 &_velocity) {
      m_velocity = m_device_scale*_velocity;
    }
    
    // set velocity
    void setVelocity(Float x, Float y, Float z) {
      setVelocity( vec3(x,y,z) );
    }
    
    // get velocity
    vec3 getVelocity() {
      return (1/m_device_scale)*m_velocity;
    }
    
    // set orientation
    void setOrientation(const mat3 &_orientation) {
      m_orientation = _orientation;
    }
    
//     // set orientation
//     void setOrientation(Float x, Float y, Float z) {
//       setOrientation( vec3(x,y,z) );
//     }
    
    // get orientation
    mat3 getOrientation() {
      return m_orientation;
    }
    
    // set device scale factor
    void setDeviceScale(const Float &_device_scale) {
      m_device_scale = _device_scale;
    }
    
    // get device scale factor
    Float getDeviceScale() {
      return m_device_scale;
    }
    
    // set local reference frame (in world coordinates)
    //
    // important that it is orthonormal
    void setLocalReferenceFrame(const vec3 &e0,
				const vec3 &e1,
				const vec3 &e2) {
      m_lrf[0] = e0;
      m_lrf[1] = e1;
      m_lrf[2] = e2;
    }
    
    void setLocalReferenceFrame(int i, const vec3 &e) {
      m_lrf[i] = e;
    }
    
    void setLRF(const vec3 &e0,
		const vec3 &e1,
		const vec3 &e2) {
      setLocalReferenceFrame(e0,e1,e2);
    }
    
    void setLRF(int i,
		const vec3 &e) {
      setLocalReferenceFrame(i, e);
    }
    
    // get local reference frame
    vec3 getLocalReferenceFrame(int i) const {
      return m_lrf[i];
    }
    
    // get local reference frame
    vec3 getLRF(int i) const {
      return getLocalReferenceFrame(i);
    }
    
    // set motion rules
    void setMotionRuleTypes(MotionRuleType m0,
			    MotionRuleType m1,
			    MotionRuleType m2) {
      m_mrl[0] = m0;
      m_mrl[1] = m1;
      m_mrl[2] = m2;
    }
    
    void setMotionRuleType(int i, MotionRuleType m) {
      m_mrl[i] = m;
    }
    
    // get motion rule
    MotionRuleType getMotionRuleType(int i) const {
      return m_mrl[i];
    }
    
    // set strengths (thresholds)
    //
    // OBS given as forces (most natural) and recomputed 
    // to distances by dividing by the stiffness.
    void setMotionRuleStrengths(Float s0,
				Float s1,
				Float s2) {
      m_str[0] = s0/m_stiffness;
      m_str[1] = s1/m_stiffness;
      m_str[2] = s2/m_stiffness;
    }
    
    void setMotionRuleStrength(int i, Float s) {
      m_str[i] = s/m_stiffness;
    }
    
    // get motion rule strength (as a force)
    Float getMotionRuleStrength(int i) const {
      return m_str[i]*m_stiffness;
    }
    
    // set the stiffness
    void setStiffness(Float stiff) {
      m_stiffness = stiff;
    }
    
    // get the stiffness
    Float getStiffness() const {
      return m_stiffness;
    }

    // set the damping
    void setDamping(Float damp) {
      m_damping = damp;
    }
    
    // get the damping
    Float getDamping() const {
      return m_damping;
    }
    
    // get proxy velocity
    vec3 getProxyVelocity() const {
      Float dt = m_time-m_old_time;
      if( dt < 1e-9 ) dt = 1.0;
      return (m_proxy-m_old_proxy)/dt;
    }
    
    // get probe velocity
    vec3 getProbeVelocity() const {
     Float dt = m_time-m_old_time;
     if( dt < 1e-9 ) dt = 1.0;
     return (m_probe-m_old_probe)/dt;
    }
    
    // get the distance vector between probe and proxy  
    vec3 getDelta() const {
      return m_probe-m_proxy;
    }
    
    // get the force feedback (in world coordinates)
    vec3 getForceFeedback() const {
      return -m_stiffness*getDelta(); 
      //- m_damping*(getProbeVelocity()-getProxyVelocity()) );
    }
    
    // resets the lrf, mrls, and mrl strengths
    void deactivate() {
      setLocalReferenceFrame(vec3(1,0,0), vec3(0,1,0), vec3(0,0,1));
      setMotionRuleTypes(NO_FORCE, NO_FORCE, NO_FORCE);
      setMotionRuleStrengths(Float(0.0), Float(0.0), Float(0.0));
    }
    
    // update proxy position
    //    
    // use the noforce = true to just move the proxy to the probe
    // can be nice when initializing or when directions are impossible to 
    // compute
    //
    // Typical order of execution:
    //
    //  Initialization:
    //   setStiffness()
    //
    //  In haptic loop: 
    //   setProbe - must know where we are, get position from API
    //   setOrientation - sometimes needed, device orientation from API
    //   setVelocity - sometimes needed, device velocity from API
    //
    //   setLocalReferenceFrame - the directions, probably data dependent
    //   setMotionRuleTypes - which constraints, probably data dependent
    //   setMotionRuleStrengths - can be static, but probably data dependent
    //    
    //   updateProxy - move the proxy according to the motion rules
    //   getForceFeedback - use this and send to haptic device
    virtual void updateProxy() {
      
      static bool first_time = true;
      
      if( first_time ) {
	deactivate();
	first_time = false;
      }
      
      // displacement vector
      vec3 delta = getDelta();
      
      // displacement in each axis of the lrf
      Float dx[3];
      dx[0] = Float( delta*m_lrf[0] );
      dx[1] = Float( delta*m_lrf[1] );
      dx[2] = Float( delta*m_lrf[2] );
      
      // compute proxy displacement
      Float dp[3];
      
      computeDP(0,dx,dp);
      computeDP(1,dx,dp);
      computeDP(2,dx,dp);
      
      // now its time to actually move the proxy
      vec3 mv = 
	dp[0]*m_lrf[0] +
        dp[1]*m_lrf[1] + 
	dp[2]*m_lrf[2];
      
      moveProxy(mv);
      
    }
    
  };
  
} // end namespace wish

#endif
