///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_VOXEL_ITERATOR_HH_
#define _WISH_VOXEL_ITERATOR_HH_

#include "wishImage.hh"
#include "wishBoundingBox.hh"

namespace wish {
  
  /*! 
    A VoxelIterator is an "iterator" that can traverse a part of 
    the image using
    the next(), prev(), and moveto() functions. The part is defined by 
    the bounding box.
  */
  template<class ImageType> 
  class VoxelIterator {
  public:
  
    //! typedefs
    typedef auto_ptr< VoxelIterator<ImageType> > Pointer;
    typedef typename ImageType::Pointer ImagePointer;
    typedef typename ImageType::VoxelType ImageVoxelType;
    
  protected:
    //! pointer to image
    ImageType *img;
    
    //! pointer to voxel at current position
    ImageVoxelType *ptr;
    
    //! current position
    Vec3i pos;
    
    //! box
    BoundingBox box;
    
  public:
    //! constructor that takes pointer to an image
    /*!
      Constructor that takes pointer to an image.
      The default bounding box is the whole image.
    */
    VoxelIterator(ImageType *_img=0) {
      img = _img;
      ptr = 0;
      
      // default bounding box is the whole image
      if(img) {
	box.start.x = box.start.y = box.start.z = 0;
	box.stop.x = img -> width()-1; 
	box.stop.y = img -> height()-1;
	box.stop.z = img -> depth()-1;;
	first();
      }
      
      else {
	//cerr << "Warning: Constructing VoxelIterator with 
	//nullpointer to image\n";
      }
    }
  
    //! copy constructor
    VoxelIterator(const VoxelIterator &v) {
      img = v.img;
      ptr = v.ptr;
      pos = v.pos;
      box.start = v.box.start;
      box.stop = v.box.stop;
    }
  
    //! assigment operator
    VoxelIterator& operator=(const VoxelIterator &v) {
      if( this != &v ) {
	img = v.img;
	ptr = v.ptr;
	pos = v.pos;
	box.start = v.box.start;
	box.stop = v.box.stop;
      }
      return *this;
    }
    
    //! destructor
    virtual ~VoxelIterator() {
      img = 0;
      ptr = 0;
    }
    
    // get and set functions
    //! get the Image pointer
    ImageType* getImagePointer() const { 
      return img; 
    }
    
    //! set the Image pointer
    void setImagePointer(ImageType *imgptr) {
      img = imgptr;
    }
    
    //! get pointer to current voxel
    ImageVoxelType* getVoxelPointer() const { 
      return ptr; 
    }
    
    //! set current voxel pointer
    void setVoxelPointer(ImageVoxelType *t) { 
      ptr = t; 
    }
    
    //! operator for access to any voxel
    ImageVoxelType& operator()(int x=0, int y=0, int z=0) const {
      return img->operator()(x,y,z);
    }
    
    //! operator for access to any voxel
    ImageVoxelType& operator()(const Vec3i &v) const {
      return img->operator()(v);
    }
    
    //! operator for access to image buffer
    ImageVoxelType& operator[](int i) const {
      return img->operator[](i);
    }
    
    //! get current position
    Vec3i getPosition() const { 
      return pos; 
    }
    
    //! get bounding box
    BoundingBox getBox() const {
      return box;
    }
    
    //! get bounding box start position
    Vec3i getBoxStart() const { 
      return box.start; 
    }
    
    //! get bounding box stop position
    Vec3i getBoxStop() const { 
      return box.stop; 
    }
    
    //! get box dimensions
    Vec3i getBoxSize() const {
      return box.getSize();
    }
    
    //! get number of voxels in bounding box
    int getTotalBoxSize() const { 
      return box.getTotalSize();
    }
    
    //! set the bounding box (remember to use first() afterwards)
    void setBox(const BoundingBox &_box) {
      
      // no weird bounding boxes!
      assert( _box.start.x >= 0 && _box.start.y >= 0 && _box.start.z >= 0 );
      assert( _box.start.x < img->width() && 
	      _box.start.y < img->height() && 
	      _box.start.z < img->depth() );
      assert( _box.stop.x >= 0 && 
	      _box.stop.y >= 0 && 
	      _box.stop.z >= 0 );    
      assert( _box.stop.x < img->width()  && 
	      _box.stop.y < img->height() && 
	      _box.stop.z < img->depth() );
      
      box = _box;
    }

    void setBox(const Vec3i &start, 
		const Vec3i &stop) {
      setBox(BoundingBox(start,stop));
    }
    
    //! reset bounding box to whole image (remember to use first())
    void resetBox() {
      box.start.x = box.start.y = box.start.z = 0;
      box.stop.x = img -> width()-1;
      box.stop.y = img -> height()-1;
      box.stop.z = img -> depth()-1;
    }
    
    //! check if we are inside image
    bool isInside() const {
      return img -> isInside(pos);
    }
    
    //! check if we are inside box
    bool isInsideBox(int x, int y, int z) const {
      return (x >= box.start.x && x <= box.stop.x &&
	      y >= box.start.y && y <= box.stop.y &&
	      z >= box.start.z && z <= box.stop.z);
    }
    
    //! check if we are inside box
    bool isInsideBox(const Vec3i &p) const {
      return (p.x >= box.start.x && p.x <= box.stop.x &&
	      p.y >= box.start.y && p.y <= box.stop.y &&
	      p.z >= box.start.z && p.z <= box.stop.z);
    }
  
    //! reference to the value at the current position
    ImageVoxelType& value() const { 
      return *ptr; 
    }
  
    //! compare value at current position with another VoxelIterator
    bool operator> (const VoxelIterator &v) const {
      return (value() > v.value());
    }

    //! compare value at current position with another VoxelIterator
    bool operator< (const VoxelIterator &v) const {
      return (value() < v.value());
    }
    
    //! compare value at current position with another VoxelIterator
    bool operator>= (const VoxelIterator &v) const {
      return (value() >= v.value());
    }
    
    //! compare value at current position with another VoxelIterator
    bool operator<= (const VoxelIterator &v) const {
      return (value() <= v.value());
    }
    
    //! compare value at current position with another VoxelIterator
    bool operator== (const VoxelIterator &v) const {
      return (value() == v.value());
    }
    
    //! compare value at current position with another VoxelIterator
    bool operator!= (const VoxelIterator &v) const {
      return (value() != v.value());
    }
    
    //! jump to start position of bounding box
    virtual void first() {
      ptr = &(img -> operator()(box.start));
      pos = box.start;
    }
    //! jump to last position of bounding box
    virtual void last() {
      ptr = &(img -> operator()(box.stop));
      pos = box.stop;
    }
  
    //! move to (x,y,z)
    /*!
      Moves to (x,y,z) and returns true if it is inside the box.
    */
    virtual bool moveTo(int x, int y, int z) {
      if(isInsideBox(x,y,z)) {
	ptr = &(img -> operator()(x,y,z));
	pos.x = x;
	pos.y = y;
	pos.z = z;
	return true;
      }
      return false;
    }
    
    //! move to p
    /*!
      Moves to p and returns true if it is inside the box.
    */  
    virtual bool moveTo(const Vec3i &p) {
      if(isInsideBox(p)) {
	ptr = &(img -> operator()(p));
	pos = p;
	return true;
      }
      return false;
    }
  
    //! move to next voxel in buffer
    /*!
      Move to next voxel in buffer and return true if ok.
      If we have reached last, we move to first and return false.
    */
    virtual bool next() {
      if(++pos.x<=box.stop.x) { 
	ptr = &(img -> operator()(pos)); 
	return true; 
      }
      pos.x = box.start.x;
      if(++pos.y<=box.stop.y) { 
	ptr = &(img -> operator()(pos)); 
	return true; 
      }
      pos.y = box.start.y;
      if(++pos.z<=box.stop.z) { 
	ptr = &(img -> operator()(pos));
	return true; 
      }
    
      first();
      return false;
    }
    //! move to previous voxel in buffer
    /*!
      Move to previous voxel in buffer and return true if ok.
      If we have reached start, we move to last and return false.
    */
    virtual bool prev() {
      if(--pos.x>=box.start.x) { 
	ptr = &(img -> operator()(pos)); 
	return true; 
      }
      pos.x = box.stop.x;
      if(--pos.y>=box.start.y) { 
	ptr = &(img -> operator()(pos)); 
	return true; 
      }
      pos.y = box.stop.y;
      if(--pos.z>=box.start.z) { 
	ptr = &(img -> operator()(pos));
	return true; 
      }
    
      last();
      return false;
    }
    

  };

} // end namespace wish

#endif
 

