///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_UTIL_HH_
#define _WISH_UTIL_HH_

#include "wish.hh"

namespace wish {

  namespace util {
    //! max
    template<class T, class U>
    T maxOf(const T &a, const U &b) {
      return (a > b) ? a:(T)b;
    }
    //! min
    template<class T, class U>
    T minOf(const T &a, const U &b) {
      return (a < b) ? a:(T)b;
    }

    // check if it is a NaN
    template< typename R > // R should be float or double or long double
    bool isNaN( R r )
    {
      return r != r;
    }
    
    // split filename into dir, name, ext
    void splitFilename(const string &fn, 
		       string &dir,
		       string &name,
		       string &ext);
    
    //! systemCommand
    bool systemCommand(const std::string &cmd);
   
    //! check filesize
    int fileSize(const std::string &filename);
    
    //! convert to string
    template<class T>
    std::string typeToString(const T &t, 
			     int precision=-1, 
			     int width=-1) {
      std::ostringstream os;
      
      if( precision > -1 )
	os.precision(precision);
      
      if( width > -1 )
	os.width(width);
      
      os << setiosflags(ios::fixed) << t;
      
      return os.str();
    }
    
    template<class T>
    T stringToType(const std::string &str) {
      std::istringstream is(str);
      T t;
      is >> t;
      return t;
    }
   
    //! computes x^n
    template<class T>
    T pown(const T &x, const unsigned int &n) {
      if(n == 0)
	return 1;
      return x*pown(x,n-1);
    }
   
    //! returns the current time (seconds)
    // clockticks/(clockticks/second) = seconds
    inline double timer() {
      return double(std::clock())/CLOCKS_PER_SEC;
    }
   
    struct timerobject {
      std::clock_t tstart;
      std::clock_t tstop;
      std::clock_t telapsed;
     
      inline timerobject() {
	reset();
      }
     
      inline void reset() {
	tstart = tstop = telapsed = 0;     
      }
     
      inline void start() {
	tstart = std::clock();
      }
     
      inline void stop() {
	tstop = std::clock();
	telapsed = tstop-tstart;
      }
     
      inline double getMilliSeconds() {
	return 1000.0*double(telapsed)/CLOCKS_PER_SEC;
      }
     
      inline double getSeconds() {
	return double(telapsed)/CLOCKS_PER_SEC;
      }
     
      inline double getMinutes() {
	return 1.0/60.0*double(telapsed)/CLOCKS_PER_SEC;
      }
     
    };
   
    //! initializes the random numer generator
    inline void randInit() {
      std::srand( (unsigned int) std::time(0));
    }
   
    //! returns a random number in the interval [0,1]
    inline double rand() {
      return double(std::rand())/RAND_MAX;
    }
   
    //! returns the sign of t
    template<class T>
    T sign(const T &t) {
      if(t >= T(0))
	return T(1);
      return T(-1);
    }
   
    //! a variant 
    template<class T>
    T sign(const T &a, const T &b) {
      if(b >= 0.0)
	return std::fabs(a);
      return -std::fabs(a);
    }
   
    //! rounds a floating point
    inline int round(const double &t) {
      if(t > 0)
	return int(t+0.5);
      else
	return int(t-0.5);
    }
   
    //! computes the square of t by a simple multiplication
    template<class T>
    T square(const T &t) {
      return t*t;
    }
    
    //! 2^n
    inline int pow2(int n) {
      if( n>=0 )
	return 1 << n;
      else 
	return 0;
    }
    
    //! linear interpolation between a and b, returns a double
    template<class T>
    double interpd(const T &a, const T &b, const double &alpha) {
      return (1-alpha)*a + alpha*b; 
    }
   
    //! linear interpolation between a and b
    template<class T>
    T interp(const T &a, const T &b, const double &alpha) {
      return T((1-alpha)*a + alpha*b); 
    }
    
    //! is this architecture little endian
    inline bool isLittleEndian() {
      union u { 
	unsigned long l; 
	unsigned char c[sizeof(unsigned long)]; 
      };
      u dummy;
      dummy.l = 1;
      return dummy.c[0] == 1;
    }
    
    //! is this architecture big endian
    inline bool isBigEndian() {
      return !isLittleEndian();
    }
    
    inline unsigned int shiftLeft(unsigned int a, int n=1) {
      return a << n;
    }
    
    inline bool isPowerOfTwo(unsigned long a) {
      bitset<8*sizeof(unsigned long)> vec(a);      
      return vec.count() <= 1; 
    }
    
    inline unsigned long nextPowerOfTwo(unsigned long a) {
      bitset<8*sizeof(unsigned long)> vec(a);      
      size_t c=vec.count();
      if( c<=1 ) 
	return a;
      size_t i=vec.size();
      while( !vec.test(--i) );
      return 1U << (i+1);
    }
    
    //! swap byte order
    inline void swap2Bytes(byte* &ptr) { // uint16, int16
      byte tmp;  
      tmp = ptr[0]; ptr[0] = ptr[1]; ptr[1] = tmp;
    }
    inline void swap3Bytes(byte* &ptr) { // RGB8
      byte tmp;
      tmp = ptr[0]; ptr[0] = ptr[2]; ptr[2] = tmp;
    }
    inline void swap4Bytes(byte* &ptr) { // uint32, int32, float, RGBA8
      byte tmp; 
      tmp = ptr[0]; ptr[0] = ptr[3]; ptr[3] = tmp;
      tmp = ptr[1]; ptr[1] = ptr[2]; ptr[2] = tmp; 
    }
    inline void swap8Bytes(byte* &ptr) { // double
      byte tmp;
      tmp = ptr[0]; ptr[0] = ptr[7]; ptr[7] = tmp;
      tmp = ptr[1]; ptr[1] = ptr[6]; ptr[6] = tmp;
      tmp = ptr[2]; ptr[2] = ptr[5]; ptr[5] = tmp;
      tmp = ptr[3]; ptr[3] = ptr[4]; ptr[4] = tmp; 
    }
    template<class T>
    void swapBytes(T *in, T *out) {
      int n = sizeof(T);
      byte *pin = (byte*) in;
      byte *pout = (byte*) out;
      for(int i=0; i<n; i++)
	pout[i] = pin[n-i-1];
    }
    
    //! swap byte order for an array
    template<class T>
    void byteSwapArray(T *arr, unsigned int arrlen) {
      
      int n = sizeof(T);
      byte *p = (byte*) arr;
      switch( n ) {
      case 1:
	return;
	
      case 2:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap2Bytes(p);
	}
	break;
	
      case 3:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap3Bytes(p);
	}
	break; 
	
      case 4:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap4Bytes(p);
	}
	break; 
	
      case 8:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap8Bytes(p);
	}
	break;
	
      default:
	break;
      }
    }
    
    //! swap byte order for an array
    inline void byteSwapArray(byte *arr, int bytesPerComp, int numComp) {
      
      int n = bytesPerComp;
      byte *p = arr;
      unsigned int arrlen = bytesPerComp*numComp;
      
      switch( n ) {
      case 1:
	return;
	
      case 2:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap2Bytes(p);
	}
	break;
	
      case 3:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap3Bytes(p);
	}
	break; 
	
      case 4:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap4Bytes(p);
	}
	break; 
	
      case 8:
	for(unsigned int j=0; j<arrlen; j++, p+=n) {
	  swap8Bytes(p);
	}
	break;
	
      default:
	break;
      }
    }
    
    //! write binary with big endian byte order
    //! Here n is the number of elements, not the number of bytes
    //! The same goes for the functions below
    template<class T>
    void writeBE(ostream &os, T *ptr, unsigned int n) {
      if( isLittleEndian() ) {	
	// copy and write chunks of 1 MB = 2^20 Bytes
	unsigned int chunkByteSize=shiftLeft(1U,20);
	unsigned int chunkElems=chunkByteSize/sizeof(T);
	unsigned int elemsWritten=0;
	T *cpy = new T[chunkElems];
	while( elemsWritten < n ) {
	  chunkElems=util::minOf(n-elemsWritten, 
				 (unsigned int) (chunkByteSize/sizeof(T)));
	  memcpy(cpy, ptr+elemsWritten, chunkElems*sizeof(T));
	  byteSwapArray(cpy, chunkElems);
	  os.write((const char*) cpy, chunkElems*sizeof(T));
	  elemsWritten += chunkElems; 
	}
	delete [] cpy;
      }
      else {
	os.write((const char*) ptr, sizeof(T)*n);
      }
    }
    
    //! write binary with little endian byte order
    template<class T>
    void writeLE(ostream &os, T *ptr, unsigned int n) {
      if( isLittleEndian() ) {
	os.write((const char*) ptr, sizeof(T)*n);
      }
      else {
	// copy and write chunks of 1 MB = 2^20 Bytes
	unsigned int chunkByteSize=shiftLeft(1U,20);
	unsigned int chunkElems=chunkByteSize/sizeof(T);
	unsigned int elemsWritten=0;
	T *cpy = new T[chunkElems];
	while( elemsWritten < n ) {
	  chunkElems=util::minOf(n-elemsWritten, 
				 (unsigned int) (chunkByteSize/sizeof(T)));
	  memcpy(cpy, ptr+elemsWritten, chunkElems*sizeof(T));
	  byteSwapArray(cpy, chunkElems);
	  os.write((const char*) cpy, chunkElems*sizeof(T));
	  elemsWritten += chunkElems; 
	}
	delete [] cpy;
      }
    }
    
    //! read binary with big endian byte order
    template<class T>
    void readBE(istream &is, T *ptr, unsigned int n) {
      if( isLittleEndian() ) {
	is.read((char*) ptr, sizeof(T)*n); 	  
	byteSwapArray(ptr, n);
      }
      else {
	is.read((char*) ptr, sizeof(T)*n);      
      }    
    }
    
    //! read binary with little endian byte order
    template<class T>
    void readLE(istream &is, T *ptr, unsigned int n) {
      if( isLittleEndian() ) {
	is.read((char*) ptr, sizeof(T)*n);      
      }
      else {
	is.read((char*) ptr, sizeof(T)*n); 	  
	byteSwapArray(ptr, n);
      }
    }    
    
    //! swap
    template<class T>
    void swap(T &a, T &b) {
      T temp = a;
      a = b;
      b = temp;
    }
   
    //! sort two elements (ordered swap)
    template<class T>
    void sort2(T &a, T &b) {
      if( a > b ) {
	T temp = a;
	a = b;
	b = temp;
      }
    }
   
    //! array print
    template<class T>
    void printarray(const T array[], int n) {
      for(int i=0; i<n-1; i++)
	std::cout << i << "\t" << array[i] << std::endl;
      std::cout << n-1 << "\t" << array[n-1] << std::endl;
    }
    
    //! bubblesort in increasing order (actually sorts the index array)
    template<class T>
    void bubblesort(const T array[], int index[], int n) {
     
      for(int i=0; i<n; i++) {
	index[i] = i;
      }
     
      int right=n;
      bool unsorted=true;
      while(unsorted) {
	unsorted = false;   
	for(int i=0; i<right-1; i++) {
	  if(array[index[i]] > array[index[i+1]]) {
	    util::swap(index[i], index[i+1]);
	    unsorted = true;
	  }
	}
	right--;
      }
    }
   
    //! bubblesort (increasing order)
    template<class T>
    void bubblesort(T array[], int n) {
     
      int right=n;
      bool unsorted=true;
      while(unsorted) {
	unsorted = false;   
	for(int i=0; i<right-1; i++) {
	  if(array[i] > array[i+1]) {
	    util::swap(array[i], array[i+1]);
	    unsorted = true;
	  }
	}
	right--;
      }
    }
   
  } // end namespace util
 
} // end namespace wish

#endif
