///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

// the volume texture
uniform sampler3D volumeTexture;

// the normal texture
uniform sampler3D normalTexture;

// transfer function texture
uniform sampler2D tfTexture;

// the texture matrix and its inverse
uniform mat4 textureMatrix;
uniform mat4 textureMatrixInverse;

// sample distance along ray
uniform float rayStep;

// isovalue for isosurface mode
uniform float isoValue;

// render mode 0=mip, 1=composite, 2=isosurface
uniform int renderMode;

// ray start
varying vec3 rayStart;

// ray direction
varying vec3 rayDir;

// compute t such that r0+t*dir is the exit point of the [0,1] box
float rayexit(vec3 r0, vec3 dir, float eps) {
  vec3 q = 1.0/(dir+eps);
  vec3 t0 = -r0*q;
  vec3 tmax = max(t0,q+t0);
  return min(tmax.x,min(tmax.y,tmax.z));
}

// get the normal at pos from the rgba texture ntex
// first the mapping [0,0.5] -> [-1,0] and [0.5,1] -> [0,1]
// and then rotation from image coords to volume coords
vec3 normalFromRGBA(vec3 pos, sampler3D ntex) {
  vec3 n = 2.0*texture3D(ntex,pos).xyz-1.0;
  return vec3(n.x, -n.y, -n.z);
}

// return maximum of colors, intensity stored in alpha channel
vec4 maxHSV(vec4 c1, vec4 c2) {
  return c1.a > c2.a ? c1:c2;
}

// struct to return from traverseRay, 
// the resulting color and the point to use for depth calculation 
struct RayResult {
  vec4 color;
  vec4 zpoint;
};

// traverse ray
// Inputs:
//   r0  - ray start (xyz), ray exit time (w)
//   dir -  ray direction (xyz), step length (-w)
// Output:
//   RayResult.color  - the computed RGBA color for this fragment
//   RayResult.zpoint - the point to use for depth calculations,
//                      if zpoint.x<0, no depth is computed

// mip mode
RayResult traverseRay(vec4 r0, vec4 dir) {
  // values
  float sample = texture3D(volumeTexture, r0.xyz).r;
  vec3 pos = r0.xyz;
  
  // one step
  r0 += dir;
  
  // loop
  while( r0.w>=0.0 ) {
    float tmpsample = texture3D(volumeTexture, r0.xyz).r;
    if( tmpsample>sample ) {
      sample=tmpsample;
      pos=r0.xyz;
    }
    
    r0 += dir;
  }
  
  // return result 
  return RayResult(texture2D(tfTexture,vec2(sample,0.5)), vec4(pos,1.0));
}

// composite mode
RayResult traverseRay1(vec4 r0, vec4 dir) {
  // return value
  RayResult rr;
  
  // initial color of this fragment is black with zero alpha
  rr.color = vec4(0.0, 0.0, 0.0, 0.0);
  
  // initial depth is not defined
  rr.zpoint = vec4(-1.0, 0.0, 0.0, 0.0);
  
  // compositing loop
  while( rr.color.a<0.95 && r0.w>=0.0 ) {
    float val = texture3D(volumeTexture, r0.xyz).r;
    vec4 sample_color = texture2D(tfTexture, vec2(val,0.5));
    
    // front-to-back compositing
    sample_color.rgb *= sample_color.a; // opacity-weighted color
    rr.color += sample_color*(1.0-rr.color.a);
    
    // step forward along ray
    r0 += dir;
  }
  
  // set depth where ray terminates, good or bad?
  rr.zpoint = vec4(r0.xyz,1.0);
  
  // return result
  return rr;
}

// isosurface mode
RayResult traverseRay2(vec4 r0, vec4 dir) {
  // return color
  vec4 col=vec4(0,0,0,0), tfcol;
  
  float v, v1;
  vec3 r, r1;
  bool found_surf=false;
  while( r0.w>=0.0 && found_surf==false ) {
    r = r0.xyz;
    r1 = r0.xyz+dir.xyz;
    v = texture3D(volumeTexture, r).r; 
    v1 = texture3D(volumeTexture, r1).r; 
    found_surf = (min(v,v1)<=isoValue && isoValue<=max(v,v1));
    r0 += dir;
  }
  
  // surface found
  if( !found_surf ) { discard; }
  
  // interpolate position
  float a = clamp( (isoValue-v)/(v1-v), 0.0, 1.0 );
  r = mix(r,r1,a);
  
  // lookup transfer function
  tfcol = texture2D(tfTexture,vec2(isoValue,0.5));
  
  // set the opacity
  col.a = tfcol.a;
  
  // sample position in view coordinates
  vec3 pos = (gl_ModelViewMatrix*textureMatrixInverse*
	      vec4(r.xyz,1.0)).xyz;
  
  // lightsource position (only one lightsource so far)
  vec4 light_pos = gl_LightSource[0].position;
  
  // ambient
  col.rgb += vec3(.1,.1,.1);
  
  // normal vector 
  vec3 normal = normalFromRGBA(r,normalTexture);
  normal = normalize(gl_NormalMatrix*normal);
  
  // view vector 
  vec3 viewv = normalize(-pos); 
  
  // re-orient normal
  if( dot(viewv,normal)<0.0 )
    normal = -normal;
  
  // lightsource vector
  vec3 lightv = normalize(light_pos.xyz-pos);
  float ndotl = dot(normal,lightv);
  
  // diffuse
  col.rgb +=  tfcol.rgb * max(ndotl,0.0);
  
  // specular
  if (ndotl > 0.0) {
    vec3 reflv = normalize( reflect(-lightv,normal) );
    float rdotv = dot(reflv,viewv);
    col.rgb += vec3(.6,.6,.6) * pow(max(rdotv,0.0), 50.0);
  }
  
  
  // return result 
  return RayResult( col, vec4(r.xyz,1.0) );
}

// hsv mip
RayResult traverseRay3(vec4 r0, vec4 dir) {
  // quick fix
  float blend = isoValue;
  
  // loop
  vec4 col = vec4(0,0,0,0);
  while( r0.w>=0.0 ) {
    vec4 col1 = texture3D(volumeTexture, r0.xyz);
    vec4 col2 = texture3D(normalTexture, r0.xyz);
    vec4 tmp = (1.0-blend)*col1 + blend*vec4(col2.rgb,col2.r);
    col = maxHSV(col, tmp);
    r0 += dir;
  }
  
  // return result 
  return RayResult( col, vec4(-1,0,0,0) );
}

// main function
void main() {
  
  // initialize ray
  vec4 raydir = vec4(normalize(rayDir), -1.0);
  float texit = rayexit(rayStart,raydir.xyz,1.0e-6);
  raydir = rayStep*raydir; 
  vec4 raypos = vec4(rayStart, texit);  
  
  // traverse ray
  RayResult rr;
  if( renderMode==1 ) // composite
    rr = traverseRay1(raypos, raydir);
  else if( renderMode==2 ) // isosurface
    rr = traverseRay2(raypos, raydir);
  else if( renderMode==3 ) // hsv mip
    rr = traverseRay3(raypos, raydir);
  else // 0=mip
    rr = traverseRay(raypos, raydir);
  
  // set color
  gl_FragColor = rr.color;
  
  // set depth if computed
  if( rr.zpoint.x>=0.0 ) {
    // set depth
    vec4 tmp =
      gl_ModelViewProjectionMatrix*textureMatrixInverse*rr.zpoint;
    // in window coordinates
    gl_FragDepth = 0.5*(gl_DepthRange.diff*tmp.z/tmp.w +
			gl_DepthRange.near+gl_DepthRange.far);
  }
  
}
