///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DBilateralFilter.hh"

#include "wishBilateralFilter.hh"

namespace wishH3D {
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  BilateralFilter::database( "wishH3DBilateralFilter", 
			     &(newInstance<BilateralFilter>), 
			     typeid( BilateralFilter ),
			     &FilterObject::database );
  
  namespace BilateralFilterInternals {
    FIELDDB_ELEMENT( BilateralFilter, inputIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BilateralFilter, sigmaSpatial, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BilateralFilter, sigmaRange, INPUT_OUTPUT );
    
    template<class InputImageType, class OutputImageType>
    void executeFilter(InputImageType *in,
		       OutputImageType* &out,
		       H3DFloat sigma_spatial,
		       H3DFloat sigma_range) {
      
      wish::BilateralFilter<InputImageType, OutputImageType> bf;
      bf.setInput( in );
      bf.setSigmaDomain(sigma_spatial);
      bf.setNormalizedSigmaRange(sigma_range);
      cerr << "Executing BilateralFilter..." << endl; 
      bf.execute();
      cerr << "Execution time = " << bf.getProcessTime() << " s" << endl;
      
      out = bf.releaseOutput();
    }
    
  }
  
  BilateralFilter::BilateralFilter() :
    inputIndex( new SFInt32 ),
    sigmaSpatial( new SFFloat ),
    sigmaRange( new SFFloat )
  {  
    this->type_name = "BilateralFilter";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // defaults
    inputIndex->setValue( -1 );
    sigmaSpatial->setValue(1.0f);
    sigmaRange->setValue(10.0f);
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  wish::ImageObject* BilateralFilter::execute() {
    Volume *vol = volume->getValue();
    if( !vol ) 
      return 0;
    
    wish::ImageObject *input = vol->getImage( inputIndex->getValue() );
    if( !input )
      return 0;
    
    string dt = input->getDataType();
    if( dt=="uint8" ) {
      wish::Image8 *in = dynamic_cast<wish::Image8*>(input);
      wish::Image8 *out;
      BilateralFilterInternals::
	executeFilter( in, out, 
		       sigmaSpatial->getValue(),
		       sigmaRange->getValue() );
      return out;
    }
    else if( dt=="uint16" ) {
      wish::Image16 *in = dynamic_cast<wish::Image16*>(input);
      wish::Image16 *out;
      BilateralFilterInternals::
	executeFilter( in, out, 
		       sigmaSpatial->getValue(),
		       sigmaRange->getValue() );
      return out;
    }
	else {
		return 0;
	}
    
  }
  
}
