///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DColorTransferFunction.hh"

namespace wishH3D {
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  ColorTransferFunction::database( "wishH3DColorTransferFunction", 
				   &(newInstance<ColorTransferFunction>), 
				   typeid( ColorTransferFunction ),
				   &TransferFunction::database );
  
  namespace ColorTransferFunctionInternals {
    FIELDDB_ELEMENT( ColorTransferFunction, xsamples, INPUT_OUTPUT );
    FIELDDB_ELEMENT( ColorTransferFunction, ysamples, INPUT_OUTPUT );
  }
  
  ColorTransferFunction::ColorTransferFunction() : 
    xsamples( new MFFloat ),
    ysamples( new MFColor )
  {
    
    type_name = "ColorTransferFunction";
    database.initFields( this );
    
    xsamples -> route( modified );
    ysamples -> route( modified );
    
  }
  
  // evaluate function
  // this function is not safe, you must provide valid ranges and values
  RGB ColorTransferFunction::evaluateRGB(const H3DFloat &arg) const {
    
    vector<H3DFloat> vx = xsamples -> getValue();
    vector<RGB> vy = ysamples -> getValue();
    
    int sample_count = int( vy.size() );
    
    RGB value;
    
    if( sample_count == 0 ) // no samples
      value = RGB(1.0f, 0.0f, 0.0f);
    
    else if( sample_count == 1 ) // constant function
      value = vy[0];
    
    else if( arg < vx[0] ) // clamping
      value = vy[0];
  
    else if( arg > vx[sample_count-1] ) 
      value = vy[sample_count-1];
    
    else { // interpolate (linearly)
      for(int i=0; i<sample_count-1; i++) {
	if( vx[i] <= arg && arg <= vx[i+1] ) {
	  H3DFloat alpha = arg-vx[i];
	  value.r = vy[i].r + alpha*(vy[i+1].r-vy[i].r)/(vx[i+1]-vx[i]);
	  value.g = vy[i].g + alpha*(vy[i+1].g-vy[i].g)/(vx[i+1]-vx[i]);
	  value.b = vy[i].b + alpha*(vy[i+1].b-vy[i].b)/(vx[i+1]-vx[i]);
	}
      }
    }
    
    return value;
  }
  
}
