///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_H3D_DEFORMABLE_SIMPLEX_MESH_HH_
#define _WISH_H3D_DEFORMABLE_SIMPLEX_MESH_HH_

#include "Group.h"
#include "X3D.h"
#include "Threads.h"
#include "TimeStamp.h"

#include "wishH3DVolume.hh"
#include "wishH3DVolumeChild.hh"

#include "wishSimplexMesh.hh"
#include "wishExternalForce.hh"

namespace wishH3D {
  
  /**
     wishH3DDeformableSimplexMesh implements deformable models for 
     3D segmentation based on Delingettes Simplex mesh
  */
  class WISHH3D_API DeformableSimplexMesh : public Group, public VolumeChild {
    
  public:
    //! X3D interface
    static H3DNodeDatabase database;
    
    struct WriteMesh : public AutoUpdate<SFString> {
      virtual void update();
    };
    
    struct ReadMesh : public AutoUpdate<SFString> {
      virtual void update();
    };

    struct RasterizeMesh : public AutoUpdate<SFBool> {
      virtual void update();
    };
    
    struct Scaler : public AutoUpdate<SFFloat> {
      H3DFloat ref;
      virtual void update();
    };
    
    struct Resetter : public AutoUpdate<SFBool> {
      virtual void update();
    };
    
    struct Refiner : public AutoUpdate<SFBool> {
      virtual void update();
    };
    
    struct Translator : 
      public AutoUpdate<TypedField< SFVec3f,Types<SFVec3f,SFBool> > > {
      Vec3f ref;
      virtual void update();      
    };
    
    struct SpatialWeight : AutoUpdate<SFFloat> {
      virtual void update();
    };
    
    struct ExternalWeight : AutoUpdate<SFFloat> {
      virtual void update();
    };
    
    struct ExternalForceWeight : AutoUpdate<SFFloat> {
      wish::ExternalForce *ef;
      virtual void update();
    };
    
    //! constructor
    DeformableSimplexMesh();
    
    //! initialize function
    virtual void initialize();
    
    //! render
    virtual void render();
    
    //! traverse SG (activate the model)
    virtual void traverseSG( TraverseInfo &ti );
    
    //! select/deselect mesh vertices/faces
    wish::SimplexVertex* selectVertices(const Vec3f &pos, const int &nbsize);
    void deselectVertices();
    wish::SimplexFace* selectFaces(const Vec3f &pos, const int &nbsize);
    void deselectFaces();
    
    //! reset mesh
    void resetMesh() { 
      if( simplex->nVertices()>0 )
	simplex->deleteMesh();
      else
	simplex->createSphere(64.0, 1); 
    }
    
    //! refine mesh
    void refineMesh() { simplex->refine(); }
    
    //! update external forces
    void updateExternalForces();
    
    //! the button
    auto_ptr<SFBool> activateButton;
    auto_ptr<SFBool> interactionButton;
    
    //! reset mesh
    auto_ptr<SFBool> resetButton;
    auto_ptr<Resetter> resetter;
    
    //! refine mesh
    auto_ptr<SFBool> refineButton;
    auto_ptr<Refiner> refiner;
    
    //! render mode
    auto_ptr<SFInt32> renderMode;
    
    //! scaling
    auto_ptr<SFFloat> scaling;
    auto_ptr<Scaler> scaler;
    
    //! translation
    auto_ptr<SFVec3f> translation;
    auto_ptr<SFBool> translationButton;
    auto_ptr<Translator> translator;
    
    //! I/O
    auto_ptr<WriteMesh> writeMesh;
    auto_ptr<ReadMesh> readMesh;
    auto_ptr<RasterizeMesh> rasterizeMesh;
    
    // these indices are used to fetch the correct
    // image from the volume for the different external forces
    auto_ptr<SFInt32> imageIndex; // not added yet
    auto_ptr<SFInt32> outputIndex; // rasterization output
    auto_ptr<SFInt32> gradientMagnitudeIndex;
    auto_ptr<SFInt32> potentialIndex;
    auto_ptr<SFInt32> propagationIndex;
    auto_ptr<SFInt32> vectorIndex;
    
    auto_ptr<wish::InteractionForce> interactionForce;
    auto_ptr<wish::GradientMagnitudeForce> gradientMagnitudeForce;
    auto_ptr<wish::PotentialFieldForce> potentialFieldForce;
    auto_ptr<wish::PropagationForce> propagationForce;
    auto_ptr<wish::VectorFieldForce> vectorFieldForce;
    
    //! weights
    auto_ptr<SpatialWeight> spatialWeight;
    auto_ptr<ExternalWeight> externalWeight;
    auto_ptr<ExternalForceWeight> interactionWeight;
    auto_ptr<ExternalForceWeight> gradientMagnitudeWeight;
    auto_ptr<ExternalForceWeight> potentialWeight;
    auto_ptr<ExternalForceWeight> propagationWeight;
    auto_ptr<ExternalForceWeight> vectorWeight;
    
    auto_ptr<SFFloat> propagationZeroLevel;
    auto_ptr<SFInt32> interactionSize;
    
    //! the mesh
    auto_ptr<wish::SimplexMesh> simplex;
    
  protected:
    class ForceEffect : public HapticForceEffect {
      
    protected:
      Vec3f setforce;
      Vec3f force;
      H3DFloat stiffness;
      H3DFloat device_scale;
      
    public:
      static const H3DFloat MAX_FORCE;
      
      inline ForceEffect() : 
	HapticForceEffect( Matrix4f(), true ),
	setforce(Vec3f(0,0,0)),
	force(Vec3f(0,0,0)),
	stiffness(400.0f),
	device_scale(1.0f)
      {}
      
      EffectOutput calculateForces( const EffectInput &input );
      
      void computeForceFeedback() {
	force = setforce;
      }
      
      void setForce( const Vec3f &f ) {
	setforce = f;
      }
      
      void setStiffness(H3DFloat s) {
	stiffness = s;
      }
      H3DFloat getStiffness() const {
	return stiffness;
      }
      
      void setDeviceScale(H3DFloat s) {
	device_scale = s;
      }
      H3DFloat getDeviceScale() const {
	return device_scale;
      }

      Vec3f getForce() const {
	return force;
      }
      
    };
    
    AutoRef<ForceEffect> forceEffect;
    
  };
  
  const H3DFloat DeformableSimplexMesh::ForceEffect::MAX_FORCE = 6.0;


  
}
#endif
