///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DFastMarchingFilter.hh"

#include "wishFastMarchingFilter.hh"

namespace wishH3D {
    
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  FastMarchingFilter::database( "wishH3DFastMarchingFilter", 
				&(newInstance<FastMarchingFilter>), 
				typeid( FastMarchingFilter ),
				&FilterObject::database );
  
  namespace FastMarchingFilterInternals {
    FIELDDB_ELEMENT( FastMarchingFilter, inputIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FastMarchingFilter, seedIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FastMarchingFilter, timeThreshold, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FastMarchingFilter, normalizeOutput, INPUT_OUTPUT );
    
    template<class InputImageType, class OutputImageType>
    void executeFilter(InputImageType *in,
		       OutputImageType* &out,
		       wish::Image8 *seeds,
		       H3DFloat time_threshold) {
      
      wish::FastMarchingFilter<InputImageType, OutputImageType> ff;
      ff.setInput( in );
      ff.setSeedsFromImage( seeds );
      if( time_threshold > 0.0f )
	ff.setTimeMax( time_threshold );
      cerr << "Executing FastMarchingFilter..." << endl; 
      ff.execute();
      cerr << "Execution time = " << ff.getProcessTime() << " s" << endl;
      
      out = ff.releaseOutput();
    }
    
  }
  
  FastMarchingFilter::FastMarchingFilter() :
    inputIndex( new SFInt32 ),
    seedIndex( new SFInt32 ),
    timeThreshold( new SFFloat ),
    normalizeOutput( new SFBool )
  {  
    this->type_name = "FastMarchingFilter";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // defaults
    inputIndex->setValue( -1 );
    seedIndex->setValue( -1 );
    timeThreshold->setValue( 0.0f );
    normalizeOutput->setValue( true );
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  wish::ImageObject* FastMarchingFilter::execute() {
    Volume *vol = volume->getValue();
    if( !vol ) 
      return 0;
    
    wish::ImageObject *input = vol->getImage( inputIndex->getValue() );
    if( !input )
      return 0;
    
    wish::Image8 *seeds = 
      dynamic_cast<wish::Image8*>( vol->getImage( seedIndex->getValue() ) );
    if( !seeds ) 
      return 0;
    
    // output
    wish::Imagef *out = 0;
        
    string dt = input->getDataType();
    if( dt=="uint8" ) {
      FastMarchingFilterInternals::
	executeFilter( dynamic_cast<wish::Image8*>(input),
		       out,
		       seeds,
		       timeThreshold->getValue() );
    }
    else if( dt=="uint16" ) {
      FastMarchingFilterInternals::
	executeFilter( dynamic_cast<wish::Image16*>(input),
		       out,
		       seeds,
		       timeThreshold->getValue() );
    }
    
    // normalize with the max value
    if( normalizeOutput->getValue() ) {
      float tt = out->getMax();
      for(int i=0; i<out->totalSize(); i++) {
	if( out->at(i) >= tt )
	  out->at(i) = 1.0f;
	else
	  out->at(i) = out->at(i)/tt;
      }
    }
    
    return out;
    
  }
  
}
