///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DFuzzyConnectednessFilter.hh"

#include "wishFuzzyConnectednessFilter.hh"

namespace wishH3D {
    
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  FuzzyConnectednessFilter::database( "wishH3DFuzzyConnectednessFilter", 
				      &(newInstance<FuzzyConnectednessFilter>), 
				      typeid( FuzzyConnectednessFilter ),
				      &FilterObject::database );
  
  namespace FuzzyConnectednessFilterInternals {
    FIELDDB_ELEMENT( FuzzyConnectednessFilter, inputIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FuzzyConnectednessFilter, seedIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FuzzyConnectednessFilter, fcThreshold, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FuzzyConnectednessFilter, inputIsAffinity, INPUT_OUTPUT );
    FIELDDB_ELEMENT( FuzzyConnectednessFilter, invertAffinity, INPUT_OUTPUT );

    template<class InputImageType, class OutputImageType>
    void executeFilter(InputImageType *in,
		       OutputImageType* &out,
		       wish::Image8 *seeds,
		       H3DFloat fc_threshold,
		       bool isaff,
		       bool ia) {
      
      wish::FuzzyConnectednessFilter<InputImageType, OutputImageType> fcf;
      fcf.setInput( in );
      fcf.setSeedsFromImage( seeds );
      fcf.setEarlyThreshold( fc_threshold );
      fcf.setInputIsAffinity( isaff );
      fcf.setInvertAffinity( ia );
      cerr << "Executing FuzzyConnectednessFilter..." << endl; 
      fcf.execute();
      cerr << "Execution time = " << fcf.getProcessTime() << " s" << endl;
      
      out = fcf.releaseOutput();
    }
    
  }
  
  FuzzyConnectednessFilter::FuzzyConnectednessFilter() :
    inputIndex( new SFInt32 ),
    seedIndex( new SFInt32 ),
    fcThreshold( new SFFloat ),
    inputIsAffinity( new SFBool ),
    invertAffinity( new SFBool )
  {  
    this->type_name = "FuzzyConnectednessFilter";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // defaults
    inputIndex->setValue( -1 );
    seedIndex->setValue( -1 );
    fcThreshold->setValue( 0.8f );
    inputIsAffinity->setValue( false );
    invertAffinity->setValue( false );
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  wish::ImageObject* FuzzyConnectednessFilter::execute() {
    Volume *vol = volume->getValue();
    if( !vol ) 
      return 0;
    
    wish::ImageObject *input = vol->getImage( inputIndex->getValue() );
    if( !input )
      return 0;
    
    wish::Image8 *seeds = 
      dynamic_cast<wish::Image8*>( vol->getImage( seedIndex->getValue() ) );
    if( !seeds ) 
      return 0;
    
    // output
    wish::Imagef *out = 0;
    
    string dt = input->getDataType();
    if( dt=="uint8" ) {
      FuzzyConnectednessFilterInternals::
	executeFilter( dynamic_cast<wish::Image8*>(input),
		       out,
		       seeds,
		       fcThreshold->getValue(),
		       inputIsAffinity->getValue(),
		       invertAffinity->getValue() );
    }
    else if( dt=="uint16" ) {
      FuzzyConnectednessFilterInternals::
	executeFilter( dynamic_cast<wish::Image16*>(input),
		       out,
		       seeds,
		       fcThreshold->getValue(),
		       inputIsAffinity->getValue(),
		       invertAffinity->getValue() );
    }
    
    return out;
    
  }
  
}
