///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DGVFFilter.hh"

#include "wishGradientVectorFlowFilter.hh"

namespace wishH3D {
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  GVFFilter::database( "wishH3DGVFFilter", 
		       &(newInstance<GVFFilter>), 
		       typeid( GVFFilter ),
		       &FilterObject::database );
  
  namespace GVFFilterInternals {
    FIELDDB_ELEMENT( GVFFilter, inputIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( GVFFilter, mu, INPUT_OUTPUT );
    FIELDDB_ELEMENT( GVFFilter, outputRGBA, INPUT_OUTPUT );
    
    template<class InputImageType>
    void executeFilter(InputImageType *in,
		       wish::Image<wish::Vec3f>* &out,
		       H3DFloat mu) {
      
      wish::GradientVectorFlowFilter<InputImageType> gvf;
      gvf.setInput( in );
      gvf.setMu( mu );
      cerr << "Executing GVFFilter..." << endl; 
      gvf.execute();
      cerr << "Execution time = " << gvf.getProcessTime() << " s" << endl;
      
      out = gvf.releaseOutput();
    }
    
    wish::Image<wish::RGBA8>* 
    convertToRGBA(wish::Image<wish::Vec3f> *in) {
      
      if( in==0 ) 
	return 0;
      
      wish::Image<wish::RGBA8> *out = new wish::Image<wish::RGBA8>;
      out->resize( *in );
      out->setLabel( in->getLabel() + "_RGBA" );
      
      int w = in->width();
      int h = in->height();
      int d = in->depth();
      
      // convert gradients
      for(int z=0; z<d; z++) {
	for(int y=0; y<h; y++) {
	  for(int x=0; x<w; x++) {
	    
	    // gradient
	    wish::Vec3f g = in->at(x,y,z);
	    
	    // normalize
	    g.normalizeSafe();
	    
	    // convert from [-1,1] to [0,255] 
	    g.x = 255.0f*0.5f*(g.x+1.0f);
	    g.x = wish::util::maxOf(g.x, 0.0f); 
	    g.x = wish::util::minOf(g.x, 255.0f); 
	    
	    g.y = 255.0f*0.5f*(g.y+1.0f);
	    g.y = wish::util::maxOf(g.y, 0.0f); 
	    g.y = wish::util::minOf(g.y, 255.0f); 
	    
	    g.z = 255.0f*0.5f*(g.z+1.0f);
	    g.z = wish::util::maxOf(g.z, 0.0f); 
	    g.z = wish::util::minOf(g.z, 255.0f);
	    
	    // set the RGB components
	    out->at(x,y,z).r = (wish::uint8) ( g.x );
	    out->at(x,y,z).g = (wish::uint8) ( g.y );
	    out->at(x,y,z).b = (wish::uint8) ( g.z );
	    
	    // the magnitude
	    out->at(x,y,z).a = 0;
	  }
	}
      }
      
      return out;
    }
    
  }
  
  GVFFilter::GVFFilter() :
    inputIndex( new SFInt32 ),
    mu( new SFFloat ),
    outputRGBA( new SFBool )
  {  
    this->type_name = "GVFFilter";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // defaults
    inputIndex->setValue( -1 );
    mu->setValue(0.5f);
    outputRGBA->setValue( false );
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  wish::ImageObject* GVFFilter::execute() {
    Volume *vol = volume->getValue();
    if( !vol ) 
      return 0;
    
    wish::ImageObject *input = vol->getImage( inputIndex->getValue() );
    if( !input )
      return 0;
    
    string dt = input->getDataType();
    if( dt=="uint8" ) {
      wish::Image8 *in = dynamic_cast<wish::Image8*>(input);
      wish::Image<wish::Vec3f> *out;
      GVFFilterInternals::
	executeFilter( in, out, 
		       mu->getValue() );
      if( outputRGBA->getValue() ) {
	wish::Image<wish::RGBA8> *rgba = 
	  GVFFilterInternals::convertToRGBA(out);
	delete out;
	return rgba;
      }
      else
	return out;
    }
    else if( dt=="uint16" ) {
      wish::Image16 *in = dynamic_cast<wish::Image16*>(input);
      wish::Image<wish::Vec3f> *out;
      GVFFilterInternals::
	executeFilter( in, out, mu->getValue() );
      if( outputRGBA->getValue() ) {
	wish::Image<wish::RGBA8> *rgba = 
	  GVFFilterInternals::convertToRGBA(out);
	delete out;
	return rgba;
      }
      else
	return out;
    }
    else if( dt=="float" ) {
      wish::Imagef *in = dynamic_cast<wish::Imagef*>(input);
      wish::Image<wish::Vec3f> *out;
      GVFFilterInternals::
	executeFilter( in, out, mu->getValue() );
      if( outputRGBA->getValue() ) {
	wish::Image<wish::RGBA8> *rgba = 
	  GVFFilterInternals::convertToRGBA(out);
	delete out;
	return rgba;
      }
      else
	return out;
    }
    else
      return 0;
    
  }
  
}
