///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DHapticMode.hh"

using namespace H3D;

namespace wishH3D {
  
  //! weights for computing the ZuckerHummel gradient
  static const float ZHWeights[27] = 
    {2.0f, 4.0f, 2.0f,
     4.0f, 6.0f, 4.0f,
     2.0f, 4.0f, 2.0f,
     
     4.0f, 6.0f, 4.0f,
     6.0f, 0.0f, 6.0f,
     4.0f, 6.0f, 4.0f,
     
     2.0f, 4.0f, 2.0f,
     4.0f, 6.0f, 4.0f,
     2.0f, 4.0f, 2.0f};
  
  // gradient functions
  wish::Vec3f HapticMode::gradZH(wish::ImageObject *im, int x, int y, int z) {
    
    float temp[27];
    int p=0;
    for(int k=-1; k<=1; k++) 
      for(int j=-1; j<=1; j++) 
	for(int i=-1; i<=1; i++, p++)
	  temp[p] = float( im->getSafeNormalized(x+i,y+j,z+k)*ZHWeights[p] );
    wish::Vec3f g;
    
    g.x = 
      temp[2]  + temp[11] + temp[20] +
      temp[5]  + temp[14] + temp[23] + 
      temp[8]  + temp[17] + temp[26] -
      
      temp[0]  - temp[9]  - temp[18] -
      temp[3]  - temp[12] - temp[21] - 
      temp[6]  - temp[15] - temp[24];
    
    g.y = 
      temp[24] + temp[25] + temp[26] +
      temp[15] + temp[16] + temp[17] + 
      temp[6]  + temp[7]  + temp[8]  -
      
      temp[18] - temp[19] - temp[20] -
      temp[9]  - temp[10] - temp[11] - 
      temp[0]  - temp[1]  - temp[2];
    
    g.z = 
      temp[18] + temp[19] + temp[20] +
      temp[21] + temp[22] + temp[23] + 
      temp[24] + temp[25] + temp[26] -
      
      temp[0]  - temp[1]  - temp[2]  -
      temp[3]  - temp[4]  - temp[5]  - 
      temp[6]  - temp[7]  - temp[8];
    
    g.x /= 30.0;
    g.y /= 30.0;
    g.z /= 30.0;
    
    return g;
  
  }
  
  wish::Vec3f HapticMode::gradZH(wish::ImageObject *im, wish::Vec3f pos) {
    
    // integer part
    int X = (int) floor(pos.x);
    int Y = (int) floor(pos.y);
    int Z = (int) floor(pos.z);
    
    // decimal part
    float ax = pos.x-X;
    float ay = pos.y-Y;
    float az = pos.z-Z;
    
    return 
      wish::util::interp(
			 wish::util::interp(
					    wish::util::interp(gradZH(im,X,Y,Z),
							       gradZH(im,X+1,Y,Z), 
							       ax),
					    wish::util::interp(gradZH(im,X,Y+1,Z),
							       gradZH(im,X+1,Y+1,Z), 
							       ax),
					    ay),
			 wish::util::interp(
					    wish::util::interp(gradZH(im,X,Y,Z+1), 
							       gradZH(im,X+1,Y,Z+1), 
							       ax),
					    wish::util::interp(gradZH(im,X,Y+1,Z+1), 
							       gradZH(im,X+1,Y+1,Z+1), 
							       ax),
					    ay),
			 az
			 );
    
    
  }
  
  wish::Vec3f HapticMode::binGradZH(wish::ImageObject *im, int x, int y, int z) {
    
    float temp[27];
    int p=0;
    for(int k=-1; k<=1; k++) 
      for(int j=-1; j<=1; j++) 
	for(int i=-1; i<=1; i++, p++)
	  temp[p] = 
	    (im->getSafeAsDouble(x+i,y+j,z+k) > 0.0f) ? ZHWeights[p] : 0.0f;
    
    wish::Vec3f g;
    
    g.x = 
      temp[2]  + temp[11] + temp[20] +
      temp[5]  + temp[14] + temp[23] + 
      temp[8]  + temp[17] + temp[26] -
      
      temp[0]  - temp[9]  - temp[18] -
      temp[3]  - temp[12] - temp[21] - 
      temp[6]  - temp[15] - temp[24];
    
    g.y = 
      temp[24] + temp[25] + temp[26] +
      temp[15] + temp[16] + temp[17] + 
      temp[6]  + temp[7]  + temp[8]  -
      
      temp[18] - temp[19] - temp[20] -
      temp[9]  - temp[10] - temp[11] - 
      temp[0]  - temp[1]  - temp[2];
    
    g.z = 
      temp[18] + temp[19] + temp[20] +
      temp[21] + temp[22] + temp[23] + 
      temp[24] + temp[25] + temp[26] -
      
      temp[0]  - temp[1]  - temp[2]  -
      temp[3]  - temp[4]  - temp[5]  - 
      temp[6]  - temp[7]  - temp[8];
    
    return g;
    
  }   
  
  // HAPTIC MODE
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase HapticMode::database( "wishH3DHapticMode", 
					&(newInstance<HapticMode>), 
					typeid( HapticMode ),
					&Node::database );
  
  namespace HapticModeInternals {
  }
  
  HapticMode::HapticMode() :
    image(0) {
    
    type_name = "HapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    WISH_H3D_CREATE_STOP();
  }
  
  void HapticMode::computeParameters( wish::VolumeHaptics &vh ) const {
    
    // this will produce zero force
    vh.deactivate();
    
  }
  
  // VISCOSITY HAPTIC MODE
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  ViscosityHapticMode::database( "wishH3DViscosityHapticMode", 
				 &(newInstance<ViscosityHapticMode>), 
				 typeid( ViscosityHapticMode ),
				 &HapticMode::database );
  
  namespace ViscosityHapticModeInternals {
    FIELDDB_ELEMENT( ViscosityHapticMode, viscosityTF, INPUT_OUTPUT );
  }
  
  ViscosityHapticMode::ViscosityHapticMode() :
    viscosityTF  ( new SFThreadSafeTransferFunction )
  {
    
    type_name = "ViscosityHapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    ConstantTransferFunction *tf = new ConstantTransferFunction;
    tf->value->setValue(0.5f);
    viscosityTF->setValue( tf );
    
    WISH_H3D_CREATE_STOP();
  }
  
  void ViscosityHapticMode::computeParameters(wish::VolumeHaptics &vh) const {
    
    // get input data
    TransferFunction *tf = viscosityTF->getValue();
    
    if( !(image && tf) ) {
      vh.deactivate();
      return;
    }
    
    // get positions
    wish::Vec3f xt = Mw2i*(vh.getProbe());
    wish::Vec3f xp = Mw2i*(vh.getProxy());
    
    // lrf based on the xyz-coords of the device
    wish::Vec3f e0 = wish::Vec3f(1,0,0);
    wish::Vec3f e1 = wish::Vec3f(0,1,0);
    wish::Vec3f e2 = wish::Vec3f(0,0,1);
    vh.setLocalReferenceFrame( e0, e1, e2 );
    
    vh.setMotionRuleTypes(wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL);
    
    float val = float( image->getNormalizedInterpd(xp) ); 
    float str = tf->evaluate(val);
    
    vh.setMotionRuleStrengths(str, str, str);
    
    return;
  }


  // SURFACE HAPTIC MODE
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  SurfaceHapticMode::database( "wishH3DSurfaceHapticMode", 
			       &(newInstance<SurfaceHapticMode>), 
			       typeid( SurfaceHapticMode ),
			       &HapticMode::database );
  
  namespace SurfaceHapticModeInternals {
    FIELDDB_ELEMENT( SurfaceHapticMode, negativeGradientAsNormal, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SurfaceHapticMode, frictionTF, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SurfaceHapticMode, distinctnessTF, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SurfaceHapticMode, strengthTF, INPUT_OUTPUT );
  }
  
  SurfaceHapticMode::SurfaceHapticMode() :
    negativeGradientAsNormal  ( new ThreadSafeSField<SFBool> ),
    frictionTF  ( new SFThreadSafeTransferFunction ),
    distinctnessTF  ( new SFThreadSafeTransferFunction ),
    strengthTF  ( new SFThreadSafeTransferFunction )
  {
    
    type_name = "SurfaceHapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    negativeGradientAsNormal->setValue( true );
    
    ConstantTransferFunction *ftf = new ConstantTransferFunction;
    ftf->value->setValue(0.12f);
    frictionTF->setValue( ftf );
    
    ConstantTransferFunction *dtf = new ConstantTransferFunction;
    dtf->value->setValue(1.0f);
    distinctnessTF->setValue( dtf );
    
    ConstantTransferFunction *stf = new ConstantTransferFunction;
    stf->value->setValue(2.0f);
    strengthTF->setValue( stf );
    
    WISH_H3D_CREATE_STOP();
  }
  
  void SurfaceHapticMode::computeParameters(wish::VolumeHaptics &vh) const {

    // get input data
    TransferFunction *fTF = frictionTF->getValue();
    TransferFunction *sTF = strengthTF->getValue();
    TransferFunction *dTF = distinctnessTF->getValue();
    bool neg = negativeGradientAsNormal->getValue();
    
    if( !(image && fTF && sTF && dTF) ) {
      vh.deactivate();
      return;
    }
    
    // get positions
    wish::Vec3f xt = Mw2i*(vh.getProbe());
    wish::Vec3f xp = Mw2i*(vh.getProxy());
    
    // get gradient
    wish::Vec3f grad;
    
    grad = gradZH( image, xp );
    
    if( neg )
      grad = -grad;
    
    // norm
    // normalized to [0,1]
    float nrm = float( grad.norm() );
    
    // if to small
    if( nrm < 1e-9 ) {
      vh.deactivate();
      return;
    }
    
    // surface mode
    vh.setMotionRuleTypes(wish::VolumeHaptics::UNILATERAL, 
			  wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL);
    
    // compute and set lrf
    wish::Vec3f e0, e1, e2;
    e0 = Mi2w.getRotation()*(grad);
    e0.normalize();
    e1 = e0.findOrthogonal();
    e1.normalize();
    e2 = e0.cross(e1);
    vh.setLRF( e0, e1, e2 );
    
    // get scalar value
    float scal = float( image->getNormalizedInterpd(xp) );
    
    // strength (threshold for penetration)
    float str1 = strengthTF->getValue()->evaluate(scal);
        
    // friction coefficient
    float fric = frictionTF->getValue()->evaluate(scal);
    float str2 = float( fabs( vh.getForceFeedback()*e0 )*fric );
    
    // distinctness
    float dst = distinctnessTF->getValue()->evaluate(nrm);
    
    vh.setMotionRuleStrengths(dst*str1,dst*str2,dst*str2);
        
  }

  // POTENTIAL HAPTIC MODE
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  PotentialHapticMode::database( "wishH3DPotentialHapticMode", 
				 &(newInstance<PotentialHapticMode>), 
				 typeid( PotentialHapticMode ),
				 &HapticMode::database );
  
  namespace PotentialHapticModeInternals {
    FIELDDB_ELEMENT( PotentialHapticMode, negativeGradient, INPUT_OUTPUT );
    FIELDDB_ELEMENT( PotentialHapticMode, viscosityTF, INPUT_OUTPUT );
    FIELDDB_ELEMENT( PotentialHapticMode, strengthTF, INPUT_OUTPUT );
  }
  
  PotentialHapticMode::PotentialHapticMode() :
    negativeGradient ( new ThreadSafeSField<SFBool> ),
    viscosityTF  ( new SFThreadSafeTransferFunction ),
    strengthTF  ( new SFThreadSafeTransferFunction )
  {
    
    type_name = "PotentialHapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    negativeGradient->setValue( true );
    
    ConstantTransferFunction *stf = new ConstantTransferFunction;
    stf->value->setValue(0.7f);
    strengthTF->setValue( stf );
    
    ConstantTransferFunction *vtf = new ConstantTransferFunction;
    vtf->value->setValue(0.3f);
    viscosityTF->setValue( vtf );
    
    WISH_H3D_CREATE_STOP();
  }
  
  void PotentialHapticMode::computeParameters(wish::VolumeHaptics &vh) const {
    
    // get input data
    TransferFunction *sTF = strengthTF->getValue();
    TransferFunction *vTF = viscosityTF->getValue();
    bool neg = negativeGradient->getValue();
    
    if( !(image && sTF && vTF) ) {
      vh.deactivate();
      return;
    }
    
    // get positions
    wish::Vec3f xt = Mw2i*(vh.getProbe());
    wish::Vec3f xp = Mw2i*(vh.getProxy());
    
    // get gradient
    wish::Vec3f grad = gradZH( image, xp );
    if( neg )
      grad = -grad;
    
    // norm
    float nrm = float( grad.norm() );
    
    // if to small
    if( nrm < 1e-9 ) {
      vh.deactivate();
      return;
    }
    
    // surface mode
    vh.setMotionRuleTypes(wish::VolumeHaptics::UNILATERAL, 
			  wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL);

    // compute and set lrf
    wish::Vec3f e0, e1, e2;
    e0 = Mi2w.getRotation()*(grad);
    e0.normalize();
    e1 = e0.findOrthogonal();
    e1.normalize();
    e2 = e0.cross(e1);
    vh.setLRF( e0, e1, e2 );
    
    // get scalar value
    float scal = float( image->getNormalizedInterpd(xp) );
    
    // strengths
    float str1 = strengthTF->getValue()->evaluate(nrm);
    float str2 = viscosityTF->getValue()->evaluate(scal);
    
    vh.setMotionRuleStrengths(str1,str2,str2);
        
  }


  // VECTOR HAPTIC MODE
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  VectorHapticMode::database( "wishH3DVectorHapticMode", 
			      &(newInstance<VectorHapticMode>), 
			      typeid( VectorHapticMode ),
			      &HapticMode::database );
  
  namespace VectorHapticModeInternals {
    FIELDDB_ELEMENT( VectorHapticMode, negativeDirection, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VectorHapticMode, viscosityTF, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VectorHapticMode, strengthTF, INPUT_OUTPUT );
  }
  
  VectorHapticMode::VectorHapticMode() :
    negativeDirection ( new ThreadSafeSField<SFBool> ),
    viscosityTF  ( new SFThreadSafeTransferFunction ),
    strengthTF  ( new SFThreadSafeTransferFunction )
  {
    
    type_name = "VectorHapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    negativeDirection->setValue( false );
    
    ConstantTransferFunction *stf = new ConstantTransferFunction;
    stf->value->setValue(1.0f);
    strengthTF->setValue( stf );
    
    ConstantTransferFunction *vtf = new ConstantTransferFunction;
    vtf->value->setValue(0.3f);
    viscosityTF->setValue( vtf );
    
    WISH_H3D_CREATE_STOP();
  }
  
  void VectorHapticMode::computeParameters(wish::VolumeHaptics &vh) const {
    
    // get input data
    wish::Image<wish::Vec3f> *vecim = 
      dynamic_cast<wish::Image<wish::Vec3f>*>(image);
    
    TransferFunction *sTF = strengthTF->getValue();
    TransferFunction *vTF = viscosityTF->getValue();
    bool neg = negativeDirection->getValue();
    
    if( !(vecim && sTF && vTF) ) {
      vh.deactivate();
      return;
    }
    
    // get positions
    wish::Vec3f xt = Mw2i*(vh.getProbe());
    wish::Vec3f xp = Mw2i*(vh.getProxy());
    
    // get vector
    wish::Vec3f vec = vecim->getInterp( xp );
    if( neg )
      vec = -vec;
    
    // norm
    float nrm = float( vec.norm() );
    
    // if to small
    if( nrm < 1e-9 ) {
      vh.deactivate();
      return;
    }
    
    // surface mode
    vh.setMotionRuleTypes(wish::VolumeHaptics::UNILATERAL, 
			  wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL);
    
    // compute and set lrf
    wish::Vec3f e0, e1, e2;
    e0 = Mi2w.getRotation()*(vec);
    e0.normalize();
    e1 = e0.findOrthogonal();
    e1.normalize();
    e2 = e0.cross(e1);
    vh.setLRF( e0, e1, e2 );
    
    // strengths
    float str1 = strengthTF->getValue()->evaluate(nrm);
    float str2 = viscosityTF->getValue()->evaluate(nrm);
    
    vh.setMotionRuleStrengths(str1,str2,str2);
        
  }
  

  // DIRECTED HAPTIC MODE
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  DirectedHapticMode::database( "wishH3DDirectedHapticMode", 
				&(newInstance<DirectedHapticMode>), 
				typeid( DirectedHapticMode ),
				&HapticMode::database );
  
  namespace DirectedHapticModeInternals {
    FIELDDB_ELEMENT( DirectedHapticMode, negativeDirection, INPUT_OUTPUT );
    FIELDDB_ELEMENT( DirectedHapticMode, viscosityTF, INPUT_OUTPUT );
    FIELDDB_ELEMENT( DirectedHapticMode, strengthTF, INPUT_OUTPUT );
  }
  
  DirectedHapticMode::DirectedHapticMode() :
    negativeDirection ( new ThreadSafeSField<SFBool> ),
    viscosityTF  ( new SFThreadSafeTransferFunction ),
    strengthTF  ( new SFThreadSafeTransferFunction )
  {
    
    type_name = "DirectedHapticMode";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    negativeDirection->setValue( false );
    
    ConstantTransferFunction *stf = new ConstantTransferFunction;
    stf->value->setValue(1.0f);
    strengthTF->setValue( stf );
    
    ConstantTransferFunction *vtf = new ConstantTransferFunction;
    vtf->value->setValue(0.3f);
    viscosityTF->setValue( vtf );
    
    WISH_H3D_CREATE_STOP();
  }
  
  void DirectedHapticMode::computeParameters(wish::VolumeHaptics &vh) const {
    
    // get input data
    wish::Image<wish::Vec3f> *vecim = 
      dynamic_cast<wish::Image<wish::Vec3f>*>(image);
    
    TransferFunction *sTF = strengthTF->getValue();
    TransferFunction *vTF = viscosityTF->getValue();
    bool neg = negativeDirection->getValue();
    
    if( !(vecim && sTF && vTF) ) {
      vh.deactivate();
      return;
    }
    
    // get positions
    wish::Vec3f xt = Mw2i*(vh.getProbe());
    wish::Vec3f xp = Mw2i*(vh.getProxy());
    
    // get vector
    wish::Vec3f vec = vecim->getInterp( xp );
    if( neg )
      vec = -vec;
    
    // norm
    float nrm = float( vec.norm() );
    
    // if to small
    if( nrm < 1e-9 ) {
      vh.deactivate();
      return;
    }
    
    // surface mode
    vh.setMotionRuleTypes(wish::VolumeHaptics::DIRECTED_FORCE, 
			  wish::VolumeHaptics::BILATERAL, 
			  wish::VolumeHaptics::BILATERAL);
    
    // compute and set lrf
    wish::Vec3f e0, e1, e2;
    e0 = Mi2w.getRotation()*(vec);
    e0.normalize();
    e1 = e0.findOrthogonal();
    e1.normalize();
    e2 = e0.cross(e1);
    vh.setLRF( e0, e1, e2 );
    
    // strengths
    float str1 = strengthTF->getValue()->evaluate(nrm);
    float str2 = viscosityTF->getValue()->evaluate(nrm);
    
    vh.setMotionRuleStrengths(str1,str2,str2);
    
  }
  
}
