///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __WISH_H3D_HAPTIC_MODE_HH__
#define __WISH_H3D_HAPTIC_MODE_HH__

#include "Node.h"
#include "ThreadSafeFields.h"
#include "wishH3D.hh"
#include "wishH3DTransferFunction.hh"
#include "wishImage.hh"
#include "wishVolumeHaptics.hh"

using namespace H3D;

namespace wishH3D {
  
  class WISHH3D_API HapticMode : public Node {
  public:
    
    // typedefs
    typedef Node BaseClass;
    typedef ThreadSafeRefSField<SFTransferFunction> SFThreadSafeTransferFunction;
    
    // gradient functions useful for haptic rendering
    static wish::Vec3f gradZH(wish::ImageObject *im, int x, int y, int z);
    static wish::Vec3f gradZH(wish::ImageObject *im, wish::Vec3f pos);
    static wish::Vec3f binGradZH(wish::ImageObject *im, int x, int y, int z);
    static wish::Vec3f binGradZH(wish::ImageObject *im, wish::Vec3f pos);
    
    /// Constructor.
    HapticMode();
    
    virtual string defaultXMLContainerField() {
      return "hapticMode";
    }
    
    virtual void setImage( wish::ImageObject *_im ) {
      image = _im;
      if( image ) {
	Mw2i = image->getWorldToImageMatrix();
	Mi2w = image->getImageToWorldMatrix();
      }
    }
    
    virtual wish::ImageObject* getImage() const {
      return image;
    }
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  protected:
    wish::ImageObject *image;
    wish::Matrix4f Mw2i;
    wish::Matrix4f Mi2w;
  };
  
  typedef TypedSFNode<HapticMode> SFHapticMode;
  
  class WISHH3D_API ViscosityHapticMode : public HapticMode {
  public:
    
    typedef HapticMode BaseClass;
    
    /// Constructor.
    ViscosityHapticMode();
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    // the transfer function
    auto_ptr< SFThreadSafeTransferFunction > viscosityTF;
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  };

  class WISHH3D_API SurfaceHapticMode : public HapticMode {
  public:
    
    typedef HapticMode BaseClass;
    
    /// Constructor.
    SurfaceHapticMode();
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    // surface direction
    auto_ptr< ThreadSafeSField<SFBool> > negativeGradientAsNormal;
    
    // the transfer functions
    auto_ptr< SFThreadSafeTransferFunction > frictionTF; // friction coeff
    auto_ptr< SFThreadSafeTransferFunction > strengthTF; // threshold for penetration
    auto_ptr< SFThreadSafeTransferFunction > distinctnessTF; // modulate force by gradient magnitude

    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  };
  
  class WISHH3D_API PotentialHapticMode : public HapticMode {
  public:
    
    typedef HapticMode BaseClass;
    
    /// Constructor.
    PotentialHapticMode();
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    // surface direction
    auto_ptr< ThreadSafeSField<SFBool> > negativeGradient;
    
    // the transfer functions
    auto_ptr< SFThreadSafeTransferFunction > strengthTF; 
    auto_ptr< SFThreadSafeTransferFunction > viscosityTF; // viscosity in the plane
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  };

  class WISHH3D_API VectorHapticMode : public HapticMode {
  public:
    
    typedef HapticMode BaseClass;
    
    /// Constructor.
    VectorHapticMode();
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    // surface direction
    auto_ptr< ThreadSafeSField<SFBool> > negativeDirection;
    
    // the transfer functions
    auto_ptr< SFThreadSafeTransferFunction > strengthTF; 
    auto_ptr< SFThreadSafeTransferFunction > viscosityTF; // viscosity in the plane
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  };

  class WISHH3D_API DirectedHapticMode : public HapticMode {
  public:
    
    typedef HapticMode BaseClass;
    
    /// Constructor.
    DirectedHapticMode();
    
    virtual void computeParameters( wish::VolumeHaptics &vh ) const;
    
    // surface direction
    auto_ptr< ThreadSafeSField<SFBool> > negativeDirection;
    
    // the transfer functions
    auto_ptr< SFThreadSafeTransferFunction > strengthTF; 
    auto_ptr< SFThreadSafeTransferFunction > viscosityTF; // viscosity in the plane
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  };

}

#endif
