///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "Shape.h"
#include "Appearance.h"
#include "Material.h"
#include "SmoothSurface.h"
#include "H3DHapticsDevice.h"

#include "wishH3DLiveMarching.hh"

#include "wishGaussianGradientMagnitudeFilter.hh"
#include "wishSigmoidFilter.hh"

namespace wishH3D {
  
  void LiveMarching::AddPoint::update() {
    SFBool::update();
    if( value ) {
      cerr << "LiveMarching::AddPoint::update()" << endl;
      LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
      Volume *vol = lm->volume->getValue();
      
      // safety
      if( lm->active->getValue()==false || vol==0 || vol->getImage(0)==0 )
	return;
      
      // update if we have a new slice
      Vec3f slctr = lm->sliceTranslation->getValue();
      int slc = 
	int(floor(vol->getImage(0)->worldToImage(Vec3Cast(slctr)).z));
      if( slc != lm->current_slice ) {
	lm->imageSlice->update();
      }
      
      // put seed
      Vec3f pos = vol->proxyPosition->getValue();
      Vec3f voxpos = vol->worldToImage(pos);
      voxpos.z = 0.0f;
      
      if( lm->cost2D && 
	  lm->cost2D->isInside(int(floor(voxpos.x)), 
			       int(floor(voxpos.y)), 
			       int(floor(voxpos.z))) ) {
	lm->live_marching->setCursor( Vec3Cast(voxpos) );
	lm->live_marching->putSeed();
      }
    }
  }
  
  void LiveMarching::DeletePoint::update() {
    SFBool::update();
    if( value ) {
      cerr << "LiveMarching::DeletePoint::update()" << endl;
      LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
      lm->live_marching->removeSeed();
    }
  }
  
  void LiveMarching::CloseContour::update() {
    SFBool::update();
    if( value ) {
      cerr << "LiveMarching::CloseContour::update()" << endl;
      LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
      if( lm->cost2D ) {
	// for now we do it all here, closing means close, rasterize, and clear
	lm->live_marching->closeCurve();
	lm->rasterizeButton->setValue( true );
	lm->clearButton->setValue( true );
      }
    }
  }
  
  void LiveMarching::ClearPoints::update() {
    SFBool::update();
    if( value ) {
      cerr << "LiveMarching::ClearPoints::update()" << endl;
      LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
      lm->live_marching->clearSeeds();
    }
  }
  
  void LiveMarching::ImageSlice::update() {
    int ind = static_cast<SFInt32*>(routes_in[0]) -> getValue();
    
    LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
    Volume *vol = lm->volume->getValue();
    
    wish::ImageObject *im = vol->getImage( ind );
    if( im ) {
      Vec3f slctr = lm->sliceTranslation->getValue();
      int slc = 
	int(floor(im->worldToImage(Vec3Cast(slctr)).z));
      
      if( 0 <= slc && slc < im->depth() ) {
	lm->current_slice = slc;
	if( im->getDataType()=="uint8" ) {
	  wish::Image8 *im2 = new wish::Image8(im->width(), im->height());
	  im2->setData( &(static_cast<wish::Image8*>(im)->at(0,0,slc)) );
	  im2->setDataOwner( false );
	  lm->setCostSlice( im2 );
	}
	else if( im->getDataType()=="uint16" ) {
	  wish::Image16 *im2 = new wish::Image16(im->width(), im->height());
	  im2->setData( &(static_cast<wish::Image16*>(im)->at(0,0,slc)) );
	  im2->setDataOwner( false );
	  lm->setCostSlice( im2 );
	}
	else if( im->getDataType()=="float" ) {
	  wish::Imagef *im2 = new wish::Imagef(im->width(), im->height());
	  im2->setData( &(static_cast<wish::Imagef*>(im)->at(0,0,slc)) );
	  im2->setDataOwner( false );
	  lm->setCostSlice( im2 );
	}
      }
    }
  }
  
  void LiveMarching::Rasterizer::update() {
    
    SFBool::update();
    if( value ) {
      cerr << "LiveMarching::Rasterizer::update()" << endl;
      LiveMarching *lm = static_cast<LiveMarching*>( this->getOwner() );
      Volume *vol = lm->volume->getValue();
      
      int oind = lm->outputIndex->getValue();
      wish::Image8 *out = dynamic_cast<wish::Image8*>(vol->getImage(oind));
      if( !out ) {
	cerr << "LiveMarching::Rasterizer - Output image is not valid!" 
	     << endl;
	return;
      }
      
      lm->live_marching->rasterize( out, lm->current_slice, false );
    }      
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase LiveMarching::database( "wishH3DLiveMarching", 
					  &(newInstance<LiveMarching>), 
					  typeid( LiveMarching ),
					  &Group::database );
  
  namespace LiveMarchingInternals {
    FIELDDB_ELEMENT( LiveMarching, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, addButton, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, deleteButton, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, closeButton, INPUT_OUTPUT );    
    FIELDDB_ELEMENT( LiveMarching, clearButton, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, interactionButton, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, rasterizeButton, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, imageIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, outputIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, sliceTranslation, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, active, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LiveMarching, minBacktrackDistance, INPUT_OUTPUT );
  }
  
  LiveMarching::LiveMarching() :
    addButton( new SFBool ),
    addPoint( new AddPoint ),
    deleteButton( new SFBool ),
    deletePoint( new DeletePoint ),
    closeButton( new SFBool ),
    closeContour( new CloseContour ),
    clearButton( new SFBool ),
    clearPoints( new ClearPoints ),
    interactionButton( new SFBool ),
    rasterizeButton( new SFBool ),
    rasterizer( new Rasterizer ),
    imageIndex( new SFInt32 ),
    outputIndex( new SFInt32 ),
    imageSlice( new ImageSlice ),
    sliceTranslation( new SFVec3f ),
    live_marching  ( new wish::LiveMarching(1024*1024) ),
    active( new SFBool ),
    minBacktrackDistance( new SFFloat ),
    cost2D(0),
    current_slice(0)
  {

    type_name = "LiveMarching";
    database.initFields(this);

    WISH_H3D_CREATE_START();
    
    // dependencies
    addPoint -> setOwner(this);
    deletePoint -> setOwner(this);
    closeContour -> setOwner(this);
    clearPoints -> setOwner(this);
    imageSlice -> setOwner(this);
    rasterizer -> setOwner(this);

    // defaults
    addButton -> setValue(false);
    deleteButton -> setValue(false);
    closeButton -> setValue( false );
    clearButton -> setValue( false );
    interactionButton -> setValue( false );
    rasterizeButton -> setValue( false );
    imageIndex -> setValue( -1 );
    outputIndex -> setValue( -1 );
    sliceTranslation -> setValue( Vec3f(0,0,0) );
    active->setValue( false );
    minBacktrackDistance->setValue(5.0f);
    
    // routings
    addButton -> routeNoEvent( addPoint );
    deleteButton -> routeNoEvent( deletePoint );
    closeButton -> routeNoEvent( closeContour );
    clearButton -> routeNoEvent( clearPoints );
    imageIndex -> routeNoEvent( imageSlice );
    rasterizeButton ->routeNoEvent( rasterizer );
    
    addButton -> routeNoEvent( displayList );
    deleteButton -> routeNoEvent( displayList );
    closeButton -> routeNoEvent( displayList );
    clearButton -> routeNoEvent( displayList );
    interactionButton -> routeNoEvent( displayList );
    imageSlice -> routeNoEvent( displayList );
    
    WISH_H3D_CREATE_STOP();
  };

  void LiveMarching::initialize() {
   
    if( isInitialized() )
      return;
    
    WISH_H3D_INIT_START();
    
    Group::initialize();
    
    WISH_H3D_INIT_STOP();
  }
  
  void LiveMarching::traverseSG( TraverseInfo &ti ) {
    // base class
    Group::traverseSG(ti);
    
    // get the volume
    Volume *vol = volume -> getValue();
    
    // safety
    if( active->getValue()==false || vol==0 || vol->getImage(0)==0 )
      return;
    
    // interaction button pressed
    if( interactionButton->getValue() && cost2D!=0 ) {
      
      // get the probe position
      Vec3f pos = vol->proxyPosition->getValue();
      Vec3f voxpos = vol->worldToImage(pos);
      voxpos.z = 0.0f;
      
      // if inside, set cursor position
      if( cost2D && 
	  cost2D->isInside(int(floor(voxpos.x)), 
			   int(floor(voxpos.y)), 
			   int(floor(voxpos.z))) ) {
	live_marching->
	  setMinimumBacktrackDistance(minBacktrackDistance->getValue());
	live_marching->setCursor( Vec3Cast(voxpos) );
	live_marching->update();
	interactionButton->touch();
      }
      
    }
    
    // interaction button released
    else {
      
    } 
    
  }
  
  // render the contour
  void LiveMarching::render() {
    
    // the volume
    Volume *vol = volume->getValue();
    
    // safety
    if( active->getValue()==false || vol==0 || vol->getImage(0)==0 )
      return;
    
    Vec3f slctr = sliceTranslation->getValue();
    
    const vector<wish::Vec3i> &sds = live_marching->getSeeds();
    const vector<wish::Vec3f> &btc = live_marching->getBacktrackCursor();
    const vector< vector<wish::Vec3f> > &bt = live_marching->getBacktrack();
    
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glDisable( GL_LIGHTING );
    glDisable( GL_TEXTURE_2D );
    glDisable( GL_TEXTURE_3D );
    
    // offset to render on top of slice
    float offset = 0.0f;
    glDepthFunc( GL_ALWAYS ); // makes the curve always visible
    
    wish::Matrix4f Mi2w = vol->getImage(0)->getImageToWorldMatrix();
    
    // cursor
    glLineWidth(2.0f);    
    glColor3f(0.0f,1.0f,0.0f);
    glBegin(GL_LINE_STRIP);
    for(unsigned int i=0; i<btc.size(); i++) {
      wish::Vec3f v = btc[i];
      v = Mi2w*v;
      v.z = slctr.z+offset;
      glVertex3fv(v);
    }
    glEnd();
    
    // old
    glColor3f(1.0f,0.0f,0.0f);
    for(unsigned int j=0; j<bt.size(); j++) {
      glBegin(GL_LINE_STRIP);
      for(unsigned int k=0; k<bt[j].size(); k++) {
	wish::Vec3f v = bt[j][k];
	v = Mi2w*v;
	v.z = slctr.z+offset;
	glVertex3fv(v);
      }
      glEnd();
    }
    
    // seeds
    glPointSize(5.0f);
    glColor3f(0.0f,0.0f,1.0f);    
    glBegin(GL_POINTS);
    for(unsigned int l=0; l<sds.size(); l++) {
      wish::Vec3f v = sds[l];
      v = Mi2w*v;
      v.z = slctr.z+offset;
      glVertex3fv(v);
    }
    glEnd();
    
    glPopAttrib();
  }
  
  
}  


  
