///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DTransferFunction.hh"

namespace wishH3D {
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  TransferFunction::database( "wishH3DTransferFunction", 
			      &(newInstance<TransferFunction>), 
			      typeid( TransferFunction ),
			      &Node::database );
  
  namespace TransferFunctionInternals {
  }
  
  TransferFunction::TransferFunction() : modified( new Modified )
  {
    
    type_name = "TransferFunction";
    database.initFields( this );

  }
  

  // CONTRAST/BRIGHTNESS
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  CBTransferFunction::database( "wishH3DCBTransferFunction", 
				&(newInstance<CBTransferFunction>), 
				typeid( CBTransferFunction ),
				&TransferFunction::database );
  
  namespace CBTransferFunctionInternals {
    FIELDDB_ELEMENT( CBTransferFunction, contrast, INPUT_OUTPUT );
    FIELDDB_ELEMENT( CBTransferFunction, brightness, INPUT_OUTPUT );
    FIELDDB_ELEMENT( CBTransferFunction, maxvalue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( CBTransferFunction, minvalue, INPUT_OUTPUT );
  }
  
  CBTransferFunction::CBTransferFunction() :
    contrast   (new SFFloat),
    brightness (new SFFloat),
    minvalue   (new SFFloat),
    maxvalue   (new SFFloat)
  {
    type_name = "CBTransferFunction";
    database.initFields( this );
    
    contrast -> route( modified );
    brightness -> route( modified );
    contrast -> setValue(1.0);
    brightness -> setValue(0.0);
  
    // max and min (0,1), (0,255), (-1,1), ...
    minvalue -> setValue(0.0);
    maxvalue -> setValue(1.0);
  }

  // evaluate function
  H3DFloat CBTransferFunction::evaluate(const H3DFloat &arg) const {
    H3DFloat a = minvalue -> getValue();
    H3DFloat b = maxvalue -> getValue();
    
    // Remember that we adjust 1/deltaX, where deltaX is the contrast
    // It should be (b-a)/ctr, but we use (b-a)*ctr
    H3DFloat ctr = contrast -> getValue();
    
    H3DFloat brt = brightness -> getValue();
  
    H3DFloat f = (arg-(a+(b-a)/2)+brt)*(b-a)*ctr+(a+(b-a)/2);
    
    f = H3DMin(f,1.0f);
    f = H3DMin(f,0.0f);
    
    return f;
  }

  // WindowLevel
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  WindowLevelTransferFunction::database( "wishH3DWindowLevelTransferFunction", 
					 &(newInstance<WindowLevelTransferFunction>), 
					 typeid( WindowLevelTransferFunction ),
					 &TransferFunction::database );
  
  namespace WindowLevelTransferFunctionInternals {
    FIELDDB_ELEMENT( WindowLevelTransferFunction, window, INPUT_OUTPUT );
    FIELDDB_ELEMENT( WindowLevelTransferFunction, level, INPUT_OUTPUT );
    FIELDDB_ELEMENT( WindowLevelTransferFunction, displaymin, INPUT_OUTPUT );
    FIELDDB_ELEMENT( WindowLevelTransferFunction, displaymax, INPUT_OUTPUT );
  }
  
  WindowLevelTransferFunction::WindowLevelTransferFunction() :
    window   (new SFFloat),
    level    (new SFFloat),
    displaymin   (new SFFloat),
    displaymax   (new SFFloat)
  {
    type_name = "WindowLevelTransferFunction";
    database.initFields( this );
    
    window -> route( modified );
    level -> route( modified );
    window -> setValue(1.0);
    level -> setValue(0.5);
    
    displaymin -> setValue(0.0);
    displaymax -> setValue(1.0);
  }
  
  // evaluate function
  H3DFloat WindowLevelTransferFunction::evaluate(const H3DFloat &arg) const {
    H3DFloat D0 = displaymin -> getValue();
    H3DFloat D1 = displaymax -> getValue();
    
    H3DFloat w = window -> getValue();
    H3DFloat l = level -> getValue();
    
    //H3DFloat f = (arg-l)/w + 0.5f;
    H3DFloat f = D0 + (D1-D0)*(arg-l+w/2.0f)/w;
    
    f = H3DMax(f,D0);
    f = H3DMin(f,D1);
    
    return f;
  }
  

  // SIGMOID
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  SigmoidTransferFunction::database( "wishH3DSigmoidTransferFunction", 
				     &(newInstance<SigmoidTransferFunction>), 
				     typeid( SigmoidTransferFunction ),
				     &TransferFunction::database );
  
  namespace SigmoidTransferFunctionInternals {
    FIELDDB_ELEMENT( SigmoidTransferFunction, alpha, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SigmoidTransferFunction, beta, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SigmoidTransferFunction, maxvalue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( SigmoidTransferFunction, minvalue, INPUT_OUTPUT );
  }
  
  SigmoidTransferFunction::SigmoidTransferFunction() :
    alpha      (new SFFloat),
    beta       (new SFFloat),
    minvalue   (new SFFloat),
    maxvalue   (new SFFloat)
  {

    type_name = "SigmoidTransferFunction";
    database.initFields( this );
    
    alpha -> route( modified );
    beta -> route( modified );
    alpha -> setValue(1.0);
    beta -> setValue(0.0);
  
    // max and min (0,1), (0,255), (-1,1), ...
    minvalue -> setValue(0.0);
    maxvalue -> setValue(1.0);
  }

  // evaluate function
  H3DFloat SigmoidTransferFunction::evaluate(const H3DFloat &arg) const {
  
    H3DFloat mn = minvalue -> getValue();
    H3DFloat mx = maxvalue -> getValue();
    H3DFloat a  = alpha -> getValue();
    H3DFloat b  = beta -> getValue();
  
    H3DFloat f = (mx-mn)*1.0f/(1+exp(-(arg-b)/(a+1e-3f)))+mn;
  
    return f;
  }

  // GAUSSIAN
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  GaussianTransferFunction::database( "wishH3DGaussianTransferFunction", 
				     &(newInstance<GaussianTransferFunction>), 
				     typeid( GaussianTransferFunction ),
				     &TransferFunction::database );
  
  namespace GaussianTransferFunctionInternals {
    FIELDDB_ELEMENT( GaussianTransferFunction, mean, INPUT_OUTPUT );
    FIELDDB_ELEMENT( GaussianTransferFunction, sigma, INPUT_OUTPUT );
    FIELDDB_ELEMENT( GaussianTransferFunction, maxvalue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( GaussianTransferFunction, minvalue, INPUT_OUTPUT );
  }
  
  GaussianTransferFunction::GaussianTransferFunction() :
    mean      (new SFFloat),
    sigma     (new SFFloat),
    minvalue  (new SFFloat),
    maxvalue  (new SFFloat)
  {

    type_name = "GaussianTransferFunction";
    database.initFields( this );
    
    mean -> route( modified );
    sigma -> route( modified );
    mean -> setValue(0.0);
    sigma -> setValue(1.0);
  
    // max and min (0,1), (0,255), (-1,1), ...
    minvalue -> setValue(0.0);
    maxvalue -> setValue(1.0);
  }
  
  // evaluate function
  H3DFloat GaussianTransferFunction::evaluate(const H3DFloat &arg) const {
  
    H3DFloat mn = minvalue -> getValue();
    H3DFloat mx = maxvalue -> getValue();
    H3DFloat sig = sigma -> getValue();
    H3DFloat avg = mean -> getValue();
    
    H3DFloat f = (mx-mn)*exp(-0.5f*(arg-avg)*(arg-avg)/sig/sig)+mn;
    
    return f;
  }
  
  // PWL
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  PWLTransferFunction::database( "wishH3DPWLTransferFunction", 
				 &(newInstance<PWLTransferFunction>), 
				 typeid( PWLTransferFunction ),
				 &TransferFunction::database );
  
  namespace PWLTransferFunctionInternals {
    FIELDDB_ELEMENT( PWLTransferFunction, xsamples, INPUT_OUTPUT );
    FIELDDB_ELEMENT( PWLTransferFunction, ysamples, INPUT_OUTPUT );
  }
  
  PWLTransferFunction::PWLTransferFunction() :
    xsamples ( new MFFloat ),
    ysamples ( new MFFloat ) {
    
    type_name = "PWLTransferFunction";
    database.initFields( this );
    
    xsamples -> route( modified );
    ysamples -> route( modified );
    
  }

  // evaluate function
  // this function is not safe, you must provide valid ranges and values
  H3DFloat PWLTransferFunction::evaluate(const H3DFloat &arg) const {
  
    vector<H3DFloat> vx = xsamples -> getValue();
    vector<H3DFloat> vy = ysamples -> getValue();
    
    int sample_count = int( vy.size() );
    
    H3DFloat value;
  
    if( sample_count == 0 ) // no samples
      value = 0.0;
  
    else if( sample_count == 1 ) // constant function
      value = vy[0];
    
    else if( arg < vx[0] ) // clamping
      value = vy[0];
  
    else if( arg > vx[sample_count-1] ) 
      value = vy[sample_count-1];
    
    else { // interpolate (linearly)
      for(int i=0; i<sample_count-1; i++) {
	if( vx[i] <= arg && arg <= vx[i+1] ) {
	  H3DFloat alpha = arg-vx[i];
	  value = vy[i] + alpha*(vy[i+1]-vy[i])/(vx[i+1]-vx[i]);
	}
      }
    }
    
    return value;
  }

  // CONSTANT  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  ConstantTransferFunction::database( "wishH3DConstantTransferFunction", 
				      &(newInstance<ConstantTransferFunction>), 
				      typeid( ConstantTransferFunction ),
				      &TransferFunction::database );
  
  namespace ConstantTransferFunctionInternals {
    FIELDDB_ELEMENT( ConstantTransferFunction, value, INPUT_OUTPUT );
  }
  
  ConstantTransferFunction::ConstantTransferFunction() : 
    value (new SFFloat) 
  {
    
    type_name = "ConstantTransferFunction";
    database.initFields( this );
    
    value -> route( modified );
    value -> setValue(1.0);
  }

  // evaluate function
  H3DFloat ConstantTransferFunction::evaluate(const H3DFloat &arg) const {
    return value->getValue();
  }


  // LINEAR  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  LinearTransferFunction::database( "wishH3DLinearTransferFunction", 
				    &(newInstance<LinearTransferFunction>), 
				    typeid( LinearTransferFunction ),
				    &TransferFunction::database );
  
  namespace LinearTransferFunctionInternals {
    FIELDDB_ELEMENT( LinearTransferFunction, slope, INPUT_OUTPUT );
    FIELDDB_ELEMENT( LinearTransferFunction, offset, INPUT_OUTPUT );
  }
  
  LinearTransferFunction::LinearTransferFunction() : 
    slope(new SFFloat),
    offset( new SFFloat )
  {
    
    type_name = "LinearTransferFunction";
    database.initFields( this );
    
    // routings
    slope -> route( modified );
    offset -> route( modified );
    
    // settings
    slope -> setValue( 1.0f );
    offset -> setValue( 0.0f );
  }
  
  // evaluate function
  H3DFloat LinearTransferFunction::evaluate(const H3DFloat &arg) const {
    return offset->getValue() + slope->getValue()*arg;
  }
  
  // BUTTERWORTH
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  BWTransferFunction::database( "wishH3DBWTransferFunction", 
				&(newInstance<BWTransferFunction>), 
				typeid( BWTransferFunction ),
				&TransferFunction::database );
  
  namespace BWTransferFunctionInternals {
    FIELDDB_ELEMENT( BWTransferFunction, offset, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BWTransferFunction, degree, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BWTransferFunction, base, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BWTransferFunction, factor, INPUT_OUTPUT );
    FIELDDB_ELEMENT( BWTransferFunction, reject, INPUT_OUTPUT );
  }
  
  BWTransferFunction::BWTransferFunction() :
    offset ( new SFFloat ),
    degree ( new SFInt32 ),
    base   ( new SFFloat ),
    factor ( new SFFloat ),
    reject ( new SFBool  )
  {
    
    type_name = "BWTransferFunction";
    database.initFields( this );
    
    offset -> route( modified );
    degree -> route( modified );
    base -> route( modified );
    factor -> route( modified );
    reject -> route( modified );
    
    // default settings
    offset -> setValue(0);
    degree -> setValue(5);
    base -> setValue(.5);
    factor -> setValue(1.0);
    reject -> setValue( true );
  }

  // evaluate function
  H3DFloat BWTransferFunction::evaluate(const H3DFloat &arg) const {
    
    if( reject -> getValue() )
      return factor->getValue()*
	(1.0f - 1.0f/(1.0f+pow( (arg-offset->getValue())/base->getValue(), 
				2.0f*degree->getValue()) ) );
    
    return factor->getValue()*
      (1.0f/(1.0f+pow( (arg-offset->getValue())/base->getValue(), 
		       2.0f*degree->getValue()) ) );
  }
  
}
