///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_H3D_TRANSFER_FUNCTION_HH_
#define _WISH_H3D_TRANSFER_FUNCTION_HH_

#include "SFNode.h"
#include "MFNode.h"
#include "DependentNodeFields.h"
#include "SFFloat.h"
#include "MFFloat.h"
#include "SFBool.h"
#include "SFInt32.h"
#include "wishH3D.hh"

using namespace H3D;

namespace wishH3D {
  
  class WISHH3D_API TransferFunction : public Node {
  public:
    struct Modified : 
      AutoUpdate<Field> {
    };
    
    // the interface
    static H3DNodeDatabase database;
    
    // constructor
    TransferFunction();  
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const {
      return 1.0f;
    }
    
    auto_ptr<Modified> modified;  
  };
  
  /// Contrast brigtness transfer function
  class WISHH3D_API CBTransferFunction : public TransferFunction {
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    CBTransferFunction();
      
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
    
    // contrast and brightness
    auto_ptr<SFFloat> contrast;
    auto_ptr<SFFloat> brightness;
  
    // max and min (0,1), (0,255), (-1,1),...
    auto_ptr<SFFloat> minvalue;
    auto_ptr<SFFloat> maxvalue;
  };
  
  
  /// Window/Level transfer function
  /// Common contrast stretching for medical images, the level is the center 
  /// intensity and window is the width of the intensity window. D0 and D1 are
  /// the minimum and maximum display values, respectively. 
  /// Example: A 1:1 mapping of a 0-255 grayscale image
  /// will have a level of 127.5 and a window of 255. 
  class WISHH3D_API WindowLevelTransferFunction : public TransferFunction {
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    WindowLevelTransferFunction();
      
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
    
    // contrast and brightness
    auto_ptr<SFFloat> window;
    auto_ptr<SFFloat> level;
  
    // max and min (0,1), (0,255), (-1,1),...
    auto_ptr<SFFloat> displaymin;
    auto_ptr<SFFloat> displaymax;
  };


  
  /// Sigmoid transfer function
  class WISHH3D_API SigmoidTransferFunction : public TransferFunction {

    // out(x) = (max-min)*1.0/(1+exp(-(in(x)-beta)/alpha))+min

  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    SigmoidTransferFunction();
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
  
    // contrast and brightness
    auto_ptr<SFFloat> alpha;
    auto_ptr<SFFloat> beta;
  
    // max and min (0,1), (0,255), (-1,1),...
    auto_ptr<SFFloat> minvalue;
    auto_ptr<SFFloat> maxvalue;
  };

  /// Gaussian transfer function
  class WISHH3D_API GaussianTransferFunction : public TransferFunction {

    // out(x) = (max-min)*exp(-0.5*(in(x)-mean)^2/sigma^2)+min

  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    GaussianTransferFunction();
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
  
    // contrast and brightness
    auto_ptr<SFFloat> mean;
    auto_ptr<SFFloat> sigma;
  
    // max and min (0,1), (0,255), (-1,1),...
    auto_ptr<SFFloat> minvalue;
    auto_ptr<SFFloat> maxvalue;
  };

  /// Constant transfer function
  class WISHH3D_API ConstantTransferFunction : public TransferFunction {
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    ConstantTransferFunction();
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
  
    // value
    auto_ptr<SFFloat> value;
  };

  /// Linear transfer function
  class WISHH3D_API LinearTransferFunction : public TransferFunction {
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    LinearTransferFunction();
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
    
    // value
    auto_ptr<SFFloat> slope;
    auto_ptr<SFFloat> offset;
  };
  
  /// Piecewise linear transfer function
  class WISHH3D_API PWLTransferFunction : public TransferFunction {
  
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    PWLTransferFunction();
    
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
  
    // samples
    auto_ptr<MFFloat> xsamples;
    auto_ptr<MFFloat> ysamples;
  
  };

  // Transfer function based on the function of a Butterworth filter
  class WISHH3D_API BWTransferFunction : public TransferFunction {
  public:
    // the interface
    static H3DNodeDatabase database;
  
    // constructor
    BWTransferFunction();
  
    // evaluate function
    virtual H3DFloat evaluate(const H3DFloat &arg) const;
  
    // offset and slope
    auto_ptr<SFFloat> offset;  // x-x0
    auto_ptr<SFInt32> degree;  // n
    auto_ptr<SFFloat> base;    // w0
    auto_ptr<SFFloat> factor;  // 0 < f(x) < factor 
    auto_ptr<SFBool>  reject;  // pass or reject
  };
  
  typedef 
  DependentSFNode< FieldRef<TransferFunction,
			    TransferFunction::Modified,
			    &TransferFunction::modified> > SFTransferFunction;
  

}
#endif
