///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DTransferFunctionTexture.hh"

#include "TextureProperties.h"

using namespace H3D;

namespace wishH3D {
  
  // number of elements i the 1D texture 
  const int TransferFunctionTexture::NTEX = 256;
  
  // add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  TransferFunctionTexture::database( "wishH3DTransferFunctionTexture", 
				     &(newInstance<TransferFunctionTexture>), 
				     typeid( TransferFunctionTexture ),
				     &X3DTexture2DNode::database );

  namespace TransferFunctionTextureInternals {
    FIELDDB_ELEMENT( TransferFunctionTexture, red, INPUT_OUTPUT );
    FIELDDB_ELEMENT( TransferFunctionTexture, green, INPUT_OUTPUT );
    FIELDDB_ELEMENT( TransferFunctionTexture, blue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( TransferFunctionTexture, alpha, INPUT_OUTPUT );
  }
  
  TransferFunctionTexture::TransferFunctionTexture() :
    updateImage( new UpdateImage ),
    red( new SFTransferFunction ),
    green( new SFTransferFunction ),
    blue( new SFTransferFunction ),
    alpha( new SFTransferFunction )
  {
    
    this->type_name = "TransferFunctionTexture";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // dependencies
    updateImage->setOwner( this );
    
    // set the texture image      
    unsigned char data[NTEX][4];
    for(int i=0; i<NTEX; i++) { 
      data[i][0] = data[i][1] = data[i][2] = i; 
      data[i][3] = 255;
    }
    this->image->setValue( new PixelImage(NTEX,1,1,
					  32,
					  Image::RGBA,
					  Image::UNSIGNED,
					  &data[0][0], true) );
    
    // the texture properties
    TextureProperties *tp = new TextureProperties;
    tp->boundaryModeS->setValue("CLAMP_TO_EDGE");
    tp->boundaryModeT->setValue("CLAMP_TO_EDGE");
    tp->boundaryModeR->setValue("CLAMP_TO_EDGE");
    tp->minificationFilter->setValue("AVG_PIXEL");
    tp->magnificationFilter->setValue("AVG_PIXEL");
    tp->textureCompression->setValue("DEFAULT");
    this->textureProperties->setValue(tp);
    
    // routings
    red->route( updateImage );
    green->route( updateImage );
    blue->route( updateImage );
    alpha->route( updateImage );
    
    WISH_H3D_CREATE_STOP();
  }
  
  // update image if there are any transfer function updates
  void TransferFunctionTexture::UpdateImage::update() {
    
    typedef TransferFunction* TransferFunctionPointer;
    
    // nr of routes
    int rsize = int( routes_in.size() );
    
    // pointers to tf:s
    TransferFunctionPointer tf[4];
    tf[0] = tf[1] = tf[2] = tf[3] = 0;
    for( int i=0; i<rsize; i++ ) {
      tf[i] = static_cast<SFTransferFunction*>(routes_in[i])->getValue();
    }
    
    // check that we have four valid tf:s, otherwise skip
    if( !(tf[0] && tf[1] && tf[2] && tf[3]) ) 
      return;
    
    // image
    TransferFunctionTexture *tex = 
      static_cast<TransferFunctionTexture*>( this->getOwner() );
    PixelImage *im = static_cast<PixelImage*>( tex->image->getValue() );
    
    // loop image and evaluate tf:s
    unsigned int *ptr = (unsigned int*)(im->getImageData());
    for(int i=0; i<TransferFunctionTexture::NTEX; i++) {
      
      unsigned char *p = (unsigned char*)(&ptr[i]);      
      float t = float(i)/(TransferFunctionTexture::NTEX-1);
      float val;
      
      for( int j=0; j<rsize; j++ ) {
	val = tf[j]->evaluate(t);
	if( val < 0.0 ) val = 0.0;
	if( val > 1.0 ) val = 1.0;
	p[j] = (unsigned char) ( 255.0f*val );
      }
      
    }
    
    // make things update
    tex->image->setValue( im );
    
  }
  
}
