///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "H3DHapticsDevice.h"
#include "wishH3DVolume.hh"
#include "wishH3DVolumeChild.hh"

namespace wishH3D {
  
  void Volume::TotalExtent::update()  {
    int w =   static_cast<SFInt32*>( routes_in[0] )->getValue();
    int h =   static_cast<SFInt32*>( routes_in[1] )->getValue();
    int d =   static_cast<SFInt32*>( routes_in[2] )->getValue();
    Vec3f s = static_cast<SFVec3f*>( routes_in[3] )->getValue();
    value = Vec3f( w*s.x, h*s.y, d*s.z );
  }
  
  void Volume::SmartScaling::update() {
    float s = static_cast<SFFloat*>( routes_in[0] )->getValue();
    float w = static_cast<SFVec3f*>( routes_in[1] )->getValue().x;
    float tmp = s/w;
    value = Vec3f(tmp,tmp,tmp);
  }
  
  void Volume::LoadImage::update() {
    SFString::update();
    Volume *vol = static_cast<Volume*>( getOwner() );
    if( vol && value!="" ) {
      int k = vol->addImage( value ); 
      if( k>=0 ) {
	vol->currentImage->setValue( k );
      }
    }
  }
  
  void Volume::SaveImage::update() {
    SFString::update();
    Volume *vol = static_cast<Volume*>( getOwner() );
    if( vol && value!="" ) {
      wish::ImageObject *ci = vol->getCurrentImage();
      if( ci )
	ci->writeImage( value ); 
    }
  }
  
  void Volume::NewImage::update() {
    SFBool::update();
    Volume *vol = static_cast<Volume*>( getOwner() );
    if( vol && value ) {
      if( vol->getImage(0) ) {
	int W = vol->getImage(0)->getWidth();
	int H = vol->getImage(0)->getHeight();
	int D = vol->getImage(0)->getDepth();
	wish::Vec3f spc = vol->getImage(0)->getSpacing();
	wish::Image8 *newim = new wish::Image8(W,H,D,spc.x,spc.y,spc.z);
	newim->setLabel( "NewImage" );
	vol->addImage( newim );      
      }
      else {
	int W = vol->uwidth->getValue();
	int H = vol->uheight->getValue();
	int D = vol->udepth->getValue();
	Vec3f spc = vol->uspacing->getValue();
	wish::Image8 *newim = new wish::Image8(W,H,D,spc.x,spc.y,spc.z);
	newim->setLabel( "NewImage" );
	vol->addImage( newim );
	
      }
    }
  }
  
  void Volume::ClearImages::update() {
    bool b = static_cast<SFBool*>(routes_in[0])->getValue();
    Volume *vol = static_cast<Volume*>( getOwner() );
    if( vol && b ) {
      vol->removeImages();
    }
    value = b;
  }
  
  void Volume::QuitAPI::update() {
    bool b = static_cast<SFBool*>(routes_in[0])->getValue();
    if( b ) {
      throw wishH3D_QuitAPIException();
    }
    value = b;
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase Volume::database( "wishH3DVolume", 
				    &(newInstance<Volume>), 
				    typeid( Volume ),
				    &Group::database );
  
  namespace VolumeInternals {
    FIELDDB_ELEMENT( Volume, url, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, imageLabels, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, currentImage, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, nrImages, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, deriveSpacing, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, width, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, height, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, depth, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, uwidth, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, uheight, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, udepth, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, spacing, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, uspacing, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, totalExtent, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, voxelType, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, desiredSize, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, smartScaling, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, probePosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, proxyPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, voxelPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, voxelValue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, voxelValueNormalized, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, loadImage, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, saveImage, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, newImage, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, clearImages, INPUT_OUTPUT );
    FIELDDB_ELEMENT( Volume, quitAPI, INPUT_OUTPUT );
  }
  
  // constructor
  Volume::Volume() :
    imageLabels    ( new MFString ),
    currentImage   ( new SFInt32 ),
    nrImages       ( new SFInt32 ),
    deriveSpacing  ( new SFBool  ),
    width          ( new SFInt32 ),
    height         ( new SFInt32 ),
    depth          ( new SFInt32 ),
    uwidth         ( new SFInt32 ),
    uheight        ( new SFInt32 ),
    udepth         ( new SFInt32 ),
    spacing        ( new SFVec3f ),
    uspacing       ( new SFVec3f ),
    totalExtent    ( new TotalExtent ),
    voxelType      ( new SFString ),    
    desiredSize    ( new SFFloat ),
    smartScaling   ( new SmartScaling ),
    probePosition  ( new SFVec3f ),
    proxyPosition  ( new SFVec3f ),
    voxelPosition  ( new SFVec3f ),
    voxelValue     ( new SFFloat ),
    voxelValueNormalized( new SFFloat ),
    loadImage      ( new LoadImage ),
    saveImage      ( new SaveImage ),
    newImage       ( new NewImage ), 
    clearImages    ( new ClearImages ),
    quitAPI        ( new QuitAPI )
  {
    type_name = "Volume";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();    
    
    // dependencies
    loadImage->setOwner( this );
    saveImage->setOwner( this );
    newImage->setOwner( this );
    clearImages->setOwner( this );
    
    // the total extent
    width -> route( totalExtent );
    height -> route( totalExtent );
    depth -> route( totalExtent );
    spacing -> route( totalExtent );
    desiredSize -> route( smartScaling );
    totalExtent -> route( smartScaling );
    
    // can be overridden by haptics
    probePosition -> route( proxyPosition );
    
    // default settings
    width -> setValue( 128 );
    height -> setValue( 128 );
    depth -> setValue( 128 );
    uwidth -> setValue( 128 );
    uheight -> setValue( 128 );
    udepth -> setValue( 128 );
    spacing -> setValue( Vec3f(1.0,1.0,1.0) );
    uspacing -> setValue( Vec3f(1.0,1.0,1.0) );
    desiredSize -> setValue( 0.1f );
    deriveSpacing -> setValue( true );
    voxelType -> setValue( "uint8" );
    currentImage -> setValue(-1);
    nrImages -> setValue(0);
    imageLabels -> setValue( vector<string>(0) );
    images.resize(0);
    
    WISH_H3D_CREATE_STOP();
  };
  
  // destructor  
  Volume::~Volume() {
    removeImages();
  }

  // remove images
  void Volume::removeImages() {
    vector<wish::ImageObject*>::iterator it;
    for(it=images.begin(); it!=images.end(); it++) {
      delete (*it);
      (*it) = 0;
    }
    images.resize(0); 
    nrImages->setValue(0);
    
    imageLabels -> setValue( vector<string>(0) );
    currentImage -> setValue( -1 );
    
    width -> setValue( 1 );
    height -> setValue( 1 );
    depth -> setValue( 1 );
    spacing -> setValue( Vec3f(1.0,1.0,1.0) );  
  }
  
  wish::ImageObject* Volume::getImage( const string &label ) const {
    wish::ImageObject *im = 0;
    for( unsigned int i=0; i<images.size(); i++ ) {
      if( imageLabels->getValue()[i] == label ) {
	im = images[i];
	break;
      }
    }
    
    return im;
  }
  
  wish::ImageObject* Volume::getImage( unsigned int i ) const {    
    if( i >= 0 && i < images.size() )
      return images[i];
    else 
      return 0;
  }
  
  int Volume::addImage(wish::ImageObject *image) {
    if( !image ) { 
      std::cerr << "Image is NULL!" << std::endl;
      return -1;
    }
    
    // the first one sets the dimensions
    if( nrImages->getValue() == 0 ) {
      images.push_back( image );
      nrImages->setValue( 1 );
      
      // set the width
      width -> setValue( image->getWidth() );
      height -> setValue( image->getHeight() );
      depth -> setValue( image->getDepth() );
      
      // set the voxel size
      spacing -> setValue( Vec3Cast(image->getSpacing()) );
    }
    
    else {
      if( image->getWidth()  != width->getValue() ||
	  image->getHeight() != height->getValue() ||
	  image->getDepth()  != depth->getValue() ) {
	std::cerr << "Dimension mismatch!" << std::endl;
	delete image; // dangerous!?
	return -1;
      }
      
      images.push_back( image );
      nrImages->setValue( nrImages->getValue()+1 );
      
      if( deriveSpacing -> getValue() ) {
	image->setSpacing(images[0]->getSpacing());
      }
    }
    
    // set label if not defined
    int N = nrImages->getValue();
    if( N > (int)imageLabels->size() ) {
      imageLabels->push_back( image->getLabel() );
    }
    
    image->printSelf();
    
    cerr << "This label: " << imageLabels->getValue()[N-1] << endl;
    
    currentImage->setValue( N-1 );
    
    return N-1;
  }
  
  int Volume::addImage(const string &filename) {
    // read file
    // Here we use the dimension and spacing information that might 
    // have been given by the user in case he wants to read a raw image. 
    // If we read another image these values are overwritten.
    // One limitation is that if we have several raw images they
    // must be of the same datatype... Maybe I will fix this
    // by having an array of voxeltypes.
    wish::ImageObject *im = 
      wish::ImageObject::createFromFile(filename,
					uwidth->getValue(),
					uheight->getValue(),
					udepth->getValue(),
					uspacing->getValue().x,
					uspacing->getValue().y,
					uspacing->getValue().z,
					voxelType->getValue());
    
    // give it another try by looking in the data folder
    if( !im ) {
      cerr << "Looking in $(WISH_ROOT)/data/" << endl;
      string filename2 = getEnv( "WISH_ROOT" ) + "data/" + filename;
      im = wish::ImageObject::createFromFile(filename2,
					     uwidth->getValue(),
					     uheight->getValue(),
					     udepth->getValue(),
					     uspacing->getValue().x,
					     uspacing->getValue().y,
					     uspacing->getValue().z,
					     voxelType->getValue());
    }
   
    return addImage( im );
  }
  
  bool Volume::setImage( int i, wish::ImageObject *image ) {
    if( !image ) {
      std::cerr << "Image is NULL!" << std::endl;
      return false;
    }
    
    if( !(0 <= i && i<nrImages->getValue()) ) {
      std::cerr << "Index out of range!" << std::endl;
      return false;
    }
    
    if( image->getWidth()  != width->getValue() ||
	image->getHeight() != height->getValue() ||
	image->getDepth()  != depth->getValue() ) {
      std::cerr << "Dimension mismatch!" << std::endl;
      return false;
    }
    
    if( images[i] )
      delete images[i];
    
    images[i] = image;
    imageLabels->setValue( i, image->getLabel() );
    
    currentImage->setValue( i );

    return true;
  }

  void Volume::initialize() {
    WISH_H3D_INIT_START();
    
    // loop all urls 
    MFString *urls = static_cast< MFString* >( url.get() );    
    for(unsigned int i=0; i<urls->getValue().size(); i++) {
      // evaluate the url
      string filename = urls->getValue()[i];
      cerr << "Filename " << i << ": " << filename << endl;
      addImage( filename );
    }
    
    Group::initialize();
    
    WISH_H3D_INIT_STOP();
  }
  
  void Volume::traverseSG( TraverseInfo &ti ) {
    
    H3DHapticsDevice *hd = ti.getHapticsDevice(0);
    if( hd ) {
      Matrix4f T = ti.getAccInverseMatrix();
      Vec3f tp = hd->trackerPosition->getValue();
      Vec3f probepos = T*tp;
      
      // first set probe position
      probePosition->setValue( probepos );
      
      // compute voxel position from proxy position, 
      // if there are no volume haptics node, 
      // the probe position is equal to the proxy position
      Vec3f voxelpos = Vec3f(-1.0f,-1.0f,-1.0f);
      if( images.size()>0 ) {
	voxelpos = worldToImage( proxyPosition->getValue() );
      }
      voxelPosition->setValue( voxelpos );
      
      H3DFloat val = 0.0;
      H3DFloat valn = 0.0;
      wish::ImageObject *curim = getCurrentImage();      
      if( curim ) {
	val = H3DFloat( curim->getSafeAsDouble(Vec3Cast(voxelpos)) );
	valn = H3DFloat( curim->getSafeNormalized(Vec3Cast(voxelpos)) );
      }
      voxelValue->setValue( val );
      voxelValueNormalized->setValue( valn );
    }
    
    BaseClass::traverseSG( ti );
    
  }
  
  
}
