///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_H3D_VOLUME_HH_
#define _WISH_H3D_VOLUME_HH_

#include "Group.h"
#include "X3DUrlObject.h"
#include "X3DTextureNode.h"
#include "PixelImage.h"
#include "PeriodicUpdate.h"
#include "Exception.h"
#include "SFFloat.h"
#include "SFInt32.h"
#include "SFVec3f.h"
#include "SFMatrix4f.h"
#include "SFString.h"
#include "SFBool.h"

#include "wishH3D.hh"
#include "wishImage.hh"

namespace wishH3D {

  using namespace H3D;
  
  class Volume;
  
  typedef TypedSFNode<Volume> SFVolume;
  
  class WISHH3D_API Volume : public Group, public X3DUrlObject { 
  public:
    
    typedef Group BaseClass;
    
    // The Volume is the class that keeps all images that we want to use.
    // The vector images contain pointers to ImageObjects and
    // the integer current_image can be used to control which image is 
    // active now.
    
    // database
    static H3DNodeDatabase database;
    
    /// TotalExtent is the size of the volume according to spacing and 
    /// dimensions
    /// routes_in[0] = width
    /// routes_in[1] = height
    /// routes_in[2] = depth
    /// routes_in[3] = spacing
    /// value is the total extent
    class TotalExtent : 
      public TypedField< PeriodicUpdate<SFVec3f>, 
      Types<SFInt32, SFInt32, SFInt32, SFVec3f> > {
      virtual void update();
    };
    
    /// SmartScaling is a class derived from a SFVec3f that
    /// can be routed to make the width of the volume become screen_size in
    /// world coordinates
    /// routes_in[0] = desiredSize
    /// routes_in[1] = totalExtent
    /// value is the associated scaling
    class SmartScaling : 
      public TypedField< SFVec3f, Types<SFFloat,SFVec3f> > {
      virtual void update();
    };
    
    // Image loader
    // routes_in[0] = the filename to load
    class LoadImage : public AutoUpdate<SFString> {
      virtual void update();
    };

    // Image saver
    // routes_in[0] = the filename to save current image to
    class SaveImage : public AutoUpdate<SFString> {
      virtual void update();
    };

    // New image
    // this test version requires that an image is loaded,
    // this field will add a new wish::Image8 with the same dimensions 
    // as images[0]
    class NewImage : public AutoUpdate<SFBool> {
      virtual void update();
    };
    
    // Clear images
    class ClearImages : public AutoUpdate<SFBool> {
      virtual void update();
    };
    
    // To quit API
    H3D_API_EXCEPTION( wishH3D_QuitAPIException );
    class QuitAPI : public AutoUpdate< SFBool > {
      virtual void update();
    };
    
    // constructor
    Volume();
    
    // destructor
    ~Volume();
    
    // initialize function
    //bool initialized;
    virtual void initialize();
        
    // traverse scene graph
    virtual void traverseSG( TraverseInfo &ti );
    
    // check if position is inside volume (v in world coords)
    bool isInside( const Vec3f &v ) {
      Vec3f te = totalExtent->getValue()/2.0f;
      return ( -te.x <= v.x && v.x <= te.x &&
	       -te.y <= v.y && v.y <= te.y &&
	       -te.z <= v.z && v.z <= te.z );
    }
    
    // add an image, return value is the index
    int addImage(wish::ImageObject *im);
    
    // add an image, return value is the index
    int addImage(const string &filename);
    
    // set image with index i
    bool setImage( int i, wish::ImageObject *im );
    
    // remove images
    void removeImages();
    
    // return image from a given label or index
    wish::ImageObject* getImage( const string &label ) const;
    wish::ImageObject* getImage( unsigned int i ) const;
    wish::ImageObject* getCurrentImage() const {
      return getImage( currentImage->getValue() );
    }
    
    // convert from world to image and vice versa
    Vec3f worldToImage(const Vec3f &v) const {
      Vec3f ret(0,0,0);
      if( getImage(0) ) {
	ret = Vec3Cast( getImage(0)->worldToImage(Vec3Cast(v)) );
      }
      return ret;
    }
    Vec3f imageToWorld(const Vec3f &v) const {
      Vec3f ret(0,0,0);
      if( getImage(0) ) {
	ret = Vec3Cast( getImage(0)->imageToWorld(Vec3Cast(v)) );
      }
      return ret;
    }
    
    // images (the first one sets the dimensions)
    // maybe a map should be used since we use labels...
    vector<wish::ImageObject*> images;
    
    // image labels
    auto_ptr<MFString> imageLabels;
    
    // index of the current image
    auto_ptr<SFInt32> currentImage;
    
    // nr of images
    auto_ptr<SFInt32> nrImages;
    
    // if we want to derive spacing from the first image (default true)
    auto_ptr<SFBool> deriveSpacing;
    
    // voxel type if we want to read a raw image
    // should be uint8, uint16, uint32, float, double, RGBA8, RGB8, or Vec3f
    auto_ptr<SFString> voxelType;
    
    // dimensions of the volume
    auto_ptr<SFInt32> width;
    auto_ptr<SFInt32> height;
    auto_ptr<SFInt32> depth;  

    // specified by the user for raw image reading
    auto_ptr<SFInt32> uwidth;
    auto_ptr<SFInt32> uheight;
    auto_ptr<SFInt32> udepth;  
    
    // voxel size in meters
    auto_ptr<SFVec3f> spacing;
    
    // specified by the user
    auto_ptr<SFVec3f> uspacing;

    // total extent (dims times voxelsize)
    auto_ptr<TotalExtent> totalExtent;
    
    // screen scaling (can be routed from to get a good scaling)
    auto_ptr<SFFloat> desiredSize;
    auto_ptr<SmartScaling> smartScaling;
    
    // keep track of probe and proxy
    auto_ptr<SFVec3f> probePosition;
    auto_ptr<SFVec3f> proxyPosition;
    
    // TODO: add voxel position for both probe and proxy
    auto_ptr<SFVec3f> voxelPosition;
    auto_ptr<SFFloat> voxelValue;
    auto_ptr<SFFloat> voxelValueNormalized;
    
    // load image
    auto_ptr<LoadImage> loadImage;

    // save image
    auto_ptr<SaveImage> saveImage;

    // new image
    auto_ptr<NewImage> newImage;

    // clear images
    auto_ptr<ClearImages> clearImages;
    
    // quit API
    auto_ptr<QuitAPI> quitAPI;

  };
}
#endif
