///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DVolumeHaptics.hh"
#include "wishColor.hh"

using namespace H3D;

namespace wishH3D {
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase VolumeHaptics::database( "wishH3DVolumeHaptics", 
					   &(newInstance<VolumeHaptics>), 
					   typeid( VolumeHaptics ),
					   &H3DForceEffect::database );
  
  namespace VolumeHapticsInternals {
    FIELDDB_ELEMENT( VolumeHaptics, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, imageIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, proxyPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, probePosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, proxyWorldPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, probeWorldPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, force, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, forceMagnitude, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, stiffness, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, hapticMode, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, active, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeHaptics, onlyXY, INPUT_OUTPUT );
  }
  
  VolumeHaptics::VolumeHaptics() :
    imageIndex     ( new SFInt32 ),
    proxyPosition  ( new SFVec3f ),
    probePosition  ( new SFVec3f ),
    proxyWorldPosition  ( new SFVec3f ),
    probeWorldPosition  ( new SFVec3f ),
    force          ( new SFVec3f ),
    forceMagnitude ( new SFFloat ),
    stiffness      ( new SFFloat ),
    hapticMode     ( new SFHapticMode ),
    forceEffect    ( new ForceEffect ),
    active         ( new SFBool ),
    onlyXY         ( new SFBool )
  {
    
    type_name = "VolumeHaptics";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    proxyPosition->setValue( Vec3f(0,0,0) );
    probePosition->setValue( Vec3f(0,0,0) );
    proxyWorldPosition->setValue( Vec3f(0,0,0) );
    probeWorldPosition->setValue( Vec3f(0,0,0) );
    force->setValue( Vec3f(0,0,0) );
    forceMagnitude->setValue( 0.0f );
    stiffness->setValue( 400.0f );
    hapticMode->setValue( new HapticMode );
    imageIndex->setValue( 0 );
    active->setValue( true );
    onlyXY->setValue( false );

    WISH_H3D_CREATE_STOP();
  }
  
  void VolumeHaptics::initialize() {
    
    if( isInitialized() )
      return;
    
    WISH_H3D_INIT_START();
    
    BaseClass::initialize();
        
    // override the route in Volume 
    proxyPosition -> route( volume->getValue()->proxyPosition );
    
    WISH_H3D_INIT_STOP();
  }
  
  void VolumeHaptics::traverseSG( TraverseInfo &ti ) {
    
    BaseClass::traverseSG( ti );
    
    // get the probe
    Vec3f probe = getLocalProbe(ti);
    
    // the volume
    Volume *vol = volume->getValue();
    
    // voxel position
    bool inside = vol->isInside( probe );
    
    // set the things needed
    forceEffect->vh.setDeviceScale( getLocalToDeviceScale(ti) );
    forceEffect->vh.setStiffness( stiffness->getValue() );
    forceEffect->vh.setDamping(0.0);
    forceEffect->transform = ti.getAccForwardMatrix();
    
    // set the image dataset for the haptic mode  
    wish::ImageObject *im = vol->getImage(imageIndex->getValue());
    hapticMode->getValue()->setImage( im );
    
    // set the haptic mode
    forceEffect->haptic_mode = hapticMode->getValue();

    // set if we should apply force only in the xy plane
    forceEffect->only_xy = onlyXY->getValue();
    
    // add force if we are inside the volume, the image pointer is valid, and
    // active is true
    if( inside && im && active->getValue() ) { 
      ti.addForceEffectToAll( forceEffect.get() );
    }
    else { // must move the proxy to the probe
      forceEffect->vh.setProbe( Vec3Cast(probe) );
      forceEffect->vh.deactivate();
      forceEffect->vh.updateProxy();
    }
    
    // update the variables
    proxyPosition->setValue( Vec3Cast( forceEffect->vh.getProxy() ), id );
    probePosition->setValue( Vec3Cast( forceEffect->vh.getProbe() ), id );
    
    proxyWorldPosition->setValue( ti.getAccForwardMatrix()*
				  proxyPosition->getValue(), id );
    probeWorldPosition->setValue( ti.getAccForwardMatrix()*
				  probePosition->getValue(), id );
    
    force->setValue( Vec3Cast( forceEffect->vh.getForceFeedback() ) );
    Vec3f f = force->getValue();
    forceMagnitude->setValue( f.length() );
    
  }
  
  HapticForceEffect::EffectOutput 
  VolumeHaptics::ForceEffect::
  calculateForces( const HapticForceEffect::EffectInput &input ) {
    
    // lock
    lock.lock();
    
    // inverse of acc_forward_transform (global -> local)
    Matrix4f Tinv = transform.inverse();
    
    // rotational part of inverse transform
    Matrix3f Rinv = Tinv.getRotationPart();
    
    // rotational part of acc_forward_transform
    Matrix3f R( transform.getRotationPart() );
    
    // THE CURRENT STATE
    wish::Vec3f local_pos = Vec3Cast( Tinv*input.position );
    wish::Matrix3f local_orn = Matrix3Cast( Rinv*input.orientation );
    wish::Vec3f local_vel = Vec3Cast( Tinv*input.velocity );
    vh.setProbe( local_pos );
    vh.setOrientation( local_orn );
    vh.setVelocity( local_vel );
    
    // let the haptic mode set the LRF and strengths
    haptic_mode->computeParameters( vh );
    
    // update proxy position
    vh.updateProxy();
    
    // check if force is too large
    Vec3f force = Vec3Cast( vh.getForceFeedback() );
    if( force.length() > ForceEffect::MAX_FORCE ) {
      force = Vec3f(0,0,0);
      cerr << "FORCE TOO LARGE" << endl;
    }
    
    // unlock
    lock.unlock();
    
    // return force in world coordinates
    Vec3f ret = R*force;
    
    if( only_xy )
      ret.z = 0;
    
    return EffectOutput( ret );
    
  }
  
}
