///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __WISH_H3D_VOLUME_HAPTICS_HH__
#define __WISH_H3D_VOLUME_HAPTICS_HH__

#include "Transform.h"
#include "H3DForceEffect.h"
#include "H3DHapticsDevice.h"
#include "Threads.h"

#include "wishH3DVolumeChild.hh"
#include "wishH3DHapticMode.hh"

#include "wishImage.hh"
#include "wishVolumeHaptics.hh"

using namespace H3D;

namespace wishH3D {

  class WISHH3D_API VolumeHaptics : public H3DForceEffect, public VolumeChild {
  public:
    
    typedef H3DForceEffect BaseClass;
    
    /// Constructor.
    VolumeHaptics();
    
    /// initialize
    virtual void initialize();
    
    /// Traverse the scenegraph.
    virtual void traverseSG( TraverseInfo &ti );
    
    /// Get the scaling between local coords and device coords
    H3DFloat getLocalToDeviceScale( TraverseInfo &ti ) const {
      
      // device to local matrix
      Matrix4f D2L = 
	ti.getAccInverseMatrix()*
	ti.getHapticsDevice(0)->positionCalibration->getValue();
      
      // rotation and scale part
      Matrix3f D2L_rs = D2L.getScaleRotationPart();
      
      // find scale factor (OBS! only uniform scaling allowed)
      float scl = 1.0e-3f / ( D2L_rs*Vec3f(1,0,0) ).length();
      
      return scl;
    }

    /// Get the scaling between global (world) coords and device coords
    H3DFloat getGlobalToDeviceScale( TraverseInfo &ti ) const {
      
      // device to local matrix
      Matrix4f D2G = ti.getHapticsDevice(0)->positionCalibration->getValue();
      
      // rotation and scale part
      Matrix3f D2G_rs = D2G.getScaleRotationPart();
      
      // find scale factor (OBS! only uniform scaling allowed)
      float scl = 1.0e-3f / ( D2G_rs*Vec3f(1,0,0) ).length();
      
      return scl;
    }
    
    /// Get the probe in local coordinates
    Vec3f getLocalProbe( TraverseInfo &ti ) const {
      Vec3f probe = 
	ti.getAccInverseMatrix()*
	ti.getHapticsDevice(0)->
	trackerPosition->getValue();    
      return probe;
    }
    
    /// Get the probe in global (world) coordinates
    Vec3f getGlobalProbe( TraverseInfo &ti ) const {
      Vec3f probe = 
	ti.getHapticsDevice(0)->
	trackerPosition->getValue();    
      return probe;
    }
    
    /// data
    auto_ptr< SFVec3f > proxyPosition; // local
    auto_ptr< SFVec3f > probePosition; // local
    auto_ptr< SFVec3f > proxyWorldPosition;
    auto_ptr< SFVec3f > probeWorldPosition;
    auto_ptr< SFVec3f > force;
    auto_ptr< SFFloat > forceMagnitude;
    auto_ptr< SFFloat > stiffness; // N/m
    auto_ptr< SFBool > active; // render force or not
    
    /// the haptic mode
    auto_ptr< SFHapticMode > hapticMode;
    
    /// the image index
    auto_ptr< SFInt32 > imageIndex;
    
    /// only apply in xy plane
    auto_ptr< SFBool > onlyXY;
    
    /// The H3DNodeDatabase for this node.
    static H3DNodeDatabase database;
    
  protected:
    
    struct ForceEffect : HapticForceEffect {
      inline ForceEffect() : 
	HapticForceEffect( Matrix4f(), false )
      {}
      static const H3DFloat MAX_FORCE;
      EffectOutput calculateForces( const EffectInput &input );
      wish::VolumeHaptics vh;
      HapticMode *haptic_mode;
      bool only_xy;
      MutexLock lock;
    };
    
    AutoRef<ForceEffect> forceEffect;
    
  };
  
  const H3DFloat VolumeHaptics::ForceEffect::MAX_FORCE = 5.0;
}

#endif
