///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "Appearance.h"
#include "Material.h"
#include "FillProperties.h"
#include "Box.h"

#include "wishH3DVolumeOutline.hh"

#include <GL/glut.h>

using namespace H3D;

namespace wishH3D {
  
  VolumeOutlineGeometry::VolumeOutlineGeometry(Inst< SFNode > _metadata,
					       Inst< SFBound > _bound,
					       Inst< SFVec3f > _totalExtent ) :
    X3DGeometryNode( _metadata, _bound ),
    totalExtent( _totalExtent ) 
  {
    
    type_name = "VolumeOutlineGeometry";
   
    WISH_H3D_CREATE_START();
    
    totalExtent->setValue( Vec3f(0,0,0) );
    totalExtent->route(bound);
    totalExtent->route(displayList);
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  void VolumeOutlineGeometry::initialize() {
    
    WISH_H3D_INIT_START();

    BaseClass::initialize();
    
    WISH_H3D_INIT_STOP();
  }
  
  void VolumeOutlineGeometry::render() {
    
    Vec3f te = totalExtent->getValue();
    float xp = te.x;
    float yp = te.y;
    float zp = te.z;
    
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glLineWidth(2.0);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        
    glPushMatrix();
    glScalef(1.01f*xp,1.01f*yp,1.01f*zp);
    glutWireCube(1.0);
    glPopMatrix();
    
    glPopAttrib();
    
  }

  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase VolumeOutline::database( "wishH3DVolumeOutline", 
					   &(newInstance<VolumeOutline>), 
					   typeid( VolumeOutline ),
					   &Shape::database );
  
  namespace VolumeOutlineInternals {
    FIELDDB_ELEMENT( VolumeOutline, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeOutline, color, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeOutline, thickness, INPUT_OUTPUT );
  }
  
  VolumeOutline::VolumeOutline() :
    volume   ( new SFVolume ), 
    color    ( new SFColor ),
    thickness( new SFFloat )
  {
    
    type_name = "VolumeOutline";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    color->setValue( RGB(1,1,1) );
    thickness->setValue(1.0);
    
    WISH_H3D_CREATE_STOP();
    
  }
  
  void VolumeOutline::initialize() {
    
    if( isInitialized() ) 
      return;
    
    WISH_H3D_INIT_START();
    
    BaseClass::initialize();
    
    Volume *vol = volume -> getValue();
    
    Appearance *app = new Appearance;
    Material *mat = new Material;
    VolumeOutlineGeometry *vog = new VolumeOutlineGeometry;
    LineProperties *lp = new LineProperties;
    
    // appearance and material settings and routings
    thickness -> route( lp->linewidthScaleFactor );
    color -> route( mat->emissiveColor );
    mat -> diffuseColor -> setValue( RGB(0.1f,0.1f,0.1f) );
    app -> material -> setValue( mat );
    app -> lineProperties -> setValue( lp );
    
    // route volume total extent to geometry size
    vol -> totalExtent -> route( vog->totalExtent );	
    
    this -> appearance -> setValue( app );
    this -> geometry -> setValue( vog );
    
    WISH_H3D_INIT_STOP();
  }
  
}
