///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DVolumeRenderer.hh"

#include "PixelTexture.h"
#include "TextureProperties.h"

namespace wishH3D {
  
  void VolumeRenderer::UpdateDimensions::update() {
    SFVec3f::update();
    
    VolumeRenderer *vr = static_cast<VolumeRenderer*>( getOwner() );
    
    // dimensions
    Vec3f te = value;
    
    // setup texture matrices
    // rotation about x
    Rotation R  = Rotation( Vec3f(1.0f, 0.0f, 0.0f), H3DFloat(Constants::pi) );
    
    // scaling according to total extent => [-0.5, 0.5]
    Matrix4f S;
    if( te != Vec3f(0,0,0) )
      S = Matrix4f(1.0f/te.x, 0, 0, 0,
		   0, 1.0f/te.y, 0, 0,
		   0, 0, 1.0f/te.z, 0,
		   0, 0, 0, 1.0f);
    
    // center => [0, 1]
    Matrix4f  T  = Matrix4f(1, 0, 0, 0.5,
			    0, 1, 0, 0.5,
			    0, 0, 1, 0.5,
			    0, 0, 0, 1.0);
    
    Matrix4f M_tex = T*S*R;
    vr->textureMatrix->setValue( M_tex );
    vr->textureMatrixInverse->setValue( M_tex.inverse() );
  }
  
  void VolumeRenderer::SFBound::update() {
    Vec3f size = static_cast<SFVec3f*>(routes_in[0])->getValue();
    BoxBound *bb = new BoxBound;
    bb->size->setValue( size );
    value = bb;
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  VolumeRenderer::database( "wishH3DVolumeRenderer", 
			    &(newInstance<VolumeRenderer>), 
			    typeid( VolumeRenderer ),
			    &VolumeRenderer::BaseClass::database );
  
  namespace VolumeRendererInternals {
    FIELDDB_ELEMENT( VolumeRenderer, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, texture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, normalTexture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, tfTexture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, rayStep, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, isoValue, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, renderMode, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeRenderer, active, INPUT_OUTPUT );
  }
  
  VolumeRenderer::VolumeRenderer() :
    volume( new SFVolume ),
    texture( new SFTexture3DNode ),
    normalTexture( new SFTexture3DNode ),
    tfTexture( new SFTexture2DNode ),
    rayStep( new SFFloat ),    
    isoValue( new SFFloat ),
    renderMode( new SFInt32 ),
    active( new SFBool ),
    textureMatrix( new SFMatrix4f ),
    textureMatrixInverse( new SFMatrix4f ),
    updateDimensions( new UpdateDimensions ),
    shader( new ComposedShader )
  {
    
    this->type_name = "VolumeRenderer";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // dependencies
    updateDimensions->setOwner( this );
    
    // defaults    
    renderMode->setValue( 0 );
    rayStep->setValue( 0.01f );
    isoValue->setValue( 1.0 );
    active->setValue( true );

    // specialized bound
    bound.reset( new SFBound );
    
    // routes
    updateDimensions->route( bound );
        
    texture->route( displayList );
    normalTexture->route( displayList );
    rayStep->route( displayList );
    tfTexture->route( displayList );
    isoValue->route( displayList );
    renderMode->route( displayList );
    active->route( displayList );

    // shader
    shader->language->setValue( "GLSL" );
    string shadersdir = getEnv( "WISH_ROOT" ) + "/wishH3D/src/shaders/";
    
    // vertex shader
    ShaderPart *vs = new ShaderPart;
    vs->type->setValue( "VERTEX" );
    vs->url->push_back( shadersdir + "VolumeRenderer_VS.glsl" );
    shader->parts->push_back( vs );
    
    // fragment shader
    ShaderPart *fs = new ShaderPart;
    fs->type->setValue( "FRAGMENT" );
    fs->url->push_back( shadersdir + "VolumeRenderer_FS.glsl" );
    shader->parts->push_back( fs );
    
    // the render mode
    shader->addField("renderMode", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(renderMode));
    
    // the ray step
    shader->addField("rayStep", 
 		     H3D::Field::INPUT_OUTPUT, 
 		     copyAndRouteField(rayStep));

    // the ray step
    shader->addField("isoValue", 
 		     H3D::Field::INPUT_OUTPUT, 
 		     copyAndRouteField(isoValue));
    
    // texture matrices
    shader->addField("textureMatrix", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(textureMatrix));
    shader->addField("textureMatrixInverse", 
 		     H3D::Field::INPUT_OUTPUT, 
 		     copyAndRouteField(textureMatrixInverse));
    
    // add the volume texture to the shader
    shader->addField("volumeTexture", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(texture) );
    
    // add the normal texture to the shader
    shader->addField("normalTexture", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(normalTexture) );

    // add the transfer function texture to the shader
    shader->addField("tfTexture", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(tfTexture) );
    
    shader->activate->setValue( true );
    shader->displayList->route( displayList );
    
    WISH_H3D_CREATE_STOP();
 
  }
  
  void VolumeRenderer::initialize() {
    WISH_H3D_INIT_START();
    
    // route volumes total extent to update dimensions
    Volume *vol = volume->getValue();
    if( !vol ) {
      cerr << "Volume field is NULL!" << endl;
      exit(1);
    }
    vol->totalExtent->route( updateDimensions );
    
    // default transfer function
    if( tfTexture->getValue() == 0 ) {
      unsigned char data[256][4];
      for(int i=0; i<256; i++) { 
	data[i][0] = data[i][1] = data[i][2] = i; 
	data[i][3] = 255;
      }
      PixelTexture *pt = new PixelTexture;
      pt->image->setValue( new PixelImage(256,1,1,
					  32,
					  Image::RGBA,
					  Image::UNSIGNED,
					  &data[0][0], true) );
      
      TextureProperties *tp = new TextureProperties;
      tp->boundaryModeS->setValue("CLAMP_TO_EDGE");
      tp->boundaryModeT->setValue("CLAMP_TO_EDGE");
      tp->boundaryModeR->setValue("CLAMP_TO_EDGE");
      tp->minificationFilter->setValue("AVG_PIXEL");
      tp->magnificationFilter->setValue("AVG_PIXEL");
      tp->textureCompression->setValue("DEFAULT");
      pt->textureProperties->setValue(tp);
      
      tfTexture->setValue( pt );
    }
    
    BaseClass::initialize();
    
    WISH_H3D_INIT_STOP();
  }
  
  void VolumeRenderer::traverseSG( TraverseInfo &ti ) {
    BaseClass::traverseSG(ti);
  }
  
  void VolumeRenderer::renderBox(const Vec3f& te) const { 
    // A box representing the volume image. The vertices are given in
    // volume coordinates and the texture coordinates are correct,
    // no rotation needed.
    
    float xp = te.x/2.0f;
    float yp = te.y/2.0f;
    float zp = te.z/2.0f;
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_CULL_FACE);
    
    glBegin(GL_QUADS);
    
    // front
    glNormal3f(0,0,1);
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp, zp);
    glTexCoord3f(1,1,0); glVertex3f( xp, -yp, zp);
    glTexCoord3f(1,0,0); glVertex3f( xp,  yp, zp);
    glTexCoord3f(0,0,0); glVertex3f(-xp,  yp, zp);
    
    // back
    glNormal3f(0,0,-1);
    glTexCoord3f(0,0,1); glVertex3f(-xp,  yp, -zp);  
    glTexCoord3f(1,0,1); glVertex3f( xp,  yp, -zp);
    glTexCoord3f(1,1,1); glVertex3f( xp, -yp, -zp);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);
  
    // left
    glNormal3f(-1,0,0);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);  
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp,  zp);
    glTexCoord3f(0,0,0); glVertex3f(-xp,  yp,  zp);
    glTexCoord3f(0,0,1); glVertex3f(-xp,  yp, -zp);
  
    // right
    glNormal3f(1,0,0);
    glTexCoord3f(1,0,1); glVertex3f(xp,  yp, -zp);
    glTexCoord3f(1,0,0); glVertex3f(xp,  yp,  zp);
    glTexCoord3f(1,1,0); glVertex3f(xp, -yp,  zp);
    glTexCoord3f(1,1,1); glVertex3f(xp, -yp, -zp);
    
    // top
    glNormal3f(0,1,0);
    glTexCoord3f(0,0,0); glVertex3f(-xp, yp,  zp);
    glTexCoord3f(1,0,0); glVertex3f( xp, yp,  zp);
    glTexCoord3f(1,0,1); glVertex3f( xp, yp, -zp);
    glTexCoord3f(0,0,1); glVertex3f(-xp, yp, -zp);
  
    // bottom
    glNormal3f(0,-1,0);
    glTexCoord3f(0,1,1); glVertex3f(-xp, -yp, -zp);  
    glTexCoord3f(1,1,1); glVertex3f( xp, -yp, -zp);
    glTexCoord3f(1,1,0); glVertex3f( xp, -yp,  zp);
    glTexCoord3f(0,1,0); glVertex3f(-xp, -yp,  zp);
    
    glEnd();
    
    glPopAttrib();
  }
  
  void VolumeRenderer::render() {

    if( !(active->getValue()) )
      return;

    // get the volume
    Volume *vol = volume->getValue();
    
    // total extent
    Vec3f te = vol->totalExtent->getValue();
    
    // render bounding box as proxy geometry and perform raycasting
    shader->setSelected( true );
    shader->preRender();
    shader->displayList->callList();
    
    renderBox(te);
    
    shader->postRender();
  }


}
