///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DVolumeSlicer.hh"

#include "ImageTexture.h"

namespace wishH3D {
  
  HapticForceEffect::EffectOutput 
  SlicerForceEffect::calculateForces( const EffectInput &input ) {
    // rotational part of acc_forward_transform
    Matrix3f R( transform.getRotationPart() );
    
    // inverse => global to local
    Matrix4f T( transform.inverse() );
    
    // local pos
    Vec3f localpos = T*input.position;
    
    // update positions
    this->setProbe( localpos );
    
    // compute force
    this->computeForceFeedback();
    Vec3f f = this->getForce();
        
    // safety
    H3DFloat l = f.length();
    if( l > 6.0f ) {
      cerr << "FORCE MAGNITUDE: " << l << endl;
      f = Vec3f(0,0,0);
    }   
    
    // return force in world coordinates
    return EffectOutput( R*f );
  }
  
  void SlicerForceEffect::computeForceFeedback() {
    Vec3f n = normal;
    n.normalizeSafe();
    H3DFloat d = (probe-position)*n;
        
    if( active ) {
      if( H3DAbs(d) >= escape_distance ) {
	active = false;
	force = Vec3f(0,0,0);
      }
      else {
	force = -d*stiffness*normal;
      }
    }
    else {
      if( H3DAbs(d) < start_distance ) {
	active = true;
	force = -d*stiffness*normal;
      }
      else {
	force = Vec3f(0,0,0);
      }
    }
  }
  
  void VolumeSlicer::UpdateDimensions::update() {
    SFVec3f::update();
    
    VolumeSlicer *vs = static_cast<VolumeSlicer*>( getOwner() );
    
    // dimensions
    Vec3f te = value;
    
    // setup texture matrices
    // rotation about x
    Rotation R  = Rotation( Vec3f(1.0f, 0.0f, 0.0f), H3DFloat(Constants::pi) );
    
    // scaling according to total extent => [-0.5, 0.5]
    Matrix4f S;
    if( te != Vec3f(0,0,0) )
      S = Matrix4f(1.0f/te.x, 0, 0, 0,
		   0, 1.0f/te.y, 0, 0,
		   0, 0, 1.0f/te.z, 0,
		   0, 0, 0, 1.0f);
    
    // center => [0, 1]
    Matrix4f  T  = Matrix4f(1, 0, 0, 0.5,
			    0, 1, 0, 0.5,
			    0, 0, 1, 0.5,
			    0, 0, 0, 1.0);
    
    Matrix4f M_tex = T*S*R;
    vs->textureMatrix->setValue( M_tex );
    vs->textureMatrixInverse->setValue( M_tex.inverse() );
  }
  
  void VolumeSlicer::Grabber::update() {
    VolumeSlicer *vs = static_cast<VolumeSlicer*>( getOwner() );
    
    // get routes
    bool pressed = static_cast<SFBool*>(routes_in[0])->getValue();
    Vec3f pos = static_cast<SFVec3f*>(routes_in[1])->getValue();	
    
    // grabbing
    if( !is_grabbed && pressed ) {
      is_grabbed = true;
      grab_pos = pos;
      cur_pos = value;
    }
     
    // releasing
    else if( is_grabbed && !pressed ) {
      is_grabbed = false;
    }
    
    // moving
    else if( is_grabbed && pressed ) {
      Volume *vol = vs->volume->getValue();
      if( vol ) {
	Vec3f te = vol->totalExtent->getValue()/2.0f;
	Vec3f delta = pos-grab_pos;
	value = cur_pos+delta;
	value.x = H3DMin(value.x,  te.x);
	value.x = H3DMax(value.x, -te.x);
	value.y = H3DMin(value.y,  te.y);
	value.y = H3DMax(value.y, -te.y);
	value.z = H3DMin(value.z,  te.z);
	value.z = H3DMax(value.z, -te.z);
      }
    }

  }
  
  void VolumeSlicer::SFBound::update() {
    Vec3f size = static_cast<SFVec3f*>(routes_in[0])->getValue();
    BoxBound *bb = new BoxBound;
    bb->size->setValue( size );
    value = bb;
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  VolumeSlicer::database( "wishH3DVolumeSlicer", 
			  &(newInstance<VolumeSlicer>), 
			  typeid( VolumeSlicer ),
			  &VolumeSlicer::BaseClass::database );
  
  namespace VolumeSlicerInternals {
    FIELDDB_ELEMENT( VolumeSlicer, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, texture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, secondaryTexture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, colormap, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, overlay, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, renderMode, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, hapticsEnabled, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, activeX, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, activeY, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, activeZ, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, window, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, level, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, outlineColor, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, grab, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, grabPosition, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeSlicer, sliceTranslation, INPUT_OUTPUT ); 
  }
  
  VolumeSlicer::VolumeSlicer() :
    volume( new SFVolume ),
    window( new SFFloat ),
    level( new SFFloat ),
    texture( new SFTexture3DNode ),
    secondaryTexture( new SFTexture3DNode ),
    overlay( new SFTexture3DNode ),
    colormap( new SFTexture2DNode ),
    renderMode( new SFInt32 ),
    hapticsEnabled( new SFBool ),
    activeX( new SFBool ),
    activeY( new SFBool ),
    activeZ( new SFBool ),
    textureMatrix( new SFMatrix4f ),
    textureMatrixInverse( new SFMatrix4f ),
    updateDimensions( new UpdateDimensions ),
    outlineColor( new SFColor ),
    grabber( new Grabber ),
    grab( new SFBool ),
    grabPosition( new SFVec3f ),
    sliceTranslation( new SFVec3f ),
    shader( new ComposedShader ),
    slicer_force_effectX( new SlicerForceEffect ),
    slicer_force_effectY( new SlicerForceEffect ),
    slicer_force_effectZ( new SlicerForceEffect )
  {
    
    this->type_name = "VolumeSlicer";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // dependencies
    updateDimensions->setOwner( this );
    grabber->setOwner( this );
    
    // defaults    
    window->setValue(1.0f);
    level->setValue(0.5f);
    outlineColor->setValue( RGB(.7f,.7f,.7f) );
    grab->setValue(false);
    grabPosition->setValue(Vec3f(0,0,0));
    sliceTranslation->setValue( Vec3f(0,0,0) );
    hapticsEnabled->setValue( false );
    activeX->setValue( true );
    activeY->setValue( true );
    activeZ->setValue( true );
    renderMode->setValue( 0 );
    
    // specialized bound
    bound.reset( new SFBound );
    
    // routes
    updateDimensions->route( bound );
    
    grab->route( grabber );
    grabPosition->route( grabber );
    grabber->route( sliceTranslation );
    
    window->route( displayList );
    level->route( displayList );
    sliceTranslation->route( displayList );
    texture->route( displayList );
    secondaryTexture->route( displayList );
    overlay->route( displayList );
    colormap->route( displayList );
    renderMode->route( displayList );
    activeX->route( displayList );
    activeY->route( displayList );
    activeZ->route( displayList );

    // shader
    shader->language->setValue( "GLSL" );
    string shadersdir = getEnv( "WISH_ROOT" ) + "/wishH3D/src/shaders/";
    
    // vertex shader
    ShaderPart *vs = new ShaderPart;
    vs->type->setValue( "VERTEX" );
    vs->url->push_back( shadersdir + "VolumeSlicer_VS.glsl" );
    shader->parts->push_back( vs );
    
    // fragment shader
    ShaderPart *fs = new ShaderPart;
    fs->type->setValue( "FRAGMENT" );
    fs->url->push_back( shadersdir + "VolumeSlicer_FS.glsl" );
    shader->parts->push_back( fs );
    
    // add the window and level to the shader
    shader->addField("window", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(window));
    shader->addField("level", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(level));
    
    // the render mode
    shader->addField("renderMode", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(renderMode));
    
    // texture matrices
    shader->addField("textureMatrix", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(textureMatrix));

    // add the volume texture to the shader
    shader->addField("texture", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(texture) );

    // add the volume texture to the shader
    shader->addField("secondaryTexture", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(secondaryTexture) );
    
    // add the overlay to the shader
    shader->addField("overlay", 
     		     H3D::Field::INPUT_OUTPUT, 
     		     copyAndRouteField(overlay) );
    
    // add the colormap to the shader
    shader->addField("colormap", 
     		     H3D::Field::INPUT_OUTPUT, 
     		     copyAndRouteField(colormap) );
    
    shader->activate->setValue( true );
    shader->displayList->route( displayList );
    
    WISH_H3D_CREATE_STOP();
 
  }
  
  void VolumeSlicer::initialize() {
    WISH_H3D_INIT_START();
    
    // route volumes total extent to update dimensions
    Volume *vol = volume->getValue();
    if( !vol ) {
      cerr << "Volume field is NULL!" << endl;
      exit(1);
    }
    vol->totalExtent->route( updateDimensions );
    
    // default colormap
    if( colormap->getValue() == 0 ) {
      ImageTexture *imtex = new ImageTexture;
      imtex->url->push_back( getEnv( "WISH_ROOT" ) + 
			     "/wishH3D/textures/jet.png" );
      
      // texture properties
      TextureProperties *tp = new TextureProperties;
      tp->boundaryModeS->setValue( "CLAMP_TO_EDGE" );
      tp->boundaryModeT->setValue( "CLAMP_TO_EDGE" );
      tp->boundaryModeR->setValue( "CLAMP_TO_EDGE" );
      tp->borderWidth->setValue( 0 );
      tp->magnificationFilter->setValue( "AVG_PIXEL" );
      tp->minificationFilter->setValue( "AVG_PIXEL" );
      tp->textureCompression->setValue( "DEFAULT" );
      imtex->textureProperties->setValue( tp );
      
      colormap->setValue( imtex );
    }
    
    BaseClass::initialize();
    
    WISH_H3D_INIT_STOP();
  }
  
  void VolumeSlicer::traverseSG( TraverseInfo &ti ) {
    BaseClass::traverseSG(ti);
    
    // get the volume
    Volume *vol = volume->getValue();
    
    // get the probe position
    Vec3f xtip = vol->probePosition->getValue();
    
    bool ax, ay, az;
    ax = activeX->getValue();
    ay = activeY->getValue();
    az = activeZ->getValue();
    
    if( !((ax || ay || az) &&
	  hapticsEnabled->getValue() &&
	  vol->getImage(0) && 
	  vol->isInside(xtip) && 
	  !grab->getValue()) )
      return;
    
    // device to local matrix
    Matrix4f D2L = 
      ti.getAccInverseMatrix()*
      ti.getHapticsDevice(0)->positionCalibration->getValue();
    
    // rotation and scale part
    Matrix3f D2L_rs = D2L.getScaleRotationPart();
    
    // find scale factor (OBS! only uniform scaling allowed)
    float scl = 1.0e-3f / ( D2L_rs*Vec3f(1,0,0) ).length();
    
    // set force effect params
    Vec3f te = vol->totalExtent->getValue();
    H3DFloat s = 3.0f/128.0f;
    
    // X
    if( ax ) {
      slicer_force_effectX->transform = ti.getAccForwardMatrix();
      slicer_force_effectX->setStiffness( 400.0f );
      slicer_force_effectX->setDeviceScale( scl );
      slicer_force_effectX->setPosition( sliceTranslation->getValue() );
      slicer_force_effectX->setNormal( Vec3f(1,0,0) );
      slicer_force_effectX->setStartDistance( s*te.x );
      slicer_force_effectX->setEscapeDistance( 1.2f*s*te.x );
      
      ti.addForceEffectToAll( slicer_force_effectX.get() );   
    }
    
    // Y
    if( ay ) {
      slicer_force_effectY->transform = ti.getAccForwardMatrix();
      slicer_force_effectY->setStiffness( 400.0f );
      slicer_force_effectY->setDeviceScale( scl );
      slicer_force_effectY->setPosition( sliceTranslation->getValue() );
      slicer_force_effectY->setNormal( Vec3f(0,1,0) );
      slicer_force_effectY->setStartDistance( s*te.y );
      slicer_force_effectY->setEscapeDistance( 1.2f*s*te.y );
      
      ti.addForceEffectToAll( slicer_force_effectY.get() );   
    }
    
    // Z
    if( az ) {
      slicer_force_effectZ->transform = ti.getAccForwardMatrix();
      slicer_force_effectZ->setStiffness( 400.0f );
      slicer_force_effectZ->setDeviceScale( scl );
      slicer_force_effectZ->setPosition( sliceTranslation->getValue() );
      slicer_force_effectZ->setNormal( Vec3f(0,0,1) );
      slicer_force_effectZ->setStartDistance( s*te.z );
      slicer_force_effectZ->setEscapeDistance( 1.2f*s*te.z );
      
      ti.addForceEffectToAll( slicer_force_effectZ.get() );   
    }
  }
  
  
  void VolumeSlicer::renderPlane(const int& axis, 
				 const Vec3f& te, 
				 const Vec3f& trans) const {
    
    // save attributes
    glPushAttrib( GL_TEXTURE_BIT | GL_POINT_BIT | 
		  GL_CURRENT_BIT | GL_LINE_BIT );
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glDisable( GL_CULL_FACE );
    
    // x-axis (view from the right)
    if( axis == 0 ) {
      glBegin(GL_QUADS);
      glNormal3f(1,0,0);
      
      //glTexCoord3f(pt.x,1,0);
      glTexCoord3f(trans.x, -te.y,  te.z);
      glVertex3f(trans.x, -te.y,  te.z);
      
      //glTexCoord3f(pt.x,1,1);
      glTexCoord3f(trans.x, -te.y, -te.z);
      glVertex3f(trans.x, -te.y, -te.z);
      
      //glTexCoord3f(pt.x,0,1);
      glTexCoord3f(trans.x,  te.y, -te.z);
      glVertex3f(trans.x,  te.y, -te.z);
      
      //glTexCoord3f(pt.x,0,0);
      glTexCoord3f(trans.x,  te.y,  te.z);
      glVertex3f(trans.x,  te.y,  te.z);
      
      glEnd();
    }
    else if( axis == 1 ) {
      glBegin(GL_QUADS);
      glNormal3f(0,1,0);
      
      //glTexCoord3f(0,pt.y,0);
      glTexCoord3f(-te.x, trans.y,  te.z);
      glVertex3f(-te.x, trans.y,  te.z);

      //glTexCoord3f(1,pt.y,0);
      glTexCoord3f( te.x, trans.y,  te.z);
      glVertex3f( te.x, trans.y,  te.z);
      
      //glTexCoord3f(1,pt.y,1);
      glTexCoord3f( te.x, trans.y, -te.z);
      glVertex3f( te.x, trans.y, -te.z);

      //glTexCoord3f(0,pt.y,1);
      glTexCoord3f(-te.x, trans.y, -te.z);
      glVertex3f(-te.x, trans.y, -te.z);
      
      glEnd();
    }
    else if( axis == 2 ) {
      glBegin(GL_QUADS);
      glNormal3f(0,0,1);
      
      //glTexCoord3f(0,1,pt.z);
      glTexCoord3f(-te.x, -te.y, trans.z);
      glVertex3f(-te.x, -te.y, trans.z);

      //glTexCoord3f(1,1,pt.z);
      glTexCoord3f( te.x, -te.y, trans.z);
      glVertex3f( te.x, -te.y, trans.z);

      //glTexCoord3f(1,0,pt.z);
      glTexCoord3f( te.x,  te.y, trans.z);
      glVertex3f( te.x,  te.y, trans.z);
      
      //glTexCoord3f(0,0,pt.z);
      glTexCoord3f(-te.x,  te.y, trans.z);
      glVertex3f(-te.x,  te.y, trans.z);
      
      glEnd();
    }
    
    glPopAttrib();
    
  }
  
  void VolumeSlicer::renderOutline(const int& axis, 
				   const Vec3f& te, 
				   const Vec3f& trans, 
				   const RGB& oc) const {
    
    // save attrib
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glLineWidth(1.0f);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        
    glColor3f(oc.r, oc.g, oc.b);
    glDisable( GL_TEXTURE_3D );
    glDisable( GL_LIGHTING );
    
    glPushMatrix();
    glScalef(1.001f,1.001f,1.001f);
    
    if( axis==0 ) {
      glBegin(GL_LINE_LOOP);
      glVertex3f(trans.x, -te.y,  te.z);
      glVertex3f(trans.x, -te.y, -te.z);
      glVertex3f(trans.x,  te.y, -te.z);
      glVertex3f(trans.x,  te.y,  te.z);
      glEnd();
    }
    else if( axis==1 ) {
      glBegin(GL_LINE_LOOP);
      glVertex3f(-te.x, trans.y,  te.z);
      glVertex3f( te.x, trans.y,  te.z);
      glVertex3f( te.x, trans.y, -te.z);
      glVertex3f(-te.x, trans.y, -te.z);
      glEnd();
    }
    else if( axis==2 ) {
      glBegin(GL_LINE_LOOP);
      glVertex3f(-te.x, -te.y, trans.z);
      glVertex3f( te.x, -te.y, trans.z);
      glVertex3f( te.x,  te.y, trans.z);
      glVertex3f(-te.x,  te.y, trans.z);
      glEnd();
    }
    
    glPopMatrix();
    
    glPopAttrib();
    
  }
  
  void VolumeSlicer::render() {
    
    bool ax, ay, az;
    ax = activeX->getValue();
    ay = activeY->getValue();
    az = activeZ->getValue();
    
    if( !(ax || ay || az) )
      return;
    
    // get the volume
    Volume *vol = volume->getValue();
    
    // total extent divided by 2
    Vec3f te = vol->totalExtent->getValue()/2.0f;
    
    // colors
    RGB oc = outlineColor->getValue();
    
    // translation
    Vec3f trans = sliceTranslation->getValue();
    
    // render planes using shader
    shader->setSelected( true );
    shader->preRender();
    shader->displayList->callList();
    
    if( ax )
      renderPlane(0,te,trans);
    if( ay )
      renderPlane(1,te,trans);
    if( az ) 
      renderPlane(2,te,trans);
    
    shader->postRender();
    
    // render outline
    if( ax )
      renderOutline(0,te,trans,oc);
    if( ay )
      renderOutline(1,te,trans,oc);
    if( az )
      renderOutline(2,te,trans,oc);
  }


}
