///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _WISH_H3D_VOLUME_SLICER_HH_
#define _WISH_H3D_VOLUME_SLICER_HH_

#include "X3DChildNode.h"
#include "X3DBoundedObject.h"
#include "H3DDisplayListObject.h"
#include "X3DTexture3DNode.h"
#include "X3DTexture2DNode.h"

#include "ComposedShader.h"
#include "ShaderPart.h"
#include "H3DHapticsDevice.h"
#include "HapticForceEffect.h"

#include "DependentNodeFields.h"
#include "PeriodicUpdate.h"
#include "SFFloat.h"
#include "SFBool.h"
#include "SFColor.h"
#include "SFVec3f.h"
#include "SFMatrix4f.h"

#include "wishH3DVolume.hh"

namespace wishH3D {
  
  class WISHH3D_API SlicerForceEffect : public HapticForceEffect {
  protected:
    Vec3f probe;
    Vec3f position;
    Vec3f normal;
    H3DFloat start_distance;
    H3DFloat escape_distance;
    Vec3f force;
    H3DFloat stiffness;
    H3DFloat device_scale;
    bool active;
    
  public:
    static const H3DFloat MAX_FORCE;
    
    inline SlicerForceEffect() : 
      HapticForceEffect( Matrix4f(), true ),
      probe(Vec3f(0,0,0)),
      position(Vec3f(0,0,0)),
      normal(Vec3f(0,0,1)),
      start_distance( 0.01f ),
      escape_distance( 0.02f ),
      force(Vec3f(0,0,0)),
      stiffness(100.0f),
      device_scale(1.0f),
      active( false )
    {}
    
    HapticForceEffect::EffectOutput 
    calculateForces( const EffectInput &input );
    
    void computeForceFeedback();
    
    void setProbe( const Vec3f &p ) { 
      probe = p*device_scale; 
    }
    Vec3f getProbe() const { 
      return probe/device_scale; 
    }
    
    void setPosition( const Vec3f &p ) { 
      position = p*device_scale; 
    }
    Vec3f getPosition() const { 
      return position/device_scale; 
    }
    
    void setNormal( const Vec3f &n ) { 
      normal = n; 
    }
    Vec3f getNormal() const { 
      return normal; 
    }
    
    void setStartDistance( const H3DFloat &sd ) { 
      start_distance = sd*device_scale; 
    }
    H3DFloat getStartDistance() const { 
      return start_distance/device_scale; 
    }
    
    void setEscapeDistance( const H3DFloat &ed ) { 
      escape_distance = ed*device_scale; 
    }
    H3DFloat getEscapeDistance() const { 
      return escape_distance/device_scale; 
    }
    
    void setStiffness(H3DFloat s) { 
      stiffness = s; 
    }
    H3DFloat getStiffness() const { 
      return stiffness; 
    }
    
    void setDeviceScale(H3DFloat s) { 
      device_scale = s; 
    }
    
    H3DFloat getDeviceScale() const { 
      return device_scale; 
    }
    
    Vec3f getForce() const { 
      return force; 
    }
  };
  
  // VolumeSlicer renders the volume using three axis aligned planes
  class WISHH3D_API VolumeSlicer : public Group {
  public:
    typedef Group BaseClass;
    
    /// Field that updates texture matrices from the volume
    struct UpdateDimensions : public PeriodicUpdate<SFVec3f> {
      virtual void update();
    };
    
    /// Field that controls the grab-transformation of slices
    struct Grabber : TypedField<SFVec3f, Types<SFBool, SFVec3f> > {
      bool is_grabbed;
      Vec3f grab_pos;
      Vec3f cur_pos;
      
      Grabber() : is_grabbed(false), 
		  grab_pos(Vec3f(0,0,0)),
		  cur_pos(Vec3f(0,0,0)) {}
      
      virtual void update();
    };
    
    /// Bound is updated from the total extent field
    struct SFBound : public TypedField<H3DBoundedObject::SFBound, SFVec3f> {
      virtual void update();
    };
    
    typedef DependentSFNode<X3DTexture3DNode, 
			    FieldRef<H3DDisplayListObject,
				     H3DDisplayListObject::DisplayList,
				     &H3D::H3DDisplayListObject::displayList>, 
			    true > SFTexture3DNode;
    
    typedef DependentSFNode<X3DTexture2DNode, 
			    FieldRef<H3DDisplayListObject,
				     H3DDisplayListObject::DisplayList,
				     &H3D::H3DDisplayListObject::displayList>, 
			    true > SFTexture2DNode;
    
    //! constructor
    VolumeSlicer();
    
    //! initializing function
    virtual void initialize();
    
    //! render functions
    virtual void render();
    
    void renderPlane(const int& axis, 
		     const Vec3f& te, const Vec3f& trans) const;
    
    void renderOutline(const int& axis, 
		       const Vec3f& te, const Vec3f& trans, 
		       const RGB& oc) const;
    
    //! traverse the scenegraph
    virtual void traverseSG( TraverseInfo &ti );
    
    //! the volume
    auto_ptr<SFVolume> volume;
    
    //! the textures
    auto_ptr<SFTexture3DNode> texture;
    auto_ptr<SFTexture3DNode> secondaryTexture;
    auto_ptr<SFTexture3DNode> overlay;
    auto_ptr<SFTexture2DNode> colormap;
    
    //! render mode
    //! 0 for no overlay
    //! 1 for region overlay
    //! 2 for colormap
    auto_ptr<SFInt32> renderMode;
    
    //! haptics
    auto_ptr<SFBool> hapticsEnabled;
    
    //! active
    auto_ptr<SFBool> activeX;
    auto_ptr<SFBool> activeY;
    auto_ptr<SFBool> activeZ;
    
    //! window/level contrast adjustment
    auto_ptr<SFFloat> window;
    auto_ptr<SFFloat> level;
    
    //! texture matrices
    auto_ptr<SFMatrix4f> textureMatrix; // world -> texture
    auto_ptr<SFMatrix4f> textureMatrixInverse; // texture -> world
    
    //! update texture matrices
    auto_ptr<UpdateDimensions> updateDimensions;
    
    //! the slices
    auto_ptr< SFColor > outlineColor;
    auto_ptr< Grabber > grabber;
    auto_ptr< SFBool > grab;
    auto_ptr< SFVec3f > grabPosition;
    auto_ptr< SFVec3f > sliceTranslation;
    
    //! the shader
    auto_ptr<ComposedShader> shader;
    
    //! force effect
    AutoRef<SlicerForceEffect> slicer_force_effectX;
    AutoRef<SlicerForceEffect> slicer_force_effectY;
    AutoRef<SlicerForceEffect> slicer_force_effectZ;
    
    // interface
    static H3DNodeDatabase database;
    
  };
  
}

#endif
