///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "wishH3DVolumeTexture.hh"

namespace wishH3D {
  
  void VolumeTexture::ImageChanger::update() {
    SFInt32::update();
    VolumeTexture *vt = static_cast<VolumeTexture*>( getOwner() );
    if( vt && vt->isInitialized() ) {
      vt->generateTexture();
    }
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  VolumeTexture::database( "wishH3DVolumeTexture", 
			   &(newInstance<VolumeTexture>), 
			   typeid( VolumeTexture ),
			   &X3DTexture3DNode::database );
  
  namespace VolumeTextureInternals {
    FIELDDB_ELEMENT( VolumeTexture, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTexture, imageIndex, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTexture, imageChanger, INPUT_OUTPUT );
  }
  
  VolumeTexture::VolumeTexture() : 
    imageIndex( new SFInt32 ),
    imageChanger( new ImageChanger )
  {
    
    this->type_name = "VolumeTexture";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    interpolate_string = "AVG_PIXEL";
    
    // dependencies
    imageChanger->setOwner( this );
    
    // default values
    imageIndex -> setValue(0);
    
    // routings
    imageIndex->route( imageChanger );
    
    WISH_H3D_CREATE_STOP();
 
  }
  
  wish::ImageObject* VolumeTexture::getImage() {
    return volume->getValue()->getImage( imageIndex->getValue() );
  }
  
  void VolumeTexture::generateTexture() {
    
    // the volume
    Volume *vol = volume->getValue();
    
    // the image index
    int ind = imageIndex->getValue();

    wish::ImageObject *im=vol->getImage( ind );
    
    if( !im ) {
      cerr << "Warning!!! Could not find the image with index " << ind << endl;
      this->image->setValue( 0 );
      return;
    }
    
    cerr << "Generating texture for index " << ind << endl;

    // dimensions
    int w = vol->width->getValue();
    int h = vol->height->getValue();
    int d = vol->depth->getValue();
    Vec3f te = vol->totalExtent->getValue();
    
    // the pixel type
    Image::PixelType pt;
    if( im->getDataType() == "RGBA8" ) {
      pt = Image::RGBA;
      cerr << "Image::RGBA" << endl;
    }
    else if( im->getDataType() == "RGB8" ) {
      pt = Image::RGB;
      cerr << "Image::RGB" << endl;
    }
    else if( im->getDataType() == "Vec3f" ) {
      pt = Image::VEC3;
      cerr << "Image::VEC3" << endl;
    }
    else {
      pt = Image::LUMINANCE;
      cerr << "Image::LUMINANCE" << endl;
    }
    
    // the pixel component type
    Image::PixelComponentType pct;
    if( im->isSignedIntegerType() ) {
      pct = Image::SIGNED;
      cerr << "Image::SIGNED" << endl;
    }
    else if( im->isUnsignedIntegerType() ) {
      pct = Image::UNSIGNED;
      cerr << "Image::UNSIGNED" << endl;
    }
    else {
      pct = Image::RATIONAL;
      cerr << "Image::RATIONAL" << endl;
    }
    
    // generate image
    
    // Note: Vec3 images not supported for texture, just set a null image
    if( pt==Image::VEC3 && pct==Image::RATIONAL ) {
      this->image->setValue( 0 );
      return;
    }
    
    // no copying of the data, the data will not be deleted by the 
    // VolumeImage class if copydata is false
    bool copydata = false;    
    unsigned int bitsperpixel = im->voxelBytes()*8;
    cerr << "Bits per pixel: " << bitsperpixel << endl;
    VolumeImage *texim;    
    texim = new VolumeImage(w,h,d, // dimensions
			    bitsperpixel, // bits per pixel
			    pt, // pixel type
			    pct, // pixel component type
			    im->getBytePointer(), // data
			    copydata); // copy

    // set the texture image
    this->image->setValue( texim );
    
    // texture properties
    TextureProperties *tp = new TextureProperties;
    tp->boundaryModeS->setValue( "CLAMP_TO_EDGE" );
    tp->boundaryModeT->setValue( "CLAMP_TO_EDGE" );
    tp->boundaryModeR->setValue( "CLAMP_TO_EDGE" );
    tp->borderWidth->setValue( 0 );
    tp->magnificationFilter->setValue( interpolate_string );
    tp->minificationFilter->setValue( interpolate_string );
    tp->textureCompression->setValue( "DEFAULT" );
    this->textureProperties->setValue( tp );
  }
  
  void VolumeTexture::initialize() {
    
    if( isInitialized() )
      return;
    
    WISH_H3D_INIT_START();
    
    Volume *vol = volume->getValue();
    
    BaseClass::initialize();
    
    // the current image of the volume
    //vol->currentImage->route( imageIndex );
    
    WISH_H3D_INIT_STOP();
  }

}
