///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007, Erik Vidholm
//
//
//   This file is part of WISH.
//
//   WISH is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   WISH is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with WISH.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

#include "TextureMatrixTransform.h"
#include "Appearance.h"
#include "ComposedShader.h"
#include "ShaderPart.h"
#include "ImageTexture.h"
#include "FreeImageLoader.h"
#include "MultiTexture.h"

#include "wishH3DVolumeTextureGraphics.hh"

namespace wishH3D {
  
  void VolumeTextureGraphics::UpdateDimensions::update() {
    SFVec3f::update();
    
    VolumeTextureGraphics *vtg = 
      static_cast<VolumeTextureGraphics*>(getOwner());
    
    // get volume
    Volume *vol = vtg->volume->getValue();
    if( !vol ) return;
    
    // dimensions
    int w = vol->width->getValue();
    int h = vol->height->getValue();
    int d = vol->depth->getValue();
    Vec3f te = vol->totalExtent->getValue();
    
    // setup texture matrices
    // rotation about x
    Rotation R  = Rotation( Vec3f(1.0f, 0.0f, 0.0f), H3DFloat(Constants::pi) );
    
    // scaling according to total extent => [-0.5, 0.5]
    
    Matrix4f S;
    if( te != Vec3f(0,0,0) )
      S = Matrix4f(1.0f/te.x, 0, 0, 0,
		   0, 1.0f/te.y, 0, 0,
		   0, 0, 1.0f/te.z, 0,
		   0, 0, 0, 1.0f);
    
    // center => [0, 1]
    Matrix4f  T  = Matrix4f(1, 0, 0, 0.5,
			    0, 1, 0, 0.5,
			    0, 0, 1, 0.5,
			    0, 0, 0, 1.0);
    
    Matrix4f M_tex = T*S*R;
    vtg->textureMatrix->setValue( M_tex );
    vtg->textureMatrixInverse->setValue( M_tex.inverse() );
  }
  
  // Add this node to the H3DNodeDatabase system.
  H3DNodeDatabase 
  VolumeTextureGraphics::database( "wishH3DVolumeTextureGraphics", 
				   &(newInstance<VolumeTextureGraphics>), 
				   typeid( VolumeTextureGraphics ),
				   &VolumeGraphics::database );
  
  namespace VolumeTextureGraphicsInternals {
    FIELDDB_ELEMENT( VolumeTextureGraphics, volume, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, texture, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, overlay, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, textureMatrix, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, transferFunction, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, window, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, level, INPUT_OUTPUT );
    FIELDDB_ELEMENT( VolumeTextureGraphics, intensityComponent, INPUT_OUTPUT );
  }
  
  VolumeTextureGraphics::VolumeTextureGraphics() : 
    window( new SFFloat ),
    level( new SFFloat ),
    intensityComponent( new SFInt32 ),
    texture( new SFVolumeTexture ),
    overlay( new SFVolumeTexture ),
    textureMatrix( new SFMatrix4f ),
    textureMatrixInverse( new SFMatrix4f ),
    transferFunction( new SFTransferFunction ),
    tfTexture( new SFTFTexture ),
    updateDimensions( new UpdateDimensions )
  {
    
    this->type_name = "VolumeTextureGraphics";
    database.initFields( this );
    
    WISH_H3D_CREATE_START();
    
    // dependencies
    updateDimensions->setOwner( this );
    
    // route the transfer function to the texture representation 
    transferFunction->route(tfTexture);
    
    window->route( displayList );
    level->route( displayList );
    intensityComponent->route( displayList );
    texture->route( displayList );
    overlay->route( displayList );
    
    window->setValue(1.0f);
    level->setValue(0.5f);
    intensityComponent->setValue(0);

    WISH_H3D_CREATE_STOP();
 
  }
  
  void VolumeTextureGraphics::initialize() {
    
    WISH_H3D_INIT_START();
    
    BaseClass::initialize();
    
    // route volumes total extent to update dimensions
    Volume *vol = volume->getValue();
    vol->totalExtent->route( updateDimensions );
    
    // set the appearance with texture and texture transform
    Appearance *app = dynamic_cast<Appearance*>(this->appearance->getValue());
    if( !app ) {
      app = new Appearance;
      this->appearance->setValue( app );
    }  
    
    // set the texture transform
    // this means that we just have to provide the same 
    // texture coordinates as ordinary vertices (should probably use 
    // texcoordgen instead, but not now)
    // if we use our own vertex shader we have to compute the texture coords
    // ourselves by using the texture matrix
    TextureMatrixTransform *tmt = new TextureMatrixTransform;
    this->textureMatrix->route( tmt->matrix );
    app->textureTransform->setValue( tmt );
    
    // the texture
    app->texture->setValue( this->texture->getValue() );
    this->texture->route( app->texture );
        
    // the default fragment shader just uses the transfer function
    ComposedShader *shader = new ComposedShader;
    shader->language->setValue( "GLSL" );
    
    string shadersdir = getEnv( "WISH_ROOT" ) + "wishH3D/src/shaders/";
    
    // fragment shader
    ShaderPart *fs = new ShaderPart;
    fs->type->setValue( "FRAGMENT" );
    fs->url->push_back( shadersdir + "WindowLevel_FS.glsl" );
    shader->parts->push_back( fs );
    
    // add the window and level
    shader->addField("window", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(window));
    shader->addField("level", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(level));
    shader->addField("intensityComponent", 
		     H3D::Field::INPUT_OUTPUT, 
		     copyAndRouteField(intensityComponent));
    
    // add the volume texture to the shader
    shader->addField("volumeTexture", 
    		     H3D::Field::INPUT_OUTPUT, 
    		     copyAndRouteField(texture) );
    
    app->shaders->push_back( shader );
    
    WISH_H3D_INIT_STOP();
  }

}
